PROGRAM Boot (infile, treefile, OUTPUT);
  (* version 3.1. (c) Copyright 1986, 1987, 1988 by the University of
     Washington and Joseph Felsenstein.  Written by Joseph Felsenstein.
     Permission is granted to copy and use this program provided no
     fee is charged for it and provided that this copyright
     notice is not removed. *)

CONST maxsp = 30;    (* maximum number of species.  <= bits.    *)
   maxsz =  60;      (* size of pointer array.  >= 2*maxsp - 1  *)
   maxchr = 200;     (* max. number characterstates per species *)
   bits = 30;        (* max. number of elements in a Pascal set *)
   wrds = 7;         (* wrds * bits >= maxchr                   *)
   nmlngth = 10;     (* number of characters in species name    *)
   maxtrees = 100;   (* maximum number of tied trees stored     *)
   maxgrp = 1000;    (* max. no. of groups in all trees found   *)

TYPE longer = ARRAY [0..5] OF INTEGER;
   steparray = ARRAY[1..maxchr] OF INTEGER;
   bitrange = 1..bits;
   bitset = SET OF bitrange;
   bitarray = ARRAY[1..wrds] OF bitset;
   spset = SET OF 1..maxsp;  (* in long range, make this an array *)
   nodeptr = ^node;   (* nodes will form a binary tree *)
   node = RECORD      (* describes a tip species or an ancestor *)
      ancestor, leftdesc, rtdesc : nodeptr;  (* pointers to nodes *)
      index : INTEGER;      (* number of the node *)
      tip : BOOLEAN;        (* present species are tips of tree *)
      seen : REAL;          (* number of times this set seen *)
      stateone, statezero : bitarray; (* see in PROCEDURE fillin *)
      nodeset : spset;      (* used by accumulate *)
      xcoord, ycoord : INTEGER    (* used by printree *)
      END;
   pointarray = ARRAY[1..maxsz] OF nodeptr;

VAR root : nodeptr;
   infile, treefile : TEXT;
   spp, nonodes, chars, words, outgrno,
      reps, repdiv10, i, rr, actualchars : INTEGER;
     (* spp = number of species
        nonodes = number of nodes in tree
        chars = number of binary characters
        words = number of words needed
                to represent characters of one organism
        outgrno indicates outgroup *)
   thresh, ancvar, factors, questions, jackknife,
      allsokal, allwagner, mixture, local, trout,
      noroot, outgropt, didreroot : BOOLEAN;
   seed : longer;
   extras, weight, factor : steparray;
   ancone, anczero : ARRAY[1..maxchr] OF BOOLEAN;
   treenode : pointarray;      (* pointers to all nodes in tree *)
   name : ARRAY[1..maxsp, 1..nmlngth] OF CHAR; (* names of species *)
   threshold : REAL;
   wagner : bitarray;
   threshwt : ARRAY[1..maxchr] OF REAL;
   grouping : ARRAY[1..maxgrp] OF spset; (* to store groups found *)
   timesseen : ARRAY[1..maxgrp] OF REAL; (* how often they're seen *)

   FUNCTION letter (ch : CHAR) : BOOLEAN;
      (* tests whether ch is a letter
         --  works for both ASCII and EBCDIC codes *)
   BEGIN (* letter *)
      letter := ((ch >= 'A') AND (ch <= 'I'))
                OR ((ch >= 'J') AND (ch <= 'R'))
                OR ((ch >= 'S') AND (ch <= 'Z'))
                OR ((ch >= 'a') AND (ch <= 'i'))
                OR ((ch >= 'j') AND (ch <= 'r'))
                OR ((ch >= 's') AND (ch <= 'z'));
   END; (* letter *)

   PROCEDURE uppercase (VAR ch : CHAR);
      (* convert a character to upper case -- either ASCII or EBCDIC *)
   BEGIN (* uppercase *)
      IF letter(ch) AND (('a' > 'A') AND (ch >= 'a')
                     OR  ('a' < 'A') AND (ch < 'A'))
      THEN ch := CHR(ORD(ch) + ORD('A') - ORD('a'));
   END; (* uppercase *)

   PROCEDURE newline (i, j, k : INTEGER);
      (* go to new line if i is a multiple of j, indent k spaces *)
      VAR m : INTEGER;
   BEGIN (* newline *)
      IF (((i-1) MOD j) = 0) AND (i>1)
      THEN BEGIN
         WRITELN(OUTPUT);
         FOR m := 1 TO k DO WRITE(OUTPUT, ' ');
         END;
   END; (* newline *)

   FUNCTION randum (VAR seed : longer) : REAL;
      (* random number generator -- slow but machine independent *)
      VAR i, j, k, sum : INTEGER;
         mult, newseed : longer;
         x : REAL;
   BEGIN (* randum *)
      mult[0] := 13;
      mult[1] := 24;
      mult[2] := 22;
      mult[3] :=  6;
      FOR i := 0 TO 5 DO
         newseed[i] := 0;
      FOR i := 0 TO 5 DO BEGIN
         sum := newseed[i];
         k := i;
         IF i > 3 THEN k := 3;
         FOR j := 0 TO k DO
            sum := sum + mult[j] * seed[i-j];
         newseed[i] := sum;
         FOR j := i TO 4 DO BEGIN
            newseed[j+1] := newseed[j+1] + newseed[j] DIV 64;
            newseed[j] := newseed[j] MOD 64;
            END;
         END;
      seed := newseed;
      seed[5] := seed[5] MOD 4;
      x := 0.0;
      FOR i := 0 TO 5 DO
         x := x / 64.0 + seed[i];
      x := x / 4.0;
      randum := x;
   END; (* randum *)

   PROCEDURE doinput;
      (* reads the input data *)

      PROCEDURE inputnumbers;
         (* input the numbers of species and of characters *)
         VAR i, inseed : INTEGER;
      BEGIN (* inputnumbers *)
         WRITELN(OUTPUT);
         WRITE(OUTPUT, ' Bootstrapped mixed parsimony algorithm');
         WRITELN(OUTPUT, ' version 3.1');
         WRITELN(OUTPUT);
         READ(infile, spp, chars, reps, inseed);
         WRITELN(OUTPUT, ' ',spp:2,' species, ',chars:3,'  characters');
         IF spp > maxsp
         THEN WRITELN(OUTPUT, ' TOO MANY SPECIES: adjust CONSTants');
         IF chars > maxchr
         THEN WRITELN(OUTPUT, ' TOO MANY CHARACTERS: adjust CONSTants');
         WRITELN(OUTPUT);
         WRITELN(OUTPUT, reps:5, ' replicates, random number seed =',
                   inseed:10);
         WRITELN(OUTPUT);
         FOR i := 0 TO 5 DO seed[i] := 0;
         i := 0;
         REPEAT
            seed[i] := inseed MOD 64;
            inseed := inseed DIV 64;
            i := i + 1;
         UNTIL inseed = 0;
         words := (chars DIV bits) + 1;
         nonodes := 2 * spp - 1;
      END; (* inputnumbers *)

      PROCEDURE inputoptions;
         (* input the information on the options *)
         VAR ch : CHAR;
            extranum, i : INTEGER;

         PROCEDURE inputmixture;
            (* input mixture of methods *)
            VAR i, j, k : INTEGER;
               ch : CHAR;
               wag : BOOLEAN;
         BEGIN (* inputmixture *)
            FOR i := 1 TO nmlngth-1 DO READ(infile, ch);
            FOR i := 1 TO words DO wagner[i] := [];
            j := 0;
            k := 1;
            FOR i := 1 TO chars DO BEGIN
               REPEAT
                  IF EOLN(infile) THEN READLN(infile);
                  READ(infile, ch);
               UNTIL ch <> ' ';
               uppercase (ch);
               wag := FALSE;
               IF (ch='W') OR (ch = '?')
               THEN wag := TRUE
               ELSE IF (ch='S') OR (ch='C')
                  THEN wag := FALSE
                  ELSE WRITELN(OUTPUT, ' error: ', ch);
               j := j + 1;
               IF j > bits
               THEN BEGIN
                  j := 1;
                  k := k + 1;
                  END;
               IF wag
               THEN wagner[k] := wagner[k] + [j];
               IF NOT ancvar
               THEN ancone[i] := wag;
               END;
            READLN(infile);
            allwagner := FALSE;
            allsokal := FALSE;
            mixture := TRUE;
         END; (* inputmixture *)

         PROCEDURE printmixture;
            (* print out list of parsimony methods *)
            VAR i, k, l : INTEGER;
         BEGIN (* printmixture *)
            allwagner := TRUE;
            WRITELN(OUTPUT, ' Parsimony methods:');
            l := 0;
            k := 1;
            FOR i := 1 TO nmlngth+4 DO WRITE(OUTPUT, ' ');
            FOR i := 1 TO chars DO BEGIN
               newline (i, 50, nmlngth+4);
               l := l + 1;
               IF l > bits
               THEN BEGIN
                  l := 1;
                  k := k + 1;
                  END;
               IF l IN wagner[k]
               THEN WRITE(OUTPUT, 'W')
               ELSE BEGIN
                  WRITE(OUTPUT, 'S');
                  allwagner := FALSE;
                  END;
               IF ((i MOD 5) = 0) THEN WRITE(OUTPUT, ' ');
               END;
            WRITELN(OUTPUT);
            WRITELN(OUTPUT);
         END; (* printmixture *)

         PROCEDURE inputhresh;
            (* input the threshold *)
         BEGIN (* inputhresh *)
            READLN(infile, threshold);
            thresh := TRUE;
         END; (* inputhresh *)

         PROCEDURE printhresh;
            (* print out the threshold value *)
         BEGIN (* printhresh *)
            WRITELN(OUTPUT, '    Threshold value = ', threshold:10:3);
            WRITELN(OUTPUT);
         END; (* printhresh *)

         PROCEDURE inputancestors;
            (* reads the ancestral states for each character *)
            VAR i : INTEGER;
               ch : CHAR;
         BEGIN (* inputancestors *)
            FOR i := 1 TO nmlngth-1 DO READ(infile, ch);
            ancvar := TRUE;
            FOR i := 1 TO chars DO BEGIN
               REPEAT
                  IF EOLN(infile) THEN READLN(infile);
                  READ(infile, ch);
               UNTIL ch <> ' ';
               ancone[i] := TRUE;
               anczero[i] := TRUE;
               IF ch = 'p' THEN ch := 'P';
               IF ch = 'b' THEN ch := 'B';
               CASE ch OF
                  '1' : anczero[i] := FALSE;
                  '0' : ancone[i] := FALSE;
                  'P' : BEGIN END;
                  'B' : BEGIN END;
                  '?' : BEGIN END
                  END;
               END;
            READLN(infile);
         END; (* inputancestors *)

         PROCEDURE printancestors;
            (* print out list of ancestral states *)
            VAR i : INTEGER;
         BEGIN (* printancestors *)
            WRITELN(OUTPUT, ' Ancestral states:');
            FOR i := 1 TO nmlngth+4 DO WRITE(OUTPUT, ' ');
            FOR i := 1 TO chars DO BEGIN
               newline (i, 50, nmlngth+4);
               IF ancone[i] AND anczero[i]
               THEN WRITE(OUTPUT, '?')
               ELSE IF ancone[i]
                  THEN WRITE(OUTPUT, '1')
                  ELSE WRITE(OUTPUT, '0');
               IF ((i MOD 5) = 0) THEN WRITE(OUTPUT, ' ');
               END;
            WRITELN(OUTPUT);
            WRITELN(OUTPUT);
         END; (* printancestor *)

         PROCEDURE inputfactors;
            VAR i, j : INTEGER;
               ch, prevch : CHAR;
         BEGIN (* inputfactors *)
            FOR i := 2 TO nmlngth DO READ(infile, ch);
            prevch := ' ';
            j := 0;
            FOR i := 1 TO chars DO BEGIN
               REPEAT
                  IF EOLN(infile) THEN READLN(infile);
                  READ(infile, ch);
               UNTIL ch <> ' ';
               IF ch <> prevch
               THEN j := j + 1;
               prevch := ch;
               factor[i] := j;
               END;
            READLN(infile);
            actualchars := j;
         END; (* inputfactors *)

         PROCEDURE printfactors;
            VAR i, j : integer;
               ch : CHAR;
         BEGIN (* printfactors *)
            WRITELN(OUTPUT, ' Factors (least significant digit)');
            WRITELN(OUTPUT, ' ');
            FOR i := 1 TO nmlngth+4 DO WRITE(OUTPUT, ' ');
            FOR i := 1 TO chars DO BEGIN
               IF ((i MOD 50) = 1) AND (i<>1)
               THEN BEGIN
                  WRITELN(OUTPUT, ' ');
                  FOR j := 1 TO nmlngth+4 DO WRITE(OUTPUT, ch);
                  END;
               ch := CHR(ORD((factor[i]) MOD 10) + ORD('0'));
               WRITE(OUTPUT, ch);
               IF (i MOD 5) = 0
               THEN WRITE(OUTPUT, ' ');
               END;
            WRITELN(OUTPUT);
            WRITELN(OUTPUT);
         END; (* printfactors *)

      BEGIN (* inputoptions *)
         ancvar := FALSE;
         allwagner := TRUE;
         allsokal := FALSE;
         factors := FALSE;
         FOR i := 1 TO chars DO factor[i] := i;
         jackknife := FALSE;
         local := FALSE;
         mixture := FALSE;
         FOR i := 1 TO chars DO BEGIN
            anczero[i] := TRUE;
            ancone[i] := TRUE;
            END;
         outgropt := FALSE;
         outgrno := 1;
         thresh := FALSE;
         trout := FALSE;
         actualchars := chars;
         extranum := 0;
         WHILE NOT EOLN(infile) DO BEGIN
            READ(infile, ch);
            uppercase (ch);
            IF (ch = 'A') OR (ch = 'C') OR (ch = 'F')
               OR (ch = 'J') OR (ch = 'L') OR (ch = 'M') 
               OR (ch = 'O') OR (ch = 'S') OR (ch = 'T') 
               OR (ch = 'Y')
            THEN CASE ch OF
                  'A' : extranum := extranum + 1;
                  'C' : allsokal := TRUE;
                  'F' : BEGIN
                        extranum := extranum + 1;
                        factors := TRUE;
                        END;
                  'J' : jackknife := TRUE;
                  'L' : local := TRUE;
                  'M' : extranum := extranum + 1;
                  'O' : BEGIN
                        outgropt := TRUE;
                        extranum := extranum + 1;
                        END;
                  'S' : allsokal := TRUE;
                  'T' : extranum := extranum + 1;
                  'Y' : trout := TRUE
                  END
            ELSE IF ch <> ' '
               THEN WRITELN(OUTPUT, ' Bad Option Character: ', ch);
            END;
         IF allsokal THEN allwagner := FALSE;
         FOR i := 1 TO words DO
            IF allsokal
            THEN wagner[i] := []
            ELSE wagner[i] := [1..bits];
         READLN(infile);
         threshold := spp;
         FOR i := 1 TO extranum DO BEGIN
            READ(infile, ch);
            uppercase (ch);
            IF ch = 'A'
            THEN inputancestors;
            IF ch = 'F'
            THEN inputfactors;
            IF ch = 'M'
            THEN inputmixture;
            IF ch = 'O'
            THEN READLN(infile, outgrno);
            IF ch = 'T'
            THEN inputhresh;
            END;
         IF jackknife
         THEN BEGIN
            WRITELN(OUTPUT, ' Jackknife by sampling ',
                                'half of the characters');
            WRITELN(OUTPUT);
            END;
         IF allsokal
         THEN BEGIN
            IF NOT ancvar
            THEN FOR i := 1 TO chars DO ancone[i] := FALSE;
            WRITELN(OUTPUT, ' Camin-Sokal parsimony method');
            END;
         IF allwagner
         THEN WRITELN(OUTPUT, ' Wagner parsimony method');
         WRITELN(OUTPUT);
         IF local
         THEN BEGIN 
            WRITELN(OUTPUT, ' Local rearrangements only');
            WRITELN(OUTPUT);
            END;
         IF mixture
         THEN printmixture;
         IF thresh
         THEN printhresh;
         IF ancvar
         THEN printancestors;
         IF factors
         THEN printfactors;
         noroot := allwagner OR (threshold <= 2.0);
         questions := FALSE;
         FOR i := 1 TO chars DO BEGIN
            noroot := noroot AND ancone[i] AND anczero[i];
            questions := questions OR (ancone[i] AND anczero[i]);
            END;
      END; (* inputoptions *)

      PROCEDURE inputdata;
         (* input the names and character state data for species *)
         VAR i, j, l : INTEGER;
            k : bitrange;
            charstate : CHAR;  (* possible states are
                                  '0', '1', 'P', 'B', and '?' *)
      BEGIN (* inputdata *)
         WRITELN(OUTPUT);
         j := nmlngth - 5 + chars DIV 2;
         IF j < nmlngth - 1 THEN j := nmlngth - 1;
         IF j > 39 THEN j := 39;
         WRITE(OUTPUT, ' Name');
         FOR i := 1 TO j DO WRITE(OUTPUT, ' ');
         WRITELN(OUTPUT, 'Characters');
         WRITE(OUTPUT, ' ----');
         FOR i := 1 TO j DO WRITE(OUTPUT, ' ');
         WRITELN(OUTPUT, '----------');
         WRITELN(OUTPUT);
         FOR i := 1 TO chars DO extras[i] := 0;
         FOR i := 1 TO nonodes DO BEGIN
            NEW (treenode[i]);
            treenode[i]^.leftdesc := NIL;
            treenode[i]^.rtdesc := NIL;
            treenode[i]^.ancestor := NIL;
            treenode[i]^.index := i;
            treenode[i]^.seen := 0.0;
            IF i <= spp
            THEN BEGIN
               treenode[i]^.tip := TRUE;
               FOR j := 1 TO nmlngth DO READ(infile, name[i,j]);
               WRITE(OUTPUT, ' ');
               FOR j := 1 TO nmlngth DO WRITE(OUTPUT, name[i,j]);
               WRITE(OUTPUT, '   ');
               FOR j := 1 TO words DO BEGIN
                  treenode[i]^.stateone[j] := [];
                  treenode[i]^.statezero[j] := [];
                  END;
               FOR j := 1 TO chars DO BEGIN
                  k := ((j-1) MOD bits) + 1;
                  l := ((j-1) DIV bits) + 1;
                  REPEAT
                     IF EOLN(infile) THEN READLN(infile);
                     READ(infile, charstate);
                  UNTIL charstate <> ' ';
                  IF charstate = 'b' THEN charstate := 'B';
                  IF charstate = 'p' THEN charstate := 'P';
                  IF NOT ((charstate = '0')
                         OR (charstate = '1')
                         OR (charstate = '?')
                         OR (charstate = 'P')
                         OR (charstate = 'B'))
                  THEN WRITELN(OUTPUT, ' Warning -- ',
                      'bad character state: ', charstate,
                      ' at character ', j:5, ' of species ', i:3);
                  newline (j, 60, nmlngth+4);
                  WRITE(OUTPUT, charstate);
                  IF (j MOD 5) = 0 THEN WRITE(OUTPUT, ' ');
                  IF (charstate = '1')
                  THEN treenode[i]^.stateone[l]
                             := treenode[i]^.stateone[l] + [k];
                  IF (charstate = '0')
                  THEN treenode[i]^.statezero[l]
                             := treenode[i]^.statezero[l] + [k];
                  IF (charstate = 'P') OR (charstate = 'B')
                  THEN extras[j] := extras[j] + 1;
                  END;
               READLN(infile);
               WRITELN(OUTPUT);
               END
            ELSE treenode[i]^.tip := FALSE;
            END;
         WRITELN(OUTPUT);
         WRITELN(OUTPUT);
      END; (* inputdata *)

   BEGIN (* doinput *)
      inputnumbers;
      inputoptions;
      inputdata;
   END; (* doinput *)

   PROCEDURE makeweights;
      (* sets up weights by resampling data *)
      VAR i, j : INTEGER;
         p, q, r : REAL;
         charweight : steparray;
   BEGIN (* makeweights *)
      FOR i := 1 TO actualchars DO
         charweight[i] := 0;
      IF jackknife
      THEN BEGIN
         p := 0.5;
         q := actualchars/2;
         r := actualchars;
         FOR i := 1 TO actualchars DO BEGIN
            IF randum(seed) < p
            THEN BEGIN
               charweight[i] := 1;
               q := q - 1;
               END;
            r := r - 1;
            IF i < actualchars THEN p := q / r;
            END;
         END
      ELSE FOR i := 1 TO actualchars DO BEGIN
         j := TRUNC(actualchars*randum(seed)) + 1;
         charweight[j] := charweight[j] + 1;
         END;
      FOR i := 1 TO chars DO BEGIN
         weight[i] := charweight[factor[i]];
         threshwt[i] := threshold * weight[i];
         END;
   END; (* makeweights *)

   PROCEDURE maketree;
      (* constructs a binary tree from the pointers in treenode.
         adds each node at location which yields highest "likelihood"
         then rearranges the tree for greatest "likelihood" *)
      TYPE placearray = ARRAY[1..maxsz] OF INTEGER;
      VAR i, j, temp, nextree : INTEGER;
         like, bestyet, bestlike, gotlike : REAL;
         lastrearr : BOOLEAN;
         there, where, item, nufork, dummynode : nodeptr;
         numsteps : steparray;
         enterorder : ARRAY[1..maxsp] OF INTEGER;
         place : placearray;
         bestrees : ARRAY[1..maxtrees] OF placearray;

      PROCEDURE add (below, newtip, newfork : nodeptr);
         (* inserts the nodes newfork and its left descendant, newtip,
            to the tree.  below becomes newfork's right descendant *)
      BEGIN (* add *)
         newfork^.ancestor := below^.ancestor;
         newfork^.rtdesc := below;
         newfork^.leftdesc := newtip;
         below^.ancestor := newfork;
         newtip^.ancestor := newfork;
         IF newfork^.ancestor <> NIL
         THEN IF newfork^.ancestor^.leftdesc = below
            THEN newfork^.ancestor^.leftdesc := newfork
            ELSE newfork^.ancestor^.rtdesc := newfork;
      END; (* add *)

      PROCEDURE remove (VAR item, fork : nodeptr);
         (* removes nodes item and its ancestor, fork, from the tree.
            the new descendant of fork's ancestor is made to be
            fork's second descendant (other than item).  Also
            returns pointers to the deleted nodes, item and fork *)
         VAR other, below : nodeptr;
      BEGIN (* remove *)
         IF item^.ancestor = NIL
         THEN fork := NIL
         ELSE BEGIN
            fork := item^.ancestor;
            IF item = fork^.leftdesc
            THEN other := fork^.rtdesc
            ELSE other := fork^.leftdesc;
            below := fork^.ancestor;
            other^.ancestor := below;
            IF below <> NIL
            THEN BEGIN
               IF below^.leftdesc = fork
               THEN below^.leftdesc := other
               ELSE below^.rtdesc := other;
               fork^.ancestor := NIL;
               END;
            fork^.leftdesc := NIL;
            fork^.rtdesc := NIL;
            item^.ancestor := NIL;
            END;
      END; (* remove *)

      PROCEDURE evaluate (r: nodeptr);
         (* Determines the number of steps needed for a tree.
            This is the minimum number needed to evolve chars on
            this tree *)
         VAR i, stepnum, smaller : INTEGER;
            sum : REAL;
            numsone, numszero : steparray;
            zeroanc : BOOLEAN;
            steps : bitarray;

         PROCEDURE count (VAR stps : bitarray);
            (* counts the number of steps in a fork of the tree.
               The program spends much of its time in this PROCEDURE *)
            VAR i, j, l : INTEGER;
         BEGIN (* count *)
            j := 1;
            l := 0;
            FOR i := 1 TO chars DO BEGIN
               l := l + 1;
               IF l > bits
               THEN BEGIN
                  l := 1;
                  j := j + 1;
                  END;
               IF l IN stps[j]
               THEN IF zeroanc
                  THEN numszero[i] := numszero[i] + weight[i]
                  ELSE numsone[i] := numsone[i] + weight[i];
               END;
         END; (* count *)

         PROCEDURE postorder (p : nodeptr);
            (* traverses a binary tree, calling PROCEDURE fillin at a
               node's descendants before calling fillin at the node *)

            PROCEDURE fillin (p: nodeptr);
               (* Sets up for each node in the tree two statesets.
                  stateone and statezero are the sets of character
                  states that must be 1 or must be 0, respectively,
                  in a most parsimonious reconstruction, based on the
                  information at or above this node.  Note that this
                  state assignment may change based on information
                  further down the tree.  If a character is in both sets
                  it is in state "P".  If in neither, it is "?". *)
               VAR i : INTEGER;
                  l0, l1, r0, r1, st, wa : bitset;
            BEGIN (* fillin *)
               FOR i := 1 TO words DO BEGIN
                  l0 := p^.leftdesc^.statezero[i];
                  l1 := p^.leftdesc^.stateone[i];
                  r0 := p^.rtdesc^.statezero[i];
                  r1 := p^.rtdesc^.stateone[i];
                  wa := wagner[i];
                  st := l1 * r0 + l0 * r1;
                  steps[i] := st;
                  p^.stateone[i] := l1 + r1 - (st * wa);
                  p^.statezero[i] := l0 + r0 - (st * wa);
                  IF zeroanc
                  THEN p^.stateone[i] := p^.stateone[i] - steps[i]
                  ELSE p^.statezero[i] := p^.statezero[i] - steps[i];
                  END;
            END; (* fillin *)

         BEGIN (* postorder *)
            IF NOT p^.tip
            THEN BEGIN
               postorder (p^.leftdesc);
               postorder (p^.rtdesc);
               fillin (p);
               count (steps);
               END;
         END; (* postorder *)

      BEGIN (* evaluate *)
         sum := 0;
         FOR i := 1 TO chars DO BEGIN
            numszero[i] := 0;
            numsone[i] := 0;
            END;
         zeroanc := TRUE;
         postorder (r);
         count(r^.stateone);
         zeroanc := FALSE;
         postorder (r);
         count(r^.statezero);
         FOR i := 1 TO chars DO BEGIN
            smaller := spp * weight[i];
            IF anczero[i]
            THEN BEGIN
               numsteps[i] := numszero[i];
               smaller := numszero[i];
               END;
            IF ancone[i] AND (numsone[i] < smaller)
            THEN numsteps[i] := numsone[i];
            stepnum := numsteps[i] + extras[i];
            IF stepnum <= threshwt[i]
            THEN sum := sum + stepnum
            ELSE sum := sum + threshwt[i];
            END;
         like := -sum;
      END; (* evaluate *)

      PROCEDURE reroot (outgroup : nodeptr);
         (* reorients tree, putting outgroup in desired position.
            Written by Kent Fiala *)
         VAR vertex, oldvertex, ante, otherbranch : nodeptr;
            left : BOOLEAN;
      BEGIN (* reroot *)
         IF NOT (outgroup^.ancestor = root)
         THEN BEGIN
            oldvertex := outgroup;
            vertex := outgroup^.ancestor;
            ante := vertex^.ancestor;
            IF vertex^.rtdesc = outgroup
            THEN BEGIN
               vertex^.rtdesc := vertex^.leftdesc;
               vertex^.leftdesc := outgroup;
               END;
            REPEAT
               IF oldvertex = vertex^.leftdesc
               THEN BEGIN
                  vertex^.leftdesc := ante;
                  left := TRUE;
                  END
               ELSE BEGIN
                  vertex^.rtdesc := ante;
                  left := FALSE;
                  END;
               vertex^.ancestor := oldvertex;
               oldvertex := vertex;
               vertex := ante;
               ante := vertex^.ancestor
            UNTIL vertex = root;
            IF root^.leftdesc = oldvertex
            THEN otherbranch := root^.rtdesc
            ELSE otherbranch := root^.leftdesc;
            IF left
            THEN oldvertex^.leftdesc := otherbranch
            ELSE oldvertex^.rtdesc := otherbranch;
            otherbranch^.ancestor := oldvertex;
            root^.rtdesc := outgroup^.ancestor;
            root^.leftdesc := outgroup;
            outgroup^.ancestor^.ancestor := root;
            outgroup^.ancestor := root;
            END;
      END; (* reroot *)

      PROCEDURE addpreorder (p, item, nufork : nodeptr);
         (* traverses a binary tree, calling PROCEDURE tryadd
            at a node before calling tryadd at its descendants *)
         TYPE knowed = RECORD    (* temporary storage for tree shape *)
            ancestor, leftdesc, rtdesc : nodeptr;
            END;
         VAR treeknowed : ARRAY [1..maxsz] OF knowed;

         PROCEDURE tryadd (p: nodeptr);
            (* temporarily adds one fork and one tip to the tree.
               if the location where they are added yields greater
               "likelihood" than other locations tested up to that
               time, then keeps that location as there *)
            VAR i, pos : INTEGER;
               found : BOOLEAN;
               rute : nodeptr;

            PROCEDURE savetree;
               (* record in place where each species has to be
                  added to reconstruct this tree *)
               VAR i, j : INTEGER;
                  p : nodeptr;
                  done : BOOLEAN;
            BEGIN (* savetree *)
               IF noroot
               THEN reroot (treenode[outgrno]);
               FOR i := 1 TO nonodes DO place[i] := 0;
               place[root^.index] := 1;
               FOR i := 1 TO spp DO BEGIN
                  p := treenode[i];
                  WHILE (place[p^.index] = 0) DO BEGIN
                     place[p^.index] := i;
                     p := p^.ancestor;
                     END;
                  IF i > 1
                  THEN BEGIN
                     place[i] := place[p^.index];
                     j := place[p^.index];
                     done := FALSE;
                     WHILE NOT done DO BEGIN
                        place[p^.index] := spp + i - 1;
                        p := p^.ancestor;
                        done := (p = NIL);
                        IF NOT done
                        THEN done := (place[p^.index] <> j);
                        END;
                     END;
                  END;
            END; (* savetree *)

            PROCEDURE findtree;
               (* finds tree given by ARRAY place in ARRAY
                  bestrees by binary search *)
               VAR i, lower, upper : INTEGER;
                  below : BOOLEAN;
            BEGIN (* findtree *)
               below := FALSE;
               lower := 1;
               upper := nextree - 1;
               found := FALSE;
               WHILE (NOT found) AND (lower <= upper) DO BEGIN
                  pos := (lower + upper) DIV 2;
                  i := 3;
                  WHILE (place[i] = bestrees[pos][i]) AND (i <= spp) DO
                     i := i + 1;
                  found := (i > spp);
                  below := place[i] < bestrees[pos][i];
                  IF NOT found
                  THEN BEGIN
                     IF below
                     THEN upper := pos - 1
                     ELSE lower := pos + 1;
                     END;
                  END;
               IF (NOT found) AND (NOT below)
               THEN pos := pos + 1;
            END; (* findtree *)

            PROCEDURE addtree;
               (* puts tree from ARRAY place in its proper position
                  in ARRAY bestrees *)
               VAR i : INTEGER;
            BEGIN (* addtree *)
               FOR i := (nextree - 1) DOWNTO pos DO
                  bestrees[i+1] := bestrees[i];
               bestrees[pos] := place;
               nextree := nextree + 1;
            END; (* addtree *)

         BEGIN (* tryadd *)
            add (p, item, nufork);
            IF root^.ancestor <> NIL THEN root := root^.ancestor;
            evaluate (root);
            IF lastrearr
            THEN BEGIN
               IF like >= bestlike
               THEN BEGIN
                  FOR i := 1 TO nonodes DO BEGIN
                     treeknowed[i].ancestor := treenode[i]^.ancestor;
                     treeknowed[i].leftdesc := treenode[i]^.leftdesc;;
                     treeknowed[i].rtdesc := treenode[i]^.rtdesc;
                     END;
                  rute := root;
                  savetree;
                  FOR i := 1 TO nonodes DO BEGIN
                     treenode[i]^.ancestor := treeknowed[i].ancestor;
                     treenode[i]^.leftdesc := treeknowed[i].leftdesc;;
                     treenode[i]^.rtdesc := treeknowed[i].rtdesc;
                     END;
                  root := rute;
                  IF like > bestlike
                  THEN BEGIN
                     bestlike := like;
                     pos := 1;
                     nextree := 1;
                     addtree;
                     END
                  ELSE BEGIN
                     pos := 0;
                     findtree;
                     IF NOT found
                     THEN IF nextree <= maxtrees
                        THEN addtree;
                     END;
                  END;
               END;
            IF like > bestyet
            THEN BEGIN
               bestyet := like;
               there := p;
               END;
            remove (item, nufork);
            IF nufork = root THEN root := p;
         END; (* tryadd *)

      BEGIN (* addpreorder *)
         IF p <> NIL
         THEN BEGIN
            tryadd (p);
            IF NOT (local AND lastrearr)
            THEN BEGIN
               addpreorder (p^.leftdesc, item, nufork);
               addpreorder (p^.rtdesc, item, nufork);
               END;
            END;
      END; (* addpreorder *)

      PROCEDURE rearrange (VAR r: nodeptr);
         (* traverses the tree (preorder), finding any local
            rearrangement which decreases the number of steps.
            if traversal succeeds in increasing the tree's
            "likelihood", PROCEDURE rearrange runs traversal again *)
         VAR success : BOOLEAN;

         PROCEDURE tryrearr (p : nodeptr);
            (* evaluates one rearrangement of the tree.
               if the new tree has greater "likelihood" than the old
               one sets success := TRUE and keeps the new tree.
               otherwise, restores the old tree *)
            VAR frombelow, whereto, forknode : nodeptr;
               oldlike : REAL;
         BEGIN (* tryrearr *)
            IF p^.ancestor <> NIL
            THEN BEGIN
               IF p^.ancestor^.ancestor <> NIL
               THEN BEGIN
                  oldlike := bestyet;
                  IF p^.ancestor^.leftdesc = p
                  THEN frombelow := p^.ancestor^.rtdesc
                  ELSE frombelow := p^.ancestor^.leftdesc;
                  whereto := p^.ancestor^.ancestor;
                  forknode := p^.ancestor;
                  remove (p, forknode);
                  add (whereto, p, forknode);
                  IF r^.ancestor <> NIL THEN r := r^.ancestor;
                  evaluate (r);
                  IF like <= oldlike
                  THEN BEGIN
                     remove (p, forknode);
                     add (frombelow, p, forknode);
                     IF r^.ancestor <> NIL THEN r := r^.ancestor;
                     END
                  ELSE BEGIN
                     success := TRUE;
                     bestyet := like;
                     END;
                  END;
               END;
         END; (* tryrearr *)

         PROCEDURE repreorder (p : nodeptr);
            (* traverses a binary tree, calling PROCEDURE tryrearr
               at a node before calling tryrearr at its descendants *)
         BEGIN (* repreorder *)
            IF p <> NIL
            THEN BEGIN
               tryrearr (p);
               repreorder (p^.leftdesc);
               repreorder (p^.rtdesc);
               END;
         END; (* repreorder *)

      BEGIN (* rearrange *)
         success := TRUE;
         WHILE success DO BEGIN
            success := FALSE;
            repreorder (r);
            END;
      END; (* rearrange *)

      PROCEDURE accumulate (r : nodeptr);

         PROCEDURE enterset;
            VAR s : spset;
               i : INTEGER;
               hashcode : REAL;
               done : BOOLEAN;
         BEGIN (* enterset *)
            s := r^.nodeset;
            IF s <> [1..spp]
            THEN BEGIN
               hashcode := 0.0;
               FOR i := spp DOWNTO 1 DO BEGIN
                  hashcode := hashcode/2.0;
                  IF (i IN s)
                  THEN hashcode := hashcode + 0.5;
                  END;
               i := TRUNC(0.8*maxgrp*hashcode) + 1;
               done := FALSE;
               WHILE NOT done DO BEGIN
                  IF (timesseen[i] > 0.0) AND (s = grouping[i])
                  THEN BEGIN
                     timesseen[i] := timesseen[i] + 1.0/(nextree-1);
                     done := TRUE;
                     END
                  ELSE IF (timesseen[i] <= 0.0)
                     THEN BEGIN
                        grouping[i] := s;
                        timesseen[i] := 1.0/(nextree-1);
                        done := TRUE;
                        END
                     ELSE i := i + 1;
                  IF i > maxgrp
                  THEN WRITELN(OUTPUT, 'Error! Hash Table Overflow');
                  END;
               END;
         END; (* enterset *)

      BEGIN (* accumulate *)
         IF r^.tip
         THEN r^.nodeset := [r^.index]
         ELSE BEGIN
            accumulate(r^.leftdesc);
            accumulate(r^.rtdesc);
            r^.nodeset := r^.leftdesc^.nodeset
                            + r^.rtdesc^.nodeset;
            END;
         enterset;
      END; (* accumulate *)

   BEGIN (* maketree *)
      FOR i := 1 TO spp DO
         enterorder[i] := i;
      FOR i := 1 TO spp DO BEGIN
         j := TRUNC(randum(seed)*spp) + 1;
         temp := enterorder[i];
         enterorder[i] := enterorder[j];
         enterorder[j] := temp;
         END;
      add (treenode[enterorder[2]], treenode[enterorder[1]],
                                       treenode[spp+1]);
      root := treenode[spp+1];
      lastrearr := FALSE;
      FOR i := 3 TO spp DO BEGIN
         bestyet := -32000;
         there := root;
         item := treenode[enterorder[i]];
         nufork := treenode[spp + i - 1];
         addpreorder (root, item, nufork);
         add (there, treenode[enterorder[i]], treenode[spp+i-1]);
         IF root^.ancestor <> NIL THEN root := root^.ancestor;
         like := bestyet;
         rearrange (root);
         lastrearr := (i = spp);
         IF lastrearr
         THEN BEGIN
            bestlike := bestyet;
            nextree := 1;
            REPEAT
               gotlike := bestlike;
               FOR j := 1 TO nonodes DO BEGIN
                  item := treenode[j];
                  IF item <> root
                  THEN BEGIN
                     IF local
                     THEN BEGIN
                        IF item^.ancestor^.leftdesc <> item
                        THEN where := item^.ancestor^.leftdesc
                        ELSE where := item^.ancestor^.rtdesc;
                        there := where;
                        bestyet := bestlike;
                        END
                     ELSE BEGIN
                        there := root;
                        bestyet := -32000;
                        END;
                     nufork := treenode[j]^.ancestor;
                     IF root = nufork
                     THEN BEGIN
                        IF nufork^.leftdesc = item
                        THEN root := nufork^.rtdesc
                        ELSE root := nufork^.leftdesc;
                        END;
                     remove (item, nufork);
                     IF local
                     THEN BEGIN
                        bestlike := gotlike;
                        IF j = 1
                        THEN addpreorder (where, item, nufork);
                        addpreorder (where^.leftdesc, item, nufork);
                        addpreorder (where^.rtdesc, item, nufork);
                        END
                     ELSE addpreorder (root, item, nufork);
                     add (there, item, nufork);
                     IF root^.ancestor <> NIL
                     THEN root := root^.ancestor;
                     END;
                  END;
            UNTIL bestlike <= gotlike;
            END;
         END;
      FOR i := spp DOWNTO 2 DO
         remove (treenode[i], dummynode);
      IF nextree > maxtrees + 1
      THEN nextree := maxtrees + 1;
      FOR i := 1 TO nextree-1 DO BEGIN
         add (treenode[2], treenode[1], treenode[spp+1]);
         root := treenode[spp+1];
         FOR j := 3 TO spp DO BEGIN
            add(treenode[bestrees[i][j]], treenode[j],
                                           treenode[spp+j-1]);
            IF root^.ancestor <> NIL THEN root := root^.ancestor;
            END;
         didreroot := outgropt AND noroot;
         IF noroot
         THEN reroot (treenode[outgrno]);
         accumulate(root);
         FOR j := 2 TO spp DO
            remove(treenode[j], dummynode);
         END;
   END; (* maketree *)

   PROCEDURE consense;
      VAR n : INTEGER;

      PROCEDURE compress;
         (* push all the nonempty subsets to the front end
            of their array *)
         VAR i, j : INTEGER;
      BEGIN (* compress *)
         i := 1;
         j := 1;
         REPEAT
            WHILE grouping[i] <> [] DO i := i + 1;
            IF j <= i
            THEN j := i + 1;
            WHILE ((grouping[j] = []) AND (j < maxgrp)) DO
               j := j + 1;
            IF (j < maxgrp)
            THEN BEGIN
               grouping[i] := grouping[j];
               timesseen[i] := timesseen[j];
               grouping[j] := [];
               timesseen[j] := 0;
               END
         UNTIL j = maxgrp;
         n := i - 1;
      END; (* compress *)

      PROCEDURE sort;
         (* Shell sort keeping grouping, timesseen in same order *)
         VAR gap, i, j : INTEGER;
            temp : REAL;
            stemp : spset;
      BEGIN (* sort *)
         gap := n DIV 2;
         WHILE (gap > 0) DO BEGIN
            FOR i := gap TO n-1 DO BEGIN
               j := i - gap + 1;
               WHILE (j>0) DO BEGIN
                  IF timesseen[j] < timesseen[j+gap]
                  THEN BEGIN
                     stemp := grouping[j];
                     grouping[j] := grouping[j+gap];
                     grouping[j+gap] := stemp;
                     temp := timesseen[j];
                     timesseen[j] := timesseen[j+gap];
                     timesseen[j+gap] := temp;
                     END;
                  j := j - gap;
                  END;
               END;
            gap := gap DIV 2;
            END;
      END; (* sort *)

      PROCEDURE eliminate;
         (* eliminate groups incompatible with preceding ones *)
         VAR i, j : INTEGER;
      BEGIN (* eliminate *)
         FOR i := 2 TO n DO BEGIN
            FOR j := 1 TO i-1 DO
               IF timesseen[j] > 0
               THEN IF ((grouping[i] * grouping[j]) <> [])
                  AND ((grouping[i] - grouping[j]) <> [])
                  AND ((grouping[j] - grouping[i]) <> [])
                  THEN BEGIN
                     timesseen[i] := 0;
                     grouping[i] := [];
                     END;
            END;
      END; (* eliminate *)

      PROCEDURE reconstruct;
         (* reconstruct tree from the subsets *)
         VAR i, j, nextnode : INTEGER;
            s : spset;

         PROCEDURE bigsubset (VAR st : spset);
            (* find a maximal subset of st among the groupings *)
            VAR i : INTEGER;
               su : spset;
         BEGIN (* bigsubset *)
            su := [];
            FOR i := 1 TO n DO
              IF (st >= grouping[i]) AND (grouping[i] <> st)
              THEN IF (grouping[i] >= su) AND (grouping[i] <> su)
                 THEN su := grouping[i];
            st := su;
         END; (* bigsubset *)

         PROCEDURE recontraverse (VAR p : nodeptr; st : spset);
            (* traverse to add next node of bootstrap estimate tree *)
            VAR i, j, k : INTEGER;
               tempset : spset;
         BEGIN (* recontraverse *)
            k := 0;
            FOR i := 1 TO spp DO
               IF i IN st
               THEN BEGIN
                  k := k + 1;
                  j := i;
                  END;
            IF k = 1
            THEN p := treenode[j]
            ELSE BEGIN
               p := treenode[nextnode];
               nextnode := nextnode + 1;
               p^.seen := 0.0;
               FOR i := 1 TO n DO
                  IF grouping[i] = st
                  THEN p^.seen := timesseen[i];
               tempset := st;
               bigsubset (tempset);
               recontraverse (p^.leftdesc, tempset);
               tempset := st - tempset;
               recontraverse (p^.rtdesc, tempset);
               END;
         END; (* recontraverse *)

      BEGIN (* reconstruct *)
         WRITELN(OUTPUT);
         WRITE(OUTPUT, ' Set (species in order)   ');
         FOR i := 1 TO spp-25 DO WRITE(OUTPUT, ' ');
         WRITELN(OUTPUT, '  How many times out of', reps:4);
         WRITELN(OUTPUT);
         FOR i := 1 TO n DO BEGIN
            IF timesseen[i] > 0.0
            THEN BEGIN
               WRITE(OUTPUT, ' ');
               FOR j := 1 TO spp DO
                  IF j IN grouping[i]
                  THEN WRITE(OUTPUT, '*')
                  ELSE WRITE(OUTPUT, '.');
               FOR j := 1 TO 23-spp DO WRITE(OUTPUT, ' ');
               WRITELN(OUTPUT, '    ', timesseen[i]:10:2);
               END;
            END;
         nextnode := spp + 1;
         s := [1..spp];
         recontraverse (root, s);
      END; (* reconstruct *)

      PROCEDURE printree;
         (* prints out diagram of the tree *)
         CONST down = 2;
            over = 3;
         VAR tipx, tipy, i, rover : INTEGER;

         PROCEDURE coordinates (p : nodeptr);
            (* establishes coordinates of nodes *)
         BEGIN (* coordinates *)
            IF p^.tip
            THEN BEGIN
               p^.xcoord := tipx;
               p^.ycoord := tipy;
               tipx := tipx + rover;
               tipy := tipy + down;
               END
            ELSE BEGIN
               coordinates (p^.leftdesc);
               p^.xcoord := p^.leftdesc^.xcoord;
               coordinates (p^.rtdesc);
               p^.ycoord := p^.rtdesc^.ycoord;
               END;
         END; (* coordinates *)

         PROCEDURE drawline (i : INTEGER);
            (* draws one row of the tree diagram by moving up tree *)
            VAR p, q : nodeptr;
               n, j, times : INTEGER;
         BEGIN (* drawline *)
            p := root;
            WRITE(OUTPUT, '   ');
            REPEAT
               IF p^.leftdesc^.ycoord >= i
               THEN p := p^.leftdesc
               ELSE BEGIN
                  q := p^.rtdesc;
                  n := q^.xcoord - p^.xcoord;
                  IF p^.ycoord = i
                  THEN BEGIN
                     IF ((i = (tipy-down)) AND (p^.xcoord = 0))
                     THEN BEGIN
                        IF noroot
                        THEN WRITE(OUTPUT, 'L')
                        ELSE WRITE(OUTPUT, '*');
                        n := n + 1;
                        END
                     ELSE BEGIN
                        times := TRUNC(p^.seen+0.5);
                        IF times >= 100
                        THEN BEGIN
                           WRITE(OUTPUT, times:3);
                           n := n - 1;
                           END
                        ELSE IF times >= 10
                           THEN WRITE(OUTPUT, times:2)
                           ELSE WRITE(OUTPUT, times:1, '-');
                        END;
                     FOR j := 1 TO n-2 DO WRITE(OUTPUT, '-');
                     END
                  ELSE BEGIN
                     WRITE(OUTPUT, '!');
                     FOR j := 1 TO n-1 DO WRITE(OUTPUT, ' ');
                     END;
                  p := q;
                  END;
            UNTIL p^.tip;
            IF p^.ycoord = i
            THEN FOR j := 1 TO nmlngth DO
                       WRITE(OUTPUT, name[p^.index, j]);
            WRITELN(OUTPUT);
         END; (* drawline *)

      BEGIN (* printree *)
         WRITELN(OUTPUT);
         WRITELN(OUTPUT, ' The numbers at the forks indicate the number');
         WRITELN(OUTPUT, ' of times the group consisting of the species');
         WRITELN(OUTPUT, ' above and to the right of that fork occurred');
         WRITELN(OUTPUT, ' among the ', reps:4, ' bootstrap replicates');
         tipx := 0;
         tipy := 1;
         rover := over;
         IF reps >= 100 THEN rover := over + 1;
         coordinates (root);
         WRITELN(OUTPUT);
         WRITELN(OUTPUT);
         WRITELN(OUTPUT);
         FOR i := 1 TO tipy-down DO drawline (i);
         IF noroot
         THEN BEGIN
            WRITELN(OUTPUT);
            WRITE(OUTPUT, '   remember:');
            IF didreroot
            THEN WRITE(OUTPUT, ' (although rooted by outgroup)');
            WRITELN(OUTPUT, ' this is an unrooted tree!');
            END
         ELSE BEGIN
            WRITELN(OUTPUT, '  /');
            WRITELN(OUTPUT, ' /');
            END;
         WRITELN(OUTPUT);
      END; (* printree *)

      PROCEDURE treeout (p : nodeptr);
         (* write out file with representation of final tree *)
         VAR i, n : INTEGER;
            c : CHAR;
      BEGIN (* treeout *)
         IF p^.tip
         THEN BEGIN
            n := 0;
            FOR i := 1 TO nmlngth DO
               IF name[p^.index, i] <> ' '
               THEN n := i;
            FOR i := 1 TO n DO BEGIN
               c := name[p^.index, i];
               IF c = ' '
               THEN c := '_';
               WRITE(treefile, c);
               END;
            END
         ELSE BEGIN
            WRITE(treefile, '(');
            treeout (p^.leftdesc);
            WRITE(treefile, ',');
            treeout (p^.rtdesc);
            WRITE(treefile, ')');
            END;
         n := TRUNC(p^.seen+0.5);
         IF p^.tip
         THEN n := reps;
         IF p = root
         THEN WRITELN(treefile, ';')
         ELSE IF n > 99
            THEN WRITE(treefile, ':', n:3)
            ELSE IF n > 9
               THEN WRITE(treefile, ':', n:2)
               ELSE WRITE(treefile, ':', n:1);
      END; (* treeout *)

   BEGIN (* consense *)
      compress;
      sort;
      eliminate;
      compress;
      reconstruct;
      printree;
      IF trout
      THEN BEGIN
         treeout(root);
         WRITELN(OUTPUT, ' Tree also written onto file');
         WRITELN(OUTPUT);
         END;
   END; (* consense *)

BEGIN (* Mixed parsimony bootstrapped *)
   (* reads in spp, chars, and the data; then
      cycles among makeweights, maketree *)
   (* ASSIGN(infile, 'INFILE');  Turbo Pascal only *)
   RESET(infile);
   doinput;
   IF trout
   THEN BEGIN
      (* ASSIGN(treefile, 'TREEFILE');  Turbo Pascal only *)
      REWRITE(treefile);
      END;
   FOR i := 1 TO maxgrp DO BEGIN
      timesseen [i] := 0.0;
      grouping [i] := [];
      END;
   repdiv10 := reps DIV 10;
   IF repdiv10 < 1
   THEN repdiv10 := 1;
   FOR rr := 1 To reps DO BEGIN
      makeweights;
      maketree;
      IF (rr MOD repdiv10) = 0
      THEN WRITELN(OUTPUT, ' completed replicate number ', rr:4);
      END;
   consense;
(* CLOSE(infile);  -- may need this here *)
   IF trout
   THEN BEGIN
      (* CLOSE(treefile);  -- may need this here *)
      END;
END. (* Mixed parsimony bootstrapped *)

