PROGRAM Consense (infile, OUTPUT, treefile);
  (* version 3.1. (c) Copyright 1986, 1987, 1988 by the University of
     Washington.  Written by Joseph Felsenstein.  Permission is granted
     to copy and use this program provided no fee is charged for it and
     provided that this copyright notice is not removed. *)

CONST maxsp = 30;    (* maximum number of species               *)
   maxsz =  60;      (* size of pointer array.  >= 2*maxsp - 1
                   and must be <= maximum size of a Pascal set  *)
   nmlngth = 10;     (* number of characters in species name    *)
   maxgrp = 1000;    (* max. no. of groups in all trees found   *)

TYPE spset = SET OF 1..maxsp;
   nodeptr = ^node;   (* nodes will form a binary tree *)
   node = RECORD      (* describes a tip species or an ancestor *)
      ancestor, leftdesc, rtdesc : nodeptr;  (* pointers to nodes *)
      index : INTEGER;      (* number of the node *)
      tip : BOOLEAN;        (* present species are tips of tree *)
      nodeset : spset;      (* used by accumulate *)
      xcoord, ycoord : INTEGER;   (* used by printree *)
      forreal : BOOLEAN     (* used by treeread *)
      END;
   pointarray = ARRAY[1..maxsz] OF nodeptr;

VAR root : nodeptr;
   infile, treefile : TEXT;
   spp, numopts, outgrno, i, ntrees : INTEGER;
     (* outgrno indicates outgroup *)
   trout, first, noroot, outgropt, didreroot : BOOLEAN;
   ch : CHAR;
   treenode : pointarray;      (* pointers to all nodes in tree *)
   name : ARRAY[1..maxsp, 1..nmlngth] OF CHAR; (* names of species *)
   grouping, group2 : ARRAY[1..maxgrp] OF spset;
                      (* to store groups found *)
   timesseen, times2 : ARRAY[1..maxgrp] OF INTEGER;
                       (* how often they're seen *)

   PROCEDURE setuptree;
      (* initialization of tree pointers, variables *)
      VAR i : INTEGER;
   BEGIN (* setuptree *)
      FOR i := 1 TO maxsz DO BEGIN
         NEW (treenode[i]);
         treenode[i]^.leftdesc := NIL;
         treenode[i]^.rtdesc := NIL;
         treenode[i]^.ancestor := NIL;
         treenode[i]^.index := i;
         treenode[i]^.tip := FALSE;
         END;
   END; (* setuptree *)

   PROCEDURE treeread;
      (* read in user-defined tree and set it up *)
      VAR ch : CHAR;
         nextnode : INTEGER;

      PROCEDURE findch (c : CHAR);
         (* scan forward until find character c *)
      BEGIN (* findch *)
         IF ch <> c
         THEN REPEAT
               READ(infile, ch);
            UNTIL ch = c;
      END; (* findch *)

      PROCEDURE addelement (VAR p : nodeptr);
         (* recursive procedure adds nodes to user-defined tree *)
         VAR q, r : nodeptr;
            i, n : INTEGER;
            found : BOOLEAN;
            str : ARRAY[1..nmlngth] OF CHAR;
      BEGIN (* addelement *)
         REPEAT
            IF EOLN(infile) THEN READLN(infile);
            READ(infile, ch);
         UNTIL ch <> ' ';
         IF ch = '('
         THEN BEGIN
            nextnode := nextnode + 1;
            q := treenode[nextnode];
            q^.forreal := TRUE;
            addelement (q^.leftdesc);
            q^.leftdesc^.ancestor := q;
            findch (',');
            addelement (q^.rtdesc);
            q^.rtdesc^.ancestor := q;
            REPEAT
               IF (ch <> ',') AND (ch <> ')')
               THEN REPEAT
                     READ(infile, ch);
                  UNTIL ((ch = ',') OR (ch = ')'));
               IF ch = ','
               THEN BEGIN
                  nextnode := nextnode + 1;
                  r := q;
                  q := treenode[nextnode];
                  q^.leftdesc := r;
                  r^.ancestor := q;
                  addelement (q^.rtdesc);
                  q^.rtdesc^.ancestor := q;
                  r^.forreal := FALSE;
                  END;
               p := q;
            UNTIL ch = ')';
            END
         ELSE BEGIN
            FOR i := 1 TO nmlngth DO str[i] := ' ';
            n := 1;
            REPEAT
               IF (ch = '_') THEN ch := ' ';
               str[n] := ch;
               IF EOLN(infile) THEN READLN(infile);
               READ(infile, ch);
               n := n + 1;
            UNTIL ((ch = ':') OR (ch = ',') OR (ch = ')')
                     OR (n > nmlngth));
            n := n - 1;
            IF first
            THEN BEGIN
               spp := spp + 1;
               p := treenode[spp];
               p^.tip := TRUE;
               p^.forreal := TRUE;
               FOR i := 1 TO nmlngth DO
                  name[spp, i] := ' ';
               FOR i := 1 TO n DO
                  name[spp, i] := str[i];
               WRITE(OUTPUT, '   ');
               FOR i := 1 TO n DO WRITE(OUTPUT, str[i]);
               WRITELN(OUTPUT);
               END
            ELSE BEGIN
               n := 1;
               REPEAT
                  found := TRUE;
                  FOR i := 1 TO nmlngth DO
                     found := found AND (str[i] = name[n, i]);
                  IF found
                  THEN p := treenode[n]
                  ELSE n := n + 1;
               UNTIL (n > spp) OR found;
               IF n > spp
               THEN BEGIN
                  WRITE(OUTPUT, ' Cannot find species: ');
                  FOR i := 1 TO nmlngth DO
                     WRITE(OUTPUT, str[i]);
                  WRITELN(OUTPUT);
                  END;
               END;
         END;
      END; (* addelement *)

   BEGIN (* treeread *)
      root := treenode[maxsp + 1];
      nextnode := maxsp + 1;
      root^.ancestor := NIL;
      addelement (root);
      READLN(infile);
   END; (* treeread *)

   PROCEDURE reroot (outgroup : nodeptr);
      (* reorients tree, putting outgroup in desired position.
         Written by Kent Fiala *)
      VAR vertex, oldvertex, ante, otherbranch : nodeptr;
         left : BOOLEAN;
   BEGIN (* reroot *)
      IF NOT (outgroup^.ancestor = root)
      THEN BEGIN
         oldvertex := outgroup;
         vertex := outgroup^.ancestor;
         ante := vertex^.ancestor;
         IF vertex^.rtdesc = outgroup
         THEN BEGIN
            vertex^.rtdesc := vertex^.leftdesc;
            vertex^.leftdesc := outgroup;
            END;
         REPEAT
            IF oldvertex = vertex^.leftdesc
            THEN BEGIN
               vertex^.leftdesc := ante;
               left := TRUE;
               END
            ELSE BEGIN
               vertex^.rtdesc := ante;
               left := FALSE;
               END;
            vertex^.ancestor := oldvertex;
            oldvertex := vertex;
            vertex := ante;
            ante := vertex^.ancestor
         UNTIL vertex = root;
         IF root^.leftdesc = oldvertex
         THEN otherbranch := root^.rtdesc
         ELSE otherbranch := root^.leftdesc;
         IF left
         THEN oldvertex^.leftdesc := otherbranch
         ELSE oldvertex^.rtdesc := otherbranch;
         otherbranch^.ancestor := oldvertex;
         root^.rtdesc := outgroup^.ancestor;
         root^.leftdesc := outgroup;
         outgroup^.ancestor^.ancestor := root;
         outgroup^.ancestor := root;
         END;
   END; (* reroot *)

   PROCEDURE accumulate (r : nodeptr);

      PROCEDURE enterset;
         VAR s : spset;
            i, times : INTEGER;
            hashcode : REAL;
            done : BOOLEAN;
      BEGIN (* enterset *)
         s := r^.nodeset;
         IF (s <> [1..spp])
         THEN BEGIN
            IF r^.forreal
            THEN times := 1
            ELSE times := 0;
            hashcode := 0.0;
            FOR i := spp DOWNTO 1 DO BEGIN
               hashcode := hashcode/2.0;
               IF (i IN s)
               THEN hashcode := hashcode + 0.5;
               END;
            i := TRUNC(0.8*maxgrp*hashcode) + 1;
            done := FALSE;
            WHILE NOT done DO BEGIN
               IF (s = grouping[i])
               THEN BEGIN
                  timesseen[i] := timesseen[i] + times;
                  done := TRUE;
                  END
               ELSE IF (timesseen[i] = 0) AND (grouping[i] = [])
                  THEN BEGIN
                     grouping[i] := s;
                     timesseen[i] := times;
                     done := TRUE;
                     END
               ELSE i := i + 1;
               IF i > maxgrp
               THEN BEGIN
                  WRITELN(OUTPUT, 'Error! Hash Table Overflow');
                  done := TRUE;
                  END;
               END;
            END;
      END; (* enterset *)

   BEGIN (* accumulate *)
      IF r^.tip
      THEN r^.nodeset := [r^.index]
      ELSE BEGIN
         accumulate(r^.leftdesc);
         accumulate(r^.rtdesc);
         r^.nodeset := r^.leftdesc^.nodeset + r^.rtdesc^.nodeset;
         END;
      enterset;
   END; (* accumulate *)

   PROCEDURE consensus;
      VAR i, n, n2 : INTEGER;
         fullset : spset;

      PROCEDURE compress;
         (* push all the nonempty subsets to the front end
            of their array *)
         VAR i, j : INTEGER;
      BEGIN (* compress *)
         i := 1;
         j := 1;
         REPEAT
            WHILE grouping[i] <> [] DO i := i + 1;
            IF j <= i
            THEN j := i + 1;
            WHILE ((grouping[j] = []) AND (j < maxgrp)) DO
               j := j + 1;
            IF (j < maxgrp)
            THEN BEGIN
               grouping[i] := grouping[j];
               timesseen[i] := timesseen[j];
               grouping[j] := [];
               timesseen[j] := 0;
               END
         UNTIL j = maxgrp;
         n := i - 1;
      END; (* compress *)

      PROCEDURE sort;
         (* Shell sort keeping grouping, timesseen in same order *)
         VAR gap, i, j, itemp : INTEGER;
           stemp : spset;
      BEGIN (* sort *)
         gap := n DIV 2;
         WHILE (gap > 0) DO BEGIN
            FOR i := gap TO n-1 DO BEGIN
               j := i - gap + 1;
               WHILE (j>0) DO BEGIN
                  IF timesseen[j] < timesseen[j+gap]
                  THEN BEGIN
                     stemp := grouping[j];
                     grouping[j] := grouping[j+gap];
                     grouping[j+gap] := stemp;
                     itemp := timesseen[j];
                     timesseen[j] := timesseen[j+gap];
                     timesseen[j+gap] := itemp;
                     END;
                  j := j - gap;
                  END;
               END;
            gap := gap DIV 2;
            END;
      END; (* sort *)

      PROCEDURE eliminate;
         (* eliminate groups incompatible with preceding ones *)
         VAR i, j : INTEGER;
      BEGIN (* eliminate *)
         FOR i := 2 TO n DO BEGIN
            FOR j := 1 TO i-1 DO
               IF timesseen[j] > 0
               THEN IF ((grouping[i] * grouping[j]) <> [])
                  AND ((grouping[i] - grouping[j]) <> [])
                  AND ((grouping[j] - grouping[i]) <> [])
                  AND ((fullset-grouping[i])
                     * (fullset-grouping[j]) <> [])
                  THEN BEGIN
                     n2 := n2 + 1;
                     times2[n2] := timesseen[i];
                     group2[n2] := grouping[i];
                     timesseen[i] := 0;
                     grouping[i] := [];
                     END;
            END;
      END; (* eliminate *)

      PROCEDURE printset;
         VAR i, j, size : INTEGER;
      BEGIN (* printset *)
         WRITELN(OUTPUT);
         WRITE(OUTPUT, ' Set (species in order)   ');
         FOR i := 1 TO spp-25 DO WRITE(OUTPUT, ' ');
         WRITELN(OUTPUT, '  How many times out of', ntrees:4);
         WRITELN(OUTPUT);
         FOR i := 1 TO n DO BEGIN
            IF (timesseen[i] > 0)
            THEN BEGIN
               size := 0;
               FOR j := 1 TO spp DO
                  IF j IN grouping[i]
                  THEN size := size + 1;
               IF (size <> 1) AND NOT (noroot AND (size = (spp - 1)))
               THEN BEGIN
                  WRITE(OUTPUT, ' ');
                  FOR j := 1 TO spp DO BEGIN
                     IF j IN grouping[i]
                     THEN WRITE(OUTPUT, '*')
                     ELSE WRITE(OUTPUT, '.');
                     IF (j MOD 10) = 0
                     THEN WRITE(OUTPUT, ' ');
                     END;
                  FOR j := 1 TO 23-spp DO WRITE(OUTPUT, ' ');
                  WRITELN(OUTPUT, '    ', timesseen[i]:4);
                  END;
               END;
            END;
      END; (* printset *)

      PROCEDURE reconstruct;
         (* reconstruct tree from the subsets *)
         VAR nextnode : INTEGER;
            s : spset;

         PROCEDURE bigsubset (VAR st : spset);
            (* find a maximal subset of st among the groupings *)
            VAR i : INTEGER;
               su : spset;
         BEGIN (* bigsubset *)
            su := [];
            FOR i := 1 TO n DO
              IF (st >= grouping[i]) AND (grouping[i] <> st)
              THEN IF (grouping[i] >= su) AND (grouping[i] <> su)
                 THEN su := grouping[i];
            st := su;
         END; (* bigsubset *)

         PROCEDURE recontraverse (VAR p : nodeptr; st : spset);
            (* traverse to add next node of consensus tree *)
            VAR i, j, k : INTEGER;
               tempset : spset;
         BEGIN (* recontraverse *)
            k := 0;
            FOR i := 1 TO spp DO
               IF i IN st
               THEN BEGIN
                  k := k + 1;
                  j := i;
                  END;
            IF k = 1
            THEN p := treenode[j]
            ELSE BEGIN
               p := treenode[nextnode];
               nextnode := nextnode + 1;
               p^.index := 0;
               FOR i := 1 TO n DO
                  IF grouping[i] = st
                  THEN p^.index := timesseen[i];
               tempset := st;
               IF tempset <> []
               THEN bigsubset (tempset);
               recontraverse (p^.leftdesc, tempset);
               tempset := st - tempset;
               IF tempset <> []
               THEN recontraverse (p^.rtdesc, tempset);
               END;
         END; (* recontraverse *)

      BEGIN (* reconstruct *)
         nextnode := spp + 1;
         s := [1..spp];
         recontraverse (root, s);
      END; (* reconstruct *)

      PROCEDURE printree;
         (* prints out diagram of the tree *)
         CONST down = 2;
            over = 3;
         VAR tipx, tipy, i, rover : INTEGER;

         PROCEDURE coordinates (p : nodeptr);
            (* establishes coordinates of nodes *)
         BEGIN (* coordinates *)
            IF p^.tip
            THEN BEGIN
               p^.xcoord := tipx;
               p^.ycoord := tipy;
               tipx := tipx + rover;
               tipy := tipy + down;
               END
            ELSE BEGIN
               coordinates (p^.leftdesc);
               p^.xcoord := p^.leftdesc^.xcoord;
               coordinates (p^.rtdesc);
               p^.ycoord := p^.rtdesc^.ycoord;
               END;
         END; (* coordinates *)

         PROCEDURE drawline (i : INTEGER);
            (* draws one row of the tree diagram by moving up tree *)
            VAR p, q : nodeptr;
               n, j : INTEGER;
         BEGIN (* drawline *)
            p := root;
            WRITE(OUTPUT, '   ');
            REPEAT
               IF p^.leftdesc^.ycoord >= i
               THEN p := p^.leftdesc
               ELSE BEGIN
                  q := p^.rtdesc;
                  n := q^.xcoord - p^.xcoord;
                  IF p^.ycoord = i
                  THEN BEGIN
                     IF ((i = (tipy-down)) AND (p^.xcoord = 0))
                     THEN BEGIN
                        IF noroot
                        THEN WRITE(OUTPUT, 'L')
                        ELSE WRITE(OUTPUT, '*');
                        n := n + 1;
                        END
                     ELSE BEGIN
                        IF noroot AND ((p^.xcoord = 0)
                              AND (p^.ycoord = tipy - 2*down))
                           OR ((p^.xcoord = over)
                              AND (p^.ycoord = tipy - down))
                        THEN WRITE(OUTPUT, '*-')
                        ELSE BEGIN
                           IF p^.index >= 100
                           THEN BEGIN
                              WRITE(OUTPUT, p^.index:3);
                              n := n - 1;
                              END
                           ELSE IF p^.index >= 10
                              THEN WRITE(OUTPUT, p^.index:2)
                              ELSE WRITE(OUTPUT, p^.index:1, '-');
                           END;
                        END;
                     FOR j := 1 TO n-2 DO WRITE(OUTPUT, '-');
                     END
                  ELSE BEGIN
                     WRITE(OUTPUT, '!');
                     FOR j := 1 TO n-1 DO WRITE(OUTPUT, ' ');
                     END;
                  p := q;
                  END;
            UNTIL p^.tip;
            IF p^.ycoord = i
            THEN FOR j := 1 TO nmlngth DO
                       WRITE(OUTPUT, name[p^.index, j]);
            WRITELN(OUTPUT);
         END; (* drawline *)

      BEGIN (* printree *)
         WRITELN(OUTPUT);
         WRITELN(OUTPUT, ' CONSENSUS TREE:');
         WRITELN(OUTPUT, ' The numbers at the forks indicate the number');
         WRITELN(OUTPUT, ' of times the group consisting of the species');
         WRITELN(OUTPUT, ' above and to the right of that fork occurred');
         WRITELN(OUTPUT, ' among the trees, out of ', ntrees:3, ' trees');
         tipx := 0;
         tipy := 1;
         rover := over;
         IF ntrees >= 100 THEN rover := over + 1;
         coordinates (root);
         WRITELN(OUTPUT);
         FOR i := 1 TO tipy-down DO drawline (i);
         IF noroot
         THEN BEGIN
            WRITELN(OUTPUT);
            WRITE(OUTPUT, '   remember:');
            IF didreroot
            THEN WRITE(OUTPUT, ' (though rerooted by outgroup)');
            WRITELN(OUTPUT, ' this is an unrooted tree!');
            END
         ELSE BEGIN
            WRITELN(OUTPUT, '  /');
            WRITELN(OUTPUT, ' /');
            END;
         WRITELN(OUTPUT);
      END; (* printree *)

      PROCEDURE treeout (p : nodeptr);
         (* write out file with representation of final tree *)
         VAR i, n : INTEGER;
            c : CHAR;
      BEGIN (* treeout *)
         IF p^.tip
         THEN BEGIN
            FOR i := 1 TO nmlngth DO
               IF name[p^.index, i] <> ' '
               THEN n := i;
            FOR i := 1 TO n DO BEGIN
               c := name[p^.index, i];
               IF c = ' '
               THEN c := '_';
               WRITE(treefile, c);
               END;
            END
         ELSE BEGIN
            WRITE(treefile, '(');
            treeout (p^.leftdesc);
            WRITE(treefile, ',');
            treeout (p^.rtdesc);
            WRITE(treefile, ')');
            END;
         IF p^.tip
         THEN n := ntrees
         ELSE n := p^.index;
         IF p = root
         THEN WRITELN(treefile, ';')
         ELSE IF n > 99
            THEN WRITE(treefile, ':', n:3)
            ELSE IF n > 9
               THEN WRITE(treefile, ':', n:2)
               ELSE WRITE(treefile, ':', n:1);
      END; (* treeout *)

   BEGIN (* consensus *)
      fullset := [1..spp];
      n2 := 0;
      compress;
      sort;
      eliminate;
      compress;
      reconstruct;
      WRITELN(OUTPUT);
      WRITELN(OUTPUT, ' Sets included in the consensus tree');
      printset;
      FOR i := 1 TO n2 DO BEGIN
         grouping[i] := group2[i];
         timesseen[i] := times2[i];
         END;
      n := n2;
      WRITELN(OUTPUT);
      WRITELN(OUTPUT);
      WRITELN(OUTPUT, ' Sets NOT included in consensus tree:');
      printset;
      WRITELN(OUTPUT);
      printree;
      IF trout
      THEN BEGIN
         treeout (root);
         WRITELN(OUTPUT, ' Tree also written onto file');
         WRITELN(OUTPUT);
         END;
   END; (* consensus *)

BEGIN (* main program *)
   (* ASSIGN(infile, 'INFILE');  -- Turbo Pascal only *)
   RESET(infile);
   WRITELN(OUTPUT);
   WRITE(OUTPUT, ' Majority-rule and strict consensus tree');
   WRITELN(OUTPUT, ' program, version 3.1');
   WRITELN(OUTPUT);
   READ(infile, ntrees);
   trout := FALSE;
   noroot := TRUE;
   outgropt := FALSE;
   outgrno := 1;
   numopts := 0;
   WHILE NOT EOLN(infile) DO BEGIN
      READ(infile, ch);
      IF (ch = 'O') OR (ch = 'R') OR (ch = 'Y')
      THEN CASE ch OF
         'O' : BEGIN
               outgropt := TRUE;
               numopts := numopts + 1;
               END;
         'R' : noroot := FALSE;
         'Y' : trout := TRUE
         END;
      END;
   READLN(infile);
   IF outgropt
   THEN READLN(infile, ch, outgrno);
   IF trout
   THEN BEGIN
      REWRITE(treefile);
      END;
   didreroot := FALSE;
   first := TRUE;
   spp := 0;
   setuptree;
   FOR i := 1 TO maxgrp DO BEGIN
      grouping[i] := [];
      timesseen[i] := 0;
      END;
   WRITELN(OUTPUT, ' Species in order: ');
   WRITELN(OUTPUT);
   FOR i := 1 TO ntrees DO BEGIN
      treeread;
      IF noroot
      THEN BEGIN
         reroot (treenode[outgrno]);
         didreroot := outgropt;
         END;
      accumulate (root);
      first := FALSE;
      END;
   WRITELN(OUTPUT);
   consensus;
(* CLOSE(infile);  -- may need this here *)
   IF trout
   THEN BEGIN
      (* CLOSE(treefile);  -- may need this here *)
      END;
END. (* main *)

