PROGRAM DNAMove (infile, treefile, INPUT, OUTPUT);
  (* version 3.1. (c) Copyright 1987 by the University of Washington.
     Written by Joseph Felsenstein.  Permission is granted to copy
     and use this program provided no fee is charged for it and
     provided that this copyright notice is not removed. *)

CONST maxsp = 50;    (* maximum number of species               *)
   maxsz = 100;      (* size of pointer array.  >= 2*maxsp - 1  *)
   maxsites = 300;   (* max. number base sites per sequence *)
   nmlngth = 10;     (* number of characters in species name    *)
   screenlines = 24; (* number of lines on a screen *)
   ibmpc =  TRUE;  (* is it an IBM PC-compatible? *)
   vt52 = FALSE;    (* is it a VT52-compatible terminal? *)
   ansi = FALSE;    (* is it an ANSI standard terminal? *)

TYPE steparray = ARRAY[1..maxsites] OF INTEGER;
   bases = (A, C, G, U, O);
   baset = SET OF bases;
   basearray = ARRAY [1..maxsites] OF baset;
   chartype = (horiz, vert, up, over, aa, cc, gg, tt, quest);
   nodeptr = ^node;   (* nodes will form a binary tree *)
   node = RECORD      (* describes a tip species or an ancestor *)
      ancestor, leftdesc, rtdesc : nodeptr;  (* pointers to nodes *)
      index : INTEGER;      (* number of the node *)
      tip : BOOLEAN;        (* present species are tips of tree *)
      base : basearray; (* the sequence *)
      state : CHAR;   (* stores state for printing *)
      xcoord, ycoord : INTEGER    (* used by printree *)
      END;
   pointarray = ARRAY [1..maxsz] OF nodeptr;

VAR root : nodeptr;
   infile, treefile : TEXT;
   spp, nonodes, chars, outgrno : INTEGER;
     (* spp = number of species
        nonodes = number of nodes in tree
        chars = number of sites in actual sequences
        outgrno indicates outgroup *)
   weights, thresh : BOOLEAN;
   weight : steparray;
   treenode : pointarray;      (* pointers to all nodes in tree *)
   name : ARRAY[1..maxsp, 1..nmlngth] OF CHAR; (* names of species *)
   threshold : REAL;
   threshwt : ARRAY[1..maxsites] OF REAL;
   reversed : ARRAY [chartype] OF BOOLEAN;
   graphic : ARRAY[chartype] OF BOOLEAN;
   ch : ARRAY[chartype] OF CHAR;

   FUNCTION digit (ch : CHAR) : BOOLEAN;
   BEGIN (* digit *)
      digit := (ch >= '0') AND (ch <= '9');
   END; (* digit *)

   FUNCTION letter (ch : CHAR) : BOOLEAN;
   BEGIN (* letter -- works for both ASCII and EBCDIC codes *)
      letter := ((ch >= 'A') AND (ch <= 'I'))
                OR ((ch >= 'J') AND (ch <= 'R'))
                OR ((ch >= 'S') AND (ch <= 'Z'))
                OR ((ch >= 'a') AND (ch <= 'i'))
                OR ((ch >= 'j') AND (ch <= 'r'))
                OR ((ch >= 's') AND (ch <= 'z'));
   END; (* letter *)

   PROCEDURE uppercase (VAR ch : CHAR);
      (* convert a character to upper case -- either ASCII or EBCDIC *)
   BEGIN (* uppercase *)
      IF letter(ch) AND (('a' > 'A') AND (ch >= 'a')
                     OR  ('a' < 'A') AND (ch < 'A'))
      THEN ch := CHR(ORD(ch) + ORD('A') - ORD('a'));
   END; (* uppercase *)

   PROCEDURE configure;
      (* configure to machine -- set up special characters *)
      VAR a : chartype;
   BEGIN (* configure *)
      FOR a := horiz TO quest DO reversed[a] := FALSE;
      FOR a := horiz TO quest DO graphic[a] := FALSE;
      IF ibmpc
      THEN BEGIN
         ch[horiz] := CHR(205);
         ch[vert] := CHR(186);
         ch[up] := CHR(186);
         ch[over] := CHR(205);
         ch[aa] := CHR(176);
         ch[cc] := CHR(178);
         ch[gg] := CHR(177);
         ch[tt] := CHR(219);
         ch[quest] := CHR(1);
         END
      ELSE IF vt52
         THEN BEGIN
            ch[horiz] := ' ';
            reversed[horiz] := TRUE;
            ch[vert] := ch[horiz];
            reversed[vert] := TRUE;
            ch[up] := CHR(96);
            graphic[up] := TRUE;
            ch[over] := 'a';
            graphic[over] := TRUE;
            ch[aa] := 'w';
            graphic[aa] := TRUE;
            ch[cc] := ' ';
            reversed[cc] := TRUE;
            ch[gg] := 'j';
            graphic[gg] := TRUE;
            ch[tt] := 'i';
            graphic[tt] := TRUE;
            ch[quest] := '?';
            reversed[quest] := TRUE;
            END
         ELSE IF ansi
            THEN BEGIN
               ch[horiz] := ' ';;
               reversed[horiz] := TRUE;
               ch[vert] := ch[horiz];
               reversed[vert] := TRUE;
               ch[up] := 'x';
               graphic[up] := TRUE;
               ch[over] := 'q';
               graphic[over] := TRUE;
               ch[aa] := 'a';
               reversed[aa] := TRUE;
               ch[cc] := 'c';
               reversed[cc] := TRUE;
               ch[gg] := 'g';
               reversed[gg] := TRUE;
               ch[tt] := 't';
               reversed[tt] := TRUE;
               ch[quest] := '?';
               reversed[quest] := TRUE;
               END
            ELSE BEGIN
               ch[horiz] := '=';
               ch[vert] := ' ';
               ch[up] := '!';
               ch[over] := '-';
               ch[aa] := 'a';
               ch[cc] := 'c';
               ch[gg] := 'g';
               ch[tt] := 't';
               ch[quest] := '.';
               END;
   END; (* configure *)

   PROCEDURE prefix (a : chartype);
      (* give prefix appropriate for this character *)

      PROCEDURE pregraph;
         (* turn on graphic characters *)
      BEGIN (* pregraph *)
         IF vt52
         THEN WRITE(CHR(27), 'F');
         IF ansi
         THEN WRITE(CHR(27), '(0');
      END; (* pregraph *)

      PROCEDURE prereverse;
         (* turn on reverse video *)
      BEGIN (* prereverse *)
         IF vt52
         THEN WRITE(CHR(27), 'p');
         IF ansi
         THEN WRITE(CHR(27), '[7m');
      END; (* prereverse *)

   BEGIN (* prefix *)
      IF reversed[a]
      THEN prereverse;
      IF graphic[a]
      THEN pregraph;
   END; (* prefix *)

   PROCEDURE postfix (a : chartype);
      (* give postfix appropriate for this character *)

      PROCEDURE postgraph;
         (* turn off graphic characters *)
      BEGIN (* postgraph *)
         IF vt52
         THEN WRITE(CHR(27), 'G');
         IF ansi
         THEN WRITE(CHR(27), '(B');
      END; (* postgraph *)

      PROCEDURE postreverse;
         (* turn off reverse video *)
      BEGIN (* postreverse *)
         IF vt52
         THEN WRITE(CHR(27), 'q');
         IF ansi
         THEN WRITE(CHR(27), '[0m');
      END; (* postreverse *)

   BEGIN (* postfix *)
      IF reversed[a]
      THEN postreverse;
      IF graphic[a]
      THEN postgraph;
   END; (* postfix *)

   PROCEDURE makechar (a : chartype);
      (* print out a character with appropriate prefix or postfix *)
   BEGIN (* makechar *)
      prefix (a);
      WRITE(ch[a]);
      postfix (a);
   END; (* makechar *)

   PROCEDURE logo;
      VAR i : INTEGER;
         c : CHAR;
   BEGIN (* logo *)
      IF ibmpc
      THEN ClrScr  (* Turbo Pascal only *);
      FOR i := 1 TO 10 DO WRITELN;
      WRITE('                     ');
      IF ibmpc
      THEN WRITE(CHR(201))
      ELSE WRITE(' ');
      prefix (horiz);
      FOR i := 1 TO 32 DO WRITE(ch[horiz]);
      postfix (horiz);
      IF ibmpc
      THEN WRITELN(CHR(187))
      ELSE WRITELN(' ');
      WRITE('                     ');
      makechar (vert);
      WRITE('                                ');
      makechar (vert);
      WRITELN;
      WRITE('                     ');
      makechar (vert);
      WRITE(' D N A M O V E  (in PHYLIP 3.1) ');
      makechar (vert);
      WRITELN;
      WRITE('                     ');
      makechar (vert);
      WRITE('                                ');
      makechar (vert);
      WRITELN;
      WRITE('                     ');
      IF ibmpc
      THEN WRITE(CHR(200))
      ELSE WRITE(' ');
      prefix (horiz);
      FOR i := 1 TO 32 DO WRITE(ch[horiz]);
      postfix (horiz);
      IF ibmpc
      THEN WRITELN(CHR(188))
      ELSE WRITELN(' ');
      FOR i := 1 TO 10 DO WRITELN;
      WRITELN(' TO CONTINUE, PRESS ON THE Return OR Enter KEY');
      READ(c);
   END; (* logo *)

   PROCEDURE doinput;
      (* reads the input data *)

      PROCEDURE inputnumbers;
         (* input the numbers of species and of characters *)
      BEGIN (* inputnumbers *)
         READ(infile, spp, chars);
         WRITELN(' ', spp:2, ' species, ', chars:3, '  sites');
         WRITELN;
         IF spp > maxsp
         THEN WRITELN(' TOO MANY SPECIES: adjust CONSTants');
         IF chars > maxsites
         THEN WRITELN(' TOO MANY SITES: adjust CONSTants');
         nonodes := 2 * spp - 1;
      END; (* inputnumbers *)

      PROCEDURE inputoptions;
         (* input the information on the options *)
         VAR ch : CHAR;
            extranum, i : INTEGER;

         PROCEDURE inputhresh;
            (* input the threshold *)
         BEGIN (* inputhresh *)
            READLN(infile, threshold);
            thresh := TRUE;
         END; (* inputhresh *)

         PROCEDURE printhresh;
            (* print out the threshold value *)
         BEGIN (* printhresh *)
            WRITELN('    Threshold value = ', threshold:10:3);
            WRITELN;
         END; (* printhresh *)

         PROCEDURE inputweights;
            (* input the character weights, 0-9 and A-Z for weights
               0 - 35 *)
            VAR ch : CHAR;
               i : INTEGER;
         BEGIN (* inputweights *)
            FOR i := 1 TO nmlngth-1 DO READ(infile, ch);
            FOR i := 1 TO chars DO BEGIN
               REPEAT
                  IF EOLN(infile) THEN READLN(infile);
                  READ(infile, ch);
               UNTIL ch <> ' ';
               weight[i] := 1;
               IF digit(ch)
               THEN weight[i] := ORD(ch) - ORD('0')
               ELSE IF letter(ch)
                  THEN BEGIN
                     uppercase (ch);
                     IF (ch >= 'A') AND (ch <= 'I')
                     THEN weight[i] := ORD(ch) - ORD('A') + 10
                     ELSE IF (ch >= 'J') AND (ch <= 'R')
                        THEN weight[i] := ORD(ch) - ORD('J') + 19
                        ELSE weight[i] := ORD(ch) - ORD('S') + 28;
                     END
                  ELSE WRITELN(' Bad weight character: ',ch);
               END;
            READLN(infile);
            weights := TRUE;
         END; (* inputweights *)

         PROCEDURE printweights;
            (* print out the weights of sites *)
            VAR i, j, k : INTEGER;
         BEGIN (* printweights *)
            WRITELN('Sites are weighted as follows:');
            WRITE('        ');
            FOR i := 0 TO 9 DO WRITE(i:3);
            WRITELN;
            WRITELN('     *---------------------------------');
            FOR j := 0 TO (chars DIV 10) DO BEGIN
               WRITE(10*j:5, '!  ');
               FOR i := 0 TO 9 DO BEGIN
                  k := 10*j + i;
                  IF (k>0) AND (k<=chars)
                  THEN WRITE(weight[k]:3)
                  ELSE WRITE('   ');
                  END;
               WRITELN;
               END;
            WRITELN;
         END; (* printweights *)

      BEGIN (* inputoptions *)
         thresh := FALSE;
         weights := FALSE;
         extranum := 0;
         WHILE NOT EOLN(infile) DO BEGIN
            READ(infile, ch);
            uppercase (ch);
            IF (ch = 'T') OR (ch = 'W')
            THEN CASE ch OF
                  'T' : extranum := extranum + 1;
                  'W' : extranum := extranum + 1
                  END
            ELSE IF ch <> ' '
               THEN WRITELN(' Bad Option Character: ', ch);
            END;
         READLN(infile);
         threshold := spp;
         FOR i := 1 TO chars DO weight[i] := 1;
         FOR i := 1 TO extranum DO BEGIN
            READ(infile, ch);
            uppercase (ch);
            IF ch = 'T'
            THEN inputhresh;
            IF ch = 'W'
            THEN inputweights;
            END;
         IF thresh
         THEN printhresh;
         IF weights
         THEN printweights
         ELSE FOR i := 1 TO chars DO
            threshwt[i] := threshold * weight[i];
      END; (* inputoptions *)

      PROCEDURE inputdata;
         (* input the names and sequences for each species *)
         VAR i, j : INTEGER;
            ns : baset; (* temporary base set for input *)
            charstate : CHAR;  (*  character temporary storage for
                                   input *)
      BEGIN (* inputdata *)
         FOR i := 1 TO nonodes DO BEGIN
            NEW (treenode[i]);
            treenode[i]^.leftdesc := NIL;
            treenode[i]^.rtdesc := NIL;
            treenode[i]^.ancestor := NIL;
            treenode[i]^.index := i;
            IF i <= spp
            THEN BEGIN
               treenode[i]^.tip := TRUE;
               FOR j := 1 TO nmlngth DO READ(infile, name[i,j]);
               FOR j := 1 TO chars DO BEGIN
                  REPEAT
                     IF EOLN(infile) THEN READLN(infile);
                     READ(infile, charstate);
                  UNTIL (charstate <> ' ') AND
                     ((charstate < '0') OR (charstate > '9'));
                  uppercase (charstate);
                  IF NOT ((charstate = 'A') OR (charstate = 'B')
                         OR (charstate = 'C') OR (charstate = 'D')
                         OR (charstate = 'G') OR (charstate = 'H')
                         OR (charstate = 'K') OR (charstate = 'M')
                         OR (charstate = 'N') OR (charstate = 'R')
                         OR (charstate = 'S') OR (charstate = 'T')
                         OR (charstate = 'U') OR (charstate = 'V')
                         OR (charstate = 'W') OR (charstate = 'X')
                         OR (charstate = 'Y') OR (charstate = '?')
                         OR (charstate = 'O') OR (charstate = '.')
                         OR (charstate = '-'))
                  THEN WRITELN(' Warning -- bad base: ', charstate);
                  CASE charstate OF
                     'A' : ns := [A];
                     'C' : ns := [C];
                     'G' : ns := [G];
                     'U' : ns := [U];
                     'T' : ns := [U];
                     'M' : ns := [A, C];
                     'R' : ns := [A, G];
                     'W' : ns := [A, U];
                     'S' : ns := [C, G];
                     'Y' : ns := [C, U];
                     'K' : ns := [G, U];
                     'B' : ns := [C, G, U];
                     'D' : ns := [A, G, U];
                     'H' : ns := [A, C, U];
                     'V' : ns := [A, C, G];
                     'N' : ns := [A, C, G, U];
                     'X' : ns := [A, C, G, U];
                     '?' : ns := [A, C, G, U];
                     'O' : ns := [O];
                     '.' : ns := [O];
                     '-' : ns := [O]
                     END;
                  treenode[i]^.base[j] := ns;
                  END;
               READLN(infile);
               END
            ELSE treenode[i]^.tip := FALSE;
            END;
         root := NIL;
         WRITELN;
         WRITELN;
      END; (* inputdata *)

   BEGIN (* doinput *)
      WRITELN;
      WRITELN(' Reading input file ...');
      WRITELN;
      inputnumbers;
      inputoptions;
      inputdata;
   END; (* doinput *)

   PROCEDURE treeconstruct;
      (* constructs a binary tree from the pointers in treenode. *)
      TYPE rearrtype = (rearr, flipp, reroott, none);
      VAR dispchar, atwhat, what,
            fromwhere, towhere, oldoutgrno : INTEGER;
         like, bestyet, gotlike : REAL;
         display, yours, subtree , written, oldwritten : BOOLEAN;
         numsteps : steparray;
         nuroot : nodeptr;
         lastop : rearrtype;

      PROCEDURE add (below, newtip, newfork : nodeptr);
         (* inserts the nodes newfork and its left descendant, newtip,
            to the tree.  below becomes newfork's right descendant *)
      BEGIN (* add *)
         newfork^.ancestor := below^.ancestor;
         newfork^.rtdesc := below;
         newfork^.leftdesc := newtip;
         below^.ancestor := newfork;
         newtip^.ancestor := newfork;
         IF newfork^.ancestor <> NIL
         THEN IF newfork^.ancestor^.leftdesc = below
            THEN newfork^.ancestor^.leftdesc := newfork
            ELSE newfork^.ancestor^.rtdesc := newfork;
      END; (* add *)

      PROCEDURE remove (VAR item, fork : nodeptr);
         (* removes nodes item and its ancestor, fork, from the tree.
            the new descendant of fork's ancestor is made to be
            fork's second descendant (other than item).  Also
            returns pointers to the deleted nodes, item and fork *)
         VAR other, below : nodeptr;
      BEGIN (* remove *)
         IF item^.ancestor = NIL
         THEN fork := NIL
         ELSE BEGIN
            fork := item^.ancestor;
            IF item = fork^.leftdesc
            THEN other := fork^.rtdesc
            ELSE other := fork^.leftdesc;
            below := fork^.ancestor;
            other^.ancestor := below;
            IF below <> NIL
            THEN BEGIN
               IF below^.leftdesc = fork
               THEN below^.leftdesc := other
               ELSE below^.rtdesc := other;
               fork^.ancestor := NIL;
               END;
            fork^.leftdesc := NIL;
            fork^.rtdesc := NIL;
            item^.ancestor := NIL;
            END;
      END; (* remove *)

      PROCEDURE evaluate (r: nodeptr);
         (* determines the number of steps needed for a tree. this is
            the minimum number of steps needed to evolve sequences on
            this tree *)
         VAR i, steps : INTEGER;
            sum : REAL;

         PROCEDURE postorder (p : nodeptr);
            (* traverses a binary tree, calling PROCEDURE fillin at a
               node's descendants before calling fillin at the node *)

            PROCEDURE fillin (p: nodeptr);
               (* sets up for each node in the tree the base sequence
                  at that point and counts the changes.  The program
                  spends much of its time in this PROCEDURE *)
               VAR i : INTEGER;
                  ns, rs, ls : baset;
            BEGIN (* fillin *)
               FOR i := 1 TO chars DO BEGIN
                  ls := p^.leftdesc^.base[i];
                  rs := p^.rtdesc^.base[i];
                  ns := ls * rs;
                  IF ns = []
                  THEN BEGIN
                     ns := ls + rs;
                     numsteps[i] := numsteps[i] + weight[i];
                     END;
                  p^.base[i] := ns;
                  END;
            END; (* fillin *)

         BEGIN (* postorder *)
            IF NOT p^.tip
            THEN BEGIN
               postorder (p^.leftdesc);
               postorder (p^.rtdesc);
               fillin (p);
               END;
         END; (* postorder *)

      BEGIN (* evaluate *)
         sum := 0;
         FOR i := 1 TO chars DO
            numsteps[i] := 0;
         postorder (r);
         FOR i := 1 TO chars DO BEGIN
            steps := numsteps[i];
            IF steps <= threshwt[i]
            THEN sum := sum + steps
            ELSE sum := sum + threshwt[i];
            END;
         like := -sum;
      END; (* evaluate *)

      PROCEDURE reroot (outgroup : nodeptr);
         (* reorients tree, putting outgroup in desired position.
            Written by Kent Fiala *)
         VAR vertex, oldvertex, ante, otherbranch : nodeptr;
            left : BOOLEAN;
      BEGIN (* reroot *)
         IF NOT (outgroup^.ancestor = root)
         THEN BEGIN
            oldvertex := outgroup;
            vertex := outgroup^.ancestor;
            ante := vertex^.ancestor;
            IF vertex^.rtdesc = outgroup
            THEN BEGIN
               vertex^.rtdesc := vertex^.leftdesc;
               vertex^.leftdesc := outgroup;
               END;
            REPEAT
               IF oldvertex = vertex^.leftdesc
               THEN BEGIN
                  vertex^.leftdesc := ante;
                  left := TRUE;
                  END
               ELSE BEGIN
                  vertex^.rtdesc := ante;
                  left := FALSE;
                  END;
               vertex^.ancestor := oldvertex;
               oldvertex := vertex;
               vertex := ante;
               ante := vertex^.ancestor
            UNTIL vertex = root;
            IF root^.leftdesc = oldvertex
            THEN otherbranch := root^.rtdesc
            ELSE otherbranch := root^.leftdesc;
            IF left
            THEN oldvertex^.leftdesc := otherbranch
            ELSE oldvertex^.rtdesc := otherbranch;
            otherbranch^.ancestor := oldvertex;
            root^.rtdesc := outgroup^.ancestor;
            root^.leftdesc := outgroup;
            outgroup^.ancestor^.ancestor := root;
            outgroup^.ancestor := root;
            END;
      END; (* reroot *)

      PROCEDURE printree;
         (* prints out diagram of the tree *)
         CONST downn = 2;
            overr = 4;
         VAR tipx, tipy, i, rover, dow : INTEGER;

         PROCEDURE coordinates (p : nodeptr);
            (* establishes coordinates of nodes *)
         BEGIN (* coordinates *)
            IF p^.tip
            THEN BEGIN
               p^.xcoord := tipx;
               p^.ycoord := tipy;
               tipx := tipx + rover;
               tipy := tipy + dow;
               END
            ELSE BEGIN
               coordinates (p^.leftdesc);
               p^.xcoord := p^.leftdesc^.xcoord;
               coordinates (p^.rtdesc);
               p^.ycoord := p^.rtdesc^.ycoord;
               END;
         END; (* coordinates *)

         PROCEDURE drawline (i : INTEGER);
            (* draws one row of the tree diagram by moving up tree *)
            VAR p, q : nodeptr;
               m, n, j : INTEGER;
               c : chartype;
         BEGIN (* drawline *)
            p := nuroot;
            WRITE('   ');
            REPEAT
               IF p^.leftdesc^.ycoord >= i
               THEN p := p^.leftdesc
               ELSE BEGIN
                  q := p^.rtdesc;
                  n := q^.xcoord - p^.xcoord;
                  IF p^.ycoord = i
                  THEN BEGIN
                     c := over;
                     IF display
                     THEN CASE p^.rtdesc^.state OF
                        'A' : c := aa;
                        'C' : c := cc;
                        'G' : c := gg;
                        'T' : c := tt;
                        '?' : c := quest
                        END;
                     IF p^.index >= 10
                     THEN WRITE(p^.index:2)
                     ELSE BEGIN
                        WRITE(p^.index:1);
                        makechar (c);
                        END;
                     IF n > 2
                     THEN BEGIN
                        prefix (c);
                        FOR j := 1 TO n-2 DO WRITE(ch[c]);
                        postfix (c);
                        END
                     END
                  ELSE BEGIN
                     c := up;
                     IF display
                     THEN CASE p^.leftdesc^.state OF
                        'A' : c := aa;
                        'C' : c := cc;
                        'G' : c := gg;
                        'T' : c := tt;
                        '?' : c := quest
                        END;
                     makechar (c);
                     FOR j := 1 TO n-1 DO WRITE(' ');
                     END;
                  p := q;
                  END;
            UNTIL p^.tip;
            IF p^.ycoord = i
            THEN BEGIN
               IF p^.index >= 10
               THEN m := 2
               ELSE m := 1;
               WRITE(p^.index:m, ':');
               FOR j := 1 TO nmlngth DO
                       WRITE(name[p^.index, j]);
               IF (i = 1)
               THEN BEGIN
                  WRITE('                      ');
                  IF subtree
                  THEN WRITE('      Subtree          ')
                  ELSE WRITE('                       ');
                  WRITE('  (unrooted tree)');
                  END;
               IF (i = 3) AND display
               THEN BEGIN
                  WRITE('        CHARACTER ', dispchar:3,
                           '    SYMBOLS: ');
                  makechar (aa);
                  WRITE(':A, ');
                  makechar (cc);
                  WRITE(':C, ');
                  makechar (gg);
                  WRITE(':G, ');
                  makechar (tt);
                  WRITE(':T, ');
                  makechar (quest);
                  WRITE(':?');
                  END;
               IF i = 5
               THEN BEGIN
                  WRITE('             ', -like:10:5);
                  IF NOT yours
                  THEN BEGIN
                  IF -like < bestyet
                     THEN BEGIN
                        WRITE('  BEST YET!');
                        bestyet := -like;
                        gotlike := bestyet;
                        END
                     ELSE IF ABS(-like - bestyet) < 0.000001
                        THEN WRITE(' (as good as best)')
                        ELSE BEGIN
                           IF -like < gotlike
                           THEN WRITE('  better')
                           ELSE IF -like > gotlike
                              THEN WRITE('  worse!');
                           END;
                     gotlike := -like;
                     END;
                  END;
               END;
            WRITELN;
         END; (* drawline *)

      BEGIN (* printree *)
         IF NOT subtree
         THEN nuroot := root;
         IF NOT display
         THEN evaluate(root);
         ClrScr; (* Turbo Pascal only *)
         IF ansi
         THEN WRITE(CHR(27), '[2J', CHR(27), '[H')
         ELSE IF vt52
            THEN WRITE(CHR(27), 'E', CHR(27), 'H');
         IF ibmpc
         THEN WRITELN;
         tipx := 0;
         tipy := 1;
         rover := 100 DIV spp;
         IF rover > overr THEN rover := overr;
         dow := downn;
         IF (spp*dow > screenlines) AND (NOT subtree)
         THEN BEGIN
            dow := dow-1;
            rover := rover-1;
            END;
         coordinates (nuroot);
         FOR i := 1 TO tipy-dow DO drawline (i);
         IF (spp <= (screenlines DIV 2)) OR subtree THEN WRITELN;
      END; (* printree *)

      PROCEDURE describe;
         (* makes state reconstructions in each site *)

         PROCEDURE hypstates;
            (* fill in and describe states at interior nodes *)
            VAR i : INTEGER;
               nothing : basearray;
               bottom : BOOLEAN;

            PROCEDURE ancestset (a, b : baset; VAR c : baset);
               (* make the set of ancestral states below nodes
                  whose base sets are a and b *)
            BEGIN (* ancestset *)
               c := a * b;
               IF c = []
               THEN c := a + b;
            END; (* ancestset *)

            PROCEDURE hyptrav (r : nodeptr; VAR hypset : basearray);
               (*  compute, print out states at one interior node *)
               VAR tempset, left, rt, anc : baset;
                  temparray, ancset : basearray;
            BEGIN (* hyptrav *)
               anc := hypset[i];
               IF NOT r^.tip
               THEN BEGIN
                  left := r^.leftdesc^.base[i];
                  rt := r^.rtdesc^.base[i];
                  tempset := left * rt * anc;
                  IF tempset = []
                  THEN BEGIN
                     tempset := (left * rt) + (left * anc)
                                  + (rt * anc);
                     IF tempset = []
                     THEN tempset := left + rt + anc;
                     END;
                  r^.base[i] := tempset;
                  END;
               IF NOT bottom
               THEN anc := r^.ancestor^.base[i];
               r^.state := '?';
               IF r^.base[dispchar] = [a]
               THEN r^.state := 'A';
               IF r^.base[dispchar] = [c]
               THEN r^.state := 'C';
               IF r^.base[dispchar] = [g]
               THEN r^.state := 'G';
               IF r^.base[dispchar] = [u]
               THEN r^.state := 'T';
               bottom := FALSE;
               IF NOT r^.tip
               THEN BEGIN
                  temparray := r^.leftdesc^.base;
                  ancestset(hypset[i], r^.rtdesc^.base[i],
                     ancset[i]);
                  hyptrav (r^.leftdesc, ancset);
                  ancestset(hypset[i], temparray[i],
                     ancset[i]);
                  hyptrav (r^.rtdesc, ancset);
                  END;
            END; (* hyptrav *)

         BEGIN (* hypstates *)
            i := dispchar;
            nothing[i] := [];
            bottom := TRUE;
            hyptrav (root, nothing);
         END; (* hypstates *)

      BEGIN (* describe *)
         evaluate (root);
         hypstates;
      END; (* describe *)

      PROCEDURE buildtree;
         VAR ch : CHAR;
            ok : BOOLEAN;

         PROCEDURE arbitree;
            VAR i : INTEGER;
         BEGIN (* arbitree *)
            root := treenode[1];
            add (treenode[1], treenode[2], treenode[spp+1]);
            root := treenode[spp+1];
            FOR i := 3 TO spp DO BEGIN
               add (treenode[spp+i-2], treenode[i], treenode[spp+i-1]);
               IF root^.ancestor <> NIL THEN root := root^.ancestor;
               END;
            printree;
            bestyet := -like;
            gotlike := -like;
         END; (* arbitree *)

         PROCEDURE yourtree;
            VAR i, j : INTEGER;
               ok : BOOLEAN;
         BEGIN (* yourtree *)
            root := treenode[1];
            add (treenode[1], treenode[2], treenode[spp+1]);
            root := treenode[spp+1];
            i := 2;
            REPEAT
               i := i + 1;
               printree;
               WRITE(' Add species', i:3, ': ');
               FOR j := 1 TO nmlngth DO WRITE(name[i,j]);
               REPEAT
                  WRITELN;
                  WRITE( '  before node (type number): ');
                  READLN(INPUT, j);
                  ok := ((j >= 1) AND (j < i)) OR
                          ((j > spp) AND (j < (spp + i -1)));
                  IF NOT ok
                  THEN WRITELN('Impossible number. Please try again:');
               UNTIL ok;
               add (treenode[j], treenode[i], treenode[spp+i-1]);
               WHILE root^.ancestor <> NIL DO root := root^.ancestor;
            UNTIL i = spp;
            printree;
            bestyet := -like;
            gotlike := -like;
         END; (* yourtree *)

         PROCEDURE treeread;
            (* read in user-defined tree and set it up *)
            VAR ch : CHAR;
               nextnode : INTEGER;

            PROCEDURE findch (c : CHAR);
               (* scan forward until find character c *)
            BEGIN (* findch *)
               IF ch <> c
               THEN REPEAT
                     IF EOLN(treefile) THEN READLN(treefile);
                     READ(treefile, ch);
                  UNTIL ch = c;
            END; (* findch *)

            PROCEDURE addelement (VAR p : nodeptr);
               (* recursive procedure adds nodes to user-defined tree *)
               VAR q : nodeptr;
                  i, n : INTEGER;
                  found : BOOLEAN;
                  str : ARRAY[1..nmlngth] OF CHAR;
            BEGIN (* addelement *)
               REPEAT
                  IF EOLN(treefile) THEN READLN(treefile);
                  READ(treefile, ch);
               UNTIL ch <> ' ';
               IF ch = '('
               THEN BEGIN
                  nextnode := nextnode + 1;
                  q := treenode[nextnode];
                  addelement (q^.leftdesc);
                  q^.leftdesc^.ancestor := q;
                  findch (',');
                  addelement (q^.rtdesc);
                  q^.rtdesc^.ancestor := q;
                  findch (')');
                  p := q;
                  END
               ELSE BEGIN
                  FOR i := 1 TO nmlngth DO str[i] := ' ';
                  n := 1;
                  REPEAT
                     IF (ch = '_') THEN ch := ' ';
                     str[n] := ch;
                     IF EOLN(treefile) THEN READLN(treefile);
                     READ(treefile, ch);
                     n := n + 1;
                  UNTIL ((ch = ',') OR (ch = ')') OR (n > nmlngth));
                  n := 1;
                  REPEAT
                     found := TRUE;
                     FOR i := 1 TO nmlngth DO
                        found := found AND (str[i] = name[n, i]);
                     IF found
                     THEN p := treenode[n]
                     ELSE n := n + 1;
                  UNTIL (n > spp) OR found;
                  IF n > spp
                  THEN BEGIN
                     WRITE(' Cannot find species: ');
                     FOR i := 1 TO nmlngth DO
                        WRITE(str[i]);
                     WRITELN;
                     END;
                  END;
            END; (* addelement *)

         BEGIN (* treeread *)
            RESET(treefile);
            root := treenode[spp + 1];
            nextnode := spp;
            root^.ancestor := NIL;
            addelement (root);
            WRITELN;
            printree;
            bestyet := -like;
            gotlike := -like;
         (* CLOSE(treefile); -- may need this here *)
         END; (* treeread *)

      BEGIN (* buildtree *)
         ok := FALSE;
         yours := TRUE;
         REPEAT
            WRITELN;
            WRITELN(' How should we build the initial tree?');
            WRITELN('     A   Arbitrarily');
            WRITELN('     S   as you Specify');
            WRITELN('     U   User tree read from file');
            WRITE(' Type one of those: ');
            READLN(INPUT, ch);
            UPPERCASE(ch);
            ok := (ch = 'A') OR (ch = 'S') OR (ch = 'U');
            IF NOT ok
            THEN WRITELN(' No!! Not a possible choice. Try again:');
         UNTIL ok;
         CASE ch OF
            'A' : arbitree;
            'S' : yourtree;
            'U' : treeread
            END;
         yours := FALSE;
         outgrno := root^.leftdesc^.index;
      END; (* buildtree *)

      PROCEDURE redisplay;
         VAR done : BOOLEAN;
            ch : CHAR;

         PROCEDURE clade;
            (* pick a subtree and show only that on screen *)
            VAR i : INTEGER;
               ok : BOOLEAN;
         BEGIN (* clade *)
            WRITE(' Select subtree rooted at which node');
            WRITE(' (0 for whole tree)? ');
            READLN(i);
            ok := (i >=0) AND (i <= nonodes);
            IF ok
            THEN BEGIN
               subtree := (i > 0);
               IF subtree
               THEN nuroot := treenode[i]
               ELSE nuroot := root;
               END;
            IF display
            THEN describe;
            printree;
            IF NOT ok
            THEN WRITE(' Illegal node! ');
         END; (* clade *)

         PROCEDURE flip;
            (* flip at a node left-right *)
            VAR i : INTEGER;
               ok : BOOLEAN;
               p : nodeptr;
         BEGIN (* flip *)
            WRITE('Flip branches at which node? ');
            READLN(i);
            ok := (i > spp) AND (i <= nonodes);
            IF ok
            THEN BEGIN
               p := treenode[i]^.leftdesc;
               treenode[i]^.leftdesc := treenode[i]^.rtdesc;
               treenode[i]^.rtdesc := p;
               atwhat := i;
               lastop := flipp;
               END;
            IF display
            THEN describe;
            printree;
            IF NOT ok
            THEN WRITE(' Illegal! ')
            ELSE BEGIN
               oldwritten := written;
               written := FALSE;
               END;
         END; (* flip *)

         PROCEDURE help;
            (* display help information *)
         BEGIN (* help *)
            WRITELN;
            WRITELN;
            WRITELN('  C show only one Clade (subtree)',
                       '(useful if tree is too big)');
            WRITELN('  F Flip (rotate) branches at a node');
            WRITELN('  O select an Outgroup for the tree');
            WRITELN('  R Rearrange a tree by moving a node or group');
            WRITELN('  S Show the states of a character');
            WRITELN('  +        ... of the next character');
            WRITELN('  -        ... of the previous character');
            WRITELN('  . redisplay the same tree again');
            WRITELN('  T Try all possible positions',
                         ' of a node or group');
            WRITELN('  U Undo the most recent rearrangement');
            WRITELN('  W Write tree to a file');
            WRITELN('  H Help (this screen)');
            WRITELN('  ?  "     "      "   ');
            WRITELN('  Q (Quit) Exit from program');
            WRITELN('  X Exit from program');
            WRITELN;
            WRITELN;
            WRITE(' TO CONTINUE, PRESS ON THE Return OR Enter KEY');
            READ(ch);
            IF display
            THEN describe;
            printree;
         END; (* help *)

         PROCEDURE rearrange;
            VAR i, j : INTEGER;
               ok1, ok2 : BOOLEAN;
               p, q : nodeptr;
         BEGIN (* rearrange *)
            WRITE(' Remove which node? ');
            READLN(i);
            ok1 :=  (i >= 1) AND (i <= 2*spp-1);
            IF ok1
            THEN BEGIN
               WRITE(' Add before which node? ');
               READLN(j);
               ok2 := (j >=1) AND (j <= 2*spp-1);
               IF ok2
               THEN BEGIN
                  ok2 := NOT (treenode[j] = treenode[i]^.ancestor);
                  p := treenode[j];
                  WHILE p <> root DO BEGIN
                     ok2 := ok2 AND (NOT (p = treenode[i]));
                     p := p^.ancestor;
                     END;
                  IF ok1 AND ok2
                  THEN BEGIN
                     what := i;
                     q := treenode[i]^.ancestor;
                     IF q^.leftdesc^.index = i
                     THEN fromwhere := q^.rtdesc^.index
                     ELSE fromwhere := q^.leftdesc^.index;
                     towhere := j;
                     remove (treenode[i], q);
                     add (treenode[j], treenode[i], q);
                     WHILE root^.ancestor <> NIL DO
                        root := root^.ancestor;
                     END;
                     lastop := rearr;
                  END;
               END;
            IF display
            THEN describe;
            printree;
            IF NOT (ok1 AND ok2)
            THEN WRITE(' Illegal! Try again: ')
            ELSE BEGIN
               oldwritten := written;
               written := FALSE;
               END;
         END; (* rearrange *)

         PROCEDURE changeoutgroup;
           VAR i : INTEGER;
              ok : BOOLEAN;
         BEGIN (* changeoutgroup *)
            oldoutgrno := outgrno;
            REPEAT
               WRITE('  Which node should be the new outgroup? ');
               READLN(i);
               ok := (i>=1) AND (i <= nonodes) AND (i <> root^.index);
               IF ok
               THEN outgrno := i;
            UNTIL ok;
            reroot (treenode[outgrno]);
            lastop := reroott;
            IF display
            THEN describe;
            printree;
            oldwritten := written;
            written := FALSE;
         END; (* changeoutgroup *)

         PROCEDURE show;
            VAR i : INTEGER;
               ok : BOOLEAN;
         BEGIN (* show *)
            REPEAT
               WRITE('SHOW: (Character number or 0 to see none)? ');
               READLN(i);
               ok := (i = 0) OR ((i >= 1) AND (i <= chars));
               IF ok AND (i <> 0)
               THEN BEGIN
                  display := TRUE;
                  dispchar := i;
                  END;
               IF ok AND (i = 0)
               THEN display := FALSE;
               UNTIL ok;
            IF display
            THEN describe;
            printree;
         END; (* show *)

         PROCEDURE again;
            (* show same tree again *)
         BEGIN (* again *)
            IF display
            THEN describe;
            printree;
         END; (* again *)

         PROCEDURE nextchar;
            (* show next character *)
         BEGIN (* nextchar *)
            display := TRUE;
            dispchar := dispchar + 1;
            IF dispchar > chars
            THEN dispchar := 1;
            IF display
            THEN describe;
            printree;
         END; (* nextchar *)

         PROCEDURE prevchar;
            (* show previous character *)
         BEGIN (* prevchar *)
            display := TRUE;
            dispchar := dispchar - 1;
            IF dispchar < 1
            THEN dispchar := chars;
            IF display
            THEN describe;
            printree;
         END; (* prevchar *)

         PROCEDURE try;
            (* Remove node, try it in all possible places *)
            VAR i, j : INTEGER;
               current : REAL;
               q, dummy, rute : nodeptr;
               tied, better : BOOLEAN;
               place : ARRAY [1..maxsz] OF REAL;

            PROCEDURE addpreorder (p, item, nufork : nodeptr);
               (* traverses a binary tree, calling PROCEDURE tryadd
                  at a node before calling tryadd at its descendants *)

               PROCEDURE tryadd (p: nodeptr);
                  (* temporarily adds one fork and one tip to the tree.
                     Records scores in ARRAY place *)
               BEGIN (* tryadd *)
                  add (p, item, nufork);
                  IF root^.ancestor <> NIL THEN root := root^.ancestor;
                  evaluate (root);
                  place[p^.index] := -like;
                  remove (item, nufork);
                  IF nufork = root THEN root := p;
               END; (* tryadd *)

            BEGIN (* addpreorder *)
               IF p <> NIL
               THEN BEGIN
                  tryadd (p);
                  addpreorder (p^.leftdesc, item, nufork);
                  addpreorder (p^.rtdesc, item, nufork);
                  END;
            END; (* addpreorder *)

         BEGIN (* try *)
            WRITE('Try other positions for which node? ');
            READLN(i);
            IF (i >= 1) AND (i <= nonodes) AND (i <> root^.index)
            THEN BEGIN
               WRITELN('WAIT ...');
               FOR j := 1 TO nonodes DO place[j] := -1.0;
               evaluate (root);
               current := -like;
               what := i;
               q := treenode[i]^.ancestor;
               IF q^.leftdesc^.index = i
               THEN fromwhere := q^.rtdesc^.index
               ELSE fromwhere := q^.leftdesc^.index;
               rute := root;
               IF root = treenode[i]^.ancestor
               THEN BEGIN
                  IF treenode[i]^.ancestor^.leftdesc = treenode[i]
                  THEN rute := treenode[i]^.ancestor^.rtdesc
                  ELSE rute := treenode[i]^.ancestor^.leftdesc;
                  END;
               remove (treenode[i], dummy);
               root := rute;
               addpreorder (root, treenode[i], dummy);
               add (treenode[fromwhere], treenode[what], dummy);
               IF root^.ancestor <> NIL THEN root := root^.ancestor;
               better := FALSE;
               WRITE('        BETTER: ');
               FOR j := 1 TO nonodes DO
                  IF (place[j] < current) AND (place[j] >= 0.0)
                  THEN BEGIN
                     WRITE(j:3, ':', place[j]:6:2);
                     better := TRUE;
                     END;
               IF NOT better
               THEN WRITE(' NONE');
               WRITELN;
               WRITE('        TIED:    ');
               tied := FALSE;
               FOR j := 1 TO nonodes DO
                  IF (ABS(place[j] - current) < 1.0E-6)
                     AND (j <> fromwhere)
                  THEN BEGIN
                     IF j < 10
                     THEN WRITE(j:2)
                     ELSE WRITE(j:3);
                     tied := TRUE;
                     END;
               IF tied
               THEN WRITELN(':', current:6:2)
               ELSE WRITELN('NONE');
               END
            ELSE WRITE(' Illegal choice! ');
         END; (* try *)

         PROCEDURE undo;
            (* restore to tree before last rearrangement *)
            VAR temp : INTEGER;
               btemp : BOOLEAN;
               q : nodeptr;
         BEGIN (* undo *)
            CASE lastop OF
               rearr :   BEGIN
                         remove (treenode[what], q);
                         add (treenode[fromwhere], treenode[what], q);
                         WHILE root^.ancestor <> NIL DO
                            root := root^.ancestor;
                         temp := fromwhere;
                         fromwhere := towhere;
                         towhere := temp;
                         END;
               flipp :   BEGIN
                         q := treenode[atwhat]^.leftdesc;
                         treenode[atwhat]^.leftdesc := treenode[atwhat]^.rtdesc;
                         treenode[atwhat]^.rtdesc := q;
                         END;
               reroott : BEGIN
                         temp := oldoutgrno;
                         oldoutgrno := outgrno;
                         outgrno := temp;
                         reroot (treenode[outgrno]);
                         END;
               none : BEGIN END
               END;
            IF display
            THEN describe;
            printree;
            IF lastop = none
            THEN WRITE(' No operation to undo! ')
            ELSE BEGIN
               btemp := oldwritten;
               oldwritten := written;
               written := btemp;
               END;
         END; (* undo *)

         PROCEDURE treewrite;
            (* write out tree to a file *)

            PROCEDURE treeout (p : nodeptr);
               (* write out file with representation of final tree *)
               VAR i, n : INTEGER;
                  c : CHAR;
            BEGIN (* treeout *)
               IF p^.tip
               THEN BEGIN
                  n := 0;
                  FOR i := 1 TO nmlngth DO
                     IF name[p^.index, i] <> ' '
                     THEN n := i;
                  FOR i := 1 TO n DO BEGIN
                     c := name[p^.index, i];
                     IF c = ' '
                     THEN c := '_';
                     WRITE(treefile, c);
                     END;
                  END
               ELSE BEGIN
                  WRITE(treefile, '(');
                  treeout (p^.leftdesc);
                  WRITE(treefile, ',');
                  treeout (p^.rtdesc);
                  WRITE(treefile, ')');
                  END;
               IF p = root
               THEN WRITELN(treefile, ';');
            END; (* treeout *)

         BEGIN (* treewrite *)
            REWRITE(treefile);
            treeout(root);
            WRITELN;
            WRITELN(' Tree written to file');
            WRITELN;
            CLOSE(treefile); (* -- may need this here *)
            written := TRUE;
         END; (* treewrite *)

      BEGIN (* redisplay *)
         done := FALSE;
         REPEAT
            WRITE(' NEXT? (Options: C F O R S + - . T U W H ? X) ');
            WRITE('(H or ? for Help) ');
            READLN(ch);
            uppercase (ch);
            IF (ch = 'C') OR (ch = 'F') OR (ch = 'O') OR (ch = 'R')
               OR (ch = 'S') OR (ch = 'T') OR (ch = 'U') OR (ch = 'X')
               OR (ch = 'Q') OR (ch = '+') OR (ch = '-') OR (ch = '.')
               OR (ch = 'W') OR (ch = 'H') OR (ch = '?')
            THEN CASE ch  OF
               'C' : clade;
               'F' : flip;
               'O' : changeoutgroup;
               'R' : rearrange;
               'S' : show;
               '+' : nextchar;
               '-' : prevchar;
               '.' : again;
               'T' : try;
               'U' : undo;
               'W' : treewrite;
               'Q' : done := TRUE;
               'X' : done := TRUE;
               'H' : help;
               '?' : help;
               END;
         UNTIL done;
         IF NOT written
         THEN BEGIN
            WRITE(' Do you want to write out the tree to');
            WRITE(' a file? (Y or N) ');
            READLN(ch);
            IF (ch = 'Y') OR (ch = 'y')
            THEN treewrite;
            END;
      END; (* redisplay *)

   BEGIN (* treeconstruct *)
      subtree := FALSE;
      display := FALSE;
      dispchar := 0;
      buildtree;
      lastop := none;
      redisplay;
   END; (* treeconstruct *)

BEGIN (* Interactive mixed parsimony *)
   (* reads in spp, chars, and the data. Then calls treeconstruct to
      construct the tree and query the user *)
   ASSIGN(infile, 'INFILE');  (* Turbo Pascal only *)
   RESET(infile);
   (* put any ASSIGN statement for treefile here *)
   ASSIGN(treefile, 'TREEFILE');  (* Turbo Pascal only *)
   configure;
   logo;
   doinput;
   treeconstruct;
(* CLOSE(infile);  -- may need this here *)
END. (* Interactive mixed parsimony *)

