PROGRAM DNAPenny (infile, treefile, OUTPUT);
  (* version 3.1. (c) Copyright 1986, 1987, 1988 by the University of
     Washington and Joseph Felsenstein.  Written by Joseph Felsenstein.
     Permission is granted to copy and use this program provided no
     fee is charged for it and provided that this copyright
     notice is not removed. *)

CONST maxsp = 20;    (* maximum number of species               *)
   maxsz = 39;       (* size of pointer array.  >= 2*maxsp - 1  *)
   maxsites = 300;   (* max. number base sites per sequence     *)
   nmlngth = 10;     (* number of characters in species name    *)
   maxtrees = 100;   (* maximum number of trees to be printed out *)
   howoften = 100;   (* how often to notify how many trees examined *)
   howmany = 300;    (* how many multiples of howoften before stop *)

TYPE steparray = ARRAY[1..maxsites] OF INTEGER;
   bases = (A, C, G, U, O);
   baset = SET OF bases;
   basearray = ARRAY [1..maxsites] OF baset;
   nodeptr = ^node;   (* nodes will form a binary tree *)
   node = RECORD      (* describes a tip species or an ancestor *)
      ancestor, leftdesc, rtdesc : nodeptr;
      index : INTEGER;
      tip : BOOLEAN;        (* present species are tips of tree *)
      base : basearray; (* the sequence *)
      xcoord, ycoord : INTEGER    (* used by printree *)
      END;
   pointarray = ARRAY [1..maxsz] OF nodeptr;

VAR root : nodeptr;
   infile, treefile : TEXT;
   spp, nonodes, chars, outgrno : INTEGER;
     (* spp = number of species
        nonodes = number of nodes in tree
        chars = number of sites in actual sequences
        outgrno indicates outgroup *)
   weights, thresh, simple, trout, outgropt : BOOLEAN;
   weight : steparray;
   treenode : pointarray;      (* pointers to all nodes in tree *)
   name : ARRAY[1..maxsp, 1..nmlngth] OF CHAR; (* names of species *)
   threshold, fracdone, fracinc : REAL;
   threshwt : ARRAY[1..maxsites] OF REAL;
   added : ARRAY [1..maxsz] OF BOOLEAN;
   basechar : ARRAY[1..31] OF CHAR;

   FUNCTION digit (ch : CHAR) : BOOLEAN;
   BEGIN (* digit *)
      digit := (ch >= '0') AND (ch <= '9');
   END; (* digit *)

   FUNCTION letter (ch : CHAR) : BOOLEAN;
      (* tests whether ch is a letter
         -- works for both ASCII and EBCDIC codes *)
   BEGIN (* letter *)
      letter := ((ch >= 'A') AND (ch <= 'I'))
                OR ((ch >= 'J') AND (ch <= 'R'))
                OR ((ch >= 'S') AND (ch <= 'Z'))
                OR ((ch >= 'a') AND (ch <= 'i'))
                OR ((ch >= 'j') AND (ch <= 'r'))
                OR ((ch >= 's') AND (ch <= 'z'));
   END; (* letter *)

   PROCEDURE uppercase (VAR ch : CHAR);
      (* convert ch to upper case -- either ASCII or EBCDIC *)
   BEGIN (* uppercase *)
      IF letter(ch) AND (('a' > 'A') AND (ch >= 'a')
                     OR  ('a' < 'A') AND (ch < 'A'))
      THEN ch := CHR(ORD(ch) + ORD('A') - ORD('a'));
   END; (* uppercase *)

   PROCEDURE newline (i, j, k : INTEGER);
      (* go to new line if i is a multiple of j, indent k spaces *)
      VAR m : INTEGER;
   BEGIN (* newline *)
      IF (((i-1) MOD j) = 0) AND (i>1)
      THEN BEGIN
         WRITELN(OUTPUT);
         FOR m := 1 TO k DO WRITE(OUTPUT, ' ');
         END;
   END; (* newline *)

   PROCEDURE charbase;
      (* make ARRAY basechar to hold characters for base sets using a
         binary coding with O as most significant position, A least *)
      VAR i : INTEGER;
   BEGIN (* charbase *)
      basechar[1] := 'A';
      basechar[2] := 'C';
      basechar[3] := 'M';
      basechar[4] := 'G';
      basechar[5] := 'Y';
      basechar[6] := 'S';
      basechar[7] := 'V';
      basechar[8] := 'T';
      basechar[9] := 'W';
      basechar[10] := 'R';
      basechar[11] := 'H';
      basechar[12] := 'K';
      basechar[13] := 'D';
      basechar[14] := 'B';
      basechar[15] := 'N';
      basechar[16] := 'O';
      FOR i := 17 TO 31 DO
         basechar[i] := '?';
   END; (* charbase *)

   PROCEDURE doinput;
      (* reads the input data *)

      PROCEDURE inputnumbers;
         (* input the numbers of species and of characters *)
      BEGIN (* inputnumbers *)
         WRITELN(OUTPUT);
         WRITELN(OUTPUT, ' Penny algorithm for DNA, version 3.1');
         WRITE(OUTPUT, '  branch-and-bound to find all ');
         WRITELN(OUTPUT, 'most parsimonious trees');
         WRITELN(OUTPUT);
         READ(infile, spp, chars);
         WRITELN(OUTPUT, ' ',spp:2,' species, ',chars:3,'  sites');
         IF spp > maxsp
         THEN WRITELN(OUTPUT, ' TOO MANY SPECIES: adjust CONSTants');
         IF chars > maxsites
         THEN WRITELN(OUTPUT, ' TOO MANY SITES: adjust CONSTants');
         WRITELN(OUTPUT);
         nonodes := 2 * spp - 1;
      END; (* inputnumbers *)

      PROCEDURE inputoptions;
         (* input the information on the options *)
         VAR ch : CHAR;
            extranum, i : INTEGER;

         PROCEDURE inputhresh;
            (* input the threshold *)
         BEGIN (* inputhresh *)
            READLN(infile, threshold);
            thresh := TRUE;
         END; (* inputhresh *)

         PROCEDURE printhresh;
            (* print out the threshold value *)
         BEGIN (* printhresh *)
            WRITELN(OUTPUT, '    Threshold value = ', threshold:10:3);
            WRITELN(OUTPUT);
         END; (* printhresh *)

         PROCEDURE inputweights;
            (* input the character weights, 0-9 and A-Z for weights
               0 - 35 *)
            VAR ch : CHAR;
               i : INTEGER;
         BEGIN (* inputweights *)
            FOR i := 1 TO nmlngth-1 DO READ(infile, ch);
            FOR i := 1 TO chars DO BEGIN
               REPEAT
                  IF EOLN(infile) THEN READLN(infile);
                  READ(infile, ch);
               UNTIL ch <> ' ';
               weight[i] := 1;
               IF digit(ch)
               THEN weight[i] := ORD(ch) - ORD('0')
               ELSE IF letter(ch)
                  THEN BEGIN
                     uppercase (ch);
                     IF (ch >= 'A') AND (ch <= 'I')
                     THEN weight[i] := ORD(ch) - ORD('A') + 10
                     ELSE IF (ch >= 'J') AND (ch <= 'R')
                        THEN weight[i] := ORD(ch) - ORD('J') + 19
                        ELSE weight[i] := ORD(ch) - ORD('S') + 28;
                     END
                  ELSE WRITELN(OUTPUT, ' Bad weight character: ',ch);
               END;
            READLN(infile);
            weights := TRUE;
         END; (* inputweights *)

         PROCEDURE printweights;
            (* print out the weights of sites *)
            VAR i, j, k : INTEGER;
         BEGIN (* printweights *)
            WRITELN(OUTPUT,'    Sites are weighted as follows:');
            WRITE(OUTPUT, '        ');
            FOR i := 0 TO 9 DO WRITE(OUTPUT, i:3);
            WRITELN(OUTPUT);
            WRITELN(OUTPUT,'     *---------------------------------');
            FOR j := 0 TO (chars DIV 10) DO BEGIN
               WRITE(OUTPUT, 10*j:5, '!  ');
               FOR i := 0 TO 9 DO BEGIN
                  k := 10*j + i;
                  IF (k>0) AND (k<=chars)
                  THEN WRITE(OUTPUT, weight[k]:3)
                  ELSE WRITE(OUTPUT, '   ');
                  END;
               WRITELN(OUTPUT);
               END;
            WRITELN(OUTPUT);
         END; (* printweights *)

      BEGIN (* inputoptions *)
         outgrno := 1;
         outgropt := FALSE;
         simple := TRUE;
         thresh := FALSE;
         trout := FALSE;
         weights := FALSE;
         extranum := 0;
         WHILE NOT EOLN(infile) DO BEGIN
            READ(infile, ch);
            uppercase (ch);
            IF (ch = 'O') OR (ch = 'R')
               OR (ch = 'T') OR (ch = 'W') OR (ch = 'Y')
            THEN CASE ch OF
                  'O' : BEGIN
                        outgropt := TRUE;
                        extranum := extranum + 1;
                        END;
                  'R' : simple := FALSE;
                  'T' : extranum := extranum + 1;
                  'W' : extranum := extranum + 1;
                  'Y' : trout := TRUE
                  END
            ELSE IF ch <> ' '
               THEN WRITELN(OUTPUT, ' Bad Option Character: ', ch);
            END;
         READLN(infile);
         threshold := spp;
         FOR i := 1 TO chars DO weight[i] := 1;
         FOR i := 1 TO extranum DO BEGIN
            READ(infile, ch);
            uppercase (ch);
            IF ch = 'O'
            THEN READLN(infile, outgrno);
            IF ch = 'T'
            THEN inputhresh;
            IF ch = 'W'
            THEN inputweights;
            END;
         IF thresh
         THEN printhresh;
         IF weights
         THEN printweights
         ELSE FOR i := 1 TO chars DO
            threshwt[i] := threshold * weight[i];
      END; (* inputoptions *)

      PROCEDURE inputdata;
         (* input the names and sequences for each species *)
         VAR i, j : INTEGER;
            ns : baset; (* temporary base set for input *)
            charstate : CHAR;  (*  character temporary storage for
                                   input *)
      BEGIN (* inputdata *)
         WRITELN(OUTPUT);
         j := nmlngth - 5 + chars DIV 2;
         IF j < nmlngth - 1 THEN j := nmlngth - 1;
         IF j > 39 THEN j := 39;
         WRITE(OUTPUT, ' Name');
         FOR i := 1 TO j DO WRITE(OUTPUT, ' ');
         WRITELN(OUTPUT, 'Sequences');
         WRITE(OUTPUT, ' ----');
         FOR i := 1 TO j DO WRITE(OUTPUT, ' ');
         WRITELN(OUTPUT, '---------');
         WRITELN(OUTPUT);
         FOR i := 1 TO nonodes DO BEGIN
            NEW (treenode[i]);
            treenode[i]^.leftdesc := NIL;
            treenode[i]^.rtdesc := NIL;
            treenode[i]^.ancestor := NIL;
            treenode[i]^.index := i;
            IF i <= spp
            THEN BEGIN
               treenode[i]^.tip := TRUE;
               FOR j := 1 TO nmlngth DO READ(infile, name[i,j]);
               WRITE(OUTPUT, ' ');
               FOR j := 1 TO nmlngth DO WRITE(OUTPUT, name[i,j]);
               WRITE(OUTPUT, '   ');
               FOR j := 1 TO chars DO BEGIN
                  REPEAT
                     IF EOLN(infile) THEN READLN(infile);
                     READ(infile, charstate);
                  UNTIL (charstate <> ' ') AND
                     ((charstate < '0') OR (charstate > '9'));
                  uppercase (charstate);
                  IF NOT ((charstate = 'A') OR (charstate = 'B')
                         OR (charstate = 'C') OR (charstate = 'D')
                         OR (charstate = 'G') OR (charstate = 'H')
                         OR (charstate = 'K') OR (charstate = 'M')
                         OR (charstate = 'N') OR (charstate = 'R')
                         OR (charstate = 'S') OR (charstate = 'T')
                         OR (charstate = 'U') OR (charstate = 'V')
                         OR (charstate = 'W') OR (charstate = 'X')
                         OR (charstate = 'Y') OR (charstate = '?')
                         OR (charstate = 'O') OR (charstate = '-')
                         OR (charstate = '.'))
                  THEN WRITELN(OUTPUT, ' Warning -- bad base ', charstate,
                        ' at position ', j:5, ' of species ', i:3);
                  newline (j, 60, nmlngth+4);
                  WRITE(OUTPUT, charstate);
                  IF ((j MOD 10) = 0) AND ((j MOD 60) <> 0)
                  THEN WRITE(OUTPUT, ' ');
                  CASE charstate OF
                     'A' : ns := [A];
                     'C' : ns := [C];
                     'G' : ns := [G];
                     'U' : ns := [U];
                     'T' : ns := [U];
                     'M' : ns := [A, C];
                     'R' : ns := [A, G];
                     'W' : ns := [A, U];
                     'S' : ns := [C, G];
                     'Y' : ns := [C, U];
                     'K' : ns := [G, U];
                     'B' : ns := [C, G, U];
                     'D' : ns := [A, G, U];
                     'H' : ns := [A, C, U];
                     'V' : ns := [A, C, G];
                     'N' : ns := [A, C, G, U];
                     'X' : ns := [A, C, G, U];
                     '?' : ns := [A, C, G, U];
                     'O' : ns := [O];
                     '-' : ns := [O];
                     '.' : ns := [O]
                     END;
                  treenode[i]^.base[j] := ns;
                  END;
               READLN(infile);
               WRITELN(OUTPUT);
               END
            ELSE treenode[i]^.tip := FALSE;
            END;
         WRITELN(OUTPUT);
         WRITELN(OUTPUT);
      END; (* inputdata *)

   BEGIN (* doinput *)
      inputnumbers;
      inputoptions;
      inputdata;
   END; (* doinput *)

   PROCEDURE maketree;
      (* tree construction recursively by branch and bound *)
      TYPE treenumbers = ARRAY[1..maxsp] OF INTEGER;
      VAR i, j, k, nextree, examined, mults : INTEGER;
         firsttime, done : BOOLEAN;
         like, bestyet : REAL;
         suppset : ARRAY[1..31] OF baset;
         suppno : ARRAY[1..31] OF integer;
         numsteps : steparray;
         bestorders, bestrees : ARRAY[1..maxtrees] OF treenumbers;
         current, order : treenumbers;

      PROCEDURE makesup;
         (* make up table for use by PROCEDURE supplement *)
         VAR i : INTEGER;
      BEGIN (* makesup *)
         suppset[1] := [A];
         suppset[2] := [C];
         suppset[3] := [G];
         suppset[4] := [U];
         suppset[5] := [O];
         suppset[6] := [A, C];
         suppset[7] := [A, G];
         suppset[8] := [A, U];
         suppset[9] := [A, O];
         suppset[10] := [C, G];
         suppset[11] := [C, U];
         suppset[12] := [C, O];
         suppset[13] := [G, U];
         suppset[14] := [G, O];
         suppset[15] := [U, O];
         suppset[16] := [A, C, G];
         suppset[17] := [A, C, U];
         suppset[18] := [A, C, O];
         suppset[19] := [A, G, U];
         suppset[20] := [A, G, O];
         suppset[21] := [A, U, O];
         suppset[22] := [C, G, U];
         suppset[23] := [C, G, O];
         suppset[24] := [C, U, O];
         suppset[25] := [G, U, O];
         suppset[26] := [A, C, G, U];
         suppset[27] := [A, C, G, O];
         suppset[28] := [A, C, U, O];
         suppset[29] := [A, G, U, O];
         suppset[30] := [C, G, U, O];
         suppset[31] := [A, C, G, U, O];
         FOR i := 1 TO 5 DO
            suppno[i] := 0;
         FOR i := 6 TO 15 DO
            suppno[i] := 1;
         FOR i := 16 TO 25 DO
            suppno[i] := 2;
         FOR i := 26 TO 30 DO
            suppno[i] := 3;
         suppno[31] := 4;
      END; (* makesup *)

      PROCEDURE add (VAR below, newtip, newfork : nodeptr);
         (* inserts the nodes newfork and its left descendant, newtip,
            to the tree.  below becomes newfork's right descendant.
            The global variable root is also updated *)
      BEGIN (* add *)
         newfork^.ancestor := below^.ancestor;
         newfork^.rtdesc := below;
         newfork^.leftdesc := newtip;
         below^.ancestor := newfork;
         newtip^.ancestor := newfork;
         IF newfork^.ancestor <> NIL
         THEN IF newfork^.ancestor^.leftdesc = below
            THEN newfork^.ancestor^.leftdesc := newfork
            ELSE newfork^.ancestor^.rtdesc := newfork;
         IF root = below THEN root := newfork;
      END; (* add *)

      PROCEDURE remove (VAR item, fork : nodeptr);
         (* removes nodes item and its ancestor, fork, from the tree.
            the new descendant of fork's ancestor is made to be
            fork's second descendant (other than item).  Also
            returns pointers to the deleted nodes, item and fork.
            The global variable root is also updated *)
         VAR other, below : nodeptr;
      BEGIN (* remove *)
         IF item^.ancestor = NIL
         THEN fork := NIL
         ELSE BEGIN
            fork := item^.ancestor;
            IF item = fork^.leftdesc
            THEN other := fork^.rtdesc
            ELSE other := fork^.leftdesc;
            below := fork^.ancestor;
            other^.ancestor := below;
            IF below <> NIL
            THEN BEGIN
               IF below^.leftdesc = fork
               THEN below^.leftdesc := other
               ELSE below^.rtdesc := other;
               fork^.ancestor := NIL;
               END;
            fork^.leftdesc := NIL;
            fork^.rtdesc := NIL;
            item^.ancestor := NIL;
            END;
         IF root = fork THEN root := other;
      END; (* remove *)

      PROCEDURE evaluate (r: nodeptr);
         (* determines the number of steps needed for a tree. this is
            the minimum number of steps needed to evolve sequences on
            this tree *)
         VAR i, steps : INTEGER;
            sum : REAL;

         PROCEDURE postorder (p : nodeptr);
            (* traverses a binary tree, calling PROCEDURE fillin at a
               node's descendants before calling fillin at the node *)

            PROCEDURE fillin (p: nodeptr);
               (* sets up for each node in the tree the base sequence
                  at that point and counts the changes.  The program
                  spends much of its time in this PROCEDURE *)
               VAR i : INTEGER;
                  ns, rs, ls : baset;
            BEGIN (* fillin *)
               FOR i := 1 TO chars DO BEGIN
                  ls := p^.leftdesc^.base[i];
                  rs := p^.rtdesc^.base[i];
                  ns := ls * rs;
                  IF ns = []
                  THEN BEGIN
                     ns := ls + rs;
                     numsteps[i] := numsteps[i] + weight[i];
                     END;
                  p^.base[i] := ns;
                  END;
            END; (* fillin *)

         BEGIN (* postorder *)
            IF NOT p^.tip
            THEN BEGIN
               postorder (p^.leftdesc);
               postorder (p^.rtdesc);
               fillin (p);
               END;
         END; (* postorder *)

         PROCEDURE supplement;
            (* determine minimum number of steps more which will
               be added when rest of species are put in tree *)
            VAR i, j, k, sum, sumall, sumadded : INTEGER;
               doneadded, allhave, addedhave, has : BOOLEAN;
         BEGIN (* supplement *)
            FOR i := 1 TO chars DO BEGIN
               sum := 3;
               j := 1;
               doneadded := FALSE;
               REPEAT
                  allhave := TRUE;
                  addedhave := TRUE;
                  FOR k := 1 TO spp DO BEGIN
                     has := ((treenode[k]^.base[i]) * suppset[j]) <> [];
                     IF added[k] AND (NOT doneadded)
                     THEN addedhave := addedhave AND has;
                     allhave := allhave AND has;
                     END;
                  IF allhave
                  THEN sumall := suppno[j];
                  IF addedhave
                  THEN sumadded := suppno[j];
                  doneadded := doneadded OR addedhave;
                  j := j + 1;
               UNTIL (j = 31) OR (allhave AND doneadded);
               IF addedhave AND allhave
               THEN sum := sumall - sumadded;
               numsteps[i] := numsteps[i] + sum;
               END;
         END; (* supplement *)

      BEGIN (* evaluate *)
         sum := 0;
         FOR i := 1 TO chars DO
            numsteps[i] := 0;
         postorder (r);
         supplement;
         FOR i := 1 TO chars DO BEGIN
            steps := numsteps[i];
            IF steps <= threshwt[i]
            THEN sum := sum + steps
            ELSE sum := sum + threshwt[i];
            END;
         IF (examined = 0) AND (mults = 0) THEN bestyet := -1.0;
         like := sum;
      END; (* evaluate *)

      PROCEDURE addit (m : INTEGER);
         (* adds the species one by one, recursively *)
         TYPE valarray = ARRAY[1..maxsz] OF REAL;
            placearray = ARRAY[1..maxsz] OF INTEGER;
         VAR i, j, n, n1, besttoadd : INTEGER;
            valyew, bestval : valarray;
            place, bestplace : placearray;
            oldfrac, oldfdone, sum, bestsum : REAL;

         PROCEDURE addtraverse (a, b, c : nodeptr);
            (* traverse all places to add b *)
         BEGIN (* addtraverse *)
            IF NOT done
            THEN BEGIN
               IF (m <= 2) OR NOT((a = root) OR (a = root^.leftdesc))
               THEN BEGIN
                  add (a, b, c);
                  n := n + 1;
                  evaluate (root);
                  examined := examined + 1;
                  IF examined = howoften
                  THEN BEGIN
                     examined := 0;
                     mults := mults + 1;
                     IF mults = howmany
                     THEN done := TRUE;
                     WRITE(OUTPUT, ' ', mults:7);
                     IF bestyet >= 0
                     THEN WRITE(OUTPUT,  bestyet:18:5)
                     ELSE WRITE(OUTPUT, '         -        ');
                     WRITELN(OUTPUT, nextree-1:15, fracdone*100:20:2);
                     END;
                  valyew[n] := like;
                  place[n] := a^.index;
                  remove (b, c);
                  END;
               IF NOT a^.tip
               THEN BEGIN
                  addtraverse (a^.leftdesc, b, c);
                  addtraverse (a^.rtdesc, b, c);
                  END;
               END;
         END; (* addtraverse *)

         PROCEDURE shellsort (VAR a : valarray; VAR b : placearray;
                                n : INTEGER);
            (* Shell sort keeping a, b in same order *)
            VAR gap, i, j, itemp : INTEGER;
              rtemp : REAL;
         BEGIN (* shellsort *)
            gap := n DIV 2;
            WHILE (gap > 0) DO BEGIN
               FOR i := gap TO n-1 DO BEGIN
                  j := i - gap + 1;
                  WHILE (j>0) DO BEGIN
                     IF a[j] > a[j+gap]
                     THEN BEGIN
                        rtemp := a[j];
                        a[j] := a[j+gap];
                        a[j+gap] := rtemp;
                        itemp := b[j];
                        b[j] := b[j+gap];
                        b[j+gap] := itemp;
                        END;
                     j := j - gap;
                     END;
                  END;
               gap := gap DIV 2;
               END;
         END; (* shellsort *)

      BEGIN (* addit *)
         IF simple AND (NOT firsttime)
         THEN BEGIN
            n := 0;
            added[order[m]] := TRUE;
            addtraverse (root, treenode[order[m]], treenode[spp+m-1]);
            besttoadd := order[m];
            bestplace := place;
            bestval := valyew;
            END
         ELSE BEGIN
            bestsum := -1;
            FOR i := 1 TO spp DO
               IF NOT added[i]
               THEN BEGIN
                  n := 0;
                  added[i] := TRUE;
                  addtraverse (root, treenode[i], treenode[spp+m-1]);
                  added[i] := FALSE;
                  sum := 0.0;
                  FOR j := 1 TO n DO
                     sum := sum + valyew[j];
                  IF sum > bestsum
                  THEN BEGIN
                     bestsum := sum;
                     besttoadd := i;
                     bestplace := place;
                     bestval := valyew;
                     END;
                  END;
            END;
         order[m] := besttoadd;
         place := bestplace;
         valyew := bestval;
         shellsort (valyew, place, n);
         oldfrac := fracinc;
         oldfdone := fracdone;
         n1 := 0;
         FOR i := 1 TO n DO
            IF (valyew[i] <= bestyet) OR (bestyet < 0.0)
            THEN n1 := n1 + 1;
         IF n1 > 0 THEN fracinc := fracinc / n1;
         FOR i := 1 TO n DO BEGIN
            IF (valyew[i] <= bestyet) OR (bestyet < 0.0)
            THEN BEGIN
               current[m] := place[i];
               add(treenode[place[i]], treenode[besttoadd],
                                         treenode[spp+m-1]);
               added[besttoadd] := TRUE;
               IF m < spp
               THEN addit (m+1)
               ELSE BEGIN
                  IF (valyew[i] < bestyet) OR (bestyet < 0.0)
                  THEN BEGIN
                     nextree := 1;
                     bestyet := valyew[i];
                     END;
                  IF nextree <= maxtrees
                  THEN BEGIN
                     bestorders[nextree] := order;
                     bestrees[nextree] := current;
                     END;
                  nextree := nextree + 1;
                  firsttime := FALSE;
                  END;
               remove (treenode[besttoadd], treenode[spp+m-1]);
               added[besttoadd] := FALSE;
               END;
            fracdone := fracdone + fracinc;
            END;
         fracinc := oldfrac;
         fracdone := oldfdone;
      END; (* addit *)

      PROCEDURE reroot (outgroup : nodeptr);
         (* reorients tree, putting outgroup in desired position.
            Written by Kent Fiala *)
         VAR vertex, oldvertex, ante, otherbranch : nodeptr;
            left : BOOLEAN;
      BEGIN (* reroot *)
         IF NOT (outgroup^.ancestor = root)
         THEN BEGIN
            oldvertex := outgroup;
            vertex := outgroup^.ancestor;
            ante := vertex^.ancestor;
            IF vertex^.rtdesc = outgroup
            THEN BEGIN
               vertex^.rtdesc := vertex^.leftdesc;
               vertex^.leftdesc := outgroup;
               END;
            REPEAT
               IF oldvertex = vertex^.leftdesc
               THEN BEGIN
                  vertex^.leftdesc := ante;
                  left := TRUE;
                  END
               ELSE BEGIN
                  vertex^.rtdesc := ante;
                  left := FALSE;
                  END;
               vertex^.ancestor := oldvertex;
               oldvertex := vertex;
               vertex := ante;
               ante := vertex^.ancestor
            UNTIL vertex = root;
            IF root^.leftdesc = oldvertex
            THEN otherbranch := root^.rtdesc
            ELSE otherbranch := root^.leftdesc;
            IF left
            THEN oldvertex^.leftdesc := otherbranch
            ELSE oldvertex^.rtdesc := otherbranch;
            otherbranch^.ancestor := oldvertex;
            root^.rtdesc := outgroup^.ancestor;
            root^.leftdesc := outgroup;
            outgroup^.ancestor^.ancestor := root;
            outgroup^.ancestor := root;
            END;
      END; (* reroot *)

      PROCEDURE printree;
         (* prints out diagram of the tree *)
         CONST down = 2;
            over = 3;
         VAR tipx, tipy, i, rover : INTEGER;

         PROCEDURE coordinates (p : nodeptr);
            (* establishes coordinates of nodes *)
         BEGIN (* coordinates *)
            IF p^.tip
            THEN BEGIN
               p^.xcoord := tipx;
               p^.ycoord := tipy;
               tipx := tipx + rover;
               tipy := tipy + down;
               END
            ELSE BEGIN
               coordinates (p^.leftdesc);
               p^.xcoord := p^.leftdesc^.xcoord;
               coordinates (p^.rtdesc);
               p^.ycoord := p^.rtdesc^.ycoord;
               END;
         END; (* coordinates *)

         PROCEDURE drawline (i : INTEGER);
            (* draws one row of the tree diagram by moving up tree *)
            VAR p, q : nodeptr;
               n, j : INTEGER;
         BEGIN (* drawline *)
            p := root;
            WRITE(OUTPUT, '   ');
            REPEAT
               IF p^.leftdesc^.ycoord >= i
               THEN p := p^.leftdesc
               ELSE BEGIN
                  q := p^.rtdesc;
                  n := q^.xcoord - p^.xcoord;
                  IF p^.ycoord = i
                  THEN BEGIN
                     IF rover < over
                     THEN BEGIN
                        IF (i = (tipy-down)) AND (p^.xcoord = 0)
                        THEN WRITE(OUTPUT, 'L')
                        ELSE WRITE(OUTPUT, '*');
                        n := n + 1;
                        END
                     ELSE BEGIN
                        IF p^.index - spp >= 10
                        THEN WRITE(OUTPUT, (p^.index - spp):2)
                        ELSE WRITE(OUTPUT, (p^.index - spp):1, '-');
                        END;
                     FOR j := 1 TO n-2 DO WRITE(OUTPUT, '-');
                     END
                  ELSE BEGIN
                     WRITE(OUTPUT, '!');
                     FOR j := 1 TO n-1 DO WRITE(OUTPUT, ' ');
                     END;
                  p := q;
                  END;
            UNTIL p^.tip;
            IF p^.ycoord = i
            THEN FOR j := 1 TO nmlngth DO
                       WRITE(OUTPUT, name[p^.index, j]);
            WRITELN(OUTPUT);
         END; (* drawline *)

      BEGIN (* printree *)
         tipx := 0;
         tipy := 1;
         rover := 100 DIV spp;
         IF rover > over THEN rover := over;
         coordinates (root);
         WRITELN(OUTPUT);
         WRITELN(OUTPUT);
         FOR i := 1 TO tipy-down DO drawline (i);
         WRITELN(OUTPUT);
         WRITE(OUTPUT, '   remember:');
         IF outgropt
         THEN WRITE(OUTPUT, ' (although rooted by outgroup)');
         WRITELN(OUTPUT, ' this is an unrooted tree!');
         WRITELN(OUTPUT);
      END; (* printree *)

      PROCEDURE describe;
         (* prints ancestors, steps and table of numbers of steps in
            each site *)
         VAR i, j, k : INTEGER;

         PROCEDURE hypstates;
            (* fill in and describe states at interior nodes *)
            VAR i : INTEGER;
               nothing : basearray;
               bottom : BOOLEAN;

            PROCEDURE ancestset (a, b : baset; VAR c : baset);
               (* make the set of ancestral states below nodes
                  whose base sets are a and b *)
            BEGIN (* ancestset *)
               c := a * b;
               IF c = []
               THEN c := a + b;
            END; (* ancestset *)

            PROCEDURE hyptrav (r : nodeptr; VAR hypset : basearray);
               (*  compute, print out states at one interior node *)
               VAR i : INTEGER;
                  maybe, nonzero : BOOLEAN;
                  tempset, left, rt, anc : baset;
                  temparray, ancset : basearray;

               PROCEDURE hyprint;
                  (* print out states at node *)
                  VAR i, k, n : INTEGER;
                     dot : BOOLEAN;
                     b : bases;
               BEGIN (* hyprint *)
                  IF bottom
                  THEN BEGIN
                     IF NOT outgropt
                     THEN WRITE(OUTPUT, '        ')
                     ELSE WRITE(OUTPUT, ' root   ');
                     END
                  ELSE WRITE(OUTPUT,
                       (r^.ancestor^.index - spp):4, '    ');
                  IF r^.tip
                  THEN FOR i := 1 TO nmlngth DO
                     WRITE(OUTPUT, name[r^.index, i])
                  ELSE WRITE(OUTPUT, (r^.index - spp):4, '      ');
                  IF bottom
                  THEN WRITE(OUTPUT, '          ')
                  ELSE IF nonzero
                     THEN WRITE(OUTPUT, '   yes    ')
                     ELSE IF maybe
                        THEN WRITE(OUTPUT, '  maybe   ')
                        ELSE WRITE(OUTPUT, '   no     ');
                  FOR i := 1 TO chars DO BEGIN
                     newline (i, 40, nmlngth+18);
                     tempset := r^.base[i];
                     anc := hypset[i];
                     IF NOT bottom
                     THEN anc := r^.ancestor^.base[i];
                     dot := (tempset = anc) AND (NOT bottom);
                     IF dot
                     THEN WRITE(OUTPUT, '.')
                     ELSE IF tempset = [A]
                        THEN WRITE(OUTPUT, 'A')
                        ELSE IF tempset = [C]
                           THEN WRITE(OUTPUT, 'C')
                           ELSE IF tempset = [G]
                              THEN WRITE(OUTPUT, 'G')
                              ELSE IF tempset = [U]
                                 THEN WRITE(OUTPUT, 'T')
                                 ELSE IF tempset = [O]
                                    THEN WRITE(OUTPUT, '-')
                                    ELSE BEGIN
                                       k := 1;
                                       n := 0;
                                       FOR b := A TO O DO BEGIN
                                          IF b IN tempset
                                          THEN n := n + k;
                                          k := k + k;
                                          END;
                                       WRITE(OUTPUT, basechar[n]);
                                       END;
                     IF (i MOD 10) = 0
                     THEN WRITE(OUTPUT, ' ');
                     END;
                  WRITELN(OUTPUT);
               END; (* hyprint *)

            BEGIN (* hyptrav *)
               maybe := FALSE;
               nonzero := FALSE;
               FOR i := 1 TO chars DO BEGIN
                  anc := hypset[i];
                  IF NOT r^.tip
                  THEN BEGIN
                     left := r^.leftdesc^.base[i];
                     rt := r^.rtdesc^.base[i];
                     tempset := left * rt * anc;
                     IF tempset = []
                     THEN BEGIN
                        tempset := (left * rt) + (left * anc)
                                      + (rt * anc);
                        IF tempset = []
                        THEN tempset := left + rt + anc;
                        END;
                     r^.base[i] := tempset;
                     END;
                  IF NOT bottom
                  THEN anc := r^.ancestor^.base[i];
                  nonzero := nonzero OR ((r^.base[i] * anc) = []);
                  maybe := maybe OR (r^.base[i] <> anc);
                  END;
               hyprint;
               bottom := FALSE;
               IF NOT r^.tip
               THEN BEGIN
                  temparray := r^.leftdesc^.base;
                  FOR i := 1 TO chars DO
                     ancestset(hypset[i], r^.rtdesc^.base[i],
                        ancset[i]);
                  hyptrav (r^.leftdesc, ancset);
                  FOR i := 1 TO chars DO
                     ancestset(hypset[i], temparray[i],
                        ancset[i]);
                  hyptrav (r^.rtdesc, ancset);
                  END;
            END; (* hyptrav *)

         BEGIN (* hypstates *)
            WRITELN(OUTPUT);
            WRITE(OUTPUT, ' From    To     Any Steps?    ');
            WRITELN(OUTPUT, 'State at upper node');
            WRITE(OUTPUT, '                             ');
            WRITELN(OUTPUT, ' ( . means same as in the node',
                                    ' below it on tree)');
            WRITELN(OUTPUT);
            FOR i := 1 TO chars DO nothing[i] := [];
            bottom := TRUE;
            hyptrav (root, nothing);
         END; (* hypstates *)

         PROCEDURE treeout (p : nodeptr);
            (* write out file with representation of final tree *)
            VAR i, n : INTEGER;
               c : CHAR;
         BEGIN (* treeout *)
            IF p^.tip
            THEN BEGIN
               n := 0;
               FOR i := 1 TO nmlngth DO
                  IF name[p^.index, i] <> ' '
                  THEN n := i;
               FOR i := 1 TO n DO BEGIN
                  c := name[p^.index, i];
                  IF c = ' '
                  THEN c := '_';
                  WRITE(treefile, c);
                  END;
               END
            ELSE BEGIN
               WRITE(treefile, '(');
               treeout (p^.leftdesc);
               WRITE(treefile, ',');
               treeout (p^.rtdesc);
               WRITE(treefile, ')');
               END;
            IF p = root
            THEN WRITELN(treefile, ';');
         END; (* treeout *)

      BEGIN (* describe *)
         IF weights THEN WRITE(OUTPUT, ' weighted');
         WRITELN(OUTPUT, ' steps in each site:');
         WRITE(OUTPUT, '       ');
         FOR i := 0 TO 9 DO WRITE(OUTPUT, i:4);
         WRITELN(OUTPUT);
         WRITE(OUTPUT, '      *------------------------------------');
         WRITELN(OUTPUT, '-----');
         FOR i := 0 TO (chars DIV 10) DO BEGIN
            WRITE(OUTPUT, 10*i:6);
            WRITE(OUTPUT, '!');
            FOR j := 0 TO 9 DO BEGIN
               k := 10*i + j;
               IF (k = 0) OR (k > chars)
               THEN WRITE(OUTPUT, '    ')
               ELSE WRITE(OUTPUT, numsteps[k]:4);
               END;
            WRITELN(OUTPUT);
            END;
         hypstates;
         WRITELN(OUTPUT);
         WRITELN(OUTPUT);
         IF trout
         THEN treeout (root);
      END; (* describe *)

   BEGIN (* maketree *)
      charbase;
      WRITELN(OUTPUT, ' How many');
      WRITELN(OUTPUT, ' trees looked                ',
               '                       Approximate');
      WRITELN(OUTPUT, ' at so far      Length of    ',
               '    How many           percentage');
      WRITELN(OUTPUT, ' (multiples     shortest tree',
               '    trees this long    searched');
      WRITELN(OUTPUT, ' of ', howoften:4, '):', '      found so far ',
               '    found so far       so far');
      WRITELN(OUTPUT, ' ----------     ------------ ',
               '    ------------       ------------');
      done := FALSE;
      mults := 0;
      examined := 0;
      nextree := 1;
      root := treenode[1];
      firsttime := TRUE;
      FOR i := 1 TO spp DO
         added[i] := FALSE;
      added[1] := TRUE;
      order[1] := 1;
      k := 2;
      fracdone := 0.0;
      fracinc := 1.0;
      bestyet := -1;
      makesup;
      addit (k);
      IF done
      THEN BEGIN
         WRITELN(OUTPUT, ' Search broken off!  Not guaranteed to');
         WRITELN(OUTPUT, '  have found the most parsimonious');
         WRITELN(OUTPUT, '  trees, but here is what we found:');
         END;
      WRITELN(OUTPUT);
      WRITELN(OUTPUT, ' requires a total of ',bestyet:18:3);
      WRITELN(OUTPUT);
      IF (nextree = 2)
      THEN WRITELN(OUTPUT, ' One most parsimonious tree found:')
      ELSE WRITELN(OUTPUT, nextree-1:6, ' trees in all found');
      IF nextree > maxtrees + 1
      THEN BEGIN
         WRITELN(OUTPUT, ' here are the first', maxtrees:4, ' of them');
         nextree := maxtrees + 1;
         END;
      WRITELN(OUTPUT);
      FOR i := 1 TO spp DO
         added[i] := TRUE;
      FOR i := 1 TO nextree-1 DO BEGIN
         FOR j := k TO spp DO
            add(treenode[bestrees[i][j]], treenode[bestorders[i][j]],
                                           treenode[spp+j-1]);
         reroot (treenode[outgrno]);
         evaluate (root);
         printree;
         describe;
         FOR j := k TO spp DO
            remove (treenode[bestorders[i][j]], treenode[spp+j-1]);
         END;
      IF trout
      THEN BEGIN
         WRITELN(OUTPUT, ' Trees also written onto file');
         WRITELN(OUTPUT);
         END;
   END; (* maketree *)

BEGIN (* Penny's branch-and-bound method for DNA sequences *)
   (* Reads in the number of species, number of characters,
      options and data.  Then finds all most parsimonious trees *)
   (* ASSIGN(infile, 'INFILE');  -- Turbo Pascal only *)
   RESET(infile);
   doinput;
   IF trout
   THEN BEGIN
      (* ASSIGN(treefile, 'TREEFILE');  -- Turbo Pascal only *)
      REWRITE(treefile);
      END;
   maketree;
(* CLOSE(infile);  -- may need this here *)
   IF trout
   THEN BEGIN
      (* CLOSE(treefile);  -- may need this here *)
      END;
END. (* Penny's branch-and-bound method for DNA sequences *)

