PROGRAM DolMove (infile, treefile, INPUT, OUTPUT);
  (* version 3.1. (c) Copyright 1987 by the University of Washington.
     Written by Joseph Felsenstein.  Permission is granted to copy
     and use this program provided no fee is charged for it and
     provided that this copyright notice is not removed. *)

CONST maxsp = 50;    (* maximum number of species               *)
   maxsz = 100;      (* size of pointer array.  >= 2*maxsp - 1  *)
   maxchr = 200;     (* max. number characterstates per species *)
   bits = 30;        (* max. number of elements in a Pascal set *)
   wrds = 7;         (* wrds * bits >= maxchr                   *)
   nmlngth = 10;     (* number of characters in species name    *)
   screenlines = 24; (* number of lines on a screen *)
   ibmpc = TRUE;   (* is it an IBM PC-compatible? *)
   vt52 = FALSE;    (* is it a VT52-compatible terminal? *)
   ansi = FALSE;    (* is it an ANSI standard terminal? *)

TYPE steparray = ARRAY[1..maxchr] OF INTEGER;
   bitrange = 1..bits;
   bitset = SET OF bitrange;
   chartype = (horiz, vert, up, over, one, zero, quest, polym);
   bitarray = ARRAY[1..wrds] OF bitset;
   nodeptr = ^node;   (* nodes will form a binary tree *)
   node = RECORD      (* describes a tip species or an ancestor *)
      ancestor, leftdesc, rtdesc : nodeptr;  (* pointers to nodes *)
      index : INTEGER;      (* number of the node *)
      tip : BOOLEAN;        (* present species are tips of tree *)
      stateone, statezero : bitarray; (* see in PROCEDURE fillin *)
      xcoord, ycoord : INTEGER;   (* used by printree *)
      state : CHAR   (* stores state for printing *)
      END;
   pointarray = ARRAY[1..maxsz] OF nodeptr;

VAR root : nodeptr;
   infile, treefile : TEXT;
   spp, nonodes, chars, words, outgrno : INTEGER;
     (* spp = number of species
        nonodes = number of nodes in tree
        chars = number of binary characters
        words = number of words needed
                to represent characters of one organism
        outgrno indicates outgroup *)
   weights, thresh, ancvar, questions, dollo : BOOLEAN;
   extras, weight : steparray;
   ancone, anczero : ARRAY[1..maxchr] OF BOOLEAN;
   treenode : pointarray;      (* pointers to all nodes in tree *)
   name : ARRAY[1..maxsp, 1..nmlngth] OF CHAR; (* names of species *)
   threshold : REAL;
   threshwt : ARRAY[1..maxchr] OF REAL;
   ch : ARRAY [chartype] OF CHAR;
   reversed : ARRAY [chartype] OF BOOLEAN;
   graphic : ARRAY[chartype] OF BOOLEAN;

   FUNCTION digit (ch : CHAR) : BOOLEAN;
   BEGIN (* digit *)
      digit := (ch >= '0') AND (ch <= '9');
   END; (* digit *)

   FUNCTION letter (ch : CHAR) : BOOLEAN;
   BEGIN (* letter -- works for both ASCII and EBCDIC codes *)
      letter := ((ch >= 'A') AND (ch <= 'I'))
                OR ((ch >= 'J') AND (ch <= 'R'))
                OR ((ch >= 'S') AND (ch <= 'Z'))
                OR ((ch >= 'a') AND (ch <= 'i'))
                OR ((ch >= 'j') AND (ch <= 'r'))
                OR ((ch >= 's') AND (ch <= 'z'));
   END; (* letter *)

   PROCEDURE uppercase (VAR ch : CHAR);
      (* convert a character to upper case -- either ASCII or EBCDIC *)
   BEGIN (* uppercase *)
      IF letter(ch) AND (('a' > 'A') AND (ch >= 'a')
                     OR  ('a' < 'A') AND (ch < 'A'))
      THEN ch := CHR(ORD(ch) + ORD('A') - ORD('a'));
   END; (* uppercase *)

   PROCEDURE newline (i, j, k : INTEGER);
      (* go to new line if i is a multiple of j, indent k spaces *)
      VAR m : INTEGER;
   BEGIN (* newline *)
      IF (((i-1) MOD j) = 0) AND (i>1)
      THEN BEGIN
         WRITELN;
         FOR m := 1 TO k DO WRITE(' ');
         END;
   END; (* newline *)

   PROCEDURE configure;
      (* configure to machine -- set up special characters *)
      VAR a : chartype;
   BEGIN (* configure *)
      FOR a := horiz TO polym DO reversed[a] := FALSE;
      FOR a := horiz TO polym DO graphic[a] := FALSE;
      IF ibmpc
      THEN BEGIN
         ch[horiz] := CHR(205);
         graphic[horiz] := TRUE;
         ch[vert] := CHR(186);
         graphic[vert] := TRUE;
         ch[up] := CHR(186);
         graphic[up] := TRUE;
         ch[over] := CHR(205);
         graphic[over] := TRUE;
         ch[one] := CHR(219);
         reversed[one] := TRUE;
         ch[zero] := CHR(176);
         graphic[zero] := TRUE;
         ch[quest] := CHR(178);   (* or try CHR(177) *)
         graphic[quest] := TRUE;
         ch[polym] := CHR(1);
         reversed[polym] := TRUE;
         END
      ELSE IF vt52
         THEN BEGIN
            ch[one] := ' ';
            reversed[one] := TRUE;
            ch[horiz] := ch[one];
            reversed[horiz] := TRUE;
            ch[vert] := ch[one];
            reversed[vert] := TRUE;
            ch[up] := CHR(96);
            graphic[up] := TRUE;
            ch[over] := 'a';
            graphic[over] := TRUE;
            ch[zero] := 'i';
            graphic[zero] := TRUE;
            ch[quest] := '?';
            reversed[quest] := TRUE;
            ch[polym] := '%';
            reversed[polym] := TRUE;
            END
         ELSE IF ansi
            THEN BEGIN
               ch[one] := ' ';
               reversed[one] := TRUE;
               ch[horiz] := ch[one];
               reversed[horiz] := TRUE;
               ch[vert] := ch[one];
               reversed[vert] := TRUE;
               ch[up] := 'x';
               graphic[up] := TRUE;
               ch[over] := 'q';
               graphic[over] := TRUE;
               ch[zero] := 'a';
               graphic[zero] := TRUE;
               reversed[zero] := TRUE;
               ch[quest] := '?';
               reversed[quest] := TRUE;
               ch[polym] := '%';
               reversed[polym] := TRUE;
               END
            ELSE BEGIN
               ch[horiz] := '=';
               ch[vert] := ' ';
               ch[up] := '!';
               ch[over] := '-';
               ch[one] := '*';
               ch[zero] := '=';
               ch[quest] := '.';
               ch[polym] := '%';
               END;
   END; (* configure *)

   PROCEDURE prefix (a : chartype);
      (* give prefix appropriate for this character *)

      PROCEDURE pregraph;
         (* turn on graphic characters *)
      BEGIN (* pregraph *)
         IF vt52
         THEN WRITE(CHR(27), 'F');
         IF ansi
         THEN WRITE(CHR(27), '(0');
      END; (* pregraph *)

      PROCEDURE prereverse;
         (* turn on reverse video *)
      BEGIN (* prereverse *)
         IF vt52
         THEN WRITE(CHR(27), 'p');
         IF ansi
         THEN WRITE(CHR(27), '[7m');
      END; (* prereverse *)

   BEGIN (* prefix *)
      IF reversed[a]
      THEN prereverse;
      IF graphic[a]
      THEN pregraph;
   END; (* prefix *)

   PROCEDURE postfix (a : chartype);
      (* give postfix appropriate for this character *)

      PROCEDURE postgraph;
         (* turn off graphic characters *)
      BEGIN (* postgraph *)
         IF vt52
         THEN WRITE(CHR(27), 'G');
         IF ansi
         THEN WRITE(CHR(27), '(B');
      END; (* postgraph *)

      PROCEDURE postreverse;
         (* turn off reverse video *)
      BEGIN (* postreverse *)
         IF vt52
         THEN WRITE(CHR(27), 'q');
         IF ansi
         THEN WRITE(CHR(27), '[0m');
      END; (* postreverse *)

   BEGIN (* postfix *)
      IF reversed[a]
      THEN postreverse;
      IF graphic[a]
      THEN postgraph;
   END; (* postfix *)

   PROCEDURE makechar (a : chartype);
      (* print out a character with appropriate prefix or postfix *)
   BEGIN (* makechar *)
      prefix (a);
      WRITE(ch[a]);
      postfix (a);
   END; (* makechar *)

   PROCEDURE logo;
      VAR i : INTEGER;
         c : CHAR;
   BEGIN (* logo *)
      IF ibmpc
      THEN ClrScr    (* Turbo Pascal only *);
      FOR i := 1 TO 10 DO WRITELN;
      WRITE('                     ');
      IF ibmpc
      THEN WRITE(CHR(201))
      ELSE WRITE(' ');
      prefix (horiz);
      FOR i := 1 TO 32 DO WRITE(ch[horiz]);
      postfix (horiz);
      IF ibmpc
      THEN WRITELN(CHR(187))
      ELSE WRITELN(' ');
      WRITE('                     ');
      makechar (vert);
      WRITE('                                ');
      makechar (vert);
      WRITELN;
      WRITE('                     ');
      makechar (vert);
      WRITE(' D O L M O V E  (in PHYLIP 3.1) ');
      makechar (vert);
      WRITELN;
      WRITE('                     ');
      makechar (vert);
      WRITE('                                ');
      makechar (vert);
      WRITELN;
      WRITE('                     ');
      IF ibmpc
      THEN WRITE(CHR(200))
      ELSE WRITE(' ');
      prefix (horiz);
      FOR i := 1 TO 32 DO WRITE(ch[horiz]);
      postfix (horiz);
      IF ibmpc
      THEN WRITELN(CHR(188))
      ELSE WRITELN(' ');
      FOR i := 1 TO 10 DO WRITELN;
      WRITELN(' TO CONTINUE, PRESS ON THE Return OR Enter KEY');
      READ(c);
   END; (* logo *)

   PROCEDURE doinput;
      (* reads the input data *)

      PROCEDURE inputnumbers;
         (* input the numbers of species and of characters *)
      BEGIN (* inputnumbers *)
         READ(infile, spp, chars);
         WRITE('  ',spp:2, ' Species, ', chars:3, ' Characters');
         IF spp > maxsp
         THEN WRITELN(OUTPUT, ' TOO MANY SPECIES: adjust CONSTants');
         IF chars > maxchr
         THEN WRITELN(OUTPUT, ' TOO MANY CHARACTERS: adjust CONSTants');
         WRITELN;
         WRITELN;
         words := (chars DIV bits) + 1;
         nonodes := 2 * spp - 1;
      END; (* inputnumbers *)

      PROCEDURE inputoptions;
         (* input the information on the options *)
         VAR ch : CHAR;
            extranum, i : INTEGER;

         PROCEDURE inputhresh;
            (* input the threshold *)
         BEGIN (* inputhresh *)
            READLN(infile, threshold);
            thresh := TRUE;
         END; (* inputhresh *)

         PROCEDURE printhresh;
            (* print out the threshold value *)
         BEGIN (* printhresh *)
            WRITELN('    Threshold value = ', threshold:10:3);
            WRITELN;
         END; (* printhresh *)

         PROCEDURE inputweights;
            (* input the character weights, 0-9 and A-Z for weights
               0 - 35 *)
            VAR ch : CHAR;
               i : INTEGER;
         BEGIN (* inputweights *)
            FOR i := 1 TO nmlngth-1 DO READ(infile, ch);
            FOR i := 1 TO chars DO BEGIN
               REPEAT
                  IF EOLN(infile) THEN READLN(infile);
                  READ(infile, ch);
               UNTIL ch <> ' ';
               weight[i] := 1;
               IF digit(ch)
               THEN weight[i] := ORD(ch) - ORD('0')
               ELSE IF letter(ch)
                  THEN BEGIN
                     uppercase (ch);
                     IF (ch >= 'A') AND (ch <= 'I')
                     THEN weight[i] := ORD(ch) - ORD('A') + 10
                     ELSE IF (ch >= 'J') AND (ch <= 'R')
                        THEN weight[i] := ORD(ch) - ORD('J') + 19
                        ELSE weight[i] := ORD(ch) - ORD('S') + 28;
                     END
                  ELSE WRITELN(' Bad weight character: ',ch);
               END;
            READLN(infile);
            weights := TRUE;
         END; (* inputweights *)

         PROCEDURE printweights;
            (* print out the weights of characters *)
            VAR i, j, k : INTEGER;
         BEGIN (* printweights *)
            WRITELN('    Characters are weighted as follows:');
            WRITE('        ');
            FOR i := 0 TO 9 DO WRITE(i:3);
            WRITELN;
            WRITELN('     *---------------------------------');
            FOR j := 0 TO (chars DIV 10) DO BEGIN
               WRITE(10*j:5, '!  ');
               FOR i := 0 TO 9 DO BEGIN
                  k := 10*j + i;
                  IF (k>0) AND (k<=chars)
                  THEN WRITE(weight[k]:3)
                  ELSE WRITE('   ');
                  END;
               WRITELN;
               END;
            WRITELN;
         END; (* printweights *)

         PROCEDURE inputancestors;
            (* reads the ancestral states for each character *)
            VAR i : INTEGER;
               ch : CHAR;
         BEGIN (* inputancestors *)
            FOR i := 1 TO nmlngth-1 DO READ(infile, ch);
            ancvar := TRUE;
            FOR i := 1 TO chars DO BEGIN
               ancone[i] := TRUE;
               REPEAT
                  IF EOLN(infile) THEN READLN(infile);
                  READ(infile, ch);
               UNTIL ch <> ' ';
               IF ch = 'p' THEN ch := 'P';
               IF ch = 'b' THEN ch := 'B';
               CASE ch OF
                  '1' : anczero[i] := FALSE;
                  '0' : ancone[i] := FALSE;
                  'P' : BEGIN END;
                  'B' : BEGIN END;
                  '?' : BEGIN END
                  END;
               END;
            READLN(infile);
         END; (* inputancestors *)

         PROCEDURE printancestors;
            (* print out list of ancestral states *)
            VAR i : INTEGER;
         BEGIN (* printancestors *)
            WRITELN(' Ancestral states:');
            FOR i := 1 TO nmlngth+4 DO WRITE(' ');
            FOR i := 1 TO chars DO BEGIN
               newline (i, 50, nmlngth+4);
               IF ancone[i] AND anczero[i]
               THEN WRITE('?')
               ELSE IF ancone[i]
                  THEN WRITE('1')
                  ELSE WRITE('0');
               IF ((i MOD 5) = 0) THEN WRITE(' ');
               END;
            WRITELN;
            WRITELN;
         END; (* printancestor *)

      BEGIN (* inputoptions *)
         ancvar := FALSE;
         FOR i := 1 TO chars DO BEGIN
            anczero[i] := TRUE;
            ancone[i] := FALSE;
            END;
         dollo := TRUE;
         thresh := FALSE;
         weights := FALSE;
         extranum := 0;
         WHILE NOT EOLN(infile) DO BEGIN
            READ(infile, ch);
            uppercase (ch);
            IF (ch = 'A') OR (ch = 'D') OR (ch = 'P')
               OR (ch = 'T') OR (ch = 'W')
            THEN CASE ch OF
                  'A' : extranum := extranum + 1;
                  'D' : dollo := TRUE;
                  'P' : dollo := FALSE;
                  'T' : extranum := extranum + 1;
                  'W' : extranum := extranum + 1
                  END
            ELSE IF ch <> ' '
               THEN WRITELN(' Bad Option Character: ', ch);
            END;
         READLN(infile);
         threshold := spp;
         FOR i := 1 TO chars DO weight[i] := 1;
         FOR i := 1 TO extranum DO BEGIN
            READ(infile, ch);
            uppercase (ch);
            IF ch = 'A'
            THEN inputancestors;
            IF ch = 'T'
            THEN inputhresh;
            IF ch = 'W'
            THEN inputweights;
            END;
         IF dollo
         THEN WRITE(OUTPUT, ' Dollo')
         ELSE WRITE(OUTPUT, ' Polymorphism');
         WRITELN(OUTPUT, ' parsimony method');
         WRITELN(OUTPUT);
         IF thresh
         THEN printhresh;
         IF weights
         THEN printweights;
         IF ancvar
         THEN printancestors;
         questions := FALSE;
         FOR i := 1 TO chars DO BEGIN
            questions := questions OR (ancone[i] AND anczero[i]);
            threshwt[i] := threshold * weight[i];
            END;
      END; (* inputoptions *)

      PROCEDURE inputdata;
         (* input the names and character state data for species *)
         VAR i, j, l : INTEGER;
            k : bitrange;
            charstate : CHAR;  (* possible states are
                                  '0', '1', 'P', 'B', and '?' *)
      BEGIN (* inputdata *)
         FOR i := 1 TO chars DO extras[i] := 0;
         FOR i := 1 TO nonodes DO BEGIN
            NEW (treenode[i]);
            treenode[i]^.leftdesc := NIL;
            treenode[i]^.rtdesc := NIL;
            treenode[i]^.ancestor := NIL;
            treenode[i]^.index := i;
            IF i <= spp
            THEN BEGIN
               treenode[i]^.tip := TRUE;
               FOR j := 1 TO nmlngth DO READ(infile, name[i,j]);
               FOR j := 1 TO words DO BEGIN
                  treenode[i]^.stateone[j] := [];
                  treenode[i]^.statezero[j] := [];
                  END;
               FOR j := 1 TO chars DO BEGIN
                  k := ((j-1) MOD bits) + 1;
                  l := ((j-1) DIV bits) + 1;
                  REPEAT
                     IF EOLN(infile) THEN READLN(infile);
                     READ(infile, charstate);
                  UNTIL charstate <> ' ';
                  IF charstate = 'b' THEN charstate := 'B';
                  IF charstate = 'p' THEN charstate := 'P';
                  IF NOT ((charstate = '0')
                         OR (charstate = '1')
                         OR (charstate = '?')
                         OR (charstate = 'P')
                         OR (charstate = 'B'))
                  THEN WRITELN(' Warning -- ',
                      'bad character state: ', charstate,
                      ' at character ', j:5, ' of species ', i:3);
                  IF (charstate = '1')
                  THEN treenode[i]^.stateone[l]
                             := treenode[i]^.stateone[l] + [k];
                  IF (charstate = '0')
                  THEN treenode[i]^.statezero[l]
                             := treenode[i]^.statezero[l] + [k];
                  IF (charstate = 'P') OR (charstate = 'B')
                  THEN BEGIN
                     IF dollo
                     THEN extras[j] := extras[j] + weight[j]
                     ELSE BEGIN
                        treenode[i]^.stateone[l]
                             := treenode[i]^.stateone[l] + [k];
                        treenode[i]^.statezero[l]
                             := treenode[i]^.statezero[l] + [k];
                        END;
                     END;
                  END;
               READLN(infile);
               END
            ELSE treenode[i]^.tip := FALSE;
            END;
         WRITELN;
         WRITELN;
      END; (* inputdata *)

   BEGIN (* doinput *)
      WRITELN;
      WRITELN(' Reading input file ...');
      WRITELN;
      inputnumbers;
      inputoptions;
      inputdata;
   END; (* doinput *)

   PROCEDURE treeconstruct;
      (* constructs a binary tree from the pointers in treenode. *)
      TYPE rearrtype = (rearr, flipp, reroott, none);
      VAR dispchar, dispword, dispbit, atwhat,
            what, fromwhere, towhere, oldoutgrno : INTEGER;
         like, bestyet, gotlike : REAL;
         guess : ARRAY[1..maxchr] OF CHAR;
         display, yours, subtree , written, oldwritten : BOOLEAN;
         numsteps : steparray;
         fullset : bitset;
         zeroanc, oneanc : bitarray;
         nuroot : nodeptr;
         lastop : rearrtype;

      PROCEDURE add (below, newtip, newfork : nodeptr);
         (* inserts the nodes newfork and its left descendant, newtip,
            to the tree.  below becomes newfork's right descendant *)
      BEGIN (* add *)
         newfork^.ancestor := below^.ancestor;
         newfork^.rtdesc := below;
         newfork^.leftdesc := newtip;
         below^.ancestor := newfork;
         newtip^.ancestor := newfork;
         IF newfork^.ancestor <> NIL
         THEN IF newfork^.ancestor^.leftdesc = below
            THEN newfork^.ancestor^.leftdesc := newfork
            ELSE newfork^.ancestor^.rtdesc := newfork;
      END; (* add *)

      PROCEDURE remove (VAR item, fork : nodeptr);
         (* removes nodes item and its ancestor, fork, from the tree.
            the new descendant of fork's ancestor is made to be
            fork's second descendant (other than item).  Also
            returns pointers to the deleted nodes, item and fork *)
         VAR other, below : nodeptr;
      BEGIN (* remove *)
         IF item^.ancestor = NIL
         THEN fork := NIL
         ELSE BEGIN
            fork := item^.ancestor;
            IF item = fork^.leftdesc
            THEN other := fork^.rtdesc
            ELSE other := fork^.leftdesc;
            below := fork^.ancestor;
            other^.ancestor := below;
            IF below <> NIL
            THEN BEGIN
               IF below^.leftdesc = fork
               THEN below^.leftdesc := other
               ELSE below^.rtdesc := other;
               fork^.ancestor := NIL;
               END;
            fork^.leftdesc := NIL;
            fork^.rtdesc := NIL;
            item^.ancestor := NIL;
            END;
      END; (* remove *)

      PROCEDURE fillin (p: nodeptr);
         (* Sets up for each node in the tree two statesets.
            stateone and statezero are the sets of character
            states that must be 1 or must be 0, respectively,
            in a most parsimonious reconstruction, based on the
            information at or above this node.  Note that this
            state assignment may change based on information further
            down the tree.  If a character is in both sets it is in
            state "P".  If in neither, it is "?". *)
         VAR i : INTEGER;
      BEGIN (* fillin *)
         FOR i := 1 TO words DO BEGIN
            p^.stateone[i] := p^.leftdesc^.stateone[i]
                + p^.rtdesc^.stateone[i];
            p^.statezero[i] := p^.leftdesc^.statezero[i]
                + p^.rtdesc^.statezero[i];
            END;
      END; (* fillin *)

      PROCEDURE correct (p : nodeptr);
         VAR i : INTEGER;
            z0, z1, s0, s1, temp : bitset;
      BEGIN (* get final states for intermediate nodes *)
         IF NOT p^.tip
         THEN BEGIN
            FOR i := 1 TO words DO BEGIN
               IF p^.ancestor = NIL
               THEN BEGIN
                  s0 := zeroanc[i];
                  s1 := oneanc[i];
                  END
               ELSE BEGIN
                  s0 := p^.ancestor^.statezero[i];
                  s1 := p^.ancestor^.stateone[i];
                  END;
               z0 := s0 * p^.statezero[i]
                        + p^.leftdesc^.statezero[i]
                          * p^.rtdesc^.statezero[i];
               z1 := s1 * p^.stateone[i]
                        + p^.leftdesc^.stateone[i]
                          * p^.rtdesc^.stateone[i];
               IF dollo
               THEN BEGIN
                  temp := z0 - zeroanc[i] * z1;
                  z1 := z1 - oneanc[i] * z0;
                  z0 := temp;
                  END;
               temp := fullset - z0 - z1;
               p^.statezero[i] := z0 + temp * (s0-s1);
               p^.stateone[i] := z1 + temp * (s1-s0);
               END;
            END;
      END; (* correct *)

      PROCEDURE evaluate (r: nodeptr);
         (* Determines the number of losses or polymorphisms needed
            for a tree. This is the minimum number needed to evolve
            chars on this tree *)
         VAR i, stepnum, smaller : INTEGER;
            sum : REAL;
            numsone, numszero : steparray;

         PROCEDURE postorder (p : nodeptr);
            (* traverses a binary tree, calling PROCEDURE fillin at a
               node's descendants before calling fillin at the node *)
         BEGIN (* postorder *)
            IF NOT p^.tip
            THEN BEGIN
               postorder (p^.leftdesc);
               postorder (p^.rtdesc);
               fillin (p);
               END;
         END; (* postorder *)

         PROCEDURE preorder (p : nodeptr);
            (* go back up tree setting up and counting interior node
               states *)

            PROCEDURE count (p : nodeptr);
            (* counts the number of steps in a fork of the tree.
               The program spends much of its time in this PROCEDURE *)
               VAR i, j, l : INTEGER;
                  steps : bitarray;
            BEGIN (* count *)
               IF dollo
               THEN FOR i := 1 TO words DO
                  steps[i] := (p^.ancestor^.stateone[i]
                                  * p^.statezero[i] * zeroanc[i])
                            + (p^.ancestor^.statezero[i]
                                  * p^.stateone[i]
                                  * oneanc[i])
               ELSE FOR i := 1 TO words DO
                  steps[i] := p^.ancestor^.stateone[i] *
                     p^.ancestor^.statezero[i] * p^.stateone[i]
                     * p^.statezero[i];
               j := 1;
               l := 0;
               FOR i := 1 TO chars DO BEGIN
                  l := l + 1;
                  IF l > bits
                  THEN BEGIN
                     l := 1;
                     j := j + 1;
                     END;
                  IF l IN steps[j]
                  THEN IF l IN zeroanc[j]
                     THEN numszero[i] := numszero[i] + weight[i]
                     ELSE numsone[i] := numsone[i] + weight[i];
                  END;
            END; (* count *)

         BEGIN (* preorder *)
            IF NOT p^.tip
            THEN BEGIN
               correct (p);
               preorder (p^.leftdesc);
               preorder (p^.rtdesc);
               END;
            IF p^.ancestor <> NIL
            THEN count (p);
         END; (* preorder *)

      BEGIN (* evaluate *)
         sum := 0;
         FOR i := 1 TO chars DO BEGIN
            numszero[i] := 0;
            numsone[i] := 0;
            END;
         FOR i := 1 TO words DO BEGIN
            zeroanc[i] := fullset;
            oneanc[i] := [];
            END;
         postorder (r);
         preorder (r);
         FOR i := 1 TO words DO BEGIN
            zeroanc[i] := [];
            oneanc[i] := fullset;
            END;
         postorder (r);
         preorder (r);
         FOR i := 1 TO chars DO BEGIN
            smaller := spp * weight[i];
            IF anczero[i]
            THEN BEGIN
               numsteps[i] := numszero[i];
               smaller := numszero[i];
               END;
            IF ancone[i] AND (numsone[i] < smaller)
            THEN numsteps[i] := numsone[i];
            stepnum := numsteps[i] + extras[i];
            IF stepnum <= threshwt[i]
            THEN sum := sum + stepnum
            ELSE sum := sum + threshwt[i];
            guess[i] := '?';
            IF (NOT ancone[i]) OR
               (anczero[i] AND (numszero[i] < numsone[i]))
            THEN guess[i] := '0'
            ELSE IF (NOT anczero[i]) OR
                  (ancone[i] AND (numsone[i] < numszero[i]))
               THEN guess[i] := '1';
            END;
         like := -sum;
      END; (* evaluate *)

      PROCEDURE reroot (outgroup : nodeptr);
         (* reorients tree, putting outgroup in desired position.
            Written by Kent Fiala *)
         VAR vertex, oldvertex, ante, otherbranch : nodeptr;
            left : BOOLEAN;
      BEGIN (* reroot *)
         IF NOT (outgroup^.ancestor = root)
         THEN BEGIN
            oldvertex := outgroup;
            vertex := outgroup^.ancestor;
            ante := vertex^.ancestor;
            IF vertex^.rtdesc = outgroup
            THEN BEGIN
               vertex^.rtdesc := vertex^.leftdesc;
               vertex^.leftdesc := outgroup;
               END;
            REPEAT
               IF oldvertex = vertex^.leftdesc
               THEN BEGIN
                  vertex^.leftdesc := ante;
                  left := TRUE;
                  END
               ELSE BEGIN
                  vertex^.rtdesc := ante;
                  left := FALSE;
                  END;
               vertex^.ancestor := oldvertex;
               oldvertex := vertex;
               vertex := ante;
               ante := vertex^.ancestor
            UNTIL vertex = root;
            IF root^.leftdesc = oldvertex
            THEN otherbranch := root^.rtdesc
            ELSE otherbranch := root^.leftdesc;
            IF left
            THEN oldvertex^.leftdesc := otherbranch
            ELSE oldvertex^.rtdesc := otherbranch;
            otherbranch^.ancestor := oldvertex;
            root^.rtdesc := outgroup^.ancestor;
            root^.leftdesc := outgroup;
            outgroup^.ancestor^.ancestor := root;
            outgroup^.ancestor := root;
            END;
      END; (* reroot *)

      PROCEDURE printree;
         (* prints out diagram of the tree *)
         CONST downn = 2;
            overr = 4;
         VAR tipx, tipy, i, rover, dow : INTEGER;

         PROCEDURE coordinates (p : nodeptr);
            (* establishes coordinates of nodes *)
         BEGIN (* coordinates *)
            IF p^.tip
            THEN BEGIN
               p^.xcoord := tipx;
               p^.ycoord := tipy;
               tipx := tipx + rover;
               tipy := tipy + dow;
               END
            ELSE BEGIN
               coordinates (p^.leftdesc);
               p^.xcoord := p^.leftdesc^.xcoord;
               coordinates (p^.rtdesc);
               p^.ycoord := p^.rtdesc^.ycoord;
               END;
         END; (* coordinates *)

         PROCEDURE drawline (i : INTEGER);
            (* draws one row of the tree diagram by moving up tree *)
            VAR p, q : nodeptr;
               m, n, j : INTEGER;
               s : CHAR;
               c : chartype;
         BEGIN (* drawline *)
            p := nuroot;
            WRITE('   ');
            REPEAT
               IF p^.leftdesc^.ycoord >= i
               THEN p := p^.leftdesc
               ELSE BEGIN
                  q := p^.rtdesc;
                  n := q^.xcoord - p^.xcoord;
                  IF p^.ycoord = i
                  THEN BEGIN
                     c := over;
                     s := p^.rtdesc^.state;
                     IF (s = 'P') AND (p^.state <> 'P')
                     THEN s := p^.state;
                     IF display
                     THEN CASE s OF
                        '1' : c := one;
                        '0' : c := zero;
                        '?' : c := quest;
                        'P' : c := polym
                        END;
                     IF p^.index >= 10
                     THEN WRITE(p^.index:2)
                     ELSE BEGIN
                        WRITE(p^.index:1);
                        makechar (c);
                        END;
                     IF n > 2
                     THEN BEGIN
                        prefix (c);
                        FOR j := 1 TO n-2 DO WRITE(ch[c]);
                        postfix (c);
                        END
                     END
                  ELSE BEGIN
                     c := up;
                     s := p^.leftdesc^.state;
                     IF (s = 'P') AND (p^.state <> 'P')
                     THEN s := p^.state;
                     IF display
                     THEN CASE s OF
                        '1' : c := one;
                        '0' : c := zero;
                        '?' : c := quest;
                        'P' : c := polym
                        END;
                     makechar (c);
                     FOR j := 1 TO n-1 DO WRITE(' ');
                     END;
                  p := q;
                  END;
            UNTIL p^.tip;
            IF p^.ycoord = i
            THEN BEGIN
               IF p^.index >= 10
               THEN m := 2
               ELSE m := 1;
               WRITE(p^.index:m, ':');
               FOR j := 1 TO nmlngth DO
                       WRITE(name[p^.index, j]);
               IF (i = 1)
               THEN BEGIN
                  WRITE('                      ');
                  IF subtree
                  THEN WRITE('      Subtree          ')
                  ELSE WRITE('                       ');
                  IF dollo
                  THEN WRITE('Dollo')
                  ELSE WRITE ('Polymorphism ');
                  END;
               IF (i = 3) AND display
               THEN BEGIN
                  WRITE('      CHARACTER ', dispchar:3,
                           '    SYMBOLS: ');
                  makechar (zero);
                  WRITE(':0,  ');
                  makechar (quest);
                  WRITE(':?,  ');
                  makechar (one);
                  WRITE(':1,  ');
                  IF NOT dollo
                  THEN BEGIN
                     makechar (polym);
                     WRITE(':0/1');
                     END;
                  END;
               IF i = 5
               THEN BEGIN
                  WRITE('           ', -like:10:5);
                  IF NOT yours
                  THEN BEGIN
                  IF -like < bestyet
                     THEN BEGIN
                        WRITE('  BEST YET!');
                        bestyet := -like;
                        gotlike := bestyet;
                        END
                     ELSE IF ABS(-like - bestyet) < 0.000001
                        THEN WRITE(' (as good as best)')
                        ELSE BEGIN
                           IF -like < gotlike
                           THEN WRITE('  better')
                           ELSE IF -like > gotlike
                              THEN WRITE('  worse!');
                           END;
                     gotlike := -like;
                     END;
                  END;
               END;
            WRITELN;
         END; (* drawline *)

      BEGIN (* printree *)
         IF NOT subtree
         THEN nuroot := root;
         IF NOT display
         THEN evaluate(root);
         ClrScr;   (* Turbo Pascal only *)
         IF ansi
         THEN WRITE(CHR(27), '[2J', CHR(27), '[H')
         ELSE IF vt52
            THEN WRITE(CHR(27), 'E', CHR(27), 'H');
         IF ibmpc
         THEN WRITELN;
         tipx := 0;
         tipy := 1;
         rover := 100 DIV spp;
         IF rover > overr THEN rover := overr;
         dow := downn;
         IF (spp*dow > screenlines) AND (NOT subtree)
         THEN BEGIN
            dow := dow-1;
            rover := rover-1;
            END;
         coordinates (nuroot);
         FOR i := 1 TO tipy-dow DO drawline (i);
         IF (spp <= (screenlines DIV 2)) OR subtree THEN WRITELN;
      END; (* printree *)

      PROCEDURE describe;
         (* makes state reconstructions in each character *)

         PROCEDURE hypstates;
            (* fill in and describe states at interior nodes *)
            VAR i, j, k : INTEGER;

            PROCEDURE filltrav (r : nodeptr);
               (* traverse to fill in interior node states *)
            BEGIN (* filltrav *)
               IF NOT r^.tip
               THEN BEGIN
                  filltrav (r^.leftdesc);
                  filltrav (r^.rtdesc);
                  fillin (r);
                  END;
            END; (* filltrav *)

            PROCEDURE hyptrav (r : nodeptr);
               (* compute states at interior nodes for one character *)
            BEGIN (* hyptrav *)
               IF NOT r^.tip
               THEN correct (r);
               IF dispbit IN r^.stateone[dispword]
               THEN BEGIN
                  IF dispbit IN r^.statezero[dispword]
                  THEN BEGIN
                     IF dollo
                     THEN r^.state := '?'
                     ELSE r^.state := 'P';
                     END
                  ELSE r^.state := '1';
                  END
               ELSE BEGIN
                  IF dispbit IN r^.statezero[dispword]
                  THEN r^.state := '0'
                  ELSE r^.state := '?';
                  END;
               IF NOT r^.tip
               THEN BEGIN
                  hyptrav (r^.leftdesc);
                  hyptrav (r^.rtdesc);
                  END;
            END; (* hyptrav *)

         BEGIN (* hypstates *)
            FOR i := 1 TO words DO BEGIN
               zeroanc[i] := [];
               oneanc[i] := [];
               END;
            FOR i := 1 TO chars DO BEGIN
               j := (i-1) DIV bits + 1;
               k := (i-1) MOD bits + 1;
               IF (guess[i] = '0')
               THEN zeroanc[j] := zeroanc[j] + [k];
               IF (guess[i] = '1')
               THEN oneanc[j] := oneanc[j] + [k];
               END;
            filltrav (root);
            hyptrav (root);
         END; (* hypstates *)

      BEGIN (* describe *)
         evaluate (root);
         hypstates;
      END; (* describe *)

      PROCEDURE buildtree;
         VAR ch : CHAR;
            ok : BOOLEAN;

         PROCEDURE arbitree;
            VAR i : INTEGER;
         BEGIN (* arbitree *)
            root := treenode[1];
            add (treenode[1], treenode[2], treenode[spp+1]);
            root := treenode[spp+1];
            FOR i := 3 TO spp DO BEGIN
               add (treenode[spp+i-2], treenode[i], treenode[spp+i-1]);
               IF root^.ancestor <> NIL THEN root := root^.ancestor;
               END;
            printree;
            bestyet := -like;
            gotlike := -like;
         END; (* arbitree *)

         PROCEDURE yourtree;
            VAR i, j : INTEGER;
               ok : BOOLEAN;
         BEGIN (* yourtree *)
            root := treenode[1];
            add (treenode[1], treenode[2], treenode[spp+1]);
            root := treenode[spp+1];
            i := 2;
            REPEAT
               i := i + 1;
               printree;
               WRITE(' Add species', i:3, ': ');
               FOR j := 1 TO nmlngth DO WRITE(name[i,j]);
               REPEAT
                  WRITELN;
                  WRITE( '  before node (type number): ');
                  READLN(INPUT, j);
                  ok := ((j >= 1) AND (j < i)) OR
                          ((j > spp) AND (j < (spp + i -1)));
                  IF NOT ok
                  THEN WRITELN('Impossible number. Please try again:');
               UNTIL ok;
               add (treenode[j], treenode[i], treenode[spp+i-1]);
               WHILE root^.ancestor <> NIL DO root := root^.ancestor;
            UNTIL i = spp;
            printree;
            bestyet := -like;
            gotlike := -like;
         END; (* yourtree *)

         PROCEDURE treeread;
            (* read in user-defined tree and set it up *)
            VAR ch : CHAR;
               nextnode : INTEGER;

            PROCEDURE findch (c : CHAR);
               (* scan forward until find character c *)
            BEGIN (* findch *)
               IF ch <> c
               THEN REPEAT
                     IF EOLN(treefile) THEN READLN(treefile);
                     READ(treefile, ch);
                  UNTIL ch = c;
            END; (* findch *)

            PROCEDURE addelement (VAR p : nodeptr);
               (* recursive procedure adds nodes to user-defined tree *)
               VAR q : nodeptr;
                  i, n : INTEGER;
                  found : BOOLEAN;
                  str : ARRAY[1..nmlngth] OF CHAR;
            BEGIN (* addelement *)
               REPEAT
                  IF EOLN(treefile) THEN READLN(treefile);
                  READ(treefile, ch);
               UNTIL ch <> ' ';
               IF ch = '('
               THEN BEGIN
                  nextnode := nextnode + 1;
                  q := treenode[nextnode];
                  addelement (q^.leftdesc);
                  q^.leftdesc^.ancestor := q;
                  findch (',');
                  addelement (q^.rtdesc);
                  q^.rtdesc^.ancestor := q;
                  findch (')');
                  p := q;
                  END
               ELSE BEGIN
                  FOR i := 1 TO nmlngth DO str[i] := ' ';
                  n := 1;
                  REPEAT
                     IF (ch = '_') THEN ch := ' ';
                     str[n] := ch;
                     IF EOLN(treefile) THEN READLN(treefile);
                     READ(treefile, ch);
                     n := n + 1;
                  UNTIL ((ch = ',') OR (ch = ')') OR (n > nmlngth));
                  n := 1;
                  REPEAT
                     found := TRUE;
                     FOR i := 1 TO nmlngth DO
                        found := found AND (str[i] = name[n, i]);
                     IF found
                     THEN p := treenode[n]
                     ELSE n := n + 1;
                  UNTIL (n > spp) OR found;
                  IF n > spp
                  THEN BEGIN
                     WRITE(' Cannot find species: ');
                     FOR i := 1 TO nmlngth DO
                        WRITE(str[i]);
                     WRITELN;
                     END;
                  END;
            END; (* addelement *)

         BEGIN (* treeread *)
            RESET(treefile);
            root := treenode[spp + 1];
            nextnode := spp;
            root^.ancestor := NIL;
            addelement (root);
            WRITELN;
            printree;
            bestyet := -like;
            gotlike := -like;
            CLOSE(treefile); (* -- Turbo Pascal and maybe others too *)
         END; (* treeread *)

      BEGIN (* buildtree *)
         ok := FALSE;
         yours := TRUE;
         REPEAT
            WRITELN;
            WRITELN(' How should we build the initial tree?');
            WRITELN('     A   Arbitrarily');
            WRITELN('     S   as you Specify');
            WRITELN('     U   User tree read from file');
            WRITE(' Type one of those: ');
            READLN(INPUT, ch);
            UPPERCASE(ch);
            ok := (ch = 'A') OR (ch = 'S') OR (ch = 'U');
            IF NOT ok
            THEN WRITELN(' No!! Not a possible choice. Try again:');
         UNTIL ok;
         CASE ch OF
            'A' : arbitree;
            'S' : yourtree;
            'U' : treeread
            END;
         yours := FALSE;
         outgrno := root^.leftdesc^.index;
      END; (* buildtree *)

      PROCEDURE redisplay;
         VAR done : BOOLEAN;
            ch : CHAR;

         PROCEDURE clade;
            (* pick a subtree and show only that on screen *)
            VAR i : INTEGER;
               ok : BOOLEAN;
         BEGIN (* clade *)
            WRITE(' Select subtree rooted at which node');
            WRITE(' (0 for whole tree)? ');
            READLN(i);
            ok := (i >=0) AND (i <= nonodes);
            IF ok
            THEN BEGIN
               subtree := (i > 0);
               IF subtree
               THEN nuroot := treenode[i]
               ELSE nuroot := root;
               END;
            IF display
            THEN describe;
            printree;
            IF NOT ok
            THEN WRITE(' Illegal node! ');
         END; (* clade *)

         PROCEDURE flip;
            (* flip at a node left-right *)
            VAR i : INTEGER;
               ok : BOOLEAN;
               p : nodeptr;
         BEGIN (* flip *)
            WRITE('Flip branches at which node? ');
            READLN(i);
            ok := (i > spp) AND (i <= nonodes);
            IF ok
            THEN BEGIN
               p := treenode[i]^.leftdesc;
               treenode[i]^.leftdesc := treenode[i]^.rtdesc;
               treenode[i]^.rtdesc := p;
               atwhat := i;
               lastop := flipp;
               END;
            IF display
            THEN describe;
            printree;
            IF NOT ok
            THEN WRITE(' Illegal! ')
            ELSE BEGIN
               oldwritten := written;
               written := FALSE;
               END;
         END; (* flip *)

         PROCEDURE help;
            (* display help information *)
         BEGIN (* help *)
            WRITELN;
            WRITELN;
            WRITELN('  C show only one Clade (subtree)',
                       '(useful if tree is too big)');
            WRITELN('  F Flip (rotate) branches at a node');
            WRITELN('  O select an Outgroup for the tree');
            WRITELN('  R Rearrange a tree by moving a node or group');
            WRITELN('  S Show the states of a character');
            WRITELN('  +        ... of the next character');
            WRITELN('  -        ... of the previous character');
            WRITELN('  . redisplay the same tree again');
            WRITELN('  T Try all possible positions',
                         ' of a node or group');
            WRITELN('  U Undo the most recent rearrangement');
            WRITELN('  W Write tree to a file');
            WRITELN('  H Help (this screen)');
            WRITELN('  ?  "     "      "   ');
            WRITELN('  Q (Quit) Exit from program');
            WRITELN('  X Exit from program');
            WRITELN;
            WRITELN;
            WRITE(' TO CONTINUE, PRESS ON THE Return OR Enter KEY');
            READ(ch);
            IF display
            THEN describe;
            printree;
         END; (* help *)

         PROCEDURE rearrange;
            VAR i, j : INTEGER;
               ok1, ok2 : BOOLEAN;
               p, q : nodeptr;
         BEGIN (* rearrange *)
            WRITE(' Remove which node? ');
            READLN(i);
            ok1 :=  (i >= 1) AND (i <= 2*spp-1);
            IF ok1
            THEN BEGIN
               WRITE(' Add before which node? ');
               READLN(j);
               ok2 := (j >=1) AND (j <= 2*spp-1);
               IF ok2
               THEN BEGIN
                  ok2 := NOT (treenode[j] = treenode[i]^.ancestor);
                  p := treenode[j];
                  WHILE p <> root DO BEGIN
                     ok2 := ok2 AND (NOT (p = treenode[i]));
                     p := p^.ancestor;
                     END;
                  IF ok1 AND ok2
                  THEN BEGIN
                     what := i;
                     q := treenode[i]^.ancestor;
                     IF q^.leftdesc^.index = i
                     THEN fromwhere := q^.rtdesc^.index
                     ELSE fromwhere := q^.leftdesc^.index;
                     towhere := j;
                     remove (treenode[i], q);
                     add (treenode[j], treenode[i], q);
                     WHILE root^.ancestor <> NIL DO
                        root := root^.ancestor;
                     END;
                     lastop := rearr;
                  END;
               END;
            IF display
            THEN describe;
            printree;
            IF NOT (ok1 AND ok2)
            THEN WRITE(' Illegal! Try again: ')
            ELSE BEGIN
               oldwritten := written;
               written := FALSE;
               END;
         END; (* rearrange *)

         PROCEDURE changeoutgroup;
           VAR i : INTEGER;
              ok : BOOLEAN;
         BEGIN (* changeoutgroup *)
            oldoutgrno := outgrno;
            REPEAT
               WRITE('  Which node should be the new outgroup? ');
               READLN(i);
               ok := (i>=1) AND (i <= nonodes) AND (i <> root^.index);
               IF ok
               THEN outgrno := i;
            UNTIL ok;
            reroot (treenode[outgrno]);
            lastop := reroott;
            IF display
            THEN describe;
            printree;
            oldwritten := written;
            written := FALSE;
         END; (* changeoutgroup *)

         PROCEDURE show;
            VAR i : INTEGER;
               ok : BOOLEAN;
         BEGIN (* show *)
            REPEAT
               WRITE('SHOW: (Character number or 0 to see none)? ');
               READLN(i);
               ok := (i = 0) OR ((i >= 1) AND (i <= chars));
               IF ok AND (i <> 0)
               THEN BEGIN
                  display := TRUE;
                  dispchar := i;
                  dispword := ((i-1) DIV bits) + 1;
                  dispbit := ((i-1) MOD bits) + 1;
                  END;
               IF ok AND (i = 0)
               THEN display := FALSE;
               UNTIL ok;
            IF display
            THEN describe;
            printree;
         END; (* show *)

         PROCEDURE again;
            (* show same tree again *)
         BEGIN (* again *)
            IF display
            THEN describe;
            printree;
         END; (* again *)

         PROCEDURE nextchar;
            (* show next character *)
         BEGIN (* nextchar *)
            display := TRUE;
            dispchar := dispchar + 1;
            IF dispchar > chars
            THEN dispchar := 1;
            dispword := ((dispchar-1) DIV bits) + 1;
            dispbit := ((dispchar-1) MOD bits) + 1;
            IF display
            THEN describe;
            printree;
         END; (* nextchar *)

         PROCEDURE prevchar;
            (* show previous character *)
         BEGIN (* prevchar *)
            display := TRUE;
            dispchar := dispchar - 1;
            IF dispchar < 1
            THEN dispchar := chars;
            dispword := ((dispchar-1) DIV bits) + 1;
            dispbit := ((dispchar-1) MOD bits) + 1;
            IF display
            THEN describe;
            printree;
         END; (* prevchar *)

         PROCEDURE try;
            (* Remove node, try it in all possible places *)
            VAR i, j : INTEGER;
               current : REAL;
               q, dummy, rute : nodeptr;
               tied, better : BOOLEAN;
               place : ARRAY [1..maxsz] OF REAL;

            PROCEDURE addpreorder (p, item, nufork : nodeptr);
               (* traverses a binary tree, calling PROCEDURE tryadd
                  at a node before calling tryadd at its descendants *)

               PROCEDURE tryadd (p: nodeptr);
                  (* temporarily adds one fork and one tip to the tree.
                     Records scores in ARRAY place *)
               BEGIN (* tryadd *)
                  add (p, item, nufork);
                  IF root^.ancestor <> NIL THEN root := root^.ancestor;
                  evaluate (root);
                  place[p^.index] := -like;
                  remove (item, nufork);
                  IF nufork = root THEN root := p;
               END; (* tryadd *)

            BEGIN (* addpreorder *)
               IF p <> NIL
               THEN BEGIN
                  tryadd (p);
                  addpreorder (p^.leftdesc, item, nufork);
                  addpreorder (p^.rtdesc, item, nufork);
                  END;
            END; (* addpreorder *)

         BEGIN (* try *)
            WRITE('Try other positions for which node? ');
            READLN(i);
            IF (i >= 1) AND (i <= nonodes) AND (i <> root^.index)
            THEN BEGIN
               WRITELN('WAIT ...');
               FOR j := 1 TO nonodes DO place[j] := -1.0;
               evaluate (root);
               current := -like;
               what := i;
               q := treenode[i]^.ancestor;
               IF q^.leftdesc^.index = i
               THEN fromwhere := q^.rtdesc^.index
               ELSE fromwhere := q^.leftdesc^.index;
               rute := root;
               IF root = treenode[i]^.ancestor
               THEN BEGIN
                  IF treenode[i]^.ancestor^.leftdesc = treenode[i]
                  THEN rute := treenode[i]^.ancestor^.rtdesc
                  ELSE rute := treenode[i]^.ancestor^.leftdesc;
                  END;
               remove (treenode[i], dummy);
               root := rute;
               addpreorder (root, treenode[i], dummy);
               add (treenode[fromwhere], treenode[what], dummy);
               IF root^.ancestor <> NIL THEN root := root^.ancestor;
               better := FALSE;
               WRITE('        BETTER: ');
               FOR j := 1 TO nonodes DO
                  IF (place[j] < current) AND (place[j] >= 0.0)
                  THEN BEGIN
                     WRITE(j:3, ':', place[j]:6:2);
                     better := TRUE;
                     END;
               IF NOT better
               THEN WRITE(' NONE');
               WRITELN;
               WRITE('        TIED:    ');
               tied := FALSE;
               FOR j := 1 TO nonodes DO
                  IF (ABS(place[j] - current) < 1.0E-6)
                     AND (j <> fromwhere)
                  THEN BEGIN
                     IF j < 10
                     THEN WRITE(j:2)
                     ELSE WRITE(j:3);
                     tied := TRUE;
                     END;
               IF tied
               THEN WRITELN(':', current:6:2)
               ELSE WRITELN('NONE');
               END
            ELSE WRITE(' Illegal choice! ');
         END; (* try *)

         PROCEDURE undo;
            (* restore to tree before last rearrangement *)
            VAR temp : INTEGER;
               btemp : BOOLEAN;
               q : nodeptr;
         BEGIN (* undo *)
            CASE lastop OF
               rearr :   BEGIN
                         remove (treenode[what], q);
                         add (treenode[fromwhere], treenode[what], q);
                         WHILE root^.ancestor <> NIL DO
                            root := root^.ancestor;
                         temp := fromwhere;
                         fromwhere := towhere;
                         towhere := temp;
                         END;
               flipp :   BEGIN
                         q := treenode[atwhat]^.leftdesc;
                         treenode[atwhat]^.leftdesc := treenode[atwhat]^.rtdesc;
                         treenode[atwhat]^.rtdesc := q;
                         END;
               reroott : BEGIN
                         temp := oldoutgrno;
                         oldoutgrno := outgrno;
                         outgrno := temp;
                         reroot (treenode[outgrno]);
                         END;
               none : BEGIN END
               END;
            IF display
            THEN describe;
            printree;
            IF lastop = none
            THEN WRITE(' No operation to undo! ')
            ELSE BEGIN
               btemp := oldwritten;
               oldwritten := written;
               written := btemp;
               END;
         END; (* undo *)

         PROCEDURE treewrite;
            (* write out tree to a file *)

            PROCEDURE treeout (p : nodeptr);
               (* write out file with representation of final tree *)
               VAR i, n : INTEGER;
                  c : CHAR;
            BEGIN (* treeout *)
               IF p^.tip
               THEN BEGIN
                  n := 0;
                  FOR i := 1 TO nmlngth DO
                     IF name[p^.index, i] <> ' '
                     THEN n := i;
                  FOR i := 1 TO n DO BEGIN
                     c := name[p^.index, i];
                     IF c = ' '
                     THEN c := '_';
                     WRITE(treefile, c);
                     END;
                  END
               ELSE BEGIN
                  WRITE(treefile, '(');
                  treeout (p^.leftdesc);
                  WRITE(treefile, ',');
                  treeout (p^.rtdesc);
                  WRITE(treefile, ')');
                  END;
               IF p = root
               THEN WRITELN(treefile, ';');
            END; (* treeout *)

         BEGIN (* treewrite *)
            REWRITE(treefile);
            treeout(root);
            WRITELN;
            WRITELN(' Tree written to file');
            WRITELN;
            CLOSE(treefile); (* -- Turbo Pascal and maybe others too *)
            written := TRUE;
         END; (* treewrite *)

      BEGIN (* redisplay *)
         done := FALSE;
         REPEAT
            WRITE(' NEXT? (Options: C F O R S + - . T U W H ? X) ');
            WRITE('(H or ? for Help) ');
            READLN(ch);
            uppercase (ch);
            IF (ch = 'C') OR (ch = 'F') OR (ch = 'O') OR (ch = 'R')
               OR (ch = 'S') OR (ch = 'T') OR (ch = 'U') OR (ch = 'X')
               OR (ch = 'Q') OR (ch = '+') OR (ch = '-') OR (ch = '.')
               OR (ch = 'W') OR (ch = 'H') OR (ch = '?')
            THEN CASE ch  OF
               'C' : clade;
               'F' : flip;
               'O' : changeoutgroup;
               'R' : rearrange;
               'S' : show;
               '+' : nextchar;
               '-' : prevchar;
               '.' : again;
               'T' : try;
               'U' : undo;
               'W' : treewrite;
               'Q' : done := TRUE;
               'X' : done := TRUE;
               'H' : help;
               '?' : help;
               END;
         UNTIL done;
         IF NOT written
         THEN BEGIN
            WRITE(' Do you want to write out the tree to');
            WRITE(' a file? (Y or N) ');
            READLN(ch);
            IF (ch = 'Y') OR (ch = 'y')
            THEN treewrite;
            END;
      END; (* redisplay *)

   BEGIN (* treeconstruct *)
      subtree := FALSE;
      display := FALSE;
      dispchar := 0;
      fullset := [1..bits];
      buildtree;
      lastop := none;
      redisplay;
   END; (* treeconstruct *)

BEGIN (* Interactive Dollo/polymorphism parsimony *)
   (* reads in spp, chars, and the data. Then calls treeconstruct to
      construct the tree and query the user *)
   ASSIGN(infile, 'INFILE');  (* Turbo Pascal only *)
   RESET(infile);
   (* put any ASSIGN statement for treefile here *)
   ASSIGN(treefile, 'TREEFILE');  (* Turbo Pascal only *)
   configure;
   logo;
   doinput;
   treeconstruct;
(* CLOSE(infile);  -- may need this here *)
END. (* Interactive Dollo/polymorphism parsimony *)

