PROGRAM Dotgram (treefile, parmfile, fontfile, plotfile, INPUT, OUTPUT);
   (* Copyright 1986 by Christopher A. Meacham,
      University Herbarium, University of California, Berkeley,
      California 94720, U.S.A.

      Permission is granted to copy, distribute, and modify this
      program provided that (1) this copyright message is not removed
      and (2) no fee is charged for this program.

      PHYLIP version 3.0.  Based on PLOTGRAM Version 1.3, July 1986.
      Modifications by J. Felsenstein.
      Incorporates PLOTGRAM modifications and bug-fixes suggested by
      D. L. Swofford and J. M. Humphries.  *)
   CONST smallnum = 0.0001;
      maxnodes = 100;
      maxlabels = 100;
      maxnch = 25;
      fontsize = 3200;    (* Fontsize of font1: 1165; font2: 2162;  *)
      bignum = 1000000.0; (*  font3: 3190; font4: 2173; font5: 3200 *)
      stripedepth = 4;
      stripewidth = 1350;
   TYPE fontname = (font1, font2, font3, font4, font5);
      modetype = (yes, no);
      diagramtype = (phylogram, phenogram, cladogram);
      verticaltype = (levels, lengths, top, bottom, distance);
      horizontaltype = (straightline, nearest, furthest);
      drawingtype = (rectangle, slant);
      sidetype = (rightside, leftside);
      posoptiontype = (atroot, right, left, above, below, center);
      penstatustype = (penup, pendown);
      printertype = (citoh, epson, okidata, toshiba, other);
      fonttype = ARRAY [1..fontsize] OF INTEGER;
      plotstring = ARRAY [1..maxnch] OF CHAR;
      nodeptr = ^treenode;
      labelptr = ^labelrecord;
      treenode = RECORD
         ancestor, sibling, descendant, lastdesc,
            left, right : nodeptr;
         nodenum, fromtop, frombottom : INTEGER;
         level, x, y, xstem, ystem : REAL;
         nodedone, popped : BOOLEAN;
         labels : labelptr;
         END;
      labelrecord = RECORD
         nextlab : labelptr;
         lab : plotstring;
         tipnum, nch  : INTEGER;
         xlab, ylab : REAL;
         END;

   VAR i, nnodes, nlabels, ntips, maxdepth, xnow, ynow,
         oldxhigh, oldxlow, oldyhigh, oldylow, popval,
         strpwide, strpdeep, strpdiv, totalfontsize : INTEGER;
      pi, rotation, xposition, yposition,
         xunitspercm, yunitspercm, aspectratio,
         xsize, ysize, xrange, yrange, labelheight,
         labelrotation, treeheight, treewidth, spacing,
         gap, scaledlabh, scaledspacing : REAL;
      stripe : ARRAY[1..stripedepth] OF ARRAY[1..stripewidth] OF CHAR;
      parmfile, treefile, fontfile, plotfile : TEXT;
      labelspresent, fontloaded, parmsok : BOOLEAN;
      diagram : diagramtype;
      vertical : verticaltype;
      horizontal : horizontaltype;
      drawing : drawingtype;
      fonttoload : fontname;
      font : fonttype;
      sizemode, leveltips, penchange, widthmode, popmode, pfilemode,
         startmode : modetype;
      printer : printertype;
      xoption, yoption : posoptiontype;
      root, garbnode : nodeptr;
      garblabel : labelptr;
      nodelist : ARRAY [1..maxnodes] OF nodeptr;
      labellist : ARRAY [1..maxlabels] OF labelptr;

   PROCEDURE initparms;
      CONST nabs = 26;
      VAR ab1, ab2 : ARRAY [0..nabs] OF CHAR;

      PROCEDURE printrparms(ch : CHAR);
      BEGIN (* printrparms *)
         CASE ch OF
            'C': BEGIN
               printer := citoh;
               xunitspercm := 28.346;
               yunitspercm := 28.346;
               xsize := 19.0;
               ysize := 22.0;
               xposition := 11.0;
               yposition := 0.0;
               xoption := center;
               yoption := above;
               rotation := 0.0;
               strpwide := 640;
               strpdeep := 8;
               strpdiv := 1;
               END;
            'E': BEGIN
               printer := epson;
               xunitspercm := 47.245;
               yunitspercm := 28.346;
               xsize := 19.0;
               ysize := 22.0;
               xposition := 9.5;
               yposition := 0.0;
               xoption := center;
               yoption := above;
               rotation := 0.0;
               strpwide := 900;
               strpdeep := 8;
               strpdiv := 1;
               END;
            'K': BEGIN
               printer := okidata;
               xunitspercm := 56.692;
               yunitspercm := 28.346;
               xsize := 19.0;
               ysize := 22.0;
               xposition := 9.5;
               yposition := 0.0;
               xoption := center;
               yoption := above;
               rotation := 0.0;
               strpwide := 1100;
               strpdeep := 8;
               strpdiv := 1;
               END;
            'T': BEGIN
               printer := toshiba;
               xunitspercm := 70.866;
               yunitspercm := 70.866;
               xsize := 19.0;
               ysize := 25.0;
               xposition := 10.5;
               yposition := 0.0;
               xoption := center;
               yoption := above;
               rotation := 0.0;
               strpwide := 1350;
               strpdeep := 24;
               strpdiv := 4;
               END;
            'O': BEGIN
               printer := other;
               (* initial parameter settings for a new
                  printer go here *)
               END
            END;
      END; (* printrparms *)

      PROCEDURE readparms;
         VAR parmnum : INTEGER;
            ch, c1, c2 : CHAR;

         PROCEDURE uppercase (VAR ch : CHAR);

            FUNCTION letter (ch : CHAR) : BOOLEAN;
            BEGIN (* letter -- works for both ASCII and EBCDIC codes *)
               letter := ((ch >= 'A') AND (ch <= 'I'))
                         OR ((ch >= 'J') AND (ch <= 'R'))
                         OR ((ch >= 'S') AND (ch <= 'Z'))
                         OR ((ch >= 'a') AND (ch <= 'i'))
                         OR ((ch >= 'j') AND (ch <= 'r'))
                         OR ((ch >= 's') AND (ch <= 'z'));
            END; (* letter *)

         BEGIN (* uppercase *)
            IF letter(ch) AND ((('A' < 'a') AND (ch >= 'a'))
                            OR (('A' > 'a') AND (ch < 'A')))
            THEN ch := CHR(ORD(ch) + ORD('A') - ORD('a'));
         END; (* uppercase *)

         PROCEDURE getp(VAR ch : CHAR);
         BEGIN (* getp *)
            ch := ' ';
            WHILE (ch=' ') AND NOT EOLN(parmfile) DO
               READ(parmfile, ch);
            uppercase(ch);
         END; (* getp *)

         FUNCTION setmode : modetype;
         BEGIN (* setmode *)
            getp(ch);
            CASE ch OF
               'Y': setmode := yes;
               'N': setmode := no
               END;
         END; (* setmode *)

      BEGIN (* readparms *)
         WHILE NOT EOF(parmfile) DO BEGIN
            c1 := ' ';
            c2 := ' ';
            WHILE NOT EOLN(parmfile) AND (c1=' ') DO
               READ(parmfile, c1);
            IF c1<>'*' THEN BEGIN
               IF NOT EOLN(parmfile) THEN READ(parmfile, c2);
               REPEAT READ(parmfile, ch) UNTIL EOLN(parmfile)
                                                      OR (ch=' ');
               uppercase(c1);
               uppercase(c2);
               parmnum := 0;
               WHILE NOT((c1=ab1[parmnum]) AND (c2=ab2[parmnum])) AND
                  (parmnum<=nabs) DO parmnum := parmnum+1;
               IF parmnum<=nabs
                  THEN CASE parmnum OF
                     1: BEGIN
                        getp(ch);
                        printrparms(ch)
                        END;
                     2: BEGIN
                        getp(ch);
                        CASE ch OF
                           'C': diagram := cladogram;
                           'E': BEGIN
                              diagram := phenogram;
                              vertical := levels;
                              horizontal := nearest;
                              drawing := rectangle;
                              widthmode := no
                              END;
                           'Y': BEGIN
                              diagram := phylogram;
                              vertical := lengths;
                              horizontal := nearest;
                              drawing := rectangle;
                              leveltips := no;
                              widthmode := no
                              END
                           END;
                        END;
                     3: BEGIN
                        getp(ch);
                        CASE ch OF
                           'V': vertical := levels;
                           'G': vertical := lengths;
                           'T': vertical := top;
                           'B': vertical := bottom;
                           'D': vertical := distance
                           END;
                        END;
                     4: BEGIN
                        getp(ch);
                        CASE ch OF
                           'S': horizontal := straightline;
                           'N': horizontal := nearest;
                           'F': horizontal := furthest
                           END;
                        END;
                     5: leveltips := setmode;
                     6: BEGIN
                        getp(ch);
                        CASE ch OF
                           'R': drawing := rectangle;
                           'S': drawing := slant
                           END;
                        END;
                     7: READ(parmfile, rotation);
                     8: BEGIN
                        getp(ch);
                        CASE ch OF
                           '1': fonttoload := font1;
                           '2': fonttoload := font2;
                           '3': fonttoload := font3;
                           '4': fonttoload := font4;
                           '5': fonttoload := font5
                           END;
                        END;
                     9: READ(parmfile, xunitspercm);
                     10: READ(parmfile, yunitspercm);
                     11: READ(parmfile, xsize);
                     12: READ(parmfile, ysize);
                     13: READ(parmfile, xposition);
                     14: READ(parmfile, yposition);
                     15: BEGIN
                        getp(ch);
                        CASE ch OF
                           'O': xoption := atroot;
                           'C': xoption := center;
                           'R': xoption := right;
                           'L': xoption := left
                           END;
                        END;
                     16: BEGIN
                        getp(ch);
                        CASE ch OF
                           'O': yoption := atroot;
                           'C': yoption := center;
                           'A': yoption := above;
                           'B': yoption := below
                           END;
                        END;
                     17: penchange := setmode;
                     18: widthmode := setmode;
                     19: READ (parmfile, treeheight);
                     20: READ (parmfile, treewidth);
                     21: READ (parmfile, labelheight);
                     22: READ (parmfile, spacing);
                     23: READ (parmfile, gap);
                     24: READ (parmfile, labelrotation);
                     25: popmode := setmode;
                     26: startmode := setmode;
                     END
                  ELSE BEGIN
                     WRITELN(OUTPUT, ' PARAMETER LABEL THAT BEGINS ''',
                        c1, c2, ''' IS NOT A LEGITIMATE LABEL');
                     parmsok := FALSE;
                     READLN(parmfile);
                     END;
               END;
            READLN(parmfile);
            END;
      END; (* readparms *)

   BEGIN (* initparms *)
      ab1[0] := ' '; ab2[0] := ' ';
      ab1[1] := 'P'; ab2[1] := 'R'; ab1[2] := 'D'; ab2[2] := 'I';
      ab1[3] := 'V'; ab2[3] := 'E'; ab1[4] := 'H'; ab2[4] := 'O';
      ab1[5] := 'T'; ab2[5] := 'I'; ab1[6] := 'D'; ab2[6] := 'R';
      ab1[7] := 'R'; ab2[7] := 'O'; ab1[8] := 'F'; ab2[8] := 'O';
      ab1[9] := 'X'; ab2[9] := 'U'; ab1[10] := 'Y'; ab2[10] := 'U';
      ab1[11] := 'X'; ab2[11] := 'S'; ab1[12] := 'Y'; ab2[12] := 'S';
      ab1[13] := 'X'; ab2[13] := 'P'; ab1[14] := 'Y'; ab2[14] := 'P';
      ab1[15] := 'X'; ab2[15] := 'O'; ab1[16] := 'Y'; ab2[16] := 'O';
      ab1[17] := 'W'; ab2[17] := 'I'; ab1[18] := 'T'; ab2[18] := 'H';
      ab1[19] := 'T'; ab2[19] := 'W'; ab1[20] := 'L'; ab2[20] := 'H';
      ab1[21] := 'S'; ab2[21] := 'P'; ab1[22] := 'G'; ab2[22] := 'A';
      ab1[23] := 'L'; ab2[23] := 'R'; ab1[24] := 'R'; ab2[24] := 'O';
      ab1[25] := 'P'; ab2[25] := 'O'; ab1[26] := 'S'; ab2[26] := 'T';
      garbnode := NIL;
      garblabel := NIL;
      diagram := cladogram;
      vertical := distance;
      horizontal := straightline;
      leveltips := yes;
      popmode := no;
      drawing := slant;
      widthmode := yes;
      nnodes := 0;
      ntips := 0;
      nlabels := 0;
      pi := 3.1415926536;
      parmsok := TRUE;
      startmode := no;
      pfilemode := yes;
      labelspresent := FALSE;
      fontloaded := FALSE;
      sizemode := yes;
      fonttoload := font1;
      treeheight := 12.5;
      treewidth := 25.5;
      labelrotation := 0.0;
      labelheight := 0.5;
      spacing := 1.0;
      gap := 1.0;
      printrparms('E');  (* default printer = Epson *)
      readparms;
      IF popmode = yes
         THEN popval := 1
         ELSE popval := 0;
      gap := 0.5 * gap * labelheight;
      spacing := 0.75 * spacing * labelheight;
      rotation := rotation * pi/180.0;
      labelrotation := rotation - pi/2.0 + labelrotation*pi/180.0;
      aspectratio := yunitspercm/xunitspercm;
   END; (* initparms *)

   PROCEDURE newnode(VAR p : nodeptr);
   BEGIN (* newnode *)
      IF garbnode <> NIL
         THEN BEGIN
            p := garbnode;
            garbnode := garbnode^.descendant;
            END
         ELSE NEW(p);
      nnodes := nnodes + 1;
      nodelist[nnodes] := p;
      p^.nodenum := nnodes;
      p^.ancestor := NIL;
      p^.descendant := NIL;
      p^.sibling := NIL;
      p^.lastdesc := NIL;
      p^.labels := NIL;
      p^.nodedone := FALSE;
      p^.popped := FALSE;
      p^.level := 0.0;
      p^.ystem := 0.0;
   END; (* newnode *)

   PROCEDURE newlabel(VAR lp : labelptr);
   BEGIN (* newlabel *)
      IF garblabel <> NIL
         THEN BEGIN
            lp := garblabel;
            garblabel := garblabel^.nextlab;
            END
         ELSE NEW(lp);
      nlabels := nlabels+1;
      labellist[nlabels] := lp;
      lp^.nextlab := NIL;
      lp^.nch := 0;
   END; (* newlabel *)

   PROCEDURE readtree;
      VAR i : INTEGER;
         ch : CHAR;
         node : nodeptr;

      PROCEDURE addnode(anc : nodeptr; VAR linktonode,
         anclastdesc : nodeptr);
         CONST blank = ' ';
            point = '.';
            comma = ',';
            colon = ':';
            leftp = '(';
            rightp = ')';
            semi = ';';
            underline = '_';
         VAR node : nodeptr;
            doingdescendants : BOOLEAN;

         PROCEDURE getch(VAR ch : CHAR);
         BEGIN (* getch *)
            IF EOF(treefile)
               THEN BEGIN
                  WRITELN(OUTPUT, '  NO '';'' AT END OF TREE ',
                     'DESCRIPTION, '';'' ASSUMED');
                  ch := ';';
                  END
               ELSE BEGIN
                  IF EOLN(treefile) THEN READLN(treefile);
                  READ(treefile, ch);
                  END;
         END; (* getch *)

         PROCEDURE processnodelabel(node : nodeptr);
            VAR lp, lp2 : labelptr;
         BEGIN (* processnodelabel *)
            newlabel(lp);
            IF node^.labels=NIL
               THEN BEGIN
                  labelspresent := TRUE;
                  node^.labels := lp;
                  END
               ELSE BEGIN
                  lp2 := node^.labels;
                  WHILE lp2^.nextlab <> NIL DO lp2 := lp2^.nextlab;
                  lp2^.nextlab := lp;
                  END;
            REPEAT
               IF lp^.nch < maxnch THEN BEGIN
                  lp^.nch := lp^.nch + 1;
                  IF ch = underline THEN ch := blank;
                  lp^.lab[lp^.nch] := ch;
                  getch(ch);
                  END;
               UNTIL (ch=blank) OR (ch=leftp) OR (ch=comma) OR
                  (ch=rightp) OR (ch=colon) OR (ch=semi);
         END; (* processnodelabel *)

         PROCEDURE processlength(node : nodeptr);
            VAR digit, ordzero, ordnine : INTEGER;
               lenvalue, divisor : REAL;
               pointread : BOOLEAN;
         BEGIN (* processlength *)
            ordzero := ORD('0');
            ordnine := ORD('9');
            pointread := FALSE;
            lenvalue := 0.0;
            divisor := 1.0;
            IF ch='-' THEN divisor := -divisor;
            IF (ch='+') OR (ch='-') THEN getch(ch);
            REPEAT getch(ch) UNTIL ch <> blank;
            digit := ORD(ch)-ordzero;
            WHILE ((digit >= 0) AND (digit <= 9)) OR (ch=point)
               DO BEGIN
               IF ch = point
                  THEN pointread := TRUE
                  ELSE BEGIN
                     lenvalue := lenvalue*10.0 + digit;
                     IF pointread THEN divisor := divisor*10.0;
                     END;
               getch(ch);
               digit := ORD(ch)-ordzero;
               END;
            node^.level := lenvalue/divisor;
            IF node^.level<0.0 THEN node^.level := 0.0;
         END; (* processlength *)

      BEGIN (* addnode *)
         doingdescendants := FALSE;
         newnode(node);
         linktonode := node;
         anclastdesc := node;
         node^.ancestor := anc;
         WHILE ch=blank DO getch(ch);
         WHILE (ch <> semi) AND (((ch <> comma) AND
            (ch <> rightp)) OR doingdescendants)
            DO BEGIN
            IF ch = leftp
               THEN BEGIN
                  doingdescendants := TRUE;
                  getch(ch);
                  addnode(node, node^.descendant,
                     node^.lastdesc);
                  END
               ELSE IF ch = comma
                  THEN BEGIN
                     getch(ch);
                     addnode(node, node^.lastdesc^.sibling,
                        node^.lastdesc);
                     END
                  ELSE IF ch = rightp
                     THEN BEGIN
                        doingdescendants := FALSE;
                        REPEAT getch(ch) UNTIL ch <> blank;
                        END
                     ELSE IF ch = colon
                        THEN processlength(node)
                        ELSE IF ch <> semi
                           THEN processnodelabel(node);
            WHILE ch = blank DO getch(ch);
            END;
      END; (* addnode *)

   BEGIN (* readtree *)
      ch := ' ';
      addnode(NIL, root, root);
      READLN(treefile);
   END; (* readtree *)

   PROCEDURE calctree(root : nodeptr);
      VAR count : INTEGER;
         fullscale, maxlevel, minlevel : REAL;

      PROCEDURE popall(p : nodeptr);
         VAR newp, prev : nodeptr;
            lp : labelptr;
      BEGIN (* popall *)
         prev := NIL;
         lp := p^.labels;
         WHILE lp <> NIL DO BEGIN
            newnode(newp);
            newp^.popped := TRUE;
            lp := lp^.nextlab;
            IF p^.descendant = NIL
               THEN p^.lastdesc := newp
               ELSE newp^.sibling := p^.descendant;
            p^.descendant := newp;
            newp^.ancestor := p;
            IF vertical = levels THEN newp^.level := p^.level;
            END;
         lp := p^.labels;
         p^.labels := NIL;
         newp := p^.descendant;
         WHILE lp <> NIL DO BEGIN
            newp^.labels := lp;
            lp := lp^.nextlab;
            newp^.labels^.nextlab := NIL;
            newp := newp^.sibling;
            END;
      END; (* popall *)

      PROCEDURE popout(p : nodeptr);
      BEGIN (* popout *)
         IF p^.descendant <> NIL
            THEN BEGIN
               popout(p^.descendant);
               IF p^.labels <> NIL THEN popall(p);
               END
            ELSE IF p^.labels^.nextlab <> NIL
                  THEN popall(p);
         IF p^.sibling <> NIL THEN popout(p^.sibling);
      END; (* popout *)

      PROCEDURE counttips(p : nodeptr);
         VAR ptr : nodeptr;
      BEGIN (* counttips *)
         p^.fromtop := 0;
         IF p = root
            THEN p^.frombottom := 0
            ELSE BEGIN
               IF p^.popped
                  THEN p^.frombottom := p^.ancestor^.frombottom +
                                                              popval
                  ELSE p^.frombottom := p^.ancestor^.frombottom + 1;
               END;
         IF p^.frombottom > maxdepth THEN maxdepth := p^.frombottom;
         IF p^.descendant = NIL
            THEN BEGIN
               ntips := ntips + 1;
               IF p^.labels<>NIL THEN p^.labels^.tipnum := ntips;
               END
            ELSE BEGIN
               counttips(p^.descendant);
               ptr := p^.descendant;
               WHILE ptr <> NIL DO BEGIN
                  IF ptr^.popped
                     THEN BEGIN
                        IF p^.fromtop < ptr^.fromtop+popval
                           THEN p^.fromtop := ptr^.fromtop+popval;
                        END
                     ELSE BEGIN
                        IF ptr^.fromtop+1 > p^.fromtop
                           THEN p^.fromtop := ptr^.fromtop+1;
                        END;
                  ptr := ptr^.sibling;
                  END;
               ptr := p^.descendant;
               WHILE ptr <> NIL DO BEGIN
                  IF ptr^.popped THEN ptr^.fromtop := p^.fromtop-popval;
                  ptr := ptr^.sibling;
                  END;
               END;
         IF p^.sibling <> NIL THEN counttips(p^.sibling);
      END; (* counttips *)

      PROCEDURE placenodes(p : nodeptr; VAR count : INTEGER);
      BEGIN (* placenodes *)
         IF p^.descendant = NIL
            THEN BEGIN
               p^.left := p;
               p^.right := p;
               count := count+1;
               p^.x := (count-1)/(ntips-1);
               CASE vertical OF
                  lengths: BEGIN END;
                  top: p^.level := 1.0-p^.fromtop/maxdepth;
                  bottom: p^.level := p^.frombottom/maxdepth;
                  distance: p^.level := 0.0
                  END;
               p^.nodedone := TRUE;
               END
            ELSE BEGIN
               placenodes(p^.descendant, count);
               p^.left := p^.descendant^.left;
               p^.right := p^.lastdesc^.right;
               IF horizontal = furthest
                  THEN p^.x := (p^.left^.x+p^.right^.x)/2.0
                  ELSE p^.x := (p^.descendant^.x+p^.lastdesc^.x)/2.0;
               CASE vertical OF
                  lengths: BEGIN END;
                  top: p^.level := 1.0-p^.fromtop/maxdepth;
                  bottom: p^.level := p^.frombottom/maxdepth;
                  distance: p^.level := -(p^.right^.x-p^.left^.x)
                  END;
               END;
         IF p^.sibling <> NIL THEN placenodes(p^.sibling, count);
      END; (* placenodes *)

      PROCEDURE addlengths(p : nodeptr);
         VAR currentlevel : REAL;
      BEGIN (* addlengths *)
         IF p <> root THEN p^.level := p^.level + p^.ancestor^.level;
         IF p^.descendant <> NIL THEN addlengths(p^.descendant);
         IF p^.sibling <> NIL THEN addlengths(p^.sibling);
      END; (* addlengths *)

      PROCEDURE lineupnodes(p : nodeptr);

         PROCEDURE singleline(p : nodeptr; side : sidetype;
               pstart, pend : nodeptr);
         BEGIN (* singleline *)
            WHILE p <> NIL DO BEGIN
               p^.x := (p^.y-pstart^.y)*(pend^.x-pstart^.x)/
                                  (pend^.y-pstart^.y)+pstart^.x;
               p^.nodedone := TRUE;
               IF side = leftside
                  THEN p := p^.descendant
                  ELSE p := p^.lastdesc;
               END;
         END; (* singleline *)

      BEGIN (* lineupnodes *)
         IF p <> NIL THEN BEGIN
            IF NOT p^.nodedone THEN BEGIN
               IF ABS(p^.ancestor^.x-p^.right^.x) >=
                       ABS(p^.ancestor^.x-p^.left^.x)
                  THEN singleline(p, rightside, p^.ancestor,
                     p^.right)
                  ELSE singleline(p, leftside, p^.ancestor,
                     p^.left);
               END;
            lineupnodes(p^.descendant);
            lineupnodes(p^.sibling);
            END;
      END; (* lineupnodes *)

      PROCEDURE levellimits(p : nodeptr);
      BEGIN (* levellimits *)
         IF p^.descendant <> NIL THEN levellimits(p^.descendant);
         IF p^.sibling <> NIL THEN levellimits(p^.sibling);
         IF p^.level > maxlevel THEN maxlevel := p^.level;
         IF p^.level < minlevel THEN minlevel := p^.level;
      END; (* levellimits *)

      PROCEDURE scalenodes(p : nodeptr);
      BEGIN (* scalenodes *)
         IF p^.descendant <> NIL
               THEN scalenodes(p^.descendant);
         IF (leveltips = yes) AND (p^.descendant = NIL)
            THEN p^.y := 1.0
            ELSE p^.y := ABS(p^.level-root^.level)/fullscale;
         IF p^.labels<>NIL THEN p^.labels^.ylab := p^.y;
         IF p^.sibling <> NIL THEN scalenodes(p^.sibling);
      END; (* scalenodes *)

      PROCEDURE setstems(p : nodeptr);
         VAR ptr : nodeptr;
      BEGIN (* setstems *)
         IF p^.descendant <> NIL THEN setstems(p^.descendant);
         p^.xstem := p^.x;
         ptr := p^.descendant;
         WHILE ptr <> NIL DO BEGIN
            ptr^.ystem := p^.y;
            ptr := ptr^.sibling;
            END;
         IF p^.sibling <> NIL THEN setstems(p^.sibling);
      END; (* setstems *)

   BEGIN (* calctree *)
      popout(root);
      counttips(root);
      count := 0;
      placenodes(root, count);
      IF vertical = lengths THEN addlengths(root);
      maxlevel := -bignum;
      minlevel := bignum;
      levellimits(root);
      fullscale := maxlevel-minlevel;
      scalenodes(root);
      IF horizontal = straightline THEN BEGIN
         root^.x := 0.5;
         root^.nodedone := TRUE;
         lineupnodes(root);
         END;
      IF drawing = rectangle THEN setstems(root);
   END; (* calctree *)

   PROCEDURE hardwarecoordinates;
      VAR i : INTEGER;
         factor, xmax, xmin, ymax, ymin, thislength,
            dsin, dcos, lsin, lcos, xtrans, ytrans,
            xcorrect, ycorrect, dum, tail : REAL;
         lp : labelptr;
         p : nodeptr;

      FUNCTION lengthtext(pstring : plotstring; nchars : INTEGER;
         height, compress : REAL; VAR font : fonttype) : REAL;
         VAR i, j, code, coord : INTEGER;
            sumlength, heightfont, widthfont, yfactor : REAL;
      BEGIN (* lengthext *)
         sumlength := 0.0;
         FOR i := 1 TO nchars DO BEGIN
            code := ORD(pstring[i]);
            j := 1;
            WHILE (font[j+1] <> code) AND (font[j] <> 0) DO
               j := font[j];
            IF font[j+1] = code THEN BEGIN
               heightfont := font[j+2];
               widthfont := font[j+3];
               sumlength := sumlength + (height/heightfont)
                     *widthfont;
               END;
            END;
            lengthtext := sumlength*compress;
      END; (* lengthtext *)

      PROCEDURE rotate(VAR x, y : REAL; xdisp, ydisp, xcenter, ycenter,
             sinrot, cosrot : REAL);
      BEGIN (* rotate *)
         x := xcenter+xdisp*cosrot+ydisp*sinrot;
         IF xmin>x THEN xmin := x;
         IF xmax<x THEN xmax := x;
         y := ycenter-xdisp*sinrot+ydisp*cosrot;
         IF ymin>y THEN ymin := y;
         IF ymax<y THEN ymax := y;
      END; (* rotate *)

   BEGIN (* hardwarecoordinates *)
      xmin := bignum;
      xmax := -bignum;
      ymin := bignum;
      ymax := -bignum;
      dsin := SIN(rotation);
      dcos := COS(rotation);
      lsin := SIN(labelrotation);
      lcos := COS(labelrotation);
      scaledlabh := labelheight;
      scaledspacing := spacing;
      IF widthmode = yes
         THEN BEGIN
            factor := treewidth / ((ntips-1)*(labelheight+spacing));
            IF factor < 1.0
               THEN BEGIN
                  scaledlabh := labelheight*factor;
                  scaledspacing := spacing*factor;
                  END
               ELSE scaledspacing := (treewidth/(ntips-1)) - labelheight;
            END
         ELSE treewidth := (ntips-1)*(labelheight+spacing);
      xcorrect := -scaledlabh*(1+SIN(labelrotation-rotation)/2.0);
      ycorrect := -scaledlabh*COS(labelrotation-rotation)/2.0;
      tail := scaledlabh*((totalfontsize/font[3])-1);
      FOR i := 1 TO nlabels DO BEGIN
         lp := labellist[i];
         lp^.xlab := lp^.tipnum*scaledlabh +
                           (lp^.tipnum-1)*scaledspacing + xcorrect;
         lp^.ylab := lp^.ylab*treeheight + gap + ycorrect;
         thislength := lengthtext(lp^.lab,lp^.nch,scaledlabh,1.0,font);
         rotate(dum,dum,0.0,scaledlabh,lp^.xlab,lp^.ylab,lsin,lcos);
         rotate(dum,dum,thislength,scaledlabh,lp^.xlab,lp^.ylab,lsin,lcos);
         rotate(dum,dum,thislength,-tail,lp^.xlab,lp^.ylab,lsin,lcos);
         rotate(dum,dum,0.0,-tail,lp^.xlab,lp^.ylab,lsin,lcos);
         END;
      FOR i := 1 TO nnodes DO BEGIN
         p := nodelist[i];
         p^.x := p^.x * treewidth;
         p^.y := p^.y * treeheight;
         rotate(p^.x,p^.y,p^.x,p^.y,0.0,0.0,dsin,dcos);
         IF drawing = rectangle THEN BEGIN
            p^.xstem := p^.xstem * treewidth;
            p^.ystem := p^.ystem * treeheight;
            rotate(p^.xstem,p^.ystem,p^.xstem,p^.ystem,0.0,0.0,
                  dsin,dcos);
            END;
         END;
      xrange := xmax - xmin;
      yrange := ymax - ymin;
      IF sizemode = yes
         THEN BEGIN
            IF ysize/xsize > yrange/xrange
               THEN factor := xsize/xrange
               ELSE factor := ysize/yrange;
            END
         ELSE factor := 1.0;
      CASE xoption OF
         atroot: xtrans := xposition;
         right: xtrans := xposition - xmin*factor;
         left: xtrans := xposition - xmax*factor;
         center: xtrans := xposition + (xrange/2.0 - xmax)*factor
         END;
      CASE yoption OF
         atroot: ytrans := 0.0;
         above: ytrans := yposition - ymin*factor;
         below: ytrans := yposition - ymax*factor;
         center: ytrans := yposition + (yrange/2.0 - ymax)*factor
         END;
      FOR i := 1 TO nlabels DO BEGIN
         lp := labellist[i];
         lp^.xlab := (lp^.xlab*factor+xtrans)*xunitspercm;
         lp^.ylab := (lp^.ylab*factor+ytrans)*xunitspercm;
         END;
      FOR i := 1 TO nnodes DO BEGIN
         p := nodelist[i];
         p^.x := (p^.x*factor+xtrans)*xunitspercm;
         p^.y := (p^.y*factor+ytrans)*xunitspercm;
         IF (drawing = rectangle) AND (i <> 1) THEN BEGIN
            p^.xstem := (p^.xstem*factor+xtrans)*xunitspercm;
            p^.ystem := (p^.ystem*factor+ytrans)*xunitspercm;
            END;
         END;
      scaledlabh := scaledlabh * factor * xunitspercm;
      xrange := xrange * factor * xunitspercm;
      yrange := yrange * factor * xunitspercm;
   END; (* hardwarecoordinates *)

   PROCEDURE loadfont(VAR font : fonttype);
      VAR i, charstart, hersheynum : INTEGER;
         ch : CHAR;
   BEGIN (* loadfont *)
      ch := 'A';
      i := 0;
      WHILE NOT(EOF(fontfile) OR (ch = ' ')) DO BEGIN
         charstart := i+1;
         READ(fontfile, ch, ch, hersheynum,
            font[charstart+2], font[charstart+3]);
         IF (i=0) THEN READ(fontfile, totalfontsize);
         font[charstart+1] := ORD(ch);
         i := charstart+3;
         REPEAT
            IF (i-(charstart+3)) MOD 10=0 THEN READLN(fontfile);
            i := i+1;
            READ(fontfile,font[i]);
            UNTIL ABS(font[i]) >= 10000;
         READLN(fontfile);
         font[charstart] := i+1;
         END;
      font[charstart] := 0;
      WRITELN(OUTPUT, '  FONT REQUIRES', i:5, ' INTEGERS OUT OF',
         fontsize:5, ' AVAILABLE');
   END; (* loadfont *)

   PROCEDURE initplotter;
   BEGIN (* initprinter *)
      CASE printer OF
         citoh : WRITE(plotfile, CHR(27), 'T16');
         epson : BEGIN
                 WRITE(plotfile, CHR(27), 'A', CHR(8));
                 WRITE(plotfile, CHR(27), '2');
                 END;
         okidata : WRITE(plotfile, CHR(27), '%9', CHR(16));
         toshiba : BEGIN
                   WRITELN(plotfile, CHR(27), CHR(26), 'I');
                   WRITELN(plotfile);
                   WRITELN(plotfile, CHR(27), 'L06');
                   END;
         other : BEGIN
            (* initialization code for a new printer goes here *)
            END
         END;
   END; (* initprinter *)

   PROCEDURE draw(p : nodeptr);
      VAR i, j, line, strptop, strpbottom : INTEGER;
         xnow, ynow, astrptop, astrpbottom : REAL;
         null : CHAR;
         empty : BOOLEAN;
         bit : ARRAY[0..7] OF INTEGER;

      PROCEDURE plot(pen : penstatustype; xabs, yabs : REAL);
         VAR ixnow, iynow, ixabs, iyabs, ix, iy, iy1, iy2, temp,
            xdiff, ydiff, err, x1, y1, i, n : INTEGER;
            xinc, yinc, x, y : REAL;

         PROCEDURE plotdot (ix, iy : INTEGER);
           (* plot one dot at ix, iy *)
            VAR iy1, iy2 : INTEGER;
         BEGIN (* plotdot *)
            iy := strptop - iy;
            IF (iy >= 0) AND (iy < strpdeep) AND (ix>0) AND
               (ix<=strpwide)
            THEN BEGIN
               empty := FALSE;
               CASE printer OF
                  citoh: BEGIN
                     iy1 := 1;
                     iy2 := iy;
                     END;
                  epson: BEGIN
                     iy1 := 1;
                     iy2 := 7 - iy;
                     END;
                  okidata: BEGIN
                     iy1 := 1;
                     iy2 := 7 - iy;
                     END;
                  toshiba: BEGIN
                     iy1 := iy DIV 6 + 1;
                     iy2 := 5 - (iy MOD 6);
                     END;
                  other: BEGIN
                     (* code for making dot array for a new printer
                        goes here *)
                     END
                  END;
               IF ((ORD(stripe[iy1][ix]) DIV bit[iy2]) MOD 2) = 0
               THEN stripe[iy1][ix] :=
                  CHR(ORD(stripe[iy1][ix]) + bit[iy2]);
               END;
         END; (* plotdot *)

      BEGIN (* plot *)
         yabs := yabs * aspectratio;
         ixabs := ROUND(xabs);
         iyabs := ROUND(yabs);
         ixnow := ROUND(xnow);
         iynow := ROUND(ynow);
         IF (pen = pendown) AND
            (NOT (((iyabs < strpbottom) AND (iynow < strpbottom))
               OR ((iyabs > strptop) AND (iynow > strptop))))
         THEN BEGIN
            IF ixabs < ixnow
            THEN BEGIN
               temp := ixnow;
               ixnow := ixabs;
               ixabs := temp;
               temp := iynow;
               iynow := iyabs;
               iyabs := temp;
               END;
            xdiff := ixabs - ixnow;
            ydiff := iyabs - iynow;
            IF ydiff >= 0
            THEN IF xdiff >= ydiff
               THEN BEGIN
                  err := -xdiff DIV 2;
                  FOR x1 := ixnow TO ixabs DO BEGIN
                     plotdot (x1, iynow);
                     err := err + ydiff;
                     IF err > 0
                     THEN BEGIN
                        iynow := iynow + 1;
                        err := err - xdiff;
                        END;
                     END;
                  END
               ELSE BEGIN
                  err := -ydiff DIV 2;
                  FOR y1 := iynow TO iyabs DO BEGIN
                     plotdot (ixnow, y1);
                     err := err + xdiff;
                     IF err > 0
                     THEN BEGIN
                        ixnow := ixnow + 1;
                        err := err - ydiff;
                        END;
                     END;
                  END
            ELSE IF xdiff < -ydiff
               THEN BEGIN
                  err := ydiff DIV 2;
                  FOR y1 := iynow DOWNTO iyabs DO BEGIN
                     plotdot (ixnow, y1);
                     err := err + xdiff;
                     IF err > 0
                     THEN BEGIN
                        ixnow := ixnow + 1;
                        err := err + ydiff;
                        END;
                     END;
                  END
               ELSE BEGIN
                  err := -xdiff DIV 2;
                  FOR x1 := ixnow TO ixabs DO BEGIN
                     plotdot (x1, iynow);
                     err := err - ydiff;
                     IF err > 0
                     THEN BEGIN
                        iynow := iynow - 1;
                        err := err - xdiff;
                        END;
                     END;
                  END;
            END;
         xnow := xabs;
         ynow := yabs;
      END; (* plot *)

      PROCEDURE plotlabels;
         TYPE pentype = (treepen, labelpen);
         VAR i : INTEGER;
            xlab, ylab, x, y : REAL;
            lp : labelptr;

         PROCEDURE plottext(pstring : plotstring; nchars : INTEGER;
            height, compress : REAL; x, y : REAL;
            slope : REAL; VAR font : fonttype);
            CONST xstart = 10;
               ystart = 35;
            VAR i, j, code, coord : INTEGER;
               heightfont, widthfont, xfactor, yfactor, xfont, yfont,
               xplot, yplot, sinslope, cosslope : REAL;
               penstatus : penstatustype;

            PROCEDURE plotchar(VAR place : INTEGER);
               VAR spacing : REAL;
            BEGIN (* plotchar *)
               heightfont := font[place+2];
               widthfont := font[place+3];
               yfactor := height/heightfont;
               xfactor := yfactor*compress;
               spacing := yfactor*ABS(heightfont*cosslope
                             -widthfont*sinslope)/aspectratio;
               IF (y <= astrptop)
                  AND ((y + spacing) >= astrpbottom)
               THEN BEGIN
                  place := place+3;
                  REPEAT
                     place := place+1;
                     coord := font[place];
                     IF coord > 0
                        THEN penstatus := pendown
                        ELSE penstatus := penup;
                     coord := ABS(coord);
                     coord := coord MOD 10000;
                     xfont := ((coord DIV 100) - xstart)*xfactor;
                     yfont := ((coord MOD 100) - ystart)*yfactor;
                     IF slope = 0.0
                        THEN BEGIN
                           xplot := x + xfont;
                           yplot := y + yfont;
                           END
                        ELSE BEGIN
                           xplot := x + xfont*cosslope
                                       + yfont*sinslope;
                           yplot := y - xfont*sinslope
                                       + yfont*cosslope;
                           END;
                     plot(penstatus, xplot, yplot);
                     UNTIL ABS(font[place]) >= 10000;
                  x := xplot;
                  y := yplot;
                  END
               ELSE BEGIN
                  xfont := xfactor*widthfont;
                  x := x + xfont*cosslope;
                  y := y - xfont*sinslope;
                  END;
            END; (* plotchar *)

         BEGIN (* plottext *)
            sinslope := SIN(slope);
            cosslope := COS(slope);
            FOR i := 1 TO nchars DO BEGIN
               code := ORD(pstring[i]);
               j := 1;
               WHILE (font[j+1] <> code) AND (font[j] <> 0) DO
                  j := font[j];
               IF font[j+1] = code THEN plotchar(j);
               END;
         END; (* plottext *)

      BEGIN (* plotlabels *)
         FOR i := 1 TO nlabels DO BEGIN
            lp := labellist[i];
            plottext(lp^.lab,lp^.nch,scaledlabh,
               1.0,lp^.xlab,lp^.ylab,labelrotation,font);
            END;
      END; (* plotlabels *)

      PROCEDURE slantdrawing(p : nodeptr);
      BEGIN (* slantdrawing *)
         IF p <> NIL THEN BEGIN
            IF p <> root THEN BEGIN
               plot(penup, p^.ancestor^.x, p^.ancestor^.y);
               plot(pendown, p^.x, p^.y);
               END;
            slantdrawing(p^.descendant);
            slantdrawing(p^.sibling);
            END;
      END; (* slantdrawing *)

      PROCEDURE rectdrawing(p : nodeptr);
      BEGIN (* rectdrawing *)
         IF p <> NIL THEN BEGIN
            IF p <> root THEN BEGIN
               IF p = p^.ancestor^.descendant
                  THEN BEGIN
                     plot(penup, p^.ancestor^.lastdesc^.xstem,
                            p^.ancestor^.lastdesc^.ystem);
                     plot(pendown, p^.xstem, p^.ystem);
                     plot(pendown, p^.x, p^.y);
                     END
                  ELSE BEGIN
                     plot(penup, p^.xstem, p^.ystem);
                     plot(pendown, p^.x, p^.y);
                     END;
               END;
            rectdrawing(p^.descendant);
            rectdrawing(p^.sibling);
            END;
      END; (* rectdrawing *)

      PROCEDURE striprint;
         VAR i, j, width : INTEGER;
            done : BOOLEAN;
      BEGIN (* striprint *)
         done := FALSE;
         width := strpwide;
         WHILE NOT done DO BEGIN
            FOR i := 1 TO strpdiv DO
               done := done OR (stripe[i][width] <> null);
            IF NOT done
            THEN width := width - 1;
            done := done OR (width = 0);
            END;
         CASE printer OF
            citoh: BEGIN
               IF NOT empty
               THEN BEGIN
                  WRITE(plotfile, CHR(27), 'S', (width DIV 1000):1,
                     ((width DIV 100) MOD 10):1, ((width DIV 10) MOD 10):1,
                     (width MOD 10):1);
                  FOR i := 1 TO width DO
                     WRITE(plotfile, stripe[1][i]);
                  END;
               WRITELN(plotfile);
               END;
            epson: BEGIN
               IF NOT empty
               THEN BEGIN
                  WRITE(plotfile, CHR(27), 'L', CHR(width MOD 256),
                     CHR(width DIV 256));
                  FOR i := 1 TO width DO
                     WRITE(plotfile, stripe[1][i]);
                  END;
               WRITELN(plotfile);
               END;
            okidata: BEGIN
               IF NOT empty
               THEN BEGIN
                  WRITE(plotfile, CHR(27), '%1', CHR(width DIV 128),
                     CHR(width MOD 128));
                  FOR i := 1 TO width DO
                     WRITE(plotfile, stripe[1][i]);
                  END;
               WRITELN(plotfile);
               END;
            toshiba: BEGIN
               IF NOT empty
               THEN BEGIN
                  FOR i := 1 TO width DO
                     FOR j := 1 TO 4 DO
                        stripe[j][i] := CHR(ORD(stripe[j][i])+64);
                  WRITE(plotfile, CHR(27), ';', (width DIV 1000):1,
                     ((width DIV 100) MOD 10):1, ((width DIV 10) MOD 10):1,
                     (width MOD 10):1);
                  FOR i := 1 TO width DO
                     WRITE(plotfile, stripe[1][i], stripe[2][i],
                            stripe[3][i], stripe[4][i]);
                  END;
               WRITELN(plotfile);
               END;
            other : BEGIN
               (* graphics print code for a new printer goes here *)
               END
            END;
      END; (* striprint *)

   BEGIN (* draw *)
      j := 1;
      FOR i := 0 TO 7 DO BEGIN
         bit[i] := j;
         j := j + j;
         END;
      null := CHR(0);
      strptop := 80*strpdeep;
      strpbottom := strptop-strpdeep+1;
      WRITELN(OUTPUT, ' writing 80 lines...');
      FOR line := 1 TO 80 DO BEGIN
         astrptop := strptop/aspectratio;
         astrpbottom := (strpbottom-1)/aspectratio;
         FOR i := 1 TO strpdiv DO
            FOR j := 1 TO strpwide DO
               stripe[i][j] := null;
         empty := TRUE;
         xnow := strpwide / 2.0;
         ynow := 0.0;
         IF drawing = slant
            THEN slantdrawing(p)
            ELSE rectdrawing(p);
         plotlabels;
         strptop := strpbottom - 1;
         strpbottom := strpbottom - strpdeep;
         striprint;
         IF (line MOD 5) = 0
         THEN WRITELN(OUTPUT, line:3);
         END;
   END; (* draw *)

   PROCEDURE finishplotter;
   BEGIN (* finishplotter *)
      CASE printer OF
         citoh : WRITE(plotfile, CHR(27), 'A');
         epson : BEGIN
            WRITE(plotfile, CHR(27), 'A', CHR(12));
            WRITE(plotfile, CHR(27), '2');
            END;
         okidata : BEGIN END;
         toshiba : WRITELN(plotfile, CHR(27), CHR(26), 'I');
         other : BEGIN
            (* termination code for a new printer goes here *)
            END
         END;
   END; (* finishplotter *)

   PROCEDURE garbagestack;
      VAR i : INTEGER;
   BEGIN (* garbagestack *)
      FOR i := 1 TO nnodes DO BEGIN
         nodelist[i]^.descendant := garbnode;
         garbnode := nodelist[i];
         END;
      FOR i := 1 TO nlabels DO BEGIN
         labellist[i]^.nextlab := garblabel;
         garblabel := labellist[i];
         END;
   END; (* garbagestack *)

BEGIN (* plotgram *)
   WRITELN(OUTPUT, ' DOTGRAM VERSION 3.0');
   WRITELN(OUTPUT, ' Based on Plotgram 1.3 by C. Meacham');
   WRITELN(OUTPUT, ' READING OF PARAMETERS BEGINS');
   RESET(parmfile);
   initparms;
   (*  CLOSE(parmfile);   may need this here *)
   WRITELN(OUTPUT, ' READING OF PARAMETERS COMPLETED');
   IF parmsok THEN BEGIN
      RESET(treefile);
      WHILE NOT EOF(treefile) DO BEGIN
         WRITELN(OUTPUT, ' READING OF DIAGRAM BEGINS');
         readtree;
         WRITELN(OUTPUT, ' READING OF DIAGRAM COMPLETED');
         WRITELN(OUTPUT, ' CALCULATION OF DIAGRAM BEGINS');
         calctree(root);
         WRITELN(OUTPUT, ' CALCULATION OF DIAGRAM COMPLETED');
         IF NOT(fontloaded) AND labelspresent THEN BEGIN
            WRITELN(OUTPUT, ' LOADING OF FONT BEGINS');
            RESET(fontfile);
            loadfont(font);
            (* CLOSE(fontfile);   may need this here *)
            WRITELN(OUTPUT, ' LOADING OF FONT COMPLETED');
            fontloaded := TRUE;
            END;
         hardwarecoordinates;
         WRITELN(OUTPUT, ' DRAWING OF DIAGRAM BEGINS');
         REWRITE(plotfile);
         IF startmode=yes THEN BEGIN
            WRITELN(OUTPUT,
                '** HIT ENTER (RETURN) TO BEGIN PLOTTING ***');
            READLN(INPUT);
            END;
         initplotter;
         draw(root);
         finishplotter;
         WRITELN(OUTPUT, ' DRAWING OF DIAGRAM COMPLETED');
         garbagestack;
         END;
      END;
   (* CLOSE(treefile);   may need this here *)
   (* CLOSE(plotfile);   may need this here *)
   WRITELN(OUTPUT, ' EXECUTION TERMINATED');
END. (* plotgram *)

