PROGRAM Dotree (treefile, parmfile, fontfile, plotfile, INPUT, OUTPUT);
   (* Copyright (c) 1986 by Christopher A. Meacham,
      University Herbarium, University of California, Berkeley,
      California 94720, U.S.A.

      Permission is granted to copy, distribute, and modify this
      program provided that (1) this copyright message is not removed
      and (2) no fee is charged for this program.

      PHYLIP version 2.9. Based on PLOTREE Version 1.3, July 1986.
      Modifications by J. Felsenstein.
      Incorporates modifications and bug-fixes suggested by
      D. L. Swofford and J. M. Humphries.  *)

   CONST bignum = 1000000.0;
      smallnum = 0.0001;
      maxnodes =  100;
      maxlablen = 3;
      fontsize = 3200;  (* Required for font1: 1165; font2: 2162;  *)
      doswitch = TRUE;  (*   font3: 3190; font4: 2173; font5: 3200 *)
      dontswitch = FALSE;
      stripedepth = 4;  (* max. depth of the dot stripe *)
      stripewidth = 1350; (* max. width of the dot stripe *)

   TYPE fontname = (font1, font2, font3, font4, font5);
      modetype = (yes, no);
      posoptiontype = (atroot, right, left, above, below, center);
      penstatustype = (penup, pendown);
      printertype = (citoh, epson, okidata, toshiba, other);
      fonttype = ARRAY [1..fontsize] OF INTEGER;
      plotstring = ARRAY [1..maxlablen] OF CHAR;
      nodeptr = ^treenode;
      treenode = RECORD
         ancestor, sibling, descendant, lastdesc,
            left, right : nodeptr;
         c, nbranches, nch : INTEGER;
         nodesize, len, spread, r, theta, x, y,
         height, width : REAL;
         ndlabel : plotstring
         END;

   VAR nnodes, xnow, ynow, oldxhigh, oldxlow, oldyhigh,
         oldylow, longestlablen, strpwide, strpdeep, strpdiv : INTEGER;
      pi, smallnode, radius, absradius, diameter, spanunit,
         padding, spaceval, scrunchval, heightval,
         scrunchlimit, rotation, xposition, yposition,
         xunitspercm, yunitspercm, aspectratio,
         xsize, ysize, xrange, yrange, maxdiam, labelheight,
         labelsize : REAL;
      xcircle, ycircle : ARRAY [1..30] OF REAL;
      stripe : ARRAY [1..stripedepth] OF ARRAY[1..stripewidth] OF CHAR;
      parmfile, treefile, fontfile, plotfile : TEXT;
      labelspresent, fontloaded, candraw, parmsok : BOOLEAN;
      fonttoload : fontname;
      font : fonttype;
      sizemode, lengthmode, monitor, pfilemode, startmode : modetype;
      printer : printertype;
      xoption, yoption : posoptiontype;
      root, garbage : nodeptr;
      nodelist : ARRAY [1..maxnodes] OF nodeptr;

   PROCEDURE initparms;
      CONST nabs = 21;
      VAR ab1, ab2 : ARRAY [0..nabs] OF CHAR;

      PROCEDURE printrparms(ch : CHAR);
      BEGIN (* printrparms *)
         CASE ch OF
            'C': BEGIN
               printer := citoh;
               xunitspercm := 31.496;
               yunitspercm := 28.346;
               xsize := 20.32;
               ysize := 20.32;
               xposition := 10.16;
               yposition := 0.0;
               xoption := center;
               yoption := above;
               rotation := 0.0;
               strpwide := 640;
               strpdeep := 8;
               strpdiv := 1;
               END;
            'E': BEGIN
               printer := epson;
               xunitspercm := 47.24;
               yunitspercm := 28.34;
               xsize := 19.0;
               ysize := 22.0;
               xposition := 9.5;
               yposition := 0.0;
               xoption := center;
               yoption := above;
               rotation := 0.0;
               strpwide := 900;
               strpdeep := 8;
               strpdiv := 1;
               END;
            'K': BEGIN
               printer := okidata;
               xunitspercm := 56.693;
               yunitspercm := 28.346;
               xsize := 19.0;
               ysize := 22.0;
               xposition := 9.5;
               yposition := 0.0;
               xoption := center;
               yoption := above;
               rotation := 0.0;
               strpwide := 1100;
               strpdeep := 8;
               strpdiv := 1;
               END;
            'T': BEGIN
               printer := toshiba;
               xunitspercm := 70.866;
               yunitspercm := 70.866;
               xsize := 19.0;
               ysize := 25.0;
               xposition := 10.5;
               yposition := 0;
               xoption := center;
               yoption := above;
               rotation := 0.0;
               strpwide := 1350;
               strpdeep := 24;
               strpdiv := 4;
               END;
            'O': BEGIN
               printer := other;
               (* initial parameter settings for a new
                  printer go here *)
               END
            END;
         END; (* printrparms *)

      PROCEDURE readparms;
         VAR parmnum : INTEGER;
            ch, c1, c2 : CHAR;

         PROCEDURE uppercase (VAR ch : CHAR);

            FUNCTION letter (ch : CHAR) : BOOLEAN;
            BEGIN (* letter -- works for both ASCII and EBCDIC codes *)
               letter := ((ch >= 'A') AND (ch <= 'I'))
                         OR ((ch >= 'J') AND (ch <= 'R'))
                         OR ((ch >= 'S') AND (ch <= 'Z'))
                         OR ((ch >= 'a') AND (ch <= 'i'))
                         OR ((ch >= 'j') AND (ch <= 'r'))
                         OR ((ch >= 's') AND (ch <= 'z'));
            END; (* letter *)

         BEGIN (* uppercase *)
            IF letter(ch) AND ((('A' < 'a') AND (ch >= 'a'))
                            OR (('A' > 'a') AND (ch < 'A')))
            THEN ch := CHR(ORD(ch) + ORD('A') - ORD('a'));
         END; (* uppercase *)

         PROCEDURE getp(VAR ch : CHAR);
         BEGIN (* getp *)
            ch := ' ';
            WHILE (ch=' ') AND NOT EOLN(parmfile) DO
               READ(parmfile, ch);
            uppercase(ch);
         END; (* getp *)

         FUNCTION setmode : modetype;
         BEGIN (* setmode *)
            getp(ch);
            CASE ch OF
               'Y': setmode := yes;
               'N': setmode := no
               END;
         END; (* setmode *)

      BEGIN (* readparms *)
         WHILE NOT EOF(parmfile) DO BEGIN
            c1 := ' ';
            c2 := ' ';
            WHILE NOT EOLN(parmfile) AND (c1=' ') DO
               READ(parmfile, c1);
            IF c1<>'*' THEN BEGIN
               IF NOT EOLN(parmfile) THEN READ(parmfile, c2);
               REPEAT READ(parmfile, ch) UNTIL EOLN(parmfile) OR (ch=' ');
               uppercase(c1);
               uppercase(c2);
               parmnum := 0;
               WHILE NOT((c1=ab1[parmnum]) AND (c2=ab2[parmnum])) AND
                  (parmnum<=nabs) DO parmnum := parmnum+1;
               IF parmnum <= nabs
                  THEN CASE parmnum OF
                     1: BEGIN
                        getp(ch);
                        printrparms(ch);
                        END;
                     2: lengthmode := setmode;
                     3: sizemode := setmode;
                     4: READ(parmfile, diameter);
                     5: READ(parmfile, spanunit);
                     6: READ(parmfile, smallnode);
                     7: READ(parmfile, rotation);
                     8: BEGIN
                        getp(ch);
                        CASE ch OF
                          '1': fonttoload := font1;
                          '2': fonttoload := font2;
                          '3': fonttoload := font3;
                          '4': fonttoload := font4;
                          '5': fonttoload := font5
                          END;
                        END;
                     9: READ(parmfile, xunitspercm);
                     10: READ(parmfile, yunitspercm);
                     11: READ(parmfile, xsize);
                     12: READ(parmfile, ysize);
                     13: READ(parmfile, xposition);
                     14: READ(parmfile, yposition);
                     15: BEGIN
                        getp(ch);
                        CASE ch OF
                           'O': xoption := atroot;
                           'C': xoption := center;
                           'R': xoption := right;
                           'L': xoption := left
                           END;
                        END;
                     16: BEGIN
                        getp(ch);
                        CASE ch OF
                           'O': yoption := atroot;
                           'C': yoption := center;
                           'A': yoption := above;
                           'B': yoption := below
                           END;
                        END;
                     17: READ(parmfile, padding);
                     18: READ(parmfile, maxdiam);
                     19: monitor := setmode;
                     20: READ(parmfile, labelsize);
                     21: startmode := setmode
                     END
                  ELSE BEGIN
                     WRITELN(OUTPUT, ' PARAMETER LABEL THAT BEGINS ''',
                        c1, c2, ''' IS NOT A LEGITIMATE LABEL');
                     parmsok := FALSE;
                     READLN(parmfile);
                     END;
               END;
            READLN(parmfile);
            END;
      END; (* readparms *)

   BEGIN (* initparms *)
      ab1[0] := ' '; ab2[0] := ' ';
      ab1[1] := 'P'; ab2[1] := 'R'; ab1[2] := 'L'; ab2[2] := 'E';
      ab1[3] := 'S'; ab2[3] := 'I'; ab1[4] := 'N'; ab2[4] := 'O';
      ab1[5] := 'I'; ab2[5] := 'N'; ab1[6] := 'S'; ab2[6] := 'M';
      ab1[7] := 'R'; ab2[7] := 'O'; ab1[8] := 'F'; ab2[8] := 'O';
      ab1[9] := 'X'; ab2[9] := 'U'; ab1[10] := 'Y'; ab2[10] := 'U';
      ab1[11] := 'X'; ab2[11] := 'S'; ab1[12] := 'Y'; ab2[12] := 'S';
      ab1[13] := 'X'; ab2[13] := 'P'; ab1[14] := 'Y'; ab2[14] := 'P';
      ab1[15] := 'X'; ab2[15] := 'O'; ab1[16] := 'Y'; ab2[16] := 'O';
      ab1[17] := 'P'; ab2[17] := 'A'; ab1[18] := 'M'; ab2[18] := 'A';
      ab1[19] := 'M'; ab2[19] := 'O'; ab1[20] := 'L'; ab2[20] := 'A';
      ab1[21] := 'S'; ab2[21] := 'T';
      garbage := NIL;
      pi := 3.1415926536;
      heightval := 0.9;
      scrunchval := 0.8;
      scrunchlimit := 0.01;
      padding := 1.1;
      parmsok := TRUE;
      labelspresent := FALSE;
      fontloaded := FALSE;
      startmode := no;
      pfilemode := yes;
      monitor := no;
      lengthmode := no;
      sizemode := yes;
      fonttoload := font1;
      labelsize := 1.0;
      maxdiam := 4.0;
      smallnode := 0.6;
      spanunit := 3.0;
      diameter := 1.0;
      printrparms('E');    (* default printer designation *)
      readparms;
      padding := padding*diameter;
      IF (SQR(spanunit)-SQR(padding/2.0)) < 0.0
         THEN BEGIN
            WRITELN(OUTPUT, ' INTERNODE LENGTH IS TOO SMALL IN ',
               'RELATION TO NODE DIAMETER');
            parmsok := FALSE;
            END
         ELSE spaceval := 2.0*ARCTAN((padding/2.0)/SQRT(SQR(spanunit)-
            SQR(padding/2.0)));
      radius := diameter/2.0;
      labelsize := labelsize * 0.9;
      rotation := rotation * pi/180.0;
      aspectratio := yunitspercm/xunitspercm;
   END; (* initparms *)

   PROCEDURE initnodes(p, bottom : nodeptr);
   BEGIN (* initnodes *)
      IF p <> NIL
      THEN BEGIN
         initnodes(p^.descendant, bottom);
         IF p <> bottom THEN initnodes(p^.sibling, bottom);
         p^.left := p;
         p^.right := p;
         p^.height := 0.0;
         p^.width := 0.0;
         p^.r := 0.0;
         p^.theta := 0.0;
         END;
   END; (* initnodes *)

   PROCEDURE readtree;
      VAR i, count : INTEGER;
         ch : CHAR;
         node : nodeptr;

      PROCEDURE newnode(VAR p : nodeptr);
      BEGIN (* newnode *)
         IF garbage <> NIL
            THEN BEGIN
               p := garbage;
               garbage := garbage^.descendant;
               END
            ELSE NEW(p);
         count := count+1;
         nodelist[count] := p;
         p^.c := count;
         p^.ancestor := NIL;
         p^.descendant := NIL;
         p^.sibling := NIL;
         p^.lastdesc := NIL;
         p^.nch := 0;
         p^.nbranches := 0;
         p^.nodesize := 1.0;
         p^.len := spanunit;
         p^.spread := pi;
      END; (* newnode *)

      PROCEDURE addnode(anc : nodeptr; VAR linktonode,
         anclastdesc : nodeptr);
         CONST blank = ' ';
            point = '.';
            comma = ',';
            colon = ':';
            leftp = '(';
            rightp = ')';
            semi = ';';
            underline = '_';
         VAR node : nodeptr;
            doingdescendants : BOOLEAN;

         PROCEDURE getch(VAR ch : CHAR);
         BEGIN (* getch *)
            IF EOF(treefile)
               THEN BEGIN
                  WRITELN(OUTPUT, '  NO ''*'' AT END OF TREE ',
                     'DESCRIPTION, ''*'' ASSUMED');
                  WRITELN(OUTPUT, '  NO '';'' AT END OF TREE ',
                     'DESCRIPTION, '';'' ASSUMED');
                  ch := ';';
                  END
               ELSE BEGIN
                  IF EOLN(treefile) THEN READLN(treefile);
                  READ(treefile, ch);
                  END;
         END; (* getch *)

         PROCEDURE processnodelabel(node : nodeptr);
         BEGIN (* processnodelabel *)
            IF node^.nch=0
               THEN BEGIN
                  labelspresent := TRUE;
                  REPEAT
                     IF node^.nch<maxlablen THEN BEGIN
                        node^.nch := node^.nch+1;
                        IF ch = underline THEN ch := blank;
                        node^.ndlabel[node^.nch] := ch;
                        END;
                     getch(ch);
                     UNTIL (ch=blank) OR (ch=leftp) OR (ch=comma) OR
                        (ch=rightp) OR (ch=colon) OR (ch=semi);
                  END
               ELSE REPEAT getch(ch) UNTIL (ch=blank) OR (ch=leftp)
                  OR (ch=comma) OR (ch=rightp) OR (ch=colon) OR
                  (ch=semi);
         END; (* processnodelabel *)

         PROCEDURE processlength(node : nodeptr);
            VAR digit, ordzero : INTEGER;
               lenvalue, divisor : REAL;
               pointread : BOOLEAN;
         BEGIN (* processlength *)
            ordzero := ORD('0');
            pointread := FALSE;
            lenvalue := 0.0;
            divisor := 1.0;
            REPEAT getch(ch) UNTIL ch <> blank;
            IF ch='-' THEN divisor := -divisor;
            IF (ch='+') OR (ch='-') THEN getch(ch);
            digit := ORD(ch)-ordzero;
            WHILE ((digit >= 0) AND (digit <= 9)) OR (ch=point)
               DO BEGIN
               IF ch = point
                  THEN pointread := TRUE
                  ELSE BEGIN
                     lenvalue := lenvalue*10.0 + digit;
                     IF pointread THEN divisor := divisor*10.0;
                     END;
               getch(ch);
               digit := ORD(ch)-ordzero;
               END;
            IF lengthmode = yes
               THEN node^.len := (lenvalue/divisor)*spanunit;
         END; (* processlength *)

      BEGIN (* addnode *)
         doingdescendants := FALSE;
         newnode(node);
         linktonode := node;
         anclastdesc := node;
         node^.ancestor := anc;
         WHILE ch=blank DO getch(ch);
         WHILE (ch <> semi) AND (((ch <> comma) AND
            (ch <> rightp)) OR doingdescendants) DO BEGIN
            IF ch = leftp
               THEN BEGIN
                  doingdescendants := TRUE;
                  getch(ch);
                  addnode(node, node^.descendant,
                     node^.lastdesc);
                  node^.nbranches := node^.nbranches+1;
                  END
               ELSE IF ch = comma
                  THEN BEGIN
                     getch(ch);
                     addnode(node, node^.lastdesc^.sibling,
                        node^.lastdesc);
                     node^.nbranches := node^.nbranches+1;
                     END
                  ELSE IF ch = rightp
                     THEN BEGIN
                        doingdescendants := FALSE;
                        REPEAT getch(ch) UNTIL ch <> blank;
                        END
                     ELSE IF ch = colon
                        THEN processlength(node)
                        ELSE IF ch <> semi
                           THEN processnodelabel(node);
            WHILE ch = blank DO getch(ch);
            END;
      END; (* addnode *)

   BEGIN (* readtree *)
      count := 0;
      ch := ' ';
      addnode(NIL, root, root);
      READLN(treefile);
      nnodes := count;
      initnodes(root, root);
      FOR i := 1 TO nnodes DO BEGIN
         node := nodelist[i];
         IF node^.nch = 0 THEN node^.nodesize := smallnode;
         IF (node^.nch = 1) AND (node^.ndlabel[1] = '@') THEN
            node^.nch := 0;
         END;
      IF lengthmode = yes THEN
         FOR i := 1 TO nnodes DO BEGIN
            node := nodelist[i];
            IF node <> root THEN node^.len := node^.len +
               (node^.nodesize + node^.ancestor^.nodesize)*radius;
            END;
   END; (* readtree *)

   PROCEDURE calctree(p : nodeptr);
      VAR fitok : BOOLEAN;
         current : nodeptr;

         PROCEDURE fanflip(p, bottom : nodeptr; angle : REAL;
            flip : BOOLEAN);
            VAR current : nodeptr;
         BEGIN (* fanflip *)
            IF p <> NIL THEN BEGIN
               fanflip(p^.descendant, bottom, angle, flip);
               IF p <> bottom THEN fanflip(p^.sibling, bottom,
                                                 angle, flip);
               IF flip
                  THEN BEGIN
                     p^.theta := angle - p^.theta;
                     current := p^.descendant;
                     p^.descendant := p^.lastdesc;
                     WHILE current <> p^.descendant DO BEGIN
                        p^.lastdesc^.sibling := current;
                        p^.lastdesc := p^.lastdesc^.sibling;
                        current := current^.sibling;
                        p^.lastdesc^.sibling := NIL;
                        END;
                     current := p^.left;
                     p^.left := p^.right;
                     p^.right := current;
                     END
                  ELSE p^.theta := angle + p^.theta;
               IF p^.theta < bottom^.ancestor^.left^.theta
                  THEN bottom^.ancestor^.left := p;
               IF p^.theta > bottom^.ancestor^.right^.theta
                  THEN bottom^.ancestor^.right := p;
               END;
         END; (* fanflip *)

      PROCEDURE arrange(p : nodeptr; VAR fitok : BOOLEAN);
         VAR nlong : INTEGER;
            maxheight, angle, angleused, spacing : REAL;
            flip : BOOLEAN;
            previous, current, next : nodeptr;
         PROCEDURE switchbranches(VAR ance, prev, curr, next
               : nodeptr;
            switch : BOOLEAN);
         BEGIN (* switchbranches *)
            IF switch
               THEN BEGIN
                  IF prev = ance
                     THEN ance^.descendant := next
                     ELSE prev^.sibling := next;
                  curr^.sibling := next^.sibling;
                  next^.sibling := curr;
                  prev := ance;
                  curr := ance^.descendant;
                  next := curr^.sibling;
                  END
               ELSE BEGIN
                  prev := curr;
                  curr := next;
                  next := next^.sibling;
                  END;
         END; (* switchbranches *)

      BEGIN (* arrange *)
         previous := p;
         current := p^.descendant;
         next := current^.sibling;
         WHILE next <> NIL DO
            IF (current^.height < next^.height) OR
               ((current^.height = next^.height) AND
               (current^.width < next^.width))
               THEN switchbranches(p, previous, current, next, doswitch)
               ELSE switchbranches(p, previous, current,
                                                   next, dontswitch);
         nlong := 0;
         current := p^.descendant;
         maxheight := current^.height;
         WHILE current <> NIL DO BEGIN
            IF current^.height >= heightval*maxheight
               THEN nlong := nlong + 1;
            current := current^.sibling;
            END;
         angleused := 0.0;
         current := p^.descendant;
         WHILE current <> NIL DO BEGIN
            angleused := angleused + current^.width;
            current := current^.sibling;
            END;
         spacing := (p^.spread-angleused)/(p^.nbranches+1);
         fitok := spacing >= spaceval;
         current := p^.descendant;
         IF ODD(nlong)
            THEN angle := 0.0
            ELSE angle := spacing/2.0 + current^.right^.theta;
         p^.right := current^.left;
         p^.left := current^.right;
         flip := TRUE;
         fanflip(current, current, angle, flip);
         current := current^.sibling;
         WHILE current <> NIL DO BEGIN
            IF ABS(p^.left^.theta) > ABS(p^.right^.theta)
               THEN angle := p^.right^.theta
                             + spacing + current^.right^.theta
               ELSE angle := p^.left^.theta - spacing
                             - current^.right^.theta;
            flip := angle >= 0.0;
            fanflip(current, current, angle, flip);
            current := current^.sibling;
            END;
         previous := p;
         current := p^.descendant;
         next := current^.sibling;
         WHILE next <> NIL DO
            IF current^.theta > next^.theta
               THEN switchbranches(p, previous, current, next, doswitch)
               ELSE switchbranches(p, previous, current, next,
                                                     dontswitch);
         p^.lastdesc := current;
      END; (* arrange *)

      PROCEDURE scrunch(p, bottom : nodeptr);
      BEGIN (* scrunch *)
         IF p <> NIL THEN BEGIN
            IF (monitor = yes) AND (p = bottom)
               THEN WRITELN(OUTPUT, '    SCRUNCHING AT NODE', p^.c:4);
            scrunch(p^.descendant, bottom);
            IF p <> bottom THEN scrunch(p^.sibling, bottom);
            p^.spread := p^.spread * scrunchval;
            IF p^.spread < scrunchlimit THEN candraw := FALSE;
            END;
      END; (* scrunch *)

      PROCEDURE polartrans(p, bottom : nodeptr; len : REAL);
         VAR pr, ptheta : REAL;
      BEGIN (* polartrans *)
         IF p <> NIL THEN BEGIN
            polartrans(p^.descendant, bottom, len);
            IF p <> bottom THEN polartrans(p^.sibling, bottom, len);
            pr := p^.r;
            ptheta := p^.theta;
            p^.r := SQRT(SQR(len) + SQR(pr)+ 2*len*pr*COS(ptheta));
            p^.theta := ARCTAN(pr*SIN(ptheta)/(pr*COS(ptheta) + len));
            IF p^.r  > bottom^.height THEN bottom^.height := p^.r;
            IF p^.theta > bottom^.right^.theta THEN bottom^.right := p;
            IF p^.theta < bottom^.left^.theta THEN bottom^.left := p;
            END;
      END; (* polartrans *)

   BEGIN (* calctree *)
      IF (p <> NIL) AND candraw THEN BEGIN
         IF monitor = yes THEN WRITELN(OUTPUT, '     AT NODE',p^.c:4);
         REPEAT
            calctree(p^.descendant);
            IF p^.descendant <> NIL
               THEN arrange(p,fitok)
               ELSE fitok := TRUE;
            IF fitok
               THEN BEGIN
                  IF p <> root THEN BEGIN
                     p^.right := p;
                     p^.left := p;
                     polartrans(p, p, p^.len);
                     p^.width := p^.right^.theta - p^.left^.theta;
                     END;
                  END
               ELSE BEGIN
                  current := p^.descendant;
                  WHILE current <> NIL DO BEGIN
                     scrunch(current, current);
                     initnodes(current, current);
                     current := current^.sibling;
                     END;
                  END;
            UNTIL fitok OR NOT candraw;
         IF candraw THEN BEGIN
            IF ABS(p^.right^.theta) > ABS(p^.left^.theta)
               THEN fanflip(p, p, 0.0, TRUE);
            calctree(p^.sibling);
            END;
         END;
   END; (* calctree *)

   PROCEDURE hardwarecoordinates;
      VAR i : INTEGER;
         factor, angle, xmax, xmin, ymax, ymin, thisx, thisy,
         thisrad, xtrans, ytrans : REAL;
         p : nodeptr;
   BEGIN (* hardwarecoordinates *)
      IF rotation <> 0.0 THEN BEGIN
         FOR i := 1 TO nnodes DO
            nodelist[i]^.theta := nodelist[i]^.theta + rotation;
         END;
      xmax := -bignum;
      xmin := bignum;
      ymax := -bignum;
      ymin := bignum;
      FOR i := 1 TO nnodes DO BEGIN
         p := nodelist[i];
         p^.x := p^.r*SIN(p^.theta);
         p^.y := p^.r*COS(p^.theta);
         thisx := p^.x;
         thisy := p^.y;
         thisrad := p^.nodesize*radius;
         IF thisx + thisrad > xmax THEN xmax := thisx + thisrad;
         IF thisx - thisrad < xmin THEN xmin := thisx - thisrad;
         IF thisy + thisrad > ymax THEN ymax := thisy + thisrad;
         IF thisy - thisrad < ymin THEN ymin := thisy - thisrad;
         END;
      xrange := xmax - xmin;
      yrange := ymax - ymin;
      IF sizemode = yes
         THEN BEGIN
            IF ysize/xsize > yrange/xrange
               THEN factor := xsize/xrange
               ELSE factor := ysize/yrange;
            IF 2.0*radius*factor > maxdiam
               THEN factor := maxdiam/(2.0*radius);
            END
         ELSE factor := 1.0;
      CASE xoption OF
         atroot: xtrans := xposition;
         right: xtrans := xposition - xmin*factor;
         left: xtrans := xposition - xmax*factor;
         center: xtrans := xposition + (xrange/2.0 - xmax)*factor
         END;
      CASE yoption OF
         atroot: ytrans := 0.0;
         above: ytrans := yposition - ymin*factor;
         below: ytrans := yposition - ymax*factor;
         center: ytrans := yposition + (yrange/2.0 - ymax)*factor
         END;
      FOR i := 1 TO nnodes DO BEGIN
         p := nodelist[i];
         p^.x := (p^.x*factor+xtrans)*xunitspercm;
         p^.y := (p^.y*factor+ytrans)*xunitspercm
         END;
      absradius := radius * factor * xunitspercm;
      xrange := xrange * factor * xunitspercm;
      yrange := yrange * factor * xunitspercm;
      FOR i := 1 TO 30 DO BEGIN
         angle := i * pi/15.0;
         xcircle[i] := SIN(angle) * absradius;
         ycircle[i] := -COS(angle) * absradius;
         END;
   END; (* hardwarecoordinates *)

   PROCEDURE loadfont(VAR font : fonttype);
      VAR i, charstart, hersheynum : INTEGER;
         ch : CHAR;
   BEGIN (* loadfont *)
      ch := 'A';
      i := 0;
      WHILE NOT(EOF(fontfile) OR (ch=' ')) DO BEGIN
         charstart := i+1;
         READ(fontfile, ch, ch, hersheynum,
            font[charstart+2], font[charstart+3]);
         font[charstart+1] := ORD(ch);
         i := charstart+3;
         REPEAT
            IF (i-(charstart+3)) MOD 10=0 THEN READLN(fontfile);
            i := i+1;
            READ(fontfile,font[i]);
            UNTIL ABS(font[i]) >= 10000;
         READLN(fontfile);
         font[charstart] := i+1;
         END;
      font[charstart] := 0;
      WRITELN(OUTPUT, '  FONT REQUIRES', i:5, ' INTEGERS OUT OF',
         fontsize:5, ' AVAILABLE');
   END; (* loadfont *)

   PROCEDURE labelsetup;
      VAR i : INTEGER;
   BEGIN (* labelsetup *)
      longestlablen := 0;
      FOR i := 1 TO nnodes DO
         IF nodelist[i]^.nch > longestlablen
            THEN longestlablen := nodelist[i]^.nch;
      IF longestlablen > 0
         THEN labelheight := 2.0*absradius/
                                   SQRT(SQR(longestlablen)+1.0)
         ELSE labelspresent := FALSE;
   END; (* labelsetup *)

   PROCEDURE initprinter;
   BEGIN (* initprinter *)
      CASE printer OF
         citoh: WRITE(plotfile, CHR(27), 'T16');
         epson: BEGIN
                WRITE(plotfile, CHR(27), 'A', CHR(8));
                WRITE(plotfile, CHR(27), '2');
                END;
         okidata : WRITE(plotfile, CHR(27), '%9', CHR(16));
         toshiba: BEGIN
                  WRITELN(plotfile, CHR(27), CHR(26), 'I');
                  WRITELN(plotfile);
                  WRITELN(plotfile, CHR(27), 'L06');
                  END;
         other : BEGIN
            (* initialization code for a new printer goes here *)
            END
         END;
   END; (* initprinter *)

   PROCEDURE draw(p : nodeptr);
      VAR i, j, line, strptop, strpbottom : INTEGER;
         xnow, ynow, astrptop, astrpbottom : REAL;
         null : CHAR;
         empty : BOOLEAN;
         bit : ARRAY[0..7] OF INTEGER;

         PROCEDURE plot(pen : penstatustype; xabs, yabs : REAL);
            VAR ixnow, iynow, ixabs, iyabs, xdiff, ydiff,
               err, temp, x1, y1 : INTEGER;

            PROCEDURE plotdot (ix, iy : INTEGER);
               (* plot one dot at ix, iy *)
               VAR iy1, iy2 : INTEGER;
            BEGIN (* plotdot *)
               iy := strptop - iy;
               IF (iy >= 0) AND (iy < strpdeep) AND (ix>0) AND
                  (ix<=strpwide)
               THEN BEGIN
                  empty := FALSE;
                  CASE printer OF
                     citoh: BEGIN
                        iy1 := 1;
                        iy2 := iy;
                        END;
                     epson: BEGIN
                        iy1 := 1;
                        iy2 := 7 - iy;
                        END;
                     okidata: BEGIN
                        iy1 := 1;
                        iy2 := 7 - iy;
                        END;
                     toshiba: BEGIN
                        iy1 := iy DIV 6 + 1;
                        iy2 := 5 - (iy MOD 6);
                        END;
                     other: BEGIN
                        (* code for making dot array on a new printer
                           goes here *)
                        END
                     END;
                  IF ((ORD(stripe[iy1][ix]) DIV bit[iy2]) MOD 2) = 0
                  THEN stripe[iy1][ix] :=
                     CHR(ORD(stripe[iy1][ix]) + bit[iy2]);
                  END;
            END; (* plotdot *)

         BEGIN (* plot *)
            yabs := yabs * aspectratio;
            ixabs := ROUND(xabs);
            iyabs := ROUND(yabs);
            ixnow := ROUND(xnow);
            iynow := ROUND(ynow);
            IF (pen = pendown) AND
               (NOT (((iyabs < strpbottom) AND (iynow < strpbottom))
                  OR ((iyabs > strptop) AND (iynow > strptop))))
            THEN BEGIN
               IF ixabs < ixnow
               THEN BEGIN
                  temp := ixnow;
                  ixnow := ixabs;
                  ixabs := temp;
                  temp := iynow;
                  iynow := iyabs;
                  iyabs := temp;
                  END;
               xdiff := ixabs - ixnow;
               ydiff := iyabs - iynow;
               IF ydiff >= 0
               THEN IF xdiff >= ydiff
                  THEN BEGIN
                     err := -xdiff DIV 2;
                     FOR x1 := ixnow TO ixabs DO BEGIN
                        plotdot (x1, iynow);
                        err := err + ydiff;
                        IF err > 0
                        THEN BEGIN
                           iynow := iynow + 1;
                           err := err - xdiff;
                           END;
                        END;
                     END
                  ELSE BEGIN
                     err := -ydiff DIV 2;
                     FOR y1 := iynow TO iyabs DO BEGIN
                        plotdot (ixnow, y1);
                        err := err + xdiff;
                        IF err > 0
                        THEN BEGIN
                           ixnow := ixnow + 1;
                           err := err - ydiff;
                           END;
                        END;
                     END
               ELSE IF xdiff < -ydiff
                  THEN BEGIN
                     err := ydiff DIV 2;
                     FOR y1 := iynow DOWNTO iyabs DO BEGIN
                        plotdot (ixnow, y1);
                        err := err + xdiff;
                        IF err > 0
                        THEN BEGIN
                           ixnow := ixnow + 1;
                           err := err + ydiff;
                           END;
                        END;
                     END
                  ELSE BEGIN
                     err := -xdiff DIV 2;
                     FOR x1 := ixnow TO ixabs DO BEGIN
                        plotdot (x1, iynow);
                        err := err - ydiff;
                        IF err > 0
                        THEN BEGIN
                           iynow := iynow - 1;
                           err := err - xdiff;
                           END;
                        END;
                     END;
               END;
            xnow := xabs;
            ynow := yabs;
         END; (* plot *)

         PROCEDURE drawinternode(x1, y1, x2, y2, ndsize1,
                                              ndsize2 : REAL);
            VAR slope, xdiff, ydiff, xstart, ystart, xend,
                   yend : REAL;

            FUNCTION trimend(thisend, farend, amount : REAL) : REAL;
            BEGIN (* trimend *)
               IF thisend > farend
                  THEN trimend := thisend - amount
                  ELSE trimend := thisend + amount;
            END; (* trimend *)

         BEGIN (* drawinternode *)
            IF ABS(x1 - x2) - smallnum < 0.0
               THEN BEGIN
                  xdiff := 0.0;
                  ydiff := absradius;
                  END
               ELSE IF ABS(y1 - y2) - smallnum < 0.0
                  THEN BEGIN
                     xdiff := absradius;
                     ydiff := 0.0;
                     END
                  ELSE BEGIN
                     slope := (y2 - y1)/(x2 - x1);
                     xdiff := absradius/SQRT(SQR(slope) + 1.0);
                     ydiff := absradius/SQRT((1.0/SQR(slope)) + 1.0);
                     END;
            xstart := trimend(x1, x2, xdiff*ndsize1);
            ystart := trimend(y1, y2, ydiff*ndsize1);
            xend := trimend(x2, x1, xdiff*ndsize2);
            yend := trimend(y2, y1, ydiff*ndsize2);
            plot(penup, xstart, ystart);
            plot(pendown, xend, yend);
         END; (* drawinternode *)

         PROCEDURE drawnode(x, y, ndsize : REAL);
            VAR i : INTEGER;
         BEGIN (* drawnode *)
            plot(penup, x, y - absradius*ndsize);
            FOR i := 1 TO 30 DO BEGIN
               plot(pendown, x+xcircle[i]*ndsize,
                  y+ycircle[i]*ndsize);
               END;
         END; (* drawnode *)

         PROCEDURE plotlabel(p : nodeptr);
            TYPE pentype = (treepen, labelpen);
            VAR textlen, compress, xlab, ylab, x, y : REAL;

            PROCEDURE plottext(pstring : plotstring; nchars : INTEGER;
               height, compress : REAL; VAR x, y : REAL;
               slope : REAL; VAR font : fonttype);
               CONST xstart = 10;
                  ystart = 35;
               VAR i, j, code, coord : INTEGER;
                  sinslope, cosslope : REAL;
                  penstatus : penstatustype;

               PROCEDURE plotchar(VAR place : INTEGER);
                  VAR spacing, heightfont, widthfont, xfactor, yfactor, 
                     xplot, yplot, xfont, yfont : REAL;
               BEGIN (* plotchar *)
                  heightfont := font[place+2];
                  widthfont := font[place+3];
                  yfactor := height/heightfont;
                  xfactor := yfactor*compress;
                  spacing := yfactor*ABS(heightfont*cosslope
                                -widthfont*sinslope)/aspectratio;
                  IF (y <= astrptop)
                     AND ((y + spacing) >= astrpbottom)
                  THEN BEGIN
                     place := place+3;
                     REPEAT
                        place := place+1;
                        coord := font[place];
                        IF coord > 0
                           THEN penstatus := pendown
                           ELSE penstatus := penup;
                        coord := ABS(coord);
                        coord := coord MOD 10000;
                        xfont := ((coord DIV 100) - xstart)*xfactor;
                        yfont := ((coord MOD 100) - ystart)*yfactor;
                        IF slope = 0.0
                           THEN BEGIN
                              xplot := x + xfont;
                              yplot := y + yfont;
                              END
                           ELSE BEGIN
                              xplot := x + xfont*cosslope
                                          + yfont*sinslope;
                              yplot := y - xfont*sinslope
                                          + yfont*cosslope;
                              END;
                        plot(penstatus, xplot, yplot);
                        UNTIL ABS(font[place]) >= 10000;
                     x := xplot;
                     y := yplot;
                     END
                  ELSE BEGIN
                     xfont := xfactor*widthfont;
                     x := x + xfont*cosslope;
                     y := y - xfont*sinslope;
                     END;
               END; (* plotchar *)

            BEGIN (* plottext *)
               sinslope := SIN(slope);
               cosslope := COS(slope);
               FOR i := 1 TO nchars DO BEGIN
                  code := ORD(pstring[i]);
                  j := 1;
                  WHILE (font[j+1] <> code) AND (font[j] <> 0) DO
                     j := font[j];
                  IF font[j+1] = code THEN plotchar(j);
                  END;
            END; (* plottext *)

            FUNCTION lengthtext(pstring : plotstring; nchars : INTEGER;
               height, compress : REAL; VAR font : fonttype) : REAL;
               VAR i, j, code, coord : INTEGER;
                  sumlength, heightfont, widthfont, yfactor : REAL;
            BEGIN (* lengthtext *)
               sumlength := 0.0;
               FOR i := 1 TO nchars DO BEGIN
                  code := ORD(pstring[i]);
                  j := 1;
                  WHILE (font[j+1] <> code) AND (font[j] <> 0) DO
                     j := font[j];
                  IF font[j+1] = code THEN BEGIN
                     heightfont := font[j+2];
                     widthfont := font[j+3];
                     sumlength := sumlength + (height/heightfont)
                           *widthfont;
                     END;
                  END;
                  lengthtext := sumlength*compress;
            END; (* lengthtext *)

         BEGIN (* plotlabel *)
            compress := 1.0;
            textlen := lengthtext(p^.ndlabel, p^.nch,
               labelheight, compress, font);
            If textlen > p^.nch*labelheight THEN BEGIN
               compress := p^.nch*labelheight/textlen;
               textlen := p^.nch*labelheight;
               END;
            xlab := -textlen*labelsize/2.0;
            ylab := -labelheight*labelsize/2.0;
            x := p^.x + xlab*COS(rotation) + ylab*SIN(rotation);
            y := p^.y - xlab*SIN(rotation) + ylab*COS(rotation);
            plottext(p^.ndlabel, p^.nch, labelheight*labelsize,
                        compress, x, y, rotation, font);
         END; (* plotlabel *)

      PROCEDURE drawtree(p : nodeptr);
      BEGIN (* drawtree *)
         IF p <> NIL
         THEN BEGIN
            IF p^.ancestor <> NIL THEN drawinternode(p^.ancestor^.x,
               p^.ancestor^.y, p^.x, p^.y, p^.ancestor^.nodesize,
               p^.nodesize);
            IF ((astrptop - (p^.y - absradius*p^.nodesize)) > 0.0) AND
               (((p^.y + absradius*p^.nodesize) - astrpbottom) > 0.0)
            THEN BEGIN
               drawnode(p^.x, p^.y, p^.nodesize);
               IF labelspresent AND (p^.nch > 0) THEN plotlabel(p);
               END;
            drawtree(p^.descendant);
            drawtree(p^.sibling);
            END;
      END; (* drawtree *)

      PROCEDURE striprint;
         VAR i, j, width : INTEGER;
            done : BOOLEAN;
      BEGIN (* striprint *)
         done := FALSE;
         width := strpwide;
         WHILE NOT done DO BEGIN
            FOR i := 1 TO strpdiv DO
               done := done OR (stripe[i][width] <> null);
            IF NOT done
            THEN width := width - 1;
            done := done OR (width = 0);
            END;
         CASE printer OF
            citoh: BEGIN
               IF NOT empty
               THEN BEGIN
                  WRITE(plotfile, CHR(27), 'S', (width DIV 1000):1,
                     ((width DIV 100) MOD 10):1, ((width DIV 10) MOD 10):1,
                     (width MOD 10):1);
                  FOR i := 1 TO width DO
                     WRITE(plotfile, stripe[1][i]);
                  END;
               WRITELN(plotfile);
               END;
            epson: BEGIN
               IF NOT empty
               THEN BEGIN
                  WRITE(plotfile, CHR(27), 'L', CHR(width MOD 256),
                     CHR(width DIV 256));
                  FOR i := 1 TO width DO
                     WRITE(plotfile, stripe[1][i]);
                  END;
               WRITELN(plotfile);
               END;
            okidata: BEGIN
               IF NOT empty
               THEN BEGIN
                  WRITE(plotfile, CHR(27), '%1', CHR(width DIV 128),
                     CHR(width MOD 128));
                  FOR i := 1 TO width DO
                     WRITE(plotfile, stripe[1][i]);
                  END;
               WRITELN(plotfile);
               END;
            toshiba: BEGIN
               IF NOT empty
               THEN BEGIN
                  FOR i := 1 TO width DO
                     FOR j := 1 TO 4 DO
                        stripe[j][i] := CHR(ORD(stripe[j][i])+64);
                  WRITE(plotfile, CHR(27), ';', (width DIV 1000):1,
                     ((width DIV 100) MOD 10):1, ((width DIV 10)
                       MOD 10):1, (width MOD 10):1);
                  FOR i := 1 TO width DO
                     WRITE(plotfile, stripe[1][i], stripe[2][i],
                            stripe[3][i], stripe[4][i]);
                  END;
               WRITELN(plotfile);
               END;
            other : BEGIN
               (* graphics print code for a new printer goes here *)
               END
            END;
      END; (* striprint *)

   BEGIN (* draw *)
      j := 1;
      FOR i := 0 TO 7 DO BEGIN
         bit[i] := j;
         j := j + j;
         END;
      null := CHR(0);
      strptop := 80*strpdeep;
      strpbottom := strptop-strpdeep;
      WRITELN(OUTPUT, ' writing 80 lines...');
      FOR line := 1 TO 80 DO BEGIN
         astrptop := strptop/aspectratio;
         astrpbottom := strpbottom/aspectratio;
         FOR i := 1 TO strpdiv DO
            FOR j := 1 TO strpwide DO
               stripe[i][j] := null;
         empty := TRUE;
         xnow := strpwide / 2.0;
         ynow := 0.0;
         drawtree(root);
         strptop := strpbottom - 1;
         strpbottom := strpbottom - strpdeep;
         striprint;
         IF (line MOD 5) = 0
         THEN WRITELN(OUTPUT, line:3);
         END;
   END; (* draw *)

   PROCEDURE finishprinter;
   BEGIN (* finishprinter *)
      CASE printer OF
         citoh: WRITE(plotfile, CHR(27), 'A');
         epson: BEGIN
                WRITE(plotfile, CHR(27), 'A', CHR(12));
                WRITE(plotfile, CHR(27), '2');
                END;
         okidata: BEGIN END;
         toshiba: WRITELN(plotfile, CHR(27), CHR(26), 'I');
         other : BEGIN
            (* termination code for a new printer goes here *)
            END
         END;
   END; (* finishprinter *)

   PROCEDURE garbagestack;
      VAR i : INTEGER;
   BEGIN (* garbagestack *)
      FOR i := 1 TO nnodes DO BEGIN
         nodelist[i]^.descendant := garbage;
         garbage := nodelist[i];
         END;
   END; (* garbagestack *)

BEGIN (* Dotree *)
   WRITELN(OUTPUT, ' DOTREE VERSION 2.9');
   WRITELN(OUTPUT, ' based on Plotree 1.3 by Christopher Meacham');
   WRITELN(OUTPUT, ' READING OF PARAMETERS BEGINS');
   RESET(parmfile);
   initparms;
(* CLOSE(parmfile);   may have to change this *)
   WRITELN(OUTPUT, ' READING OF PARAMETERS COMPLETED');
   IF parmsok THEN BEGIN
      RESET(treefile);
      WHILE NOT EOF(treefile) DO BEGIN
         candraw := TRUE;
         WRITELN(OUTPUT, ' READING OF TREE BEGINS');
         readtree;
         WRITELN(OUTPUT, ' READING OF TREE COMPLETED');
         WRITELN(OUTPUT, ' CALCULATION OF TREE BEGINS');
         calctree(root);
         WRITELN(OUTPUT, ' CALCULATION OF TREE COMPLETED');
         IF candraw
            THEN BEGIN
               hardwarecoordinates;
               IF labelspresent THEN BEGIN
                  labelsetup;
                  IF NOT(fontloaded) AND labelspresent THEN BEGIN
                     WRITELN(OUTPUT, ' LOADING OF FONT BEGINS');
                     RESET(fontfile);
                     loadfont(font);
                  (* CLOSE(fontfile);    may have to change this *)
                     WRITELN(OUTPUT, ' LOADING OF FONT COMPLETED');
                     fontloaded := TRUE;
                     END;
                  END;
               WRITELN(OUTPUT, ' DRAWING OF TREE BEGINS');
               REWRITE(plotfile);
               IF startmode=yes THEN BEGIN
                  WRITELN(OUTPUT,
                     '*** HIT ENTER (RETURN) TO BEGIN PLOTTING ***');
                  READLN(INPUT);
                  END;
               initprinter;
               draw(root);
               finishprinter;
               WRITELN(OUTPUT, ' DRAWING OF TREE COMPLETED');
               END
            ELSE BEGIN
               WRITELN(OUTPUT, ' TREE CANNOT BE DRAWN WITH CURRENT ',
                         'PARAMETERS');
               WRITELN(OUTPUT, '  NODE DIAMETER MUST BE SMALLER OR ',
                         'INTERNODE LENGTH LARGER');
               END;
         garbagestack;
         END;
      END;
(* CLOSE(treefile);    may have to change this *)
(* CLOSE(plotfile);    may have to change this *)
   WRITELN(OUTPUT, ' EXECUTION TERMINATED');
END. (* plotree *)

