PROGRAM Factor (infile, outfile, OUTPUT);
   (* version 3.1
      A program to factor multistate character trees.
      Programmed 29 May 1983 by C. A. Meacham, Botany Department,
      University of Georgia, Athens, Georgia 30602, U.S.A.
      Additional code by Joe Felsenstein, 1988
      Copyright 1988 by Christopher A. Meacham *)

   CONST nmlngth = 10;  (* Length of species name                     *)
      maxchars = 100;   (* Max. number of characters to be factored   *)
      maxstates = 20;   (* Maximum number of states in multi chars    *)
      maxoutput = 80;   (* Maximum length of output line              *)
      sizearray = 2000; (* Size of symbarray; must be >= the sum of the
                            squares of the number of states in each
                            multi char to be factored        *)
      factchar = ':';   (* character to indicate state connections *)
      unkchar  = '?';   (* input character to indicate state unknown *)

   VAR infile, outfile : TEXT;
      neus, nchars, charindex, lastindex : INTEGER;
      ch : CHAR;
      ancestorrequest, factorrequest, rooted, terminate : BOOLEAN;
      symbarray : ARRAY [1..sizearray] OF CHAR; (* Holds multi symbols 
                                           and their factored equivs  *)
      charnum : ARRAY [1..maxchars] OF INTEGER; (* Multis   *)
      chstart : ARRAY [1..maxchars] OF INTEGER; (* Position of each *)
      numstates : ARRAY [1..maxchars] OF INTEGER; (* Number of states *)
      ancsymbol : ARRAY [1..maxchars] OF CHAR;  (* Ancestral state    *)

   PROCEDURE dotrees;
      (* Process character-state trees *)
      TYPE nodepntr = ^node;
         node = RECORD         (* Node of multifurcating tree *)
            ancestor, sibling, descendant : nodepntr;
            state : CHAR;      (* Symbol of character state *)
            edge : INTEGER     (* Number of subtending edge *)
            END;
      VAR npairs, offset, charnumber, lastchar, nstates : INTEGER;
         error : BOOLEAN;
         root : nodepntr;
         pair : ARRAY [1..maxstates, 1..2] OF CHAR;
         nodes : ARRAY [1..maxstates] OF nodepntr;

      PROCEDURE readtree;
         (* Reads a single character-state tree; puts adjacent symbol
            pairs into array 'pairs' *)

         PROCEDURE nextch (VAR ch : CHAR);
         BEGIN (* nextch *)
            ch := ' ';
            WHILE (ch = ' ') AND NOT(EOLN(infile)) DO READ(infile, ch);
         END; (* nextch *)

      BEGIN (* readtree *)
         error := FALSE;
         npairs := 0;
         WHILE NOT(EOLN(infile)) AND NOT(error) DO BEGIN
            nextch(ch);
            IF NOT(EOLN(infile))
            THEN BEGIN
               npairs := npairs + 1;
               pair[npairs,1] := ch;
               nextch(ch);
               IF EOLN(infile) OR (ch <> factchar)
               THEN error := TRUE
               ELSE BEGIN
                  nextch(pair[npairs,2]);
                  IF EOLN(infile) AND (pair[npairs,2] = ' ')
                  THEN error := TRUE;
                  END;
               END;
            END;
         READLN(infile);
         IF error
         THEN WRITELN(OUTPUT, ' CHARACTER', charnumber:4,
               ':  ERROR IN CHARACTER STATE TREE FORMAT');
      END;  (* readtree *)

      PROCEDURE construct;
         (* Puts tree together from array 'pairs' *)
         VAR rootstate, otherone : CHAR;
            i, j, k : INTEGER;
            done : BOOLEAN;
            poynter : nodepntr;

         PROCEDURE attachnodes (poynter : nodepntr);
            (* Makes linked list of all nodes to which passed node is
               ancestral.  First such node is 'descendant'; second
               such node is 'sibling' of first; third such node is
               sibling of second; etc.  *)
            VAR linker, ptr : nodepntr;
                  i, j, k : INTEGER;
         BEGIN (* attachnodes *)
            linker := poynter;
            FOR i := 1 TO npairs DO
               FOR j := 1 TO 2 DO
                  IF poynter^.state = pair[i,j]
                  THEN BEGIN
                     IF j = 1
                     THEN otherone := pair[i,2]
                     ELSE otherone := pair[i,1];
                     IF (otherone <> '.') AND
                        (otherone <> poynter^.ancestor^.state)
                     THEN BEGIN
                        k := offset+1;
                        WHILE otherone <> symbarray[k] DO
                           k := k+1;
                        IF nodes[k-offset] <> NIL THEN error := TRUE;
                        NEW(ptr);
                        ptr^.ancestor := poynter;
                        ptr^.descendant := NIL;
                        ptr^.sibling := NIL;
                        ptr^.state := otherone;
                        IF linker = poynter
                        THEN poynter^.descendant := ptr  (* If first *)
                        ELSE linker^.sibling := ptr; (* If not first *)
                        nodes[k-offset] := ptr; (* Save pntr to node *)
                        linker := ptr;
                        END;
                     END;
            END;  (* attachnodes *)

            PROCEDURE maketree (poynter : nodepntr);
               (* Recursively attach nodes *)
            BEGIN (* maketree *)
               IF (poynter <> NIL) AND (error = FALSE)
               THEN BEGIN
                  attachnodes(poynter);
                  maketree(poynter^.descendant);
                  maketree(poynter^.sibling);
                  END;
            END;  (* maketree *)

         BEGIN (* construct *)
            error := FALSE;
            rooted := FALSE;
            ancsymbol[charindex] := '?';
            rootstate := pair[1,1];
            nstates := 0;
            FOR i := 1 TO npairs DO
               FOR j := 1 TO 2 DO BEGIN
                  k := 1;
                  done := FALSE;
                  WHILE NOT done DO
                     IF k > nstates
                     THEN done := TRUE
                     ELSE IF pair[i,j] = symbarray[offset+k]
                        THEN done := TRUE
                        ELSE k := k+1;
                  IF k > nstates
                  THEN IF pair[i,j] = '.'
                     THEN BEGIN
                        IF rooted THEN error := TRUE;
                        rooted := TRUE;
                        ancsymbol[charindex] := '0';
                        IF j = 1
                        THEN rootstate := pair[i,2]
                        ELSE rootstate := pair[i,1];
                        END
                     ELSE BEGIN
                        nstates := nstates+1;
                        symbarray[offset+nstates] := pair[i,j];
                        END;
                  END;
            IF (rooted AND (nstates <> npairs)) OR
               (NOT(rooted) AND (nstates <> npairs+1))
            THEN error := TRUE;
            NEW(root);
            root^.state := ' ';
            NEW(root^.descendant);
            root^.descendant^.ancestor := root;
            root := root^.descendant;
            root^.descendant := NIL;
            root^.sibling := NIL;
            root^.state := rootstate;
            FOR i := 1 TO nstates DO nodes[i] := NIL;
            i := 1;
            WHILE symbarray[offset+i] <> rootstate DO i := i+1;
            nodes[i] := root;
            maketree(root);
            FOR i := 1 TO nstates DO
               IF nodes[i] <> root
               THEN IF nodes[i] = NIL
                  THEN error := TRUE
                  ELSE BEGIN
                     poynter := nodes[i]^.ancestor;
                     WHILE (poynter <> root) 
                       AND (poynter <> nodes[i]) DO
                        poynter := poynter^.ancestor;
                     IF poynter <> root THEN error := TRUE;
                     END;
            IF error THEN WRITELN(OUTPUT, ' CHARACTER', charnumber:4,
                     ':  INVALID CHARACTER STATE TREE DESCRIPTION');
         END;  (* construct *)

         PROCEDURE factortree;
            (* Generate the string of 0's and 1's that will be 
               substituted for each symbol of the multistate char. *)
            VAR i, j, place, edgenum, factoroffset : INTEGER;
               poynter : nodepntr;
            PROCEDURE numberedges (poynter : nodepntr);
               (* Assign to each node a number for the edge below it.
                  The root is zero *)
            BEGIN (* numberedges *)
               IF poynter <> NIL THEN BEGIN
                  poynter^.edge := edgenum;
                  edgenum := edgenum+1;
                  numberedges(poynter^.descendant);
                  numberedges(poynter^.sibling);
                  END;
            END;  (* numberedges *)

         BEGIN (* factortree *)
            edgenum := 0;
            numberedges(root);
            factoroffset := offset+nstates;
            FOR i := 1 TO nstates DO BEGIN
               place := factoroffset + (nstates-1)*(i-1);
               FOR j := place+1 TO place+nstates-1 DO
                  symbarray[j] := '0';
               poynter := nodes[i];
               WHILE poynter <> root DO BEGIN
                  symbarray[place+poynter^.edge] := '1';
                  poynter := poynter^.ancestor;
                  END;
               END;
            END;  (* factortree *)

      BEGIN (* dotrees *)
         charindex := 0;
         lastchar := 0;
         offset := 0;
         charnumber := 0;
         READ(infile, charnumber);
         WHILE charnumber < 999 DO BEGIN
            IF charnumber < lastchar
            THEN BEGIN
               WRITE(OUTPUT, ' CHARACTER',charnumber:4,
                                      ':  OUT OF ORDER');
               terminate := TRUE;
               END;
            charindex := charindex + 1;
            lastindex := charindex;
            readtree;     (* Process character-state tree  *)
            IF error = TRUE
            THEN BEGIN
               terminate := TRUE;
               nstates := 0;
               END
            ELSE IF npairs > 0
               THEN BEGIN
                  construct;   (* Link tree together  *)
                  IF error = TRUE
                  THEN terminate := TRUE
                  ELSE factortree;   (* Generate binary factors for  *)
                  END                (*    each multistate character *)
               ELSE BEGIN            (*    symbol  *)
                  nstates := 0;
                  ancsymbol[charindex] := '?';
                  END;
            lastchar := charnumber;
            charnum[charindex] := charnumber;
            chstart[charindex] := offset;
            numstates[charindex] := nstates;
            offset := offset + SQR(nstates);
            READ(infile, charnumber);
            END;
         READLN(infile);
         END;  (* dotrees *)

   PROCEDURE dodatamatrix;
      (* Reads species information and write factored data set *)
      VAR i, charindex, totalfactors, chposition : INTEGER;

      PROCEDURE writech(ch : CHAR);
         (* Writes a single character to output *)
      BEGIN (* writech *)
         IF chposition > maxoutput
         THEN BEGIN
            WRITELN(outfile);
            chposition := 1;
            END;
         WRITE(outfile, ch);
         chposition := chposition+1;
      END;  (* writech *)

      PROCEDURE writefactors;
         (* Writes 'FACTORS' line *)
         VAR i, charindex : INTEGER;
            symbol : CHAR;
      BEGIN (* writefactors *)
         WRITE(outfile, 'FACTORS   ');
         chposition := 11;
         symbol := '-';
         FOR charindex := 1 TO lastindex DO BEGIN
            IF symbol = '-'
            THEN symbol := '+'
            ELSE symbol := '-';
            IF numstates[charindex] = 0
            THEN writech(symbol)
            ELSE FOR i := 1 TO numstates[charindex]-1 DO
               writech(symbol);
            END;
         WRITELN(outfile);
         END;  (* writefactors *)

      PROCEDURE writeancestor;
         (* Writes 'ANCESTOR' line *)
         VAR i, charindex : INTEGER;
      BEGIN (* writeancestor *)
         charindex := 1;
         WHILE ancsymbol[charindex] = '?' DO charindex := charindex+1;
         IF charindex <= lastindex
         THEN BEGIN
            WRITE(outfile, 'ANCESTOR  ');
            chposition := 11;
            FOR charindex := 1 TO lastindex DO
               IF numstates[charindex] = 0
               THEN writech(ancsymbol[charindex])
               ELSE FOR i := 1 TO numstates[charindex]-1 DO
                  writech(ancsymbol[charindex]);
            WRITELN(outfile);
            END;
      END;  (* writeancestor *)

      PROCEDURE doeu;
         (* Writes factored data for a single species  *)
         VAR i, charindex, place : INTEGER;
            multichar : ARRAY [1..maxchars] OF CHAR;
      BEGIN (* doeu *)
         FOR i := 1 TO nmlngth DO BEGIN
            READ(infile, ch);
            WRITE(outfile, ch);
            END;
         chposition := 11;
         FOR i := 1 TO nchars DO BEGIN
            REPEAT
               IF EOLN(infile) THEN READLN(infile);
               READ(infile, ch);
            UNTIL ch <> ' ';
            multichar[i] := ch;
            END;
         READLN(infile);
         FOR charindex := 1 TO lastindex DO BEGIN
            IF numstates[charindex] = 0
            THEN writech(multichar[charnum[charindex]])
            ELSE BEGIN
               i := 1;
               WHILE (symbarray[chstart[charindex]+i]
                        <> multichar[charnum[charindex]])
                        AND (i <= numstates[charindex]) DO
                  i := i+1;
               IF i > numstates[charindex]
               THEN BEGIN
                  IF multichar[charnum[charindex]] = unkchar
                  THEN FOR i := 1 TO numstates[charindex]-1 DO
                          writech('?')
                  ELSE BEGIN
                     WRITELN(outfile);
                     WRITELN(OUTPUT, ' MULTISTATE CHARACTER',
                        charnum[charindex]:4,':  ''',multichar
                        [charnum[charindex]],''' IS NOT A DOCUMENTED ',
                        'STATE');
                     END
                  END
               ELSE BEGIN
                  place := chstart[charindex] + numstates[charindex]
                               + (numstates[charindex]-1)*(i-1);
                  FOR i := 1 TO numstates[charindex]-1 DO
                     writech(symbarray[place+i]);
                  END;
               END;
            END;
         WRITELN(outfile);
      END;  (* doeu *)

   BEGIN (* dodatamatrix *)
      totalfactors := 0;
      FOR charindex := 1 TO lastindex DO
         IF numstates[charindex] = 0
         THEN totalfactors := totalfactors + 1
         ELSE totalfactors := totalfactors + numstates[charindex] - 1;
      IF rooted AND ancestorrequest
      THEN WRITELN(outfile, neus+1:5, totalfactors:5, '    A')
      ELSE WRITELN(outfile, neus:5, totalfactors:5);
      IF factorrequest THEN writefactors;
      IF ancestorrequest THEN writeancestor;
      FOR i := 1 TO neus DO doeu;
   END;  (* dodatamatrix *)

BEGIN (* factor *)
   (* ASSIGN(infile, 'INFILE'); -- Turbo Pascal only *)
   RESET(infile);
   (* ASSIGN(outfile, 'OUTFILE'); -- Turbo Pascal only *)
   REWRITE(outfile);
   factorrequest := FALSE;
   ancestorrequest := FALSE;
   terminate := FALSE;
   READ(infile, neus, nchars);
   IF nchars > maxchars
   THEN WRITELN(OUTPUT, ' TOO MANY CHARACTERS -- increase CONSTants');
   WHILE NOT(EOLN(infile)) DO BEGIN    (* Read options  *)
      READ(infile, ch);
      IF (ch = 'F') OR (ch = 'f') THEN factorrequest := TRUE;
      IF (ch = 'A') OR (ch = 'a') THEN ancestorrequest := TRUE;
      END;
   READLN(infile);
   dotrees;   (* Read and factor character-state trees *)
   IF terminate
   THEN WRITELN(OUTPUT, ' EXECUTION TERMINATED BECAUSE ',
                 'OF PREVIOUS ERRORS')
   ELSE dodatamatrix;  (* Read species information and write factors *)
   (* CLOSE(infile);  May need this here *)
   (* CLOSE(outfile);  May need this here *)
END.  (* factor *)

