PROGRAM Fitch (infile, treefile, OUTPUT);
  (* version 3.1. (c) Copyright 1986, 1987, 1988 by the University of
     Washington and Joseph Felsenstein.  Written by Joseph Felsenstein.
     Permission is granted to copy and use this program provided no
     fee is charged for it and provided that this copyright
     notice is not removed. *)

CONST maxsp = 25;   (* maximum number of species *)
   maxsp2 = 48;     (* 2 x maxsp - 2 *)
   smoothings = 4;  (* number of passes through smoothing algorithm *)
   namelength = 10; (* number of characters max. in species name *)

TYPE vector = ARRAY [1..maxsp2] OF REAL;
   intvector = ARRAY [1..maxsp] OF INTEGER;
   naym = ARRAY[1..namelength] OF CHAR;
   longer = ARRAY[0..5] OF INTEGER;

VAR infile, treefile : TEXT;
   numsp, numsp1, numsp2, nums, inseed, outgrno : INTEGER;
   x : ARRAY[1..maxsp] OF vector;
   reps : ARRAY[1..maxsp] OF intvector;
   names : ARRAY[1..maxsp] OF naym;
   global, jumble, usertree, lower, upper, negallowed,
      outgropt, replicates, trout : BOOLEAN;
   power : REAL;
   seed : longer;
   enterorder : ARRAY[1..maxsp] OF INTEGER;

   FUNCTION randum (VAR seed : longer) : REAL;
      (* random number generator -- slow but machine independent *)
      VAR i, j, k, sum : INTEGER;
         mult, newseed : longer;
         x : REAL;
   BEGIN (* randum *)
      mult[0] := 13;
      mult[1] := 24;
      mult[2] := 22;
      mult[3] :=  6;
      FOR i := 0 TO 5 DO
         newseed[i] := 0;
      FOR i := 0 TO 5 DO BEGIN
         sum := newseed[i];
         k := i;
         IF i > 3 THEN k := 3;
         FOR j := 0 TO k DO
            sum := sum + mult[j] * seed[i-j];
         newseed[i] := sum;
         FOR j := i TO 4 DO BEGIN
            newseed[j+1] := newseed[j+1] + newseed[j] DIV 64;
            newseed[j] := newseed[j] MOD 64;
            END;
         END;
      seed := newseed;
      seed[5] := seed[5] MOD 4;
      x := 0.0;
      FOR i := 0 TO 5 DO
         x := x / 64.0 + seed[i];
      x := x / 4.0;
      randum := x;
   END; (* randum *)

   FUNCTION letter (ch : CHAR) : BOOLEAN;
      (* tests whether ch is a letter
         --  works for both ASCII and EBCDIC codes *)
   BEGIN (* letter *)
      letter := ((ch >= 'A') AND (ch <= 'I'))
                OR ((ch >= 'J') AND (ch <= 'R'))
                OR ((ch >= 'S') AND (ch <= 'Z'))
                OR ((ch >= 'a') AND (ch <= 'i'))
                OR ((ch >= 'j') AND (ch <= 'r'))
                OR ((ch >= 's') AND (ch <= 'z'));
   END; (* letter *)

   PROCEDURE uppercase (VAR ch : CHAR);
      (* convert a character to upper case -- either ASCII or EBCDIC *)
   BEGIN (* uppercase *)
      IF letter(ch) AND (('a' > 'A') AND (ch >= 'a')
                     OR  ('a' < 'A') AND (ch < 'A'))
      THEN ch := CHR(ORD(ch) + ORD('A') - ORD('a'));
   END; (* uppercase *)

   PROCEDURE getinput;

      PROCEDURE getnums;
      BEGIN (* getnums *)
         WRITELN(OUTPUT);
         WRITELN(OUTPUT,' Fitch-Margoliash method version 3.1');
         WRITELN(OUTPUT);
         READ(infile, numsp);
         WRITELN(OUTPUT, ' ', numsp:4, ' Populations');
         IF numsp > maxsp
         THEN WRITELN(OUTPUT, ' TOO MANY SPECIES: adjust CONSTants');
         numsp1 := numsp + 1;
         numsp2 := 2*numsp - 2;
      END; (* getnums *)

      PROCEDURE getoptions;
         VAR ch : CHAR;
            i, extranum : INTEGER;
      BEGIN (* getoptions *)
         global := FALSE;
         jumble := FALSE;
         lower := FALSE;
         negallowed := FALSE;
         outgropt := FALSE;
         outgrno := 1;
         power := 2;
         replicates := FALSE;
         trout := FALSE;
         upper := FALSE;
         usertree := FALSE;
         extranum := 0;
         WHILE NOT EOLN(infile) DO BEGIN
            READ(infile, ch);
            uppercase (ch);
            IF (ch = 'G') OR (ch = 'J') OR (ch = 'L') OR (ch = 'O')
               OR (ch = 'N') OR (ch = 'P') OR (ch = 'R')
               OR (ch = 'S') OR (ch = 'U') OR (ch = 'Y')
            THEN CASE ch OF
                  'G' : global := TRUE;
                  'J' : BEGIN
                        jumble := TRUE;
                        extranum := extranum + 1;
                        END;
                  'L' : lower := TRUE;
                  'N' : negallowed := TRUE;
                  'O' : BEGIN
                        outgropt := TRUE;
                        extranum := extranum + 1;
                        END;
                  'P' : extranum := extranum + 1;
                  'R' : upper := TRUE;
                  'S' : replicates := TRUE;
                  'U' : usertree := TRUE;
                  'Y' : trout := TRUE
                  END
            ELSE IF ch <> ' '
               THEN WRITELN(OUTPUT, ' Bad Option Character: ', ch);
            END;
         FOR i := 1 TO extranum DO BEGIN
            READLN(infile);
            READ(infile, ch);
            uppercase (ch);
            IF (ch = 'J') OR (ch = 'O') OR (ch = 'P')
            THEN CASE ch OF
               'J' : READ(infile, inseed);
               'O' : READ(infile, outgrno);
               'P' : READ(infile, power)
               END;
            END;
         WRITELN(OUTPUT, '                                     2');
         WRITELN(OUTPUT, '                   <  <   (Obs - Exp)');
         WRITELN(OUTPUT, ' Sum of squares =  <  <   ------------');
         WRITELN(OUTPUT, '                   i  j         ',
                        power:4:1);
         WRITELN(OUTPUT, '                              Obs');
         WRITELN(OUTPUT);
         WRITE(OUTPUT, ' Negative branch lengths ');
         IF NOT negallowed
         THEN WRITE(OUTPUT, 'not ');
         WRITELN(OUTPUT, 'allowed');
         WRITELN(OUTPUT);
         IF global
         THEN BEGIN
            WRITELN(OUTPUT, ' global optimization');
            WRITELN(OUTPUT);
            END;
         IF jumble
         THEN BEGIN
            WRITELN(OUTPUT, ' Random number seed = ', inseed:8);
            WRITELN(OUTPUT);
            FOR i := 0 TO 5 DO seed[i] := 0;
            i := 0;
            REPEAT
               seed[i] := inseed MOD 64;
               inseed := inseed DIV 64;
               i := i + 1;
            UNTIL inseed = 0;
            END;
      END; (* getoptions *)

      PROCEDURE getdata;
         VAR i, j, k, columns : INTEGER;
            skipit, skipother : BOOLEAN;
      BEGIN (* getdata *)
         IF replicates
         THEN columns := 4
         ELSE columns := 6;
         WRITELN(OUTPUT);
         WRITE(OUTPUT, ' Name                       Distances');
         IF replicates
         THEN WRITE(OUTPUT, ' (replicates)');
         WRITELN(OUTPUT);
         WRITE(OUTPUT, ' ----                       ---------');
         IF replicates
         THEN WRITE(OUTPUT, '-------------');
         WRITELN(OUTPUT);
         WRITELN(OUTPUT);
         FOR i := 1 TO numsp DO BEGIN
            x[i][i] := 0.0;
            READLN(infile);
            FOR j := 1 TO namelength DO
               IF EOLN(infile)
               THEN names[i][j] := ' '
               ELSE READ(infile, names[i][j]);
            FOR j := 1 TO numsp DO BEGIN
               reps[i][j] := 1;
               skipit := (lower AND (j >= i)) OR (upper AND (j <= i));
               skipother := (lower AND (i >= j)) OR
                              (upper AND (i <= j));
               IF NOT skipit
               THEN BEGIN
                  IF EOLN(infile)
                  THEN READLN(infile);
                  READ(infile, x[i][j]);
                  IF replicates
                  THEN BEGIN
                     IF EOLN(infile)
                     THEN READLN(infile);
                     READ(infile, reps[i][j]);
                     END;
                  END;
               IF (NOT skipit) AND skipother
               THEN BEGIN
                  x[j][i] := x[i][j];
                  reps[j][i] := reps[i][j];
                  END;
               END;
            END;
         FOR i := 1 TO numsp DO BEGIN
            WRITE(OUTPUT, ' ');
            FOR j := 1 TO namelength DO WRITE(OUTPUT, names[i][j]);
            WRITE(OUTPUT, ' ');
            FOR j := 1 TO numsp DO BEGIN
               WRITE(OUTPUT, x[i][j]:10:5);
               IF replicates
               THEN WRITE(OUTPUT, ' (', reps[i][j]:3, ')');
               IF ((j MOD columns) = 0) AND (j < numsp)
               THEN BEGIN
                  WRITELN(OUTPUT);
                  FOR k := 1 TO namelength+2 DO
                     WRITE(OUTPUT, ' ');
                  END;
               END;
            WRITELN(OUTPUT);
            END;
         WRITELN(OUTPUT);
      END; (* getdata *)

   BEGIN (* getinput *)
      getnums;
      getoptions;
      getdata;
   END; (* getinput *)

   PROCEDURE maketree;

      TYPE nodeptr = ^node;
         node = RECORD
            next, back : nodeptr;
            tip        : BOOLEAN;
            number     : INTEGER;
            nayme      : naym;
            d, w       : vector;
            v          : REAL;
            dist       : REAL;
            xcoord,
             ycoord    : INTEGER
            END;
         tree = RECORD
            nodep      : ARRAY[1..maxsp2] OF nodeptr;
            likelihood : REAL;
            start      : nodeptr
            END;

      VAR curtree, priortree, bestree : tree;
         i, j, k, nextsp, numtrees, which : INTEGER;
         succeeded : BOOLEAN;

      PROCEDURE setuptree (VAR a : tree);
        VAR i, j : INTEGER;
           p, q : nodeptr;
      BEGIN (* setuptree *)
         WITH a DO BEGIN
            FOR i := 1 TO numsp DO BEGIN
               NEW(nodep[i]);
               nodep[i]^.tip := TRUE;
               nodep[i]^.number := i;
               END;
            FOR i := numsp1 TO numsp2 DO BEGIN
               q := NIL;
               FOR j := 1 TO 3 DO BEGIN
                  NEW(p);
                  p^.tip := FALSE;
                  p^.number := i;
                  p^.next := q;
                  q := p;
                  END;
               p^.next^.next^.next := p;
               nodep[i] := p;
               END;
            likelihood := -1.0;
            start := nodep[1];
            END;
      END; (* setuptree *)

      PROCEDURE hookup (p, q : nodeptr);
      BEGIN (* hookup *)
         p^.back := q;
         q^.back := p;
      END; (* hookup *)

      FUNCTION evaluate (VAR t : tree ) : REAL;
         VAR sum : REAL;

         PROCEDURE firsttraverse (p : nodeptr);
            VAR nx : INTEGER;

            PROCEDURE secondtraverse (q : nodeptr; y : REAL);
               VAR z : REAL;
            BEGIN (* secondtraverse *)
               z := y + q^.v;
               IF q^.tip
               THEN sum := sum + q^.w[nx] * SQR(q^.d[nx] - z)
               ELSE BEGIN
                  secondtraverse (q^.next^.back, z);
                  secondtraverse (q^.next^.next^.back, z);
                  END;
            END; (* secondtraverse *)

         BEGIN (* firsttraverse *)
            IF p^.tip
            THEN BEGIN
               nx := p^.number;
               secondtraverse (p^.back, 0.0);
               END
            ELSE BEGIN
               firsttraverse (p^.next^.back);
               firsttraverse (p^.next^.next^.back);
               END;
         END; (* firsttraverse *)

      BEGIN (* evaluate *)
         sum := 0.0;
         firsttraverse (t.start^.back^.back);
         firsttraverse (t.start^.back);
         IF replicates AND (lower OR upper)
         THEN sum := sum / 2;
         t.likelihood := -sum;
         evaluate := -sum;
      END; (* evaluate *)

      PROCEDURE nudists (x, y : nodeptr);
         VAR nq, nr, nx, ny : INTEGER;
            dil, djl, wil, wjl, vi, vj : REAL;
            qprime, rprime : nodeptr;
      BEGIN (* nudists *)
         qprime := x^.next;
         rprime := qprime^.next^.back;
         qprime := qprime^.back;
         ny := y^.number;
         dil := qprime^.d[ny];
         djl := rprime^.d[ny];
         wil := qprime^.w[ny];
         wjl := rprime^.w[ny];
         vi := qprime^.v;
         vj := rprime^.v;
         x^.w[ny] := wil + wjl;
         IF (wil+wjl) <= 0.0
         THEN x^.d[ny] := 0.0
         ELSE x^.d[ny] := ((dil - vi) * wil + (djl - vj) * wjl)
                          /(wil + wjl);
         nx := x^.number;
         nq := qprime^.number;
         nr := rprime^.number;
         dil := y^.d[nq];
         djl := y^.d[nr];
         wil := y^.w[nq];
         wjl := y^.w[nr];
         y^.w[nx] := wil + wjl;
         IF (wil+wjl) <= 0.0
         THEN y^.d[nx] := 0.0
         ELSE y^.d[nx] := ((dil - vi) * wil + (djl - vj) * wjl)
                          /(wil + wjl);
      END; (* nudists *)

      PROCEDURE update (p : nodeptr);
         VAR negatives : BOOLEAN;

         PROCEDURE makedists (p : nodeptr);
            VAR i, nr, ns : INTEGER;
               q, r, s : nodeptr;
         BEGIN (* makedists *)
            r := p^.back;
            nr := r^.number;
            FOR i := 1 TO 3 DO BEGIN
               q := p^.next;
               s := q^.back;
               ns := s^.number;
               IF (s^.w[nr] + r^.w[ns]) <= 0.0
               THEN p^.dist := 0.0
               ELSE p^.dist := (s^.w[nr]*s^.d[nr] + r^.w[ns]*r^.d[ns])
                            / (s^.w[nr] + r^.w[ns]);
               p := q;
               r := s;
               nr := ns;
               END;
         END; (* makedists *)

         PROCEDURE makebigv (p : nodeptr);
            VAR i : INTEGER;
               temp, q, r : nodeptr;
         BEGIN (* makebigv *)
            q := p^.next;
            r := q^.next;
            negatives := FALSE;
            FOR i := 1 TO 3 DO BEGIN
               p^.v := (p^.dist + r^.dist - q^.dist)/2.0;
               IF p^.v < 0.0 THEN negatives := TRUE;
               p^.back^.v := p^.v;
               temp := p;
               p := q;
               q := r;
               r := temp;
               END;
         END; (* makebigv *)

         PROCEDURE correctv (p : nodeptr);
            VAR q, r, temp : nodeptr;
               i, j, n, nq, nr, ntemp : INTEGER;
               wq, wr : REAL;
         BEGIN (* correctv *)
            q := p^.next;
            r := q^.next;
            n := p^.back^.number;
            nq := q^.back^.number;
            nr := r^.back^.number;
            FOR i := 1 TO smoothings DO
               FOR j := 1 TO 3 DO BEGIN
                  wr := r^.back^.w[n] + p^.back^.w[nr];
                  wq := q^.back^.w[n] + p^.back^.w[nq];
                  IF (wr+wq) <= 0.0
                  THEN p^.v := 0.0
                  ELSE p^.v := ((p^.dist - q^.v) * wq
                                + (r^.dist - r^.v) * wr)
                                 / (wr + wq);
                  IF p^.v < 0 THEN p^.v := 0.0;
                  p^.back^.v := p^.v;
                  temp := p;
                  p := q;
                  q := r;
                  r := temp;
                  ntemp := n;
                  n := nq;
                  nq := nr;
                  nr := ntemp;
                  END;
         END; (* correctv *)

         PROCEDURE nuview (p : nodeptr);
            VAR i : INTEGER;
               q, r, pprime, temp : nodeptr;

            PROCEDURE alter (x, y : nodeptr);
            BEGIN (* alter *)
               nudists (x, y);
               IF NOT y^.tip
               THEN BEGIN
                  alter(x, y^.next^.back);
                  alter(x, y^.next^.next^.back);
                  END;
            END; (* alter *)

         BEGIN (* nuview *)
            q := p^.next;
            r := q^.next;
            FOR i := 1 TO 3 DO BEGIN
               temp := p;
               pprime := p^.back;
               alter (p, pprime);
               p := q;
               q := r;
               r := temp;
               END;
         END; (* nuview *)

      BEGIN (* update *)
         IF NOT p^.tip
         THEN BEGIN
            makedists (p);
            makebigv (p);
            IF (negatives) AND (NOT negallowed) THEN correctv (p);
            nuview (p);
            END;
      END; (* update *)

      PROCEDURE smooth (p : nodeptr);
      BEGIN (* smooth *)
         IF NOT p^.tip
         THEN BEGIN
            update (p);
            smooth (p^.next^.back);
            smooth (p^.next^.next^.back);
            END;
      END; (* smooth *)

      PROCEDURE insert (p, q: nodeptr; contin : BOOLEAN);
         VAR i : INTEGER;
            x : REAL;

         PROCEDURE fillin (pa, qa : nodeptr);

            PROCEDURE filltraverse (pb, qb : nodeptr);
            BEGIN (* filltraverse *)
               IF NOT qb^.tip
               THEN IF contin
                  THEN BEGIN
                     filltraverse (pb, qb^.next^.back);
                     filltraverse (pb, qb^.next^.next^.back);
                     nudists (qb, pb);
                     END
                  ELSE BEGIN
                     IF NOT qb^.next^.back^.tip
                     THEN nudists (qb^.next^.back, pb);
                     IF NOT qb^.next^.next^.back^.tip
                     THEN nudists (qb^.next^.next^.back, pb);
                     END;
            END; (* filltraverse *)

         BEGIN (* fillin *)
            IF NOT pa^.tip
            THEN BEGIN
              fillin (pa^.next^.back, qa);
              fillin (pa^.next^.next^.back, qa);
              END;
            filltraverse (pa, qa);
         END; (* fillin *)

      BEGIN (* insert *)
         hookup (p^.next^.next, q^.back);
         hookup (p^.next, q);
         x := q^.v/2.0;
         p^.v := 0.0;
         p^.back^.v := 0.0;
         p^.next^.v := x;
         p^.next^.back^.v := x;
         p^.next^.next^.back^.v := x;
         p^.next^.next^.v := x;
         fillin (p^.back, p);
         FOR i :=1 TO smoothings DO BEGIN
            smooth (p);
            smooth (p^.back);
            END;
      END; (* insert *)

      PROCEDURE copy ( VAR a, b : tree);
         VAR i : INTEGER;

         PROCEDURE copynode( VAR c, d : nodeptr);
            VAR necks : nodeptr;
         BEGIN (* copynode *)
            necks := d^.next;
            d^ := c^;
            d^.next := necks;
         END; (* copynode *)

      BEGIN (* copy *)
         FOR i := 1 TO numsp DO b.nodep[i]^ := a.nodep[i]^;
         FOR i := numsp1 TO numsp2 DO BEGIN
            copynode (a.nodep[i], b.nodep[i]);
            copynode (a.nodep[i]^.next, b.nodep[i]^.next);
            copynode (a.nodep[i]^.next^.next, b.nodep[i]^.next^.next);
            END;
         b.likelihood := a.likelihood;
         b.start := a.start;
      END; (* copy *)

      PROCEDURE setuptip (m : INTEGER; VAR t : tree);
         VAR i : INTEGER;
            n : intvector;
      BEGIN (* setuptip *)
         WITH t.nodep[m]^ DO BEGIN
            d := x[m];
            n := reps[m];
            FOR i := 1 TO numsp DO
               IF (i <> m) AND (n[i] > 0)
               THEN w[i] := n[i] / EXP(power * LN(d[i]))
               ELSE BEGIN
                  w[m] := 0.0;
                  d[m] := 0.0;
                  END;
            FOR i := numsp+1 TO numsp2 DO BEGIN
               w[i] := 1.0;
               d[i] := 0.0;
               END;
            number := m;
            nayme := names[m];
            v := 0.0;
            END;
      END; (* setuptip *)

      PROCEDURE buildnewtip (m : INTEGER; VAR t : tree);
         VAR p : nodeptr;
      BEGIN (* buildnewtip *)
         setuptip (m, t);
         p := t.nodep[nextsp + numsp - 2];
         hookup (t.nodep[m], p);
      END; (* buildnewtip *)

      PROCEDURE buildsimpletree (VAR t : tree);
      BEGIN (* buildsimpletree *)
         setuptip (enterorder[1], t);
         setuptip (enterorder[2], t);
         hookup (t.nodep[enterorder[1]], t.nodep[enterorder[2]]);
         buildnewtip (enterorder[3], t);
         insert(t.nodep[enterorder[3]]^.back,
                   t.nodep[enterorder[1]], FALSE);
      END; (* buildsimpletree *)

      PROCEDURE addtraverse (p, q : nodeptr; contin : BOOLEAN);
      BEGIN (* addtraverse *)
         insert(p, q, TRUE);
         numtrees := numtrees + 1;
         IF evaluate(curtree) > bestree.likelihood
         THEN copy (curtree, bestree);
         copy (priortree, curtree);
         IF (NOT q^.tip) AND contin
         THEN BEGIN
            addtraverse (p, q^.next^.back, contin);
            addtraverse (p, q^.next^.next^.back, contin);
            END;
      END; (* addtraverse *)

      PROCEDURE rearrange (p : nodeptr);
         (* rearranges the tree, globally or locally *)
         VAR q, r : nodeptr;

         PROCEDURE remove (VAR p, q : nodeptr);
            (* remove p and record in q where it was *)
         BEGIN (* remove *)
            q := p^.next^.back;
            hookup (q, p^.next^.next^.back);
            p^.next^.back := NIL;
            p^.next^.next^.back := NIL;
            update(q);
            update(q^.back);
         END; (* remove *)

      BEGIN (* rearrange *)
         IF (NOT p^.tip) AND (NOT p^.back^.tip)
         THEN BEGIN
            r := p^.next^.next;
            remove (r, q);
            copy (curtree, priortree);
            addtraverse (r, q^.next^.back,
                            global AND (nextsp = numsp));
            addtraverse (r, q^.next^.next^.back,
                            global AND (nextsp = numsp));
            copy (bestree, curtree);
            IF global AND (nextsp = numsp)
            THEN WRITE(OUTPUT, '.');
            IF global AND (nextsp = numsp) AND (NOT succeeded)
            THEN BEGIN
               IF r^.back^.tip
               THEN BEGIN
                  r := r^.next^.next;
                  remove (r, q);
                  q := q^.back;
                  copy (curtree, priortree);
                  IF NOT q^.tip
                  THEN BEGIN
                     addtraverse (r, q^.next^.back, TRUE);
                     addtraverse (r, q^.next^.next^.back, TRUE);
                     END;
                  q := q^.back;
                  IF NOT q^.tip
                  THEN BEGIN
                     addtraverse (r, q^.next^.back, TRUE);
                     addtraverse (r, q^.next^.next^.back, TRUE);
                     END;
                  copy (bestree, curtree);
                  END;
               END;
            END;
         IF NOT p^.tip
         THEN BEGIN
            rearrange (p^.next^.back);
            rearrange (p^.next^.next^.back);
            END;
      END; (* rearrange *)

      PROCEDURE printree;
         CONST down = 2;
            over = 3;
         VAR tipx, tipy, i, rover : INTEGER;

         PROCEDURE coordinates (p : nodeptr);
         BEGIN (* coordinates *)
            IF p^.tip
            THEN BEGIN
               p^.xcoord := tipx;
               p^.ycoord := tipy;
               tipx := tipx + over;
               tipy := tipy + down;
               END
            ELSE BEGIN
               coordinates (p^.next^.back);
               p^.xcoord := p^.next^.back^.xcoord;
               coordinates (p^.next^.next^.back);
               p^.ycoord := p^.next^.next^.back^.ycoord;
               END;
         END; (* coordinates *)

         PROCEDURE drawline (i : INTEGER);
            VAR p, q : nodeptr;
               n, j : INTEGER;
         BEGIN (* drawline *)
            p := curtree.start^.back;
            WRITE(OUTPUT, ' ');
            REPEAT
               IF p^.next^.back^.ycoord >= i
               THEN p := p^.next^.back
               ELSE BEGIN
                  q := p^.next^.next^.back;
                  n := q^.xcoord - p^.xcoord;
                  IF p^.ycoord = i
                  THEN BEGIN
                     IF rover < over
                     THEN BEGIN
                        IF (i = (tipy-down)) AND (p^.xcoord = 0)
                        THEN WRITE(OUTPUT, 'L')
                        ELSE WRITE(OUTPUT, '*');
                        n := n + 1;
                        END
                     ELSE BEGIN
                        IF p^.number - numsp >= 10
                        THEN WRITE(OUTPUT, (p^.number - numsp):2)
                        ELSE WRITE(OUTPUT, (p^.number - numsp):1, '-');
                        END;
                     FOR j := 1 TO n-2 DO WRITE(OUTPUT, '-');
                     END
                  ELSE BEGIN
                     WRITE(OUTPUT, '!');
                     FOR j := 1 TO n-1 DO WRITE(OUTPUT, ' ');
                     END;
                  p := q;
                  END;
            UNTIL p^.tip;
            IF p^.ycoord = i
            THEN FOR j := 1 TO namelength DO
               WRITE(OUTPUT, p^.nayme[j]);
            WRITELN(OUTPUT);
         END; (* drawline *)

      BEGIN (* printree *)
         rover := 100 DIV numsp;
         IF rover > over THEN rover := over;
         tipx := 0;
         tipy := 1;
         coordinates (curtree.start^.back);
         WRITELN(OUTPUT);
         WRITELN(OUTPUT);
         WRITELN(OUTPUT);
         FOR i := 1 TO tipy-down DO drawline (i);
         FOR i := 1 TO down DO WRITELN(OUTPUT, ' !');
         WRITE(OUTPUT, ' ');
         FOR i := 1 TO namelength DO WRITE(OUTPUT,
            curtree.start^.nayme[i]);
         WRITELN(OUTPUT);
      END; (* printree *)

      PROCEDURE summarize;
         VAR i, j, totalnum : INTEGER;

         PROCEDURE describe (p : nodeptr);
            VAR i : INTEGER;
               q : nodeptr;
         BEGIN (* describe *)
            q := p^.back;
            WRITE(OUTPUT, ' ');
            IF q^.tip
            THEN FOR i := 1 TO namelength DO
               WRITE(OUTPUT, q^.nayme[i])
            ELSE WRITE(OUTPUT, (q^.number - numsp):4, '      ');
            WRITE(OUTPUT, '    ');
            IF p^.tip
            THEN FOR i := 1 TO namelength DO
               WRITE(OUTPUT, p^.nayme[i])
            ELSE WRITE(OUTPUT, (p^.number - numsp):4, '      ');
            WRITE(OUTPUT, q^.v:15:5);
            WRITELN(OUTPUT);
            IF NOT p^.tip
            THEN BEGIN
               describe (p^.next^.back);
               describe (p^.next^.next^.back);
               END;
         END; (* describe *)

      BEGIN (* summarize *)
         WRITELN(OUTPUT);
         WRITE(OUTPUT, ' remember:');
         IF outgropt
         THEN WRITE(OUTPUT, ' (although rooted by outgroup)');
         WRITELN(OUTPUT, ' this is an unrooted tree!');
         WRITELN(OUTPUT);
         WRITELN(OUTPUT, ' Sum of squares = ',
                     -curtree.likelihood:11:5);
         WRITELN(OUTPUT);
         IF power = 2.0
         THEN BEGIN
            totalnum := 0;
            FOR i := 1 TO nums DO
               FOR j := 1 TO nums DO
                  IF i <> j
                  THEN totalnum := totalnum + reps[i][j];
            WRITE(OUTPUT, ' Average percent standard deviation = ');
            WRITELN(OUTPUT, 100*SQRT(-curtree.likelihood/
                              (totalnum-2)):11:5);
            WRITELN(OUTPUT);
            END;
         IF NOT usertree
         THEN BEGIN
            WRITELN(OUTPUT, ' examined ',numtrees:4, ' trees');
            WRITELN(OUTPUT);
            END;
         WRITELN(OUTPUT, '  Between        And            Length');
         WRITELN(OUTPUT, '  -------        ---            ------');
         describe (curtree.start^.back);
         WRITELN(OUTPUT);
         WRITELN(OUTPUT);
      END; (* summarize *)

      PROCEDURE treeout (p : nodeptr);
         (* write out file with representation of final tree *)
         VAR i, n, w : INTEGER;
            c : CHAR;
            x : REAL;
      BEGIN (* treeout *)
         IF p^.tip
         THEN BEGIN
            n := 0;
            FOR i := 1 TO namelength DO
               IF p^.nayme[i] <> ' '
               THEN n := i;
            FOR i := 1 TO n DO BEGIN
               c := p^.nayme[i];
               IF c = ' '
               THEN c := '_';
               WRITE(treefile, c);
               END;
            END
         ELSE BEGIN
            WRITE(treefile, '(');
            treeout (p^.next^.back);
            WRITE(treefile, ',');
            treeout (p^.next^.next^.back);
            IF p = curtree.start^.back
            THEN BEGIN
               WRITE(treefile, ',');
               treeout (p^.back);
               END;
            WRITE(treefile, ')');
            END;
         x := p^.v;
         IF x > 0.0
         THEN w := TRUNC(0.4343*LN(x))
         ELSE IF x = 0.0
            THEN w := 0
            ELSE w := TRUNC(0.4343*LN(-x)) + 1;
         IF p = curtree.start^.back
         THEN WRITELN(treefile, ';')
         ELSE WRITE(treefile, ':', x:w+7:5);
      END; (* treeout *)

      PROCEDURE treeread;
         VAR ch : CHAR;
            p : nodeptr;
            i, nextnode : INTEGER;

         PROCEDURE findch (c : CHAR);
         BEGIN (* findch *)
            IF ch <> c
            THEN REPEAT
                  IF EOLN(infile) THEN READLN(infile);
                  READ(infile, ch);
               UNTIL ch = c;
         END; (* findch *)

         PROCEDURE addelement (p : nodeptr);
            VAR q : nodeptr;
               i, n : INTEGER;
               found : BOOLEAN;
               str : ARRAY[1..namelength] OF CHAR;
         BEGIN (* addelement *)
            REPEAT
               IF EOLN(infile) THEN READLN(infile);
               READ(infile, ch);
            UNTIL ch <> ' ';
            IF ch = '('
            THEN BEGIN
               nextnode := nextnode + 1;
               q := curtree.nodep[nextnode];
               hookup(p, q);
               addelement (q^.next);
               findch (',');
               addelement (q^.next^.next);
               findch (')');
               READ(infile, ch);
               END
            ELSE BEGIN
               FOR i := 1 TO namelength DO str[i] := ' ';
               n := 1;
               REPEAT
                  IF (ch = '_') THEN ch := ' ';
                  str[n] := ch;
                  IF EOLN(infile) THEN READLN(infile);
                  READ(infile, ch);
                  n := n + 1;
               UNTIL ((ch = ':') OR (ch = ',') OR (ch = ')')
                        OR (n > namelength));
               n := 1;
               REPEAT
                  found := TRUE;
                  FOR i := 1 TO namelength DO
                     found := found AND (str[i] = names[n][i]);
                  IF NOT found
                  THEN n := n + 1;
               UNTIL (n > numsp) OR found;
               IF n > numsp
               THEN BEGIN
                  WRITE(OUTPUT, ' Cannot find species: ');
                  FOR i := 1 TO namelength DO
                     WRITE(OUTPUT, str[i]);
                  WRITELN(OUTPUT);
                  END;
               nums := nums + 1;
               hookup (curtree.nodep[n], p);
               IF curtree.start^.number > n
               THEN curtree.start := curtree.nodep[n];
               END;
         END; (* addelement *)

      BEGIN (* treeread *)
         nums := 0;
         curtree.start := curtree.nodep[numsp];
         REPEAT
            READ(infile, ch);
         UNTIL ch <> ' ';
         IF ch = '('
         THEN BEGIN
            nextnode := numsp + 1;
            p := curtree.nodep[nextnode];
            FOR i := 1 TO 2 DO BEGIN
               addelement (p);
               p := p^.next;
               findch (',');
               END;
            addelement (p);
            END;
         READLN(infile);
      END; (* treeread *)

      PROCEDURE treevaluate;
         VAR i : INTEGER;
            dummy : REAL;

         PROCEDURE initrav (p : nodeptr);

            PROCEDURE nodeinit (p : nodeptr);
               VAR i, j : INTEGER;
            BEGIN (* nodeinit *)
               FOR i := 1 TO 3 DO BEGIN
                  FOR j := 1 TO numsp2 DO BEGIN
                     p^.w[j] := 1.0;
                     p^.d[j] := 0.0;
                     END;
                  p := p^.next;
                  END;
               p^.v := 1.0;
               p^.back^.v := 1.0;
            END; (* nodeinit *)

         BEGIN (* initrav *)
            IF NOT p^.tip
            THEN BEGIN
               nodeinit (p);
               initrav (p^.next^.back);
               initrav (p^.next^.next^.back);
               END;
         END; (* initrav *)

      BEGIN (* treevaluate *)
         FOR i := 1 TO numsp DO setuptip (i, curtree);
         initrav (curtree.start);
         initrav (curtree.start^.back);
         FOR i := 1 TO 4*smoothings DO smooth (curtree.start^.back);
         dummy := evaluate (curtree);
      END; (* treevaluate *)

   BEGIN (* maketree *)
      IF usertree
      THEN BEGIN
         setuptree (curtree);
         FOR which := 1 TO numsp DO setuptip (which, curtree);
         READLN(infile, numtrees);
         WRITE(OUTPUT, ' User-defined tree');
         IF numtrees > 1
         THEN WRITE(OUTPUT, 's');
         WRITELN(OUTPUT, ':');
         FOR which := 1 TO numtrees DO BEGIN
            treeread;
            treevaluate;
            printree;
            summarize;
            END;
         END
      ELSE BEGIN
         setuptree (curtree);
         setuptree (priortree);
         setuptree (bestree);
         FOR i := 1 TO numsp DO enterorder[i] := i;
         IF jumble
         THEN FOR i := 1 TO numsp DO BEGIN
            j := TRUNC(randum(seed)*numsp) + 1;
            k := enterorder[j];
            enterorder[j] := enterorder[i];
            enterorder[i] := k;
            END;
         nextsp := 3;
         buildsimpletree (curtree);
         curtree.start := curtree.nodep[enterorder[1]];
         numtrees := 1;
         nextsp := 4;
         WRITELN(OUTPUT, ' Adding species:');
         WRITE(OUTPUT, '    ');
         FOR i := 1 TO namelength DO
            WRITE(OUTPUT, names[enterorder[1]][i]);
         WRITELN(OUTPUT);
         WRITE(OUTPUT, '    ');
         FOR i := 1 TO namelength DO
            WRITE(OUTPUT, names[enterorder[2]][i]);
         WRITELN(OUTPUT);
         WRITE(OUTPUT, '    ');
         FOR i := 1 TO namelength DO
            WRITE(OUTPUT, names[enterorder[3]][i]);
         WRITELN(OUTPUT);
         WHILE nextsp <= numsp DO BEGIN
            nums := nextsp;
            buildnewtip (enterorder[nextsp], curtree);
            copy (curtree, priortree);
            bestree.likelihood := -99999.0;
            addtraverse (curtree.nodep[enterorder[nextsp]]^.back,
                                   curtree.start^.back, TRUE);
            copy (bestree, curtree);
            WRITE(OUTPUT, '    ');
            FOR j := 1 TO namelength DO
               WRITE(OUTPUT, names[enterorder[nextsp]][j]);
            WRITELN(OUTPUT);
            IF global AND (nextsp = numsp)
            THEN BEGIN
               WRITELN(OUTPUT, ' Doing global rearrangements');
               WRITE(OUTPUT, '    ');
               END;
            succeeded := TRUE;
            WHILE succeeded DO BEGIN
               succeeded := FALSE;
               rearrange (curtree.start^.back);
               IF global AND (nextsp = numsp)
               THEN WRITELN(OUTPUT);
               END;
            IF (nextsp = numsp)
            THEN BEGIN
               curtree.start := curtree.nodep[outgrno];
               printree;
               summarize;
               IF trout AND (nextsp = numsp)
               THEN BEGIN
                  treeout (curtree.start^.back);
                  WRITELN(OUTPUT, ' Tree also written onto file');
                  WRITELN(OUTPUT);
                  END;
               END;
            nextsp := nextsp + 1;
            END;
         END;
   END; (* maketree *)

BEGIN (* main program *)
   (* ASSIGN(infile, 'INFILE');  -- Turbo Pascal only *)
   RESET(infile);
   getinput;
   IF trout
   THEN BEGIN
      (* ASSIGN(treefile, 'TREEFILE');  -- Turbo Pascal only *)
      REWRITE(treefile);
      END;
   maketree;
(* CLOSE(infile);  -- may need this here *)
   IF trout
   THEN BEGIN
      (* CLOSE(treefile);  -- may need this here *)
      END;
END.

