PROGRAM Kitsch (infile, treefile, OUTPUT);
  (* version 3.1. (c) Copyright 1986, 1987, 1988 by the University of
     Washington and Joseph Felsenstein.  Written by Joseph Felsenstein.
     Permission is granted to copy and use this program provided no
     fee is charged for it and provided that this copyright
     notice is not removed. *)

CONST maxsp = 25;    (* maximum number of species               *)
   maxsz = 49;       (* size of pointer array.  >= 2*maxsp - 1  *)
   namelength = 10;  (* number of characters in species name    *)

TYPE vector = ARRAY [1..maxsz] OF REAL;
   naym = ARRAY[1..namelength] OF CHAR;
   nodeptr = ^node;           (* nodes will form binary tree *)
   node = RECORD      (* describes a tip species or an ancestor *)
      ancestor, leftdesc, rtdesc : nodeptr;
      index : INTEGER;
      tip : BOOLEAN;        (* present species are tips of tree *)
      d, w : vector;        (* distances and weights *)
      t : REAL;             (* time *)
      sametime : BOOLEAN;   (* bookkeeps scrunched nodes *)
      weight : REAL;        (* weight of node used by scrunch *)
      processed : BOOLEAN;  (* used by evaluate *)
      xcoord, ycoord : INTEGER  (* used by printree *)
      END;
   pointarray = ARRAY [1..maxsz] OF nodeptr;
   longer = ARRAY[0..5] OF INTEGER;

VAR root : nodeptr;
   infile, treefile : TEXT;
   numsp, numsp2, inseed, numtrees : INTEGER;
     (* numsp = number of species
        numtrees is used by usertree option part of maketree *)
   treenode : pointarray;      (* pointers to all nodes in tree *)
   names : ARRAY[1..maxsp] OF naym;
   jumble, usertree, lower, upper, negallowed,
                                   replicates, trout : BOOLEAN;
   power : REAL;
   seed : longer;
   enterorder : ARRAY[1..maxsp] OF INTEGER;

   FUNCTION randum (VAR seed : longer) : REAL;
      (* random number generator -- slow but machine independent *)
      VAR i, j, k, sum : INTEGER;
         mult, newseed : longer;
         x : REAL;
   BEGIN (* randum *)
      mult[0] := 13;
      mult[1] := 24;
      mult[2] := 22;
      mult[3] :=  6;
      FOR i := 0 TO 5 DO
         newseed[i] := 0;
      FOR i := 0 TO 5 DO BEGIN
         sum := newseed[i];
         k := i;
         IF i > 3 THEN k := 3;
         FOR j := 0 TO k DO
            sum := sum + mult[j] * seed[i-j];
         newseed[i] := sum;
         FOR j := i TO 4 DO BEGIN
            newseed[j+1] := newseed[j+1] + newseed[j] DIV 64;
            newseed[j] := newseed[j] MOD 64;
            END;
         END;
      seed := newseed;
      seed[5] := seed[5] MOD 4;
      x := 0.0;
      FOR i := 0 TO 5 DO
         x := x / 64.0 + seed[i];
      x := x / 4.0;
      randum := x;
   END; (* randum *)

   FUNCTION letter (ch : CHAR) : BOOLEAN;
      (* tests whether ch is a letter
         --  works for both ASCII and EBCDIC codes *)
   BEGIN (* letter *)
      letter := ((ch >= 'A') AND (ch <= 'I'))
                OR ((ch >= 'J') AND (ch <= 'R'))
                OR ((ch >= 'S') AND (ch <= 'Z'))
                OR ((ch >= 'a') AND (ch <= 'i'))
                OR ((ch >= 'j') AND (ch <= 'r'))
                OR ((ch >= 's') AND (ch <= 'z'));
   END; (* letter *)

   PROCEDURE uppercase (VAR ch : CHAR);
      (* convert a character to upper case -- either ASCII or EBCDIC *)
   BEGIN (* uppercase *)
      IF letter(ch) AND (('a' > 'A') AND (ch >= 'a')
                     OR  ('a' < 'A') AND (ch < 'A'))
      THEN ch := CHR(ORD(ch) + ORD('A') - ORD('a'));
   END; (* uppercase *)

   PROCEDURE getinput;

      PROCEDURE getnums;
      BEGIN (* getnums *)
         WRITELN(OUTPUT);
         WRITE(OUTPUT,' Fitch-Margoliash method ');
         WRITE(OUTPUT, 'with contemporary tips,');
         WRITELN(OUTPUT, ' version 3.1');
         WRITELN(OUTPUT);
         READ(infile, numsp);
         WRITELN(OUTPUT, ' ', numsp:4, ' Populations');
         IF numsp > maxsp
         THEN WRITELN(OUTPUT, ' TOO MANY SPECIES: adjust CONSTants');
         numsp2 := 2*numsp - 1;
      END; (* getnums *)

      PROCEDURE getoptions;
         VAR ch : CHAR;
            i, extranum : INTEGER;
      BEGIN (* getoptions *)
         jumble := FALSE;
         lower := FALSE;
         negallowed := FALSE;
         power := 2;
         replicates := FALSE;
         trout := FALSE;
         upper := FALSE;
         usertree := FALSE;
         extranum := 0;
         WHILE NOT EOLN(infile) DO BEGIN
            READ(infile, ch);
            uppercase (ch);
            IF (ch = 'J') OR (ch = 'L')
               OR (ch = 'N') OR (ch = 'P') OR (ch = 'R')
               OR (ch = 'S') OR (ch = 'U') OR (ch = 'Y')
            THEN CASE ch OF
                  'J' : BEGIN
                        jumble := TRUE;
                        extranum := extranum + 1;
                        END;
                  'L' : lower := TRUE;
                  'N' : negallowed := TRUE;
                  'P' : extranum := extranum + 1;
                  'R' : upper := TRUE;
                  'S' : replicates := TRUE;
                  'U' : usertree := TRUE;
                  'Y' : trout := TRUE
                  END
            ELSE IF ch <> ' '
               THEN WRITELN(OUTPUT, ' Bad Option Character: ', ch);
            END;
         FOR i := 1 TO extranum DO BEGIN
            READLN(infile);
            READ(infile, ch);
            uppercase (ch);
            IF ch = 'J'
            THEN READ(infile, inseed);
            IF ch = 'P'
            THEN READ(infile, power);
            END;
         WRITELN(OUTPUT, '                                     2');
         WRITELN(OUTPUT, '                   <  <   (Obs - Exp)');
         WRITELN(OUTPUT, ' Sum of squares =  <  <   ------------');
         WRITELN(OUTPUT, '                   i  j         ',
                    power:4:1);
         WRITELN(OUTPUT, '                              Obs');
         WRITELN(OUTPUT);
         WRITE(OUTPUT, ' negative branch lengths');
         IF NOT negallowed THEN WRITE(OUTPUT, ' not');
         WRITELN(OUTPUT, ' allowed');
         WRITELN(OUTPUT);
         IF jumble
         THEN BEGIN
            WRITELN(OUTPUT, ' Random number seed = ', inseed:8);
            WRITELN(OUTPUT);
            FOR i := 0 TO 5 DO seed[i] := 0;
            i := 0;
            REPEAT
               seed[i] := inseed MOD 64;
               inseed := inseed DIV 64;
               i := i + 1;
            UNTIL inseed = 0;
            END;
      END; (* getoptions *)

      PROCEDURE getdata;
         VAR i, j, k, columns, n : INTEGER;
            skipit, skipother : BOOLEAN;
            x : REAL;
      BEGIN (* getdata *)
         IF replicates
         THEN columns := 4
         ELSE columns := 6;
         WRITELN(OUTPUT);
         WRITE(OUTPUT, ' Name                       Distances');
         IF replicates
         THEN WRITE(OUTPUT, ' (replicates)');
         WRITELN(OUTPUT);
         WRITE(OUTPUT, ' ----                       ---------');
         IF replicates
         THEN WRITE(OUTPUT, '-------------');
         WRITELN(OUTPUT);
         WRITELN(OUTPUT);
         FOR i := 1 TO numsp2 DO BEGIN
            NEW(treenode[i]);
            treenode[i]^.ancestor := NIL;
            treenode[i]^.leftdesc := NIL;
            treenode[i]^.rtdesc := NIL;
            treenode[i]^.index := i;
            treenode[i]^.tip := (i <= numsp);
            treenode[i]^.t := 0.0;
            treenode[i]^.sametime := FALSE;
            END;
         FOR i := 1 TO numsp DO BEGIN
            treenode[i]^.d[i] := 0.0;
            treenode[i]^.w[i] := 0.0;
            treenode[i]^.weight := 0.0;
            READLN(infile);
            FOR j := 1 TO namelength DO
               IF EOLN(infile)
               THEN names[i][j] := ' '
               ELSE READ(infile, names[i][j]);
            FOR j := 1 TO numsp DO BEGIN
               skipit := (lower AND (j >= i)) OR (upper AND (j <= i));
               skipother := (lower AND (i >= j)) OR
                              (upper AND (i <= j));
               IF NOT skipit
               THEN BEGIN
                  IF EOLN(infile) THEN READLN(infile);
                  READ(infile, x);
                  treenode[i]^.d[j] := x;
                  IF replicates
                  THEN BEGIN
                     IF EOLN(infile)
                     THEN READLN(infile);
                     READ(infile, n);
                     END
                  ELSE n := 1;
                  treenode[i]^.w[j] := n;
                  IF skipother
                  THEN BEGIN
                     treenode[j]^.d[i] := treenode[i]^.d[j];
                     treenode[j]^.w[i] := treenode[i]^.w[j];
                     END;
                  END;
               END;
            END;
         READLN(infile);
         FOR i := 1 TO numsp DO BEGIN
            WRITE(OUTPUT, ' ');
            FOR j := 1 TO namelength DO WRITE(OUTPUT, names[i][j]);
            WRITE(OUTPUT, ' ');
            FOR j := 1 TO numsp DO BEGIN
               WRITE(OUTPUT, treenode[i]^.d[j]:10:5);
               IF replicates
               THEN WRITE(OUTPUT, ' (', TRUNC(treenode[i]^.w[j]):3, ')');
               IF ((j MOD columns) = 0) AND (j < numsp)
               THEN BEGIN
                  WRITELN(OUTPUT);
                  FOR k := 1 TO namelength+2 DO
                     WRITE(OUTPUT, ' ');
                  END;
               END;
            WRITELN(OUTPUT);
            END;
         WRITELN(OUTPUT);
         FOR i := 1 TO numsp DO
            FOR j := 1 TO numsp DO
               IF (i <> j) AND (treenode[i]^.d[j] > 0.0)
               THEN treenode[i]^.w[j] := treenode[i]^.w[j]
                        / EXP(power * LN(treenode[i]^.d[j]));
      END; (* getdata *)

   BEGIN (* getinput *)
      getnums;
      getoptions;
      getdata;
   END; (* getinput *)

   PROCEDURE maketree;
      (* constructs a binary tree from the pointers in treenode.
         adds each node at location which yields highest "likelihood"
         then rearranges the tree for greatest "likelihood" *)
      VAR i, j, k, examined : INTEGER;
         like, bestyet, gotlike : REAL;
         lastrearr : BOOLEAN;
         there, item, nufork, dummy : nodeptr;

      PROCEDURE add (below, newtip, newfork : nodeptr);
         (* inserts the nodes newfork and its left descendant, newtip,
            to the tree.  below becomes newfork's right descendant *)
      BEGIN (* add *)
         newfork^.ancestor := below^.ancestor;
         newfork^.rtdesc := below;
         newfork^.leftdesc := newtip;
         below^.ancestor := newfork;
         newtip^.ancestor := newfork;
         IF newfork^.ancestor <> NIL
         THEN IF newfork^.ancestor^.leftdesc = below
            THEN newfork^.ancestor^.leftdesc := newfork
            ELSE newfork^.ancestor^.rtdesc := newfork;
      END; (* add *)

      PROCEDURE remove (VAR item, fork : nodeptr);
         (* removes nodes item and its ancestor, fork, from the tree.
            the new descendant of fork's ancestor is made to be
            fork's second descendant (other than item).  Also
            returns pointers to the deleted nodes, item and fork *)
         VAR other, below : nodeptr;
      BEGIN (* remove *)
         IF item^.ancestor = NIL
         THEN fork := NIL
         ELSE BEGIN
            fork := item^.ancestor;
            IF item = fork^.leftdesc
            THEN other := fork^.rtdesc
            ELSE other := fork^.leftdesc;
            below := fork^.ancestor;
            other^.ancestor := below;
            IF below <> NIL
            THEN BEGIN
               IF below^.leftdesc = fork
               THEN below^.leftdesc := other
               ELSE below^.rtdesc := other;
               fork^.ancestor := NIL;
               END;
            fork^.leftdesc := NIL;
            fork^.rtdesc := NIL;
            item^.ancestor := NIL;
            END;
      END; (* remove *)

      PROCEDURE evaluate (r: nodeptr);
         (* fill in times and evaluate sum of squares for tree *)
         VAR i : INTEGER;
            sum : REAL;

         PROCEDURE firstraverse (q : nodeptr);
            (* prune the tree *)
            VAR i, j, k : INTEGER;
               u, v : nodeptr;

            PROCEDURE scrunch (s : nodeptr);
               (* see if nodes can be combined to
                  prevent negative lengths *)
               VAR tmax : REAL;
                  closest : nodeptr;
                  found : BOOLEAN;

               PROCEDURE scrunchtraverse (u : nodeptr);
                  (* traverse to find closest node to
                     the current one *)
               BEGIN (* scrunchtraverse *)
                  IF NOT u^.sametime
                  THEN BEGIN
                     IF u^.t > tmax
                     THEN BEGIN
                        closest := u;
                        tmax := u^.t;
                        END;
                     END
                  ELSE BEGIN
                     u^.t := u^.ancestor^.t;
                     IF NOT u^.tip
                     THEN BEGIN
                        scrunchtraverse (u^.leftdesc);
                        scrunchtraverse (u^.rtdesc);
                        END;
                     END;
               END; (* scrunchtraverse *)

               PROCEDURE combine (a, b : nodeptr);
                  (* put node b into the set having the same
                     time as a *)
               BEGIN (* combine *)
                  IF (a^.weight + b^.weight) <= 0.0
                  THEN a^.t := 0.0
                  ELSE a^.t := (a^.t * a^.weight + b^.t * b^.weight)
                           / (a^.weight + b^.weight);
                  a^.weight := a^.weight + b^.weight;
                  b^.sametime := TRUE;
               END; (* combine *)

            BEGIN (* scrunch *)
               closest := NIL;
               tmax := -1.0;
               REPEAT
                  scrunchtraverse (s^.leftdesc);
                  scrunchtraverse (s^.rtdesc);
                  found := (tmax > s^.t);
                  IF found THEN combine (s, closest);
                  tmax := -1.0;
               UNTIL (NOT found);
            END; (* scrunch *)

            PROCEDURE secondtraverse (a : nodeptr);
               (* recalculate distances, add to sum *)
               VAR l : INTEGER;
                  wil, wjl, wkl, wli, wlj, wlk : REAL;
            BEGIN (* secondtraverse *)
               IF NOT a^.processed
               THEN BEGIN
                  secondtraverse (a^.leftdesc);
                  secondtraverse (a^.rtdesc);
                  END
               ELSE IF a <> q
                  THEN BEGIN
                     l := a^.index;
                     wil := u^.w[l];
                     wjl := v^.w[l];
                     wkl := wil + wjl;
                     wli := a^.w[i];
                     wlj := a^.w[j];
                     wlk := wli + wlj;
                     q^.w[l] := wkl;
                     a^.w[k] := wlk;
                     IF wkl <= 0.0
                     THEN q^.d[l] := 0.0
                     ELSE q^.d[l] :=
                            (wil * u^.d[l] + wjl * v^.d[l]) / wkl;
                     IF wlk <= 0.0
                     THEN a^.d[k] := 0.0
                     ELSE a^.d[k] :=
                            (wli * a^.d[i] + wlj * a^.d[j]) / wlk;
                     IF wkl > 0.0
                     THEN sum := sum
                             + (wil*wjl/wkl)*SQR(u^.d[l]-v^.d[l]);
                     IF wlk > 0.0
                     THEN sum := sum
                             + (wli*wlj/wlk)*SQR(a^.d[i]-a^.d[j]);
                     END;
            END; (* secondtraverse *)

         BEGIN (* firsttraverse *)
            IF q <> NIL
            THEN BEGIN
               q^.sametime := FALSE;
               firstraverse (q^.leftdesc);
               firstraverse (q^.rtdesc);
               q^.processed := TRUE;
               IF NOT q^.tip
               THEN BEGIN
                  u := q^.leftdesc;
                  v := q^.rtdesc;
                  i := u^.index;
                  j := v^.index;
                  k := q^.index;
                  IF u^.w[j] + v^.w[i] <= 0.0
                  THEN q^.t := 0.0
                  ELSE q^.t := (u^.w[j]*u^.d[j] + v^.w[i]*v^.d[i])
                                     /(2.0*(u^.w[j] + v^.w[i]));
                  q^.weight := u^.weight + v^.weight + u^.w[j]
                                                      + v^.w[i];
                  IF NOT negallowed THEN scrunch (q);
                  secondtraverse (r);
                  END;
               END;
         END; (* firstraverse *)

         PROCEDURE sumtraverse (q : nodeptr);
            (* traverse to finish computation of sum
               of squares *)
            VAR i, j : INTEGER;
               u, v : nodeptr;
         BEGIN (* sumtraverse *)
            IF NOT q^.tip
            THEN BEGIN
               sumtraverse (q^.leftdesc);
               sumtraverse (q^.rtdesc);
               u := q^.leftdesc;
               v := q^.rtdesc;
               i := u^.index;
               j := v^.index;
               sum := sum + u^.w[j]*SQR(u^.d[j]-2.0*q^.t)
                          + v^.w[i]*SQR(v^.d[i]-2.0*q^.t);
               END;
         END; (* sumtraverse *)

      BEGIN (* evaluate *)
         sum := 0.0;
         FOR i := 1 TO numsp2 DO
            treenode[i]^.processed := treenode[i]^.tip;
         firstraverse (r);
         sumtraverse (r);
         examined := examined + 1;
         IF replicates AND (lower OR upper)
         THEN sum := sum / 2;
         like := -sum;
      END; (* evaluate *)

      PROCEDURE addpreorder (p, item, nufork : nodeptr);
         (* traverses a binary tree, calling PROCEDURE tryadd
            at a node before calling tryadd at its descendants *)

         PROCEDURE tryadd (p: nodeptr);
            (* temporarily adds one fork and one tip to the tree.
               if the location where they are added yields greater
               "likelihood" than other locations tested up to that
               time, then keeps that location as there *)
         BEGIN (* tryadd *)
            add (p, item, nufork);
            IF root^.ancestor <> NIL THEN root := root^.ancestor;
            evaluate (root);
            IF like > bestyet
            THEN BEGIN
               bestyet := like;
               there := p;
               END;
            remove (item, nufork);
         END; (* tryadd *)

      BEGIN (* addpreorder *)
         IF p <> NIL
         THEN BEGIN
            tryadd (p);
            addpreorder (p^.leftdesc, item, nufork);
            addpreorder (p^.rtdesc, item, nufork);
            END;
      END; (* addpreorder *)

      PROCEDURE rearrange (VAR r: nodeptr);
         (* traverses the tree (preorder), finding any local
            rearrangement which decreases the number of steps.
            if traversal succeeds in increasing the tree's
            "likelihood", PROCEDURE rearrange runs traversal again *)
         VAR success : BOOLEAN;

         PROCEDURE tryrearr (p : nodeptr);
            (* evaluates one rearrangement of the tree.
               if the new tree has greater "likelihood" than the old
               one sets success := TRUE and keeps the new tree.
               otherwise, restores the old tree *)
            VAR frombelow, whereto, forknode : nodeptr;
               oldlike : REAL;
         BEGIN (* tryrearr *)
            IF p^.ancestor <> NIL
            THEN BEGIN
               IF p^.ancestor^.ancestor <> NIL
               THEN BEGIN
                  oldlike := like;
                  IF p^.ancestor^.leftdesc = p
                  THEN frombelow := p^.ancestor^.rtdesc
                  ELSE frombelow := p^.ancestor^.leftdesc;
                  whereto := p^.ancestor^.ancestor;
                  forknode := p^.ancestor;
                  remove (p, forknode);
                  add (whereto, p, forknode);
                  IF r^.ancestor <> NIL THEN r := r^.ancestor;
                  evaluate (r);
                  IF like <= oldlike
                  THEN BEGIN
                     remove (p, forknode);
                     add (frombelow, p, forknode);
                     IF r^.ancestor <> NIL THEN r := r^.ancestor;
                     like := oldlike;
                     END
                  ELSE BEGIN
                     bestyet := like;
                     success := TRUE;
                     END;
                  END;
               END;
         END; (* tryrearr *)

         PROCEDURE repreorder (p : nodeptr);
            (* traverses a binary tree, calling PROCEDURE tryrearr
               at a node before calling tryrearr at its descendants *)
         BEGIN (* repreorder *)
            IF p <> NIL
            THEN BEGIN
               tryrearr (p);
               repreorder (p^.leftdesc);
               repreorder (p^.rtdesc);
               END;
         END; (* repreorder *)

      BEGIN (* rearrange *)
         success := TRUE;
         WHILE success DO BEGIN
            success := FALSE;
            repreorder (r);
            END;
      END; (* rearrange *)

      PROCEDURE treeread;
         (* read in user-defined tree and set it up *)
         VAR ch : CHAR;
            nextnode : INTEGER;

         PROCEDURE findch (c : CHAR);
            (* scan forward until find character c *)
         BEGIN (* findch *)
            IF ch <> c
            THEN REPEAT
                  IF EOLN(infile) THEN READLN(infile);
                  READ(infile, ch);
               UNTIL ch = c;
         END; (* findch *)

         PROCEDURE addelement (VAR p : nodeptr);
            (* recursive procedure adds nodes to user-defined tree *)
            VAR q : nodeptr;
               i, n : INTEGER;
               found : BOOLEAN;
               str : ARRAY[1..namelength] OF CHAR;
         BEGIN (* addelement *)
            REPEAT
               IF EOLN(infile) THEN READLN(infile);
               READ(infile, ch);
            UNTIL ch <> ' ';
            IF ch = '('
            THEN BEGIN
               nextnode := nextnode + 1;
               q := treenode[nextnode];
               addelement (q^.leftdesc);
               q^.leftdesc^.ancestor := q;
               findch (',');
               addelement (q^.rtdesc);
               q^.rtdesc^.ancestor := q;
               findch (')');
               p := q;
               END
            ELSE BEGIN
               FOR i := 1 TO namelength DO str[i] := ' ';
               n := 1;
               REPEAT
                  IF (ch = '_') THEN ch := ' ';
                  str[n] := ch;
                  IF EOLN(infile) THEN READLN(infile);
                  READ(infile, ch);
                  n := n + 1;
               UNTIL ((ch = ',') OR (ch = ')') OR (n > namelength));
               n := 1;
               REPEAT
                  found := TRUE;
                  FOR i := 1 TO namelength DO
                     found := found AND (str[i] = names[n][i]);
                  IF found
                  THEN p := treenode[n]
                  ELSE n := n + 1;
               UNTIL (n > numsp) OR found;
               IF n > numsp
               THEN BEGIN
                  WRITE(OUTPUT, ' Cannot find species: ');
                  FOR i := 1 TO namelength DO
                     WRITE(OUTPUT, str[i]);
                  WRITELN(OUTPUT);
                  END;
               END;
         END; (* addelement *)

      BEGIN (* treeread *)
         root := treenode[numsp + 1];
         nextnode := numsp;
         root^.ancestor := NIL;
         addelement (root);
         READLN(infile);
      END; (* treeread *)

      PROCEDURE printree;
         (* prints out diagram of the tree *)
         CONST down = 2;
            over = 3;
         VAR tipx, tipy, i, rover : INTEGER;

         PROCEDURE coordinates (p : nodeptr);
            (* establishes coordinates of nodes *)
         BEGIN (* coordinates *)
            IF p^.tip
            THEN BEGIN
               p^.xcoord := tipx;
               p^.ycoord := tipy;
               tipx := tipx + rover;
               tipy := tipy + down;
               END
            ELSE BEGIN
               coordinates (p^.leftdesc);
               p^.xcoord := p^.leftdesc^.xcoord;
               coordinates (p^.rtdesc);
               p^.ycoord := p^.rtdesc^.ycoord;
               END;
         END; (* coordinates *)

         PROCEDURE drawline (i : INTEGER);
            (* draws one row of the tree diagram by moving up tree *)
            VAR p, q : nodeptr;
               n, j : INTEGER;
         BEGIN (* drawline *)
            p := root;
            WRITE(OUTPUT, '   ');
            REPEAT
               IF p^.leftdesc^.ycoord >= i
               THEN p := p^.leftdesc
               ELSE BEGIN
                  q := p^.rtdesc;
                  n := q^.xcoord - p^.xcoord;
                  IF p^.ycoord = i
                  THEN BEGIN
                     IF p^.index - numsp >= 10
                     THEN WRITE(OUTPUT, (p^.index - numsp):2)
                     ELSE WRITE(OUTPUT, (p^.index - numsp):1, '-');
                     FOR j := 1 TO n-2 DO WRITE(OUTPUT, '-');
                     END
                  ELSE BEGIN
                     WRITE(OUTPUT, '!');
                     FOR j := 1 TO n-1 DO WRITE(OUTPUT, ' ');
                     END;
                  p := q;
                  END;
            UNTIL p^.tip;
            IF p^.ycoord = i
            THEN FOR j := 1 TO namelength DO
                  WRITE(OUTPUT, names[p^.index][j]);
            WRITELN(OUTPUT);
         END; (* drawline *)

      BEGIN (* printree *)
         tipx := 0;
         tipy := 1;
         rover := 100 DIV numsp;
         IF rover > over THEN rover := over;
         coordinates (root);
         WRITELN(OUTPUT);
         WRITELN(OUTPUT);
         WRITELN(OUTPUT);
         FOR i := 1 TO tipy-down DO drawline (i);
         WRITELN(OUTPUT, '  /');
         WRITELN(OUTPUT, ' /');
         WRITELN(OUTPUT);
      END; (* printree *)

      PROCEDURE describe;
         (* prints table of lengths, times, sum of squares, etc. *)
         VAR i, j : INTEGER;
            totalnum : REAL;

         PROCEDURE dtraverse (q : nodeptr);
            (* print table of lengths etc. *)
            VAR i : INTEGER;
         BEGIN (* dtraverse *)
            IF NOT q^.tip
            THEN BEGIN
               dtraverse(q^.leftdesc);
               dtraverse(q^.rtdesc);
               END;
            IF q^.ancestor <> NIL
            THEN BEGIN
               WRITE(OUTPUT, q^.ancestor^.index-numsp:4, '  ');
               IF q^.index <= numsp
               THEN FOR i := 1 TO namelength DO
                  WRITE(OUTPUT, names[q^.index][i])
               ELSE WRITE(OUTPUT, (q^.index - numsp):4, '      ');
               WRITE(OUTPUT, (q^.ancestor^.t-q^.t):13:5);
               WRITELN(OUTPUT, (root^.t-q^.t):15:5);
               END;
         END; (* dtraverse *)

      BEGIN (* describe *)
         WRITELN(OUTPUT, ' Sum of squares = ',-like:10:3);
         WRITELN(OUTPUT);
         IF ABS(power - 2) < 0.01
         THEN BEGIN
            totalnum := 0;
            FOR i := 1 TO numsp DO
               FOR j := 1 TO numsp DO
                  IF (i <> j) AND (treenode[i]^.d[j] > 0.0)
                  THEN totalnum := totalnum + treenode[i]^.w[j]
                           * SQR(treenode[i]^.d[j]);
            totalnum := totalnum - 2;
            IF replicates AND (lower OR upper)
            THEN totalnum := totalnum / 2;
            WRITE(OUTPUT, ' Average percent standard deviation =');
            WRITELN(OUTPUT,
               100*SQRT(-like/totalnum):10:5);
            WRITELN(OUTPUT);
            END;
         IF NOT usertree
         THEN BEGIN
            WRITELN(OUTPUT, ' examined ',examined:4, ' trees');
            WRITELN(OUTPUT);
            END;
         WRITELN(OUTPUT, ' from    to           length          time');
         WRITELN(OUTPUT, ' ----    --           ------          ----');
         WRITELN(OUTPUT);
         dtraverse (root);
         WRITELN(OUTPUT);
      END; (* describe *)

      PROCEDURE treeout (p : nodeptr);
         (* write out file with representation of final tree *)
         VAR i, n, w : INTEGER;
            c : CHAR;
            x : REAL;
      BEGIN (* treeout *)
         IF p^.tip
         THEN BEGIN
            n := 0;
            FOR i := 1 TO namelength DO
               IF names[p^.index][i] <> ' '
               THEN n := i;
            FOR i := 1 TO n DO BEGIN
               c := names[p^.index][i];
               IF c = ' '
               THEN c := '_';
               WRITE(treefile, c);
               END;
            END
         ELSE BEGIN
            WRITE(treefile, '(');
            treeout (p^.leftdesc);
            WRITE(treefile, ',');
            treeout (p^.rtdesc);
            WRITE(treefile, ')');
            END;
         IF p <> root
         THEN x := p^.ancestor^.t-p^.t;
         IF x > 0.0
         THEN w := TRUNC(0.4343*LN(x))
         ELSE IF x = 0.0
            THEN w := 0
            ELSE w := TRUNC(0.4343*LN(-x)) + 1;
         IF p = root
         THEN WRITELN(treefile, ';')
         ELSE WRITE(treefile, ':', x:w+7:5);
      END; (* treeout *)

   BEGIN (* maketree *)
      IF NOT usertree
      THEN BEGIN
         examined := 0;
         FOR i := 1 TO numsp DO enterorder[i] := i;
         IF jumble
         THEN FOR i := 1 TO numsp DO BEGIN
            j := TRUNC(randum(seed)*numsp) + 1;
            k := enterorder[j];
            enterorder[j] := enterorder[i];
            enterorder[i] := k;
            END;
         add (treenode[enterorder[2]],
            treenode[enterorder[1]], treenode[numsp+1]);
         root := treenode[numsp+1];
         WRITELN(OUTPUT, ' Adding species:');
         WRITE(OUTPUT, '    ');
         FOR i := 1 TO namelength DO
            WRITE(OUTPUT, names[enterorder[1]][i]);
         WRITELN(OUTPUT);
         WRITE(OUTPUT, '    ');
         FOR i := 1 TO namelength DO
            WRITE(OUTPUT, names[enterorder[2]][i]);
         WRITELN(OUTPUT);
         lastrearr := FALSE;
         FOR i := 3 TO numsp DO BEGIN
            bestyet := -10000000.0;
            there := root;
            item := treenode[enterorder[i]];
            nufork := treenode[numsp + i - 1];
            addpreorder (root, item, nufork);
            add (there, item, nufork);
            IF root^.ancestor <> NIL THEN root := root^.ancestor;
            like := bestyet;
            rearrange (root);
            evaluate (root);
            examined := examined - 1;
            WRITE(OUTPUT, '    ');
            FOR j := 1 TO namelength DO
               WRITE(OUTPUT, names[enterorder[i]][j]);
            WRITELN(OUTPUT);
            lastrearr := (i = numsp);
            IF lastrearr
            THEN BEGIN
               WRITELN(OUTPUT, ' Doing global rearrangements');
               REPEAT
                  gotlike := bestyet;
                  WRITE(OUTPUT, '    ');
                  FOR j := 1 TO numsp2 DO BEGIN
                     bestyet := -32000;
                     there := root;
                     item := treenode[j];
                     IF item <> root
                     THEN BEGIN
                        nufork := treenode[j]^.ancestor;
                        IF root = nufork
                        THEN BEGIN
                           IF nufork^.leftdesc = item
                           THEN root := nufork^.rtdesc
                           ELSE root := nufork^.leftdesc;
                           END;
                        remove (item, nufork);
                        addpreorder (root, item, nufork);
                        add (there, item, nufork);
                        IF root^.ancestor <> NIL
                        THEN root := root^.ancestor;
                        like := bestyet;
                        END;
                     WRITE(OUTPUT, '.');
                     END;
               UNTIL like <= gotlike;
               END;
            IF i = numsp
            THEN BEGIN
               evaluate (root);
               printree;
               describe;
               IF trout AND (i = numsp)
               THEN BEGIN
                  treeout (root);
                  WRITELN(OUTPUT, ' Tree also written onto file');
                  WRITELN(OUTPUT);
                  END;
               END;
            END;
         END
      ELSE BEGIN
         READLN(infile, numtrees);
         WRITELN(OUTPUT);
         WRITELN(OUTPUT);
         WRITELN(OUTPUT, ' User-defined trees:');
         FOR i := 1 TO numtrees DO BEGIN
            treeread;
            evaluate (root);
            printree;
            describe;
            END;
         END;
   END; (* maketree *)

BEGIN (* Fitch-Margoliash criterion with contemporary tips *)
   (* reads in numsp, options, and the data, then calls maketree to
      construct the tree *)
   (* ASSIGN(infile, 'INFILE');  -- Turbo Pascal only *)
   RESET(infile);
   getinput;
   IF trout
   THEN BEGIN
      (* ASSIGN(treefile, 'TREEFILE');  -- Turbo Pascal only *)
      REWRITE(treefile);
      END;
   maketree;
(* CLOSE(treefile);  -- may need this here *)
   IF trout
   THEN BEGIN
      (* CLOSE(infile);   -- may need this here *)
      END;
END. (* Fitch-Margoliash criterion with contemporary tips *)

