PROGRAM Mix (infile, treefile, OUTPUT);
  (* version 3.1. (c) Copyright 1986, 1987, 1988 by the University of
     Washington and Joseph Felsenstein.  Written by Joseph Felsenstein.
     Permission is granted to copy and use this program provided no
     fee is charged for it and provided that this copyright
     notice is not removed. *)

CONST maxsp = 50;    (* maximum number of species               *)
   maxsz = 100;      (* size of pointer array.  >= 2*maxsp - 1  *)
   maxchr = 200;     (* max. number characterstates per species *)
   bits = 30;        (* max. number of elements in a Pascal set *)
   wrds = 7;         (* wrds * bits >= maxchr                   *)
   nmlngth = 10;     (* number of characters in species name    *)
   maxtrees = 100;   (* maximum number of tied trees stored     *)
   maxuser = 10;     (* maximum number of user-defined trees    *)

TYPE steparray = ARRAY[1..maxchr] OF INTEGER;
   bitrange = 1..bits;
   bitset = SET OF bitrange;
   bitarray = ARRAY[1..wrds] OF bitset;
   nodeptr = ^node;   (* nodes will form a binary tree *)
   node = RECORD      (* describes a tip species or an ancestor *)
      ancestor, leftdesc, rtdesc : nodeptr;  (* pointers to nodes *)
      index : INTEGER;      (* number of the node *)
      tip : BOOLEAN;        (* present species are tips of tree *)
      stateone, statezero : bitarray; (* see in PROCEDURE fillin *)
      xcoord, ycoord : INTEGER    (* used by printree *)
      END;
   pointarray = ARRAY[1..maxsz] OF nodeptr;
   longer = ARRAY[0..5] OF INTEGER;

VAR root : nodeptr;
   infile, treefile : TEXT;
   spp, nonodes, chars, words, inseed, outgrno : INTEGER;
     (* spp = number of species
        nonodes = number of nodes in tree
        chars = number of binary characters
        words = number of words needed
                to represent characters of one organism
        outgrno indicates outgroup *)
   jumble, usertree, weights, thresh, ancvar, questions, allsokal,
      allwagner, mixture, trout, noroot, outgropt, didreroot : BOOLEAN;
   extras, weight : steparray;
   ancone, anczero : ARRAY[1..maxchr] OF BOOLEAN;
   treenode : pointarray;      (* pointers to all nodes in tree *)
   name : ARRAY[1..maxsp, 1..nmlngth] OF CHAR; (* names of species *)
   threshold : REAL;
   threshwt : ARRAY[1..maxchr] OF REAL;
   wagner : bitarray;
   seed : longer;
   enterorder : ARRAY[1..maxsp] OF INTEGER;

   FUNCTION randum (VAR seed : longer) : REAL;
      (* random number generator -- slow but machine independent *)
      VAR i, j, k, sum : INTEGER;
         mult, newseed : longer;
         x : REAL;
   BEGIN (* randum *)
      mult[0] := 13;
      mult[1] := 24;
      mult[2] := 22;
      mult[3] :=  6;
      FOR i := 0 TO 5 DO
         newseed[i] := 0;
      FOR i := 0 TO 5 DO BEGIN
         sum := newseed[i];
         k := i;
         IF i > 3 THEN k := 3;
         FOR j := 0 TO k DO
            sum := sum + mult[j] * seed[i-j];
         newseed[i] := sum;
         FOR j := i TO 4 DO BEGIN
            newseed[j+1] := newseed[j+1] + newseed[j] DIV 64;
            newseed[j] := newseed[j] MOD 64;
            END;
         END;
      seed := newseed;
      seed[5] := seed[5] MOD 4;
      x := 0.0;
      FOR i := 0 TO 5 DO
         x := x / 64.0 + seed[i];
      x := x / 4.0;
      randum := x;
   END; (* randum *)

   FUNCTION digit (ch : CHAR) : BOOLEAN;
   BEGIN (* digit *)
      digit := (ch >= '0') AND (ch <= '9');
   END; (* digit *)

   FUNCTION letter (ch : CHAR) : BOOLEAN;
      (* tests whether ch is a letter
         --  works for both ASCII and EBCDIC codes *)
   BEGIN (* letter *)
      letter := ((ch >= 'A') AND (ch <= 'I'))
                OR ((ch >= 'J') AND (ch <= 'R'))
                OR ((ch >= 'S') AND (ch <= 'Z'))
                OR ((ch >= 'a') AND (ch <= 'i'))
                OR ((ch >= 'j') AND (ch <= 'r'))
                OR ((ch >= 's') AND (ch <= 'z'));
   END; (* letter *)

   PROCEDURE uppercase (VAR ch : CHAR);
      (* convert a character to upper case -- either ASCII or EBCDIC *)
   BEGIN (* uppercase *)
      IF letter(ch) AND (('a' > 'A') AND (ch >= 'a')
                     OR  ('a' < 'A') AND (ch < 'A'))
      THEN ch := CHR(ORD(ch) + ORD('A') - ORD('a'));
   END; (* uppercase *)

   PROCEDURE newline (i, j, k : INTEGER);
      (* go to new line if i is a multiple of j, indent k spaces *)
      VAR m : INTEGER;
   BEGIN (* newline *)
      IF (((i-1) MOD j) = 0) AND (i>1)
      THEN BEGIN
         WRITELN(OUTPUT);
         FOR m := 1 TO k DO WRITE(OUTPUT, ' ');
         END;
   END; (* newline *)

   PROCEDURE doinput;
      (* reads the input data *)

      PROCEDURE inputnumbers;
         (* input the numbers of species and of characters *)
      BEGIN (* inputnumbers *)
         WRITELN(OUTPUT);
         WRITELN(OUTPUT, ' Mixed parsimony algorithm, version 3.1');
         WRITELN(OUTPUT);
         READ(infile, spp, chars);
         WRITELN(OUTPUT, ' ',spp:2,' species, ',chars:3,'  characters');
         IF spp > maxsp
         THEN WRITELN(OUTPUT, ' TOO MANY SPECIES: adjust CONSTants');
         IF chars > maxchr
         THEN WRITELN(OUTPUT, ' TOO MANY CHARACTERS: adjust CONSTants');
         WRITELN(OUTPUT);
         words := (chars DIV bits) + 1;
         nonodes := 2 * spp - 1;
      END; (* inputnumbers *)

      PROCEDURE inputoptions;
         (* input the information on the options *)
         VAR ch : CHAR;
            extranum, i : INTEGER;

         PROCEDURE inputmixture;
            (* input mixture of methods *)
            VAR i, j, k : INTEGER;
               ch : CHAR;
               wag : BOOLEAN;
         BEGIN (* inputmixture *)
            FOR i := 1 TO nmlngth-1 DO READ(infile, ch);
            FOR i := 1 TO words DO wagner[i] := [];
            j := 0;
            k := 1;
            FOR i := 1 TO chars DO BEGIN
               REPEAT
                  IF EOLN(infile) THEN READLN(infile);
                  READ(infile, ch);
               UNTIL ch <> ' ';
               uppercase (ch);
               wag := FALSE;
               IF (ch='W') OR (ch = '?')
               THEN wag := TRUE
               ELSE IF (ch='S') OR (ch='C')
                  THEN wag := FALSE
                  ELSE WRITELN(OUTPUT, ' error: ', ch);
               j := j + 1;
               IF j > bits
               THEN BEGIN
                  j := 1;
                  k := k + 1;
                  END;
               IF wag
               THEN wagner[k] := wagner[k] + [j];
               IF NOT ancvar
               THEN ancone[i] := wag;
               END;
            READLN(infile);
            allwagner := FALSE;
            allsokal := FALSE;
            mixture := TRUE;
         END; (* inputmixture *)

         PROCEDURE printmixture;
            (* print out list of parsimony methods *)
            VAR i, k, l : INTEGER;
         BEGIN (* printmixture *)
            allwagner := TRUE;
            WRITELN(OUTPUT, ' Parsimony methods:');
            l := 0;
            k := 1;
            FOR i := 1 TO nmlngth+4 DO WRITE(OUTPUT, ' ');
            FOR i := 1 TO chars DO BEGIN
               newline (i, 50, nmlngth+4);
               l := l + 1;
               IF l > bits
               THEN BEGIN
                  l := 1;
                  k := k + 1;
                  END;
               IF l IN wagner[k]
               THEN WRITE(OUTPUT, 'W')
               ELSE BEGIN
                  WRITE(OUTPUT, 'S');
                  allwagner := allwagner AND (weight[i] = 0);
                  END;
               IF ((i MOD 5) = 0) THEN WRITE(OUTPUT, ' ');
               END;
            WRITELN(OUTPUT);
            WRITELN(OUTPUT);
         END; (* printmixture *)

         PROCEDURE inputhresh;
            (* input the threshold *)
         BEGIN (* inputhresh *)
            READLN(infile, threshold);
            thresh := TRUE;
         END; (* inputhresh *)

         PROCEDURE printhresh;
            (* print out the threshold value *)
         BEGIN (* printhresh *)
            WRITELN(OUTPUT, '    Threshold value = ', threshold:10:3);
            WRITELN(OUTPUT);
         END; (* printhresh *)

         PROCEDURE inputweights;
            (* input the character weights, 0-9 and A-Z for weights
               0 - 35 *)
            VAR ch : CHAR;
               i : INTEGER;
         BEGIN (* inputweights *)
            FOR i := 1 TO nmlngth-1 DO READ(infile, ch);
            FOR i := 1 TO chars DO BEGIN
               REPEAT
                  IF EOLN(infile) THEN READLN(infile);
                  READ(infile, ch);
               UNTIL ch <> ' ';
               weight[i] := 1;
               IF digit(ch)
               THEN weight[i] := ORD(ch) - ORD('0')
               ELSE IF letter(ch)
                  THEN BEGIN
                     uppercase (ch);
                     IF (ch >= 'A') AND (ch <= 'I')
                     THEN weight[i] := ORD(ch) - ORD('A') + 10
                     ELSE IF (ch >= 'J') AND (ch <= 'R')
                        THEN weight[i] := ORD(ch) - ORD('J') + 19
                        ELSE weight[i] := ORD(ch) - ORD('S') + 28;
                     END
                  ELSE WRITELN(OUTPUT, ' Bad weight character: ',ch);
               END;
            READLN(infile);
            weights := TRUE;
         END; (* inputweights *)

         PROCEDURE printweights;
            (* print out the weights of characters *)
            VAR i, j, k : INTEGER;
         BEGIN (* printweights *)
            WRITELN(OUTPUT,'    Characters are weighted as follows:');
            WRITE(OUTPUT, '        ');
            FOR i := 0 TO 9 DO WRITE(OUTPUT, i:3);
            WRITELN(OUTPUT);
            WRITELN(OUTPUT,'     *---------------------------------');
            FOR j := 0 TO (chars DIV 10) DO BEGIN
               WRITE(OUTPUT, 10*j:5, '!  ');
               FOR i := 0 TO 9 DO BEGIN
                  k := 10*j + i;
                  IF (k>0) AND (k<=chars)
                  THEN WRITE(OUTPUT, weight[k]:3)
                  ELSE WRITE(OUTPUT, '   ');
                  END;
               WRITELN(OUTPUT);
               END;
            WRITELN(OUTPUT);
         END; (* printweights *)

         PROCEDURE inputancestors;
            (* reads the ancestral states for each character *)
            VAR i : INTEGER;
               ch : CHAR;
         BEGIN (* inputancestors *)
            FOR i := 1 TO nmlngth-1 DO READ(infile, ch);
            ancvar := TRUE;
            FOR i := 1 TO chars DO BEGIN
               REPEAT
                  IF EOLN(infile) THEN READLN(infile);
                  READ(infile, ch);
               UNTIL ch <> ' ';
               IF ch = 'p' THEN ch := 'P';
               IF ch = 'b' THEN ch := 'B';
               CASE ch OF
                  '1' : anczero[i] := FALSE;
                  '0' : ancone[i] := FALSE;
                  'P' : BEGIN END;
                  'B' : BEGIN END;
                  '?' : BEGIN END
                  END;
               END;
            READLN(infile);
         END; (* inputancestors *)

         PROCEDURE printancestors;
            (* print out list of ancestral states *)
            VAR i : INTEGER;
         BEGIN (* printancestors *)
            WRITELN(OUTPUT, ' Ancestral states:');
            FOR i := 1 TO nmlngth+4 DO WRITE(OUTPUT, ' ');
            FOR i := 1 TO chars DO BEGIN
               newline (i, 50, nmlngth+4);
               IF ancone[i] AND anczero[i]
               THEN WRITE(OUTPUT, '?')
               ELSE IF ancone[i]
                  THEN WRITE(OUTPUT, '1')
                  ELSE WRITE(OUTPUT, '0');
               IF ((i MOD 5) = 0) THEN WRITE(OUTPUT, ' ');
               END;
            WRITELN(OUTPUT);
            WRITELN(OUTPUT);
         END; (* printancestor *)

      BEGIN (* inputoptions *)
         ancvar := FALSE;
         FOR i := 1 TO chars DO BEGIN
            anczero[i] := TRUE;
            ancone[i] := TRUE;
            END;
         allwagner := TRUE;
         allsokal := FALSE;
         jumble := FALSE;
         mixture := FALSE;
         outgrno := 1;
         outgropt := FALSE;
         thresh := FALSE;
         trout := FALSE;
         usertree := FALSE;
         weights := FALSE;
         extranum := 0;
         WHILE NOT EOLN(infile) DO BEGIN
            READ(infile, ch);
            uppercase (ch);
            IF (ch = 'A') OR (ch = 'C') OR (ch = 'J') 
               OR (ch = 'M') OR (ch = 'O') OR (ch = 'S')
               OR (ch = 'T') OR (ch = 'U') OR (ch = 'W') OR (ch = 'Y')
            THEN CASE ch OF
                  'A' : extranum := extranum + 1;
                  'C' : allsokal := TRUE;
                  'J' : BEGIN
                        jumble := TRUE;
                        extranum := extranum + 1;
                        END;
                  'M' : extranum := extranum + 1;
                  'O' : BEGIN
                        outgropt := TRUE;
                        extranum := extranum + 1;
                        END;
                  'S' : allsokal := TRUE;
                  'T' : extranum := extranum + 1;
                  'U' : usertree := TRUE;
                  'W' : extranum := extranum + 1;
                  'Y' : trout := TRUE
                  END
            ELSE IF ch <> ' '
               THEN WRITELN(OUTPUT, ' Bad Option Character: ', ch);
            END;
         READLN(infile);
         IF allsokal THEN allwagner := FALSE;
         FOR i := 1 TO words DO
            IF allsokal
            THEN wagner[i] := []
            ELSE wagner[i] := [1..bits];
         threshold := spp;
         FOR i := 1 TO chars DO weight[i] := 1;
         FOR i := 1 TO extranum DO BEGIN
            READ(infile, ch);
            uppercase (ch);
            IF ch = 'A'
            THEN inputancestors;
            IF ch = 'J'
            THEN READLN(infile, inseed);
            IF ch = 'M'
            THEN inputmixture;
            IF ch = 'O'
            THEN READLN(infile, outgrno);
            IF ch = 'T'
            THEN inputhresh;
            IF ch = 'W'
            THEN inputweights;
            END;
         IF allsokal
         THEN BEGIN
            IF NOT ancvar
            THEN FOR i := 1 TO chars DO ancone[i] := FALSE;
            WRITELN(OUTPUT, ' Camin-Sokal parsimony method');
            END;
         IF allwagner
         THEN WRITELN(OUTPUT, ' Wagner parsimony method');
         WRITELN(OUTPUT);
         IF mixture
         THEN printmixture;
         IF jumble
         THEN BEGIN
            WRITELN(OUTPUT, ' Random number seed = ', inseed:8);
            WRITELN(OUTPUT);
            FOR i := 0 TO 5 DO seed[i] := 0;
            i := 0;
            REPEAT
               seed[i] := inseed MOD 64;
               inseed := inseed DIV 64;
               i := i + 1;
            UNTIL inseed = 0;
            END;
         IF thresh
         THEN printhresh;
         IF weights
         THEN printweights;
         IF ancvar
         THEN printancestors;
         noroot := allwagner OR (threshold <= 2.0);
         questions := FALSE;
         FOR i := 1 TO chars DO BEGIN
            noroot := noroot AND ancone[i] AND anczero[i];
            questions := questions OR (ancone[i] AND anczero[i]);
            threshwt[i] := threshold * weight[i];
            END;
      END; (* inputoptions *)

      PROCEDURE inputdata;
         (* input the names and character state data for species *)
         VAR i, j, l : INTEGER;
            k : bitrange;
            charstate : CHAR;  (* possible states are
                                  '0', '1', 'P', 'B', and '?' *)
      BEGIN (* inputdata *)
         WRITELN(OUTPUT);
         j := nmlngth - 5 + chars DIV 2;
         IF j < nmlngth - 1 THEN j := nmlngth - 1;
         IF j > 39 THEN j := 39;
         WRITE(OUTPUT, ' Name');
         FOR i := 1 TO j DO WRITE(OUTPUT, ' ');
         WRITELN(OUTPUT, 'Characters');
         WRITE(OUTPUT, ' ----');
         FOR i := 1 TO j DO WRITE(OUTPUT, ' ');
         WRITELN(OUTPUT, '----------');
         WRITELN(OUTPUT);
         FOR i := 1 TO chars DO extras[i] := 0;
         FOR i := 1 TO nonodes DO BEGIN
            NEW (treenode[i]);
            treenode[i]^.leftdesc := NIL;
            treenode[i]^.rtdesc := NIL;
            treenode[i]^.ancestor := NIL;
            treenode[i]^.index := i;
            IF i <= spp
            THEN BEGIN
               treenode[i]^.tip := TRUE;
               FOR j := 1 TO nmlngth DO READ(infile, name[i,j]);
               WRITE(OUTPUT, ' ');
               FOR j := 1 TO nmlngth DO WRITE(OUTPUT, name[i,j]);
               WRITE(OUTPUT, '   ');
               FOR j := 1 TO words DO BEGIN
                  treenode[i]^.stateone[j] := [];
                  treenode[i]^.statezero[j] := [];
                  END;
               FOR j := 1 TO chars DO BEGIN
                  k := ((j-1) MOD bits) + 1;
                  l := ((j-1) DIV bits) + 1;
                  REPEAT
                     IF EOLN(infile) THEN READLN(infile);
                     READ(infile, charstate);
                  UNTIL charstate <> ' ';
                  IF charstate = 'b' THEN charstate := 'B';
                  IF charstate = 'p' THEN charstate := 'P';
                  IF NOT ((charstate = '0')
                         OR (charstate = '1')
                         OR (charstate = '?')
                         OR (charstate = 'P')
                         OR (charstate = 'B'))
                  THEN WRITELN(OUTPUT, ' Warning -- ',
                      'bad character state: ', charstate,
                      ' at character ', j:5, ' of species ', i:3);
                  newline (j, 60, nmlngth+4);
                  WRITE(OUTPUT, charstate);
                  IF (j MOD 5) = 0 THEN WRITE(OUTPUT, ' ');
                  IF (charstate = '1')
                  THEN treenode[i]^.stateone[l]
                             := treenode[i]^.stateone[l] + [k];
                  IF (charstate = '0')
                  THEN treenode[i]^.statezero[l]
                             := treenode[i]^.statezero[l] + [k];
                  IF (charstate = 'P') OR (charstate = 'B')
                  THEN extras[j] := extras[j] + weight[j];
                  END;
               READLN(infile);
               WRITELN(OUTPUT);
               END
            ELSE treenode[i]^.tip := FALSE;
            END;
         WRITELN(OUTPUT);
         WRITELN(OUTPUT);
      END; (* inputdata *)

   BEGIN (* doinput *)
      inputnumbers;
      inputoptions;
      inputdata;
   END; (* doinput *)

   PROCEDURE maketree;
      (* constructs a binary tree from the pointers in treenode.
         adds each node at location which yields highest "likelihood"
         then rearranges the tree for greatest "likelihood" *)
      TYPE placearray = ARRAY[1..maxsz] OF INTEGER;
      VAR i, j, k, nextree, numtrees, which, minwhich : INTEGER;
         like, bestyet, bestlike, gotlike,
         minsteps, sum, sum2, sd : REAL;
         lastrearr : BOOLEAN;
         nsteps : ARRAY [1..maxuser] OF REAL;
         fsteps : ARRAY [1..maxuser] OF ARRAY [1..maxchr] OF REAL;
         guess : ARRAY[1..maxchr] OF CHAR;
         there, item, nufork, dummy : nodeptr;
         numsteps : steparray;
         fullset : bitset;
         steps, zeroanc, oneanc : bitarray;
         place : placearray;
         bestrees : ARRAY[1..maxtrees] OF placearray;

      PROCEDURE add (below, newtip, newfork : nodeptr);
         (* inserts the nodes newfork and its left descendant, newtip,
            to the tree.  below becomes newfork's right descendant *)
      BEGIN (* add *)
         newfork^.ancestor := below^.ancestor;
         newfork^.rtdesc := below;
         newfork^.leftdesc := newtip;
         below^.ancestor := newfork;
         newtip^.ancestor := newfork;
         IF newfork^.ancestor <> NIL
         THEN IF newfork^.ancestor^.leftdesc = below
            THEN newfork^.ancestor^.leftdesc := newfork
            ELSE newfork^.ancestor^.rtdesc := newfork;
      END; (* add *)

      PROCEDURE remove (VAR item, fork : nodeptr);
         (* removes nodes item and its ancestor, fork, from the tree.
            the new descendant of fork's ancestor is made to be
            fork's second descendant (other than item).  Also
            returns pointers to the deleted nodes, item and fork *)
         VAR other, below : nodeptr;
      BEGIN (* remove *)
         IF item^.ancestor = NIL
         THEN fork := NIL
         ELSE BEGIN
            fork := item^.ancestor;
            IF item = fork^.leftdesc
            THEN other := fork^.rtdesc
            ELSE other := fork^.leftdesc;
            below := fork^.ancestor;
            other^.ancestor := below;
            IF below <> NIL
            THEN BEGIN
               IF below^.leftdesc = fork
               THEN below^.leftdesc := other
               ELSE below^.rtdesc := other;
               fork^.ancestor := NIL;
               END;
            fork^.leftdesc := NIL;
            fork^.rtdesc := NIL;
            item^.ancestor := NIL;
            END;
      END; (* remove *)

      PROCEDURE fillin (p: nodeptr);
         (* Sets up for each node in the tree two statesets.
            stateone and statezero are the sets of character
            states that must be 1 or must be 0, respectively,
            in a most parsimonious reconstruction, based on the
            information at or above this node.  Note that this
            state assignment may change based on information further
            down the tree.  If a character is in both sets it is in
            state "P".  If in neither, it is "?". *)
         VAR i : INTEGER;
            l0, l1, r0, r1, st, wa, za : bitset;
      BEGIN (* fillin *)
         FOR i := 1 TO words DO BEGIN
            l0 := p^.leftdesc^.statezero[i];
            l1 := p^.leftdesc^.stateone[i];
            r0 := p^.rtdesc^.statezero[i];
            r1 := p^.rtdesc^.stateone[i];
            wa := wagner[i];
            za := zeroanc[i];
            st := l1 * r0 + l0 * r1;
            steps[i] := st;
            p^.stateone[i] := l1 + r1 - (st * (wa + za));
            p^.statezero[i] := l0 + r0 - (st * (wa + (fullset - za)));
            END;
      END; (* fillin *)

      PROCEDURE evaluate (r: nodeptr);
         (* Determines the number of steps needed for a tree.
            This is the minimum number needed to evolve chars on
            this tree *)
         VAR i, stepnum, smaller : INTEGER;
            sum, term : REAL;
            numsone, numszero : steparray;

         PROCEDURE count (VAR stps : bitarray);
            (* counts the number of steps in a fork of the tree.
               The program spends much of its time in this PROCEDURE *)
            VAR i, j, l : INTEGER;
         BEGIN (* count *)
            j := 1;
            l := 0;
            FOR i := 1 TO chars DO BEGIN
               l := l + 1;
               IF l > bits
               THEN BEGIN
                  l := 1;
                  j := j + 1;
                  END;
               IF l IN stps[j]
               THEN BEGIN
                  IF l IN zeroanc[j]
                  THEN numszero[i] := numszero[i] + weight[i]
                  ELSE numsone[i] := numsone[i] + weight[i];
                  END;
               END;
         END; (* count *)

         PROCEDURE postorder (p : nodeptr);
            (* traverses a binary tree, calling PROCEDURE fillin at a
               node's descendants before calling fillin at the node *)
         BEGIN (* postorder *)
            IF NOT p^.tip
            THEN BEGIN
               postorder (p^.leftdesc);
               postorder (p^.rtdesc);
               fillin (p);
               count (steps);
               END;
         END; (* postorder *)

      BEGIN (* evaluate *)
         sum := 0;
         FOR i := 1 TO chars DO BEGIN
            numszero[i] := 0;
            numsone[i] := 0;
            END;
         FOR i := 1 TO words DO
            zeroanc[i] := fullset;
         postorder (r);
         count (r^.stateone);
         FOR i := 1 TO words DO
            zeroanc[i] := [];
         postorder (r);
         count (r^.statezero);
         FOR i := 1 TO chars DO BEGIN
            smaller := spp * weight[i];
            IF anczero[i]
            THEN BEGIN
               numsteps[i] := numszero[i];
               smaller := numszero[i];
               END;
            IF ancone[i] AND (numsone[i] < smaller)
            THEN numsteps[i] := numsone[i];
            stepnum := numsteps[i] + extras[i];
            IF stepnum <= threshwt[i]
            THEN term := stepnum
            ELSE term := threshwt[i];
            sum := sum + term;
            IF usertree
            THEN fsteps[which][i] := term;
            guess[i] := '?';
            IF (NOT ancone[i]) OR
               (anczero[i] AND (numszero[i] < numsone[i]))
            THEN guess[i] := '0'
            ELSE IF (NOT anczero[i]) OR
                  (ancone[i] AND (numsone[i] < numszero[i]))
               THEN guess[i] := '1';
            END;
         IF usertree
         THEN BEGIN
            nsteps[which] := sum;
            IF which = 1
            THEN BEGIN
               minwhich := 1;
               minsteps := sum;
               END
            ELSE IF sum < minsteps
               THEN BEGIN
                  minwhich := which;
                  minsteps := sum;
                  END;
            END;
         like := -sum;
      END; (* evaluate *)

      PROCEDURE reroot (outgroup : nodeptr);
         (* reorients tree, putting outgroup in desired position.
            Written by Kent Fiala *)
         VAR vertex, oldvertex, ante, otherbranch : nodeptr;
            left : BOOLEAN;
      BEGIN (* reroot *)
         IF NOT (outgroup^.ancestor = root)
         THEN BEGIN
            oldvertex := outgroup;
            vertex := outgroup^.ancestor;
            ante := vertex^.ancestor;
            IF vertex^.rtdesc = outgroup
            THEN BEGIN
               vertex^.rtdesc := vertex^.leftdesc;
               vertex^.leftdesc := outgroup;
               END;
            REPEAT
               IF oldvertex = vertex^.leftdesc
               THEN BEGIN
                  vertex^.leftdesc := ante;
                  left := TRUE;
                  END
               ELSE BEGIN
                  vertex^.rtdesc := ante;
                  left := FALSE;
                  END;
               vertex^.ancestor := oldvertex;
               oldvertex := vertex;
               vertex := ante;
               ante := vertex^.ancestor
            UNTIL vertex = root;
            IF root^.leftdesc = oldvertex
            THEN otherbranch := root^.rtdesc
            ELSE otherbranch := root^.leftdesc;
            IF left
            THEN oldvertex^.leftdesc := otherbranch
            ELSE oldvertex^.rtdesc := otherbranch;
            otherbranch^.ancestor := oldvertex;
            root^.rtdesc := outgroup^.ancestor;
            root^.leftdesc := outgroup;
            outgroup^.ancestor^.ancestor := root;
            outgroup^.ancestor := root;
            END;
      END; (* reroot *)

      PROCEDURE addpreorder (p, item, nufork : nodeptr);
         (* traverses a binary tree, calling PROCEDURE tryadd
            at a node before calling tryadd at its descendants *)
         TYPE knowed = RECORD    (* temporary storage for tree shape *)
            ancestor, leftdesc, rtdesc : nodeptr;
            END;
         VAR treeknowed : ARRAY [1..maxsz] OF knowed;

         PROCEDURE tryadd (p: nodeptr);
            (* temporarily adds one fork and one tip to the tree.
               if the location where they are added yields greater
               "likelihood" than other locations tested up to that
               time, then keeps that location as there *)
            VAR i, pos : INTEGER;
               found : BOOLEAN;
               rute : nodeptr;

            PROCEDURE savetree;
               (* record in place where each species has to be
                  added to reconstruct this tree *)
               VAR i, j : INTEGER;
                  p : nodeptr;
                  done : BOOLEAN;
            BEGIN (* savetree *)
               IF noroot
               THEN reroot (treenode[outgrno]);
               FOR i := 1 TO nonodes DO place[i] := 0;
               place[root^.index] := 1;
               FOR i := 1 TO spp DO BEGIN
                  p := treenode[i];
                  WHILE (place[p^.index] = 0) DO BEGIN
                     place[p^.index] := i;
                     p := p^.ancestor;
                     END;
                  IF i > 1
                  THEN BEGIN
                     place[i] := place[p^.index];
                     j := place[p^.index];
                     done := FALSE;
                     WHILE NOT done DO BEGIN
                        place[p^.index] := spp + i - 1;
                        p := p^.ancestor;
                        done := (p = NIL);
                        IF NOT done
                        THEN done := (place[p^.index] <> j);
                        END;
                     END;
                  END;
            END; (* savetree *)

            PROCEDURE findtree;
               (* finds tree given by ARRAY place in ARRAY
                  bestrees by binary search *)
               VAR i, lower, upper : INTEGER;
                  below : BOOLEAN;
            BEGIN (* findtree *)
               below := FALSE;
               lower := 1;
               upper := nextree - 1;
               found := FALSE;
               WHILE (NOT found) AND (lower <= upper) DO BEGIN
                  pos := (lower + upper) DIV 2;
                  i := 3;
                  WHILE (place[i] = bestrees[pos][i]) AND (i <= spp) DO
                     i := i + 1;
                  found := (i > spp);
                  below := place[i] < bestrees[pos][i];
                  IF NOT found
                  THEN BEGIN
                     IF below
                     THEN upper := pos - 1
                     ELSE lower := pos + 1;
                     END;
                  END;
               IF (NOT found) AND (NOT below)
               THEN pos := pos + 1;
            END; (* findtree *)

            PROCEDURE addtree;
               (* puts tree from ARRAY place in its proper position
                  in ARRAY bestrees *)
               VAR i : INTEGER;
            BEGIN (* addtree *)
               FOR i := (nextree - 1) DOWNTO pos DO
                  bestrees[i+1] := bestrees[i];
               bestrees[pos] := place;
               nextree := nextree + 1;
            END; (* addtree *)

         BEGIN (* tryadd *)
            add (p, item, nufork);
            IF root^.ancestor <> NIL THEN root := root^.ancestor;
            evaluate (root);
            IF lastrearr
            THEN BEGIN
               IF like >= bestlike
               THEN BEGIN
                  FOR i := 1 TO nonodes DO BEGIN
                     treeknowed[i].ancestor := treenode[i]^.ancestor;
                     treeknowed[i].leftdesc := treenode[i]^.leftdesc;;
                     treeknowed[i].rtdesc := treenode[i]^.rtdesc;
                     END;
                  rute := root;
                  savetree;
                  FOR i := 1 TO nonodes DO BEGIN
                     treenode[i]^.ancestor := treeknowed[i].ancestor;
                     treenode[i]^.leftdesc := treeknowed[i].leftdesc;;
                     treenode[i]^.rtdesc := treeknowed[i].rtdesc;
                     END;
                  root := rute;
                  IF like > bestlike
                  THEN BEGIN
                     bestlike := like;
                     pos := 1;
                     nextree := 1;
                     addtree;
                     END
                  ELSE BEGIN
                     pos := 0;
                     findtree;
                     IF NOT found
                     THEN IF nextree <= maxtrees
                        THEN addtree;
                     END;
                  END;
               END;
            IF like > bestyet
            THEN BEGIN
               bestyet := like;
               there := p;
               END;
            remove (item, nufork);
            IF nufork = root THEN root := p;
         END; (* tryadd *)

      BEGIN (* addpreorder *)
         IF p <> NIL
         THEN BEGIN
            tryadd (p);
            addpreorder (p^.leftdesc, item, nufork);
            addpreorder (p^.rtdesc, item, nufork);
            END;
      END; (* addpreorder *)

      PROCEDURE rearrange (VAR r: nodeptr);
         (* traverses the tree (preorder), finding any local
            rearrangement which decreases the number of steps.
            if traversal succeeds in increasing the tree's
            "likelihood", PROCEDURE rearrange runs traversal again *)
         VAR success : BOOLEAN;

         PROCEDURE tryrearr (p : nodeptr);
            (* evaluates one rearrangement of the tree.
               if the new tree has greater "likelihood" than the old
               one sets success := TRUE and keeps the new tree.
               otherwise, restores the old tree *)
            VAR frombelow, whereto, forknode : nodeptr;
               oldlike : REAL;
         BEGIN (* tryrearr *)
            IF p^.ancestor <> NIL
            THEN BEGIN
               IF p^.ancestor^.ancestor <> NIL
               THEN BEGIN
                  oldlike := bestyet;
                  IF p^.ancestor^.leftdesc = p
                  THEN frombelow := p^.ancestor^.rtdesc
                  ELSE frombelow := p^.ancestor^.leftdesc;
                  whereto := p^.ancestor^.ancestor;
                  forknode := p^.ancestor;
                  remove (p, forknode);
                  add (whereto, p, forknode);
                  IF r^.ancestor <> NIL THEN r := r^.ancestor;
                  evaluate (r);
                  IF like <= oldlike
                  THEN BEGIN
                     remove (p, forknode);
                     add (frombelow, p, forknode);
                     IF r^.ancestor <> NIL THEN r := r^.ancestor;
                     END
                  ELSE BEGIN
                     success := TRUE;
                     bestyet := like;
                     END;
                  END;
               END;
         END; (* tryrearr *)

         PROCEDURE repreorder (p : nodeptr);
            (* traverses a binary tree, calling PROCEDURE tryrearr
               at a node before calling tryrearr at its descendants *)
         BEGIN (* repreorder *)
            IF p <> NIL
            THEN BEGIN
               tryrearr (p);
               repreorder (p^.leftdesc);
               repreorder (p^.rtdesc);
               END;
         END; (* repreorder *)

      BEGIN (* rearrange *)
         success := TRUE;
         WHILE success DO BEGIN
            success := FALSE;
            repreorder (r);
            END;
      END; (* rearrange *)

      PROCEDURE treeread;
         (* read in user-defined tree and set it up *)
         VAR ch : CHAR;
            nextnode : INTEGER;

         PROCEDURE findch (c : CHAR);
            (* scan forward until find character c *)
         BEGIN (* findch *)
            IF ch <> c
            THEN REPEAT
                  IF EOLN(infile) THEN READLN(infile);
                  READ(infile, ch);
               UNTIL ch = c;
         END; (* findch *)

         PROCEDURE addelement (VAR p : nodeptr);
            (* recursive procedure adds nodes to user-defined tree *)
            VAR q : nodeptr;
               i, n : INTEGER;
               found : BOOLEAN;
               str : ARRAY[1..nmlngth] OF CHAR;
         BEGIN (* addelement *)
            REPEAT
               IF EOLN(infile) THEN READLN(infile);
               READ(infile, ch);
            UNTIL ch <> ' ';
            IF ch = '('
            THEN BEGIN
               nextnode := nextnode + 1;
               q := treenode[nextnode];
               addelement (q^.leftdesc);
               q^.leftdesc^.ancestor := q;
               findch (',');
               addelement (q^.rtdesc);
               q^.rtdesc^.ancestor := q;
               findch (')');
               p := q;
               END
            ELSE BEGIN
               FOR i := 1 TO nmlngth DO str[i] := ' ';
               n := 1;
               REPEAT
                  IF (ch = '_') THEN ch := ' ';
                  str[n] := ch;
                  IF EOLN(infile) THEN READLN(infile);
                  READ(infile, ch);
                  n := n + 1;
               UNTIL ((ch = ',') OR (ch = ')') OR (n > nmlngth));
               n := 1;
               REPEAT
                  found := TRUE;
                  FOR i := 1 TO nmlngth DO
                     found := found AND (str[i] = name[n, i]);
                  IF found
                  THEN p := treenode[n]
                  ELSE n := n + 1;
               UNTIL (n > spp) OR found;
               IF n > spp
               THEN BEGIN
                  WRITE(OUTPUT, ' Cannot find species: ');
                  FOR i := 1 TO nmlngth DO
                     WRITE(OUTPUT, str[i]);
                  WRITELN(OUTPUT);
                  END;
               END;
         END; (* addelement *)

      BEGIN (* treeread *)
         root := treenode[spp + 1];
         nextnode := spp;
         root^.ancestor := NIL;
         addelement (root);
         READLN(infile);
      END; (* treeread *)

      PROCEDURE printree;
         (* prints out diagram of the tree *)
         CONST down = 2;
            over = 3;
         VAR tipx, tipy, i, rover : INTEGER;

         PROCEDURE coordinates (p : nodeptr);
            (* establishes coordinates of nodes *)
         BEGIN (* coordinates *)
            IF p^.tip
            THEN BEGIN
               p^.xcoord := tipx;
               p^.ycoord := tipy;
               tipx := tipx + rover;
               tipy := tipy + down;
               END
            ELSE BEGIN
               coordinates (p^.leftdesc);
               p^.xcoord := p^.leftdesc^.xcoord;
               coordinates (p^.rtdesc);
               p^.ycoord := p^.rtdesc^.ycoord;
               END;
         END; (* coordinates *)

         PROCEDURE drawline (i : INTEGER);
            (* draws one row of the tree diagram by moving up tree *)
            VAR p, q : nodeptr;
               n, j : INTEGER;
         BEGIN (* drawline *)
            p := root;
            WRITE(OUTPUT, '   ');
            REPEAT
               IF p^.leftdesc^.ycoord >= i
               THEN p := p^.leftdesc
               ELSE BEGIN
                  q := p^.rtdesc;
                  n := q^.xcoord - p^.xcoord;
                  IF p^.ycoord = i
                  THEN BEGIN
                     IF rover < over
                     THEN BEGIN
                        IF noroot AND
                           (i = (tipy-down)) AND (p^.xcoord = 0)
                        THEN WRITE(OUTPUT, 'L')
                        ELSE WRITE(OUTPUT, '*');
                        n := n + 1;
                        END
                     ELSE BEGIN
                        IF p^.index - spp >= 10
                        THEN WRITE(OUTPUT, (p^.index - spp):2)
                        ELSE WRITE(OUTPUT, (p^.index - spp):1, '-');
                        END;
                     FOR j := 1 TO n-2 DO WRITE(OUTPUT, '-');
                     END
                  ELSE BEGIN
                     WRITE(OUTPUT, '!');
                     FOR j := 1 TO n-1 DO WRITE(OUTPUT, ' ');
                     END;
                  p := q;
                  END;
            UNTIL p^.tip;
            IF p^.ycoord = i
            THEN FOR j := 1 TO nmlngth DO
                       WRITE(OUTPUT, name[p^.index, j]);
            WRITELN(OUTPUT);
         END; (* drawline *)

      BEGIN (* printree *)
         tipx := 0;
         tipy := 1;
         rover := 100 DIV spp;
         IF rover > over THEN rover := over;
         coordinates (root);
         WRITELN(OUTPUT);
         WRITELN(OUTPUT);
         WRITELN(OUTPUT);
         FOR i := 1 TO tipy-down DO drawline (i);
         IF noroot
         THEN BEGIN
            WRITELN(OUTPUT);
            WRITE(OUTPUT, '   remember:');
            IF didreroot
            THEN WRITE(OUTPUT, ' (although rooted by outgroup)');
            WRITELN(OUTPUT, ' this is an unrooted tree!');
            END
         ELSE BEGIN
            WRITELN(OUTPUT, '  /');
            WRITELN(OUTPUT, ' /');
            END;
         WRITELN(OUTPUT);
      END; (* printree *)

      PROCEDURE describe;
         (* prints ancestors, steps and table of numbers of steps in
            each character *)
         VAR i, j, k : INTEGER;

         PROCEDURE hypstates;
            (* fill in and describe states at interior nodes *)
            VAR i, j, k : INTEGER;
               unknown : BOOLEAN;
               dohyp : bitarray;

            PROCEDURE filltrav (r : nodeptr);
               (* traverse to fill in interior node states *)
            BEGIN (* filltrav *)
               IF NOT r^.tip
               THEN BEGIN
                  filltrav (r^.leftdesc);
                  filltrav (r^.rtdesc);
                  fillin (r);
                  END;
            END; (* filltrav *)

            PROCEDURE hyptrav (r : nodeptr);
               (*  compute, print out states at one interior node *)
               VAR i : INTEGER;
                  bottom, maybe, nonzero : BOOLEAN;
                  l0, l1, r0, r1, s0, s1, a0, a1, temp, dh, wa : bitset;
                  zerobelow, onebelow : bitarray;

               PROCEDURE hyprint;
                  (* print out states at node *)
                  VAR i, j, k  : INTEGER;
                     l : bitrange;
                     dot, a0, a1, s0, s1 : BOOLEAN;
               BEGIN (* hyprint *)
                  IF bottom
                  THEN BEGIN
                     IF (noroot AND (NOT didreroot))
                     THEN WRITE(OUTPUT, '        ')
                     ELSE WRITE(OUTPUT, ' root   ');
                     END
                  ELSE WRITE(OUTPUT,
                       (r^.ancestor^.index - spp):4, '    ');
                  IF r^.tip
                  THEN FOR i := 1 TO nmlngth DO
                     WRITE(OUTPUT, name[r^.index, i])
                  ELSE WRITE(OUTPUT, (r^.index - spp):4, '      ');
                  IF bottom AND (noroot AND (NOT didreroot))
                  THEN WRITE(OUTPUT, '          ')
                  ELSE IF nonzero
                     THEN WRITE(OUTPUT, '   yes    ')
                     ELSE If unknown
                        THEN WRITE(OUTPUT, '    ?     ')
                        ELSE IF maybe
                           THEN WRITE(OUTPUT, '  maybe   ')
                           ELSE WRITE(OUTPUT, '   no     ');
                  FOR j := 1 TO chars DO BEGIN
                     newline(j, 40, nmlngth+18);
                     k := (j-1) DIV bits + 1;
                     l := (j-1) MOD bits + 1;
                     dot := (NOT (l IN wagner[k])) AND (guess[j] = '?');
                     s0 := l IN r^.statezero[k];
                     s1 := l IN r^.stateone[k];
                     a0 := l IN zerobelow[k];
                     a1 := l IN onebelow[k];
                     dot := dot OR (((NOT bottom) OR (NOT noroot)
                        OR didreroot) AND ((a1 = s1) AND (a0 = s0)
                        AND (NOT (s0 = s1))));
                     IF dot
                     THEN WRITE(OUTPUT, '.')
                     ELSE BEGIN
                        IF s0
                        THEN WRITE(OUTPUT, '0')
                        ELSE IF s1
                           THEN WRITE(OUTPUT, '1')
                           ELSE WRITE(OUTPUT, '?');
                        END;
                     IF (j MOD 5) = 0
                     THEN WRITE(OUTPUT, ' ');
                     END;
                  WRITELN(OUTPUT);
               END; (* hyprint *)

            BEGIN (* hyptrav *)
               bottom := (r^.ancestor = NIL);
               maybe := FALSE;
               nonzero := FALSE;
               IF bottom
               THEN BEGIN
                  zerobelow := zeroanc;
                  onebelow := oneanc;
                  END
               ELSE BEGIN
                  zerobelow := r^.ancestor^.statezero;
                  onebelow := r^.ancestor^.stateone;
                  END;
               FOR i := 1 TO words DO BEGIN
                  dh := dohyp[i];
                  s0 := r^.statezero[i];
                  s1 := r^.stateone[i];
                  a0 := zerobelow[i];
                  a1 := onebelow[i];
                  IF NOT r^.tip
                  THEN BEGIN
                     wa := wagner[i];
                     l0 := r^.leftdesc^.statezero[i];
                     l1 := r^.leftdesc^.stateone[i];
                     r0 := r^.rtdesc^.statezero[i];
                     r1 := r^.rtdesc^.stateone[i];
                     s0 := wa * (a0 * l0 + a0 * r0 + l0 * r0)
                           + dh * (fullset-wa) * s0;
                     s1 := wa * (a1 * l1 + a1 * r1 + l1 * r1)
                           + dh * (fullset-wa) * s1;
                     temp := fullset - (s0 + s1 + l1 + l0 + r1 + r0);
                     s0 := s0 + temp * a0;
                     s1 := s1 + temp * a1;
                     r^.statezero[i] := s0;
                     r^.stateone[i] := s1;
                     END;
                  maybe := maybe OR ((dh * (s0 + s1)) <> (a0 + a1));
                  nonzero := nonzero OR ((s1 * a0 + s0 * a1) <> []);
                  END;
               hyprint;
               IF NOT r^.tip
               THEN BEGIN
                  hyptrav (r^.leftdesc);
                  hyptrav (r^.rtdesc);
                  END;
            END; (* hyptrav *)

         BEGIN (* hypstates *)
            FOR i := 1 TO words DO BEGIN
               zeroanc[i] := [];
               oneanc[i] := [];
               END;
            unknown := FALSE;
            FOR i := 1 TO chars DO BEGIN
               j := (i-1) DIV bits + 1;
               k := (i-1) MOD bits + 1;
               IF (guess[i] = '0')
               THEN zeroanc[j] := zeroanc[j] + [k];
               IF (guess[i] = '1')
               THEN oneanc[j] := oneanc[j] + [k];
               unknown := unknown OR ((NOT (k IN wagner[j]))
                                     AND (guess[i] = '?'));
               END;
            FOR i := 1 TO words DO
               dohyp[i] := wagner[i] + zeroanc[i] + oneanc[i];
            filltrav (root);
            WRITE(OUTPUT, ' From    To     Any Steps?    ');
            WRITELN(OUTPUT, 'State at upper node');
            WRITE(OUTPUT, '                             ');
            WRITELN(OUTPUT, ' ( . means same as in the node',
                                    ' below it on tree)');
            WRITELN(OUTPUT);
            hyptrav (root);
         END; (* hypstates *)

         PROCEDURE treeout (p : nodeptr);
            (* write out file with representation of final tree *)
            VAR i, n : INTEGER;
               c : CHAR;
         BEGIN (* treeout *)
            IF p^.tip
            THEN BEGIN
               n := 0;
               FOR i := 1 TO nmlngth DO
                  IF name[p^.index, i] <> ' '
                  THEN n := i;
               FOR i := 1 TO n DO BEGIN
                  c := name[p^.index, i];
                  IF c = ' '
                  THEN c := '_';
                  WRITE(treefile, c);
                  END;
               END
            ELSE BEGIN
               WRITE(treefile, '(');
               treeout (p^.leftdesc);
               WRITE(treefile, ',');
               treeout (p^.rtdesc);
               WRITE(treefile, ')');
               END;
            IF p = root
            THEN WRITELN(treefile, ';');
         END; (* treeout *)

      BEGIN (* describe *)
         WRITELN(OUTPUT);
         WRITELN(OUTPUT, ' requires a total of ',-like:10:3);
         WRITELN(OUTPUT);
         IF weights THEN WRITE(OUTPUT, ' weighted');
         WRITELN(OUTPUT, ' steps in each character:');
         WRITE(OUTPUT, '       ');
         FOR i := 0 TO 9 DO WRITE(OUTPUT, i:4);
         WRITELN(OUTPUT);
         WRITE(OUTPUT, '      *------------------------------------');
         WRITELN(OUTPUT, '-----');
         FOR i := 0 TO (chars DIV 10) DO BEGIN
            WRITE(OUTPUT, 10*i:6);
            WRITE(OUTPUT, '!');
            FOR j := 0 TO 9 DO BEGIN
               k := 10*i + j;
               IF (k = 0) OR (k > chars)
               THEN WRITE(OUTPUT, '    ')
               ELSE WRITE(OUTPUT, numsteps[k]+extras[k]:4);
               END;
            WRITELN(OUTPUT);
            END;
         WRITELN(OUTPUT);
         IF questions AND ((NOT noroot) OR didreroot)
         THEN BEGIN
            WRITELN(OUTPUT, ' best guesses of ancestral states:');
            WRITE(OUTPUT, '       ');
            FOR i := 0 TO 9 DO WRITE(OUTPUT, i:2);
            WRITELN(OUTPUT);
            WRITELN(OUTPUT, '      *--------------------');
            FOR i := 0 TO (chars DIV 10) DO BEGIN
               WRITE(OUTPUT, ' ');
               WRITE(OUTPUT, 10*i:5, '!');
               FOR j := 0 TO 9 DO
                  IF (10*i+j = 0) OR (10*i+j > chars)
                  THEN WRITE(OUTPUT, '  ')
                  ELSE WRITE(OUTPUT, ' ', guess[10*i+j]);
               WRITELN(OUTPUT);
               END;
            WRITELN(OUTPUT);
            END;
         hypstates;
         WRITELN(OUTPUT);
         IF trout
         THEN treeout (root);
      END; (* describe *)

   BEGIN (* maketree *)
      fullset := [1..bits];
      IF NOT usertree
      THEN BEGIN
         FOR i := 1 TO spp DO enterorder[i] := i;
         IF jumble
         THEN FOR i := 1 TO spp DO BEGIN
            j := TRUNC(randum(seed)*spp) + 1;
            k := enterorder[j];
            enterorder[j] := enterorder[i];
            enterorder[i] := k;
            END;
         add (treenode[enterorder[2]],
            treenode[enterorder[1]], treenode[spp+1]);
         root := treenode[spp+1];
         WRITELN(OUTPUT, ' Adding species:');
         WRITE(OUTPUT, '    ');
         FOR i := 1 TO nmlngth DO
            WRITE(OUTPUT, name[enterorder[1],i]);
         WRITELN(OUTPUT);
         WRITE(OUTPUT, '    ');
         FOR i := 1 TO nmlngth DO
            WRITE(OUTPUT, name[enterorder[2],i]);
         WRITELN(OUTPUT);
         lastrearr := FALSE;
         FOR i := 3 TO spp DO BEGIN
            bestyet := -32000;
            there := root;
            item := treenode[enterorder[i]];
            nufork := treenode[spp + i - 1];
            addpreorder (root, item, nufork);
            add (there, item, nufork);
            IF root^.ancestor <> NIL THEN root := root^.ancestor;
            like := bestyet;
            rearrange (root);
            WRITE(OUTPUT, '    ');
            FOR j := 1 TO nmlngth DO
               WRITE(OUTPUT, name[enterorder[i],j]);
            WRITELN(OUTPUT);
            lastrearr := (i = spp);
            IF lastrearr
            THEN BEGIN
               WRITELN(OUTPUT, ' Doing global rearrangements');
               nextree := 1;
               bestlike := bestyet;
               REPEAT
                  WRITE(OUTPUT, '    ');
                  gotlike := bestlike;
                  FOR j := 1 TO nonodes DO BEGIN
                     bestyet := -32000;
                     there := root;
                     item := treenode[j];
                     IF item <> root
                     THEN BEGIN
                        nufork := treenode[j]^.ancestor;
                        IF root = nufork
                        THEN BEGIN
                           IF nufork^.leftdesc = item
                           THEN root := nufork^.rtdesc
                           ELSE root := nufork^.leftdesc;
                           END;
                        remove (item, nufork);
                        addpreorder (root, item, nufork);
                        add (there, item, nufork);
                        IF root^.ancestor <> NIL
                        THEN root := root^.ancestor;
                        END;
                     WRITE(OUTPUT, '.');
                     END;
                  WRITELN(OUTPUT);
               UNTIL bestlike <= gotlike;
               END;
            END;
         FOR i := spp DOWNTO 2 DO
            remove (treenode[i], dummy);
         WRITELN(OUTPUT);
         IF (nextree = 2)
         THEN WRITELN(OUTPUT, ' One most parsimonious tree found:')
         ELSE WRITELN(OUTPUT, nextree-1:6, ' trees in all found');
         IF nextree > maxtrees + 1
         THEN BEGIN
            WRITELN(OUTPUT, ' here are the first', maxtrees:4,
                        ' of them');
            nextree := maxtrees + 1;
            END;
         WRITELN(OUTPUT);
         FOR i := 1 TO nextree-1 DO BEGIN
            add (treenode[2], treenode[1], treenode[spp+1]);
            root := treenode[spp+1];
            FOR j := 3 TO spp DO BEGIN
               add(treenode[bestrees[i][j]], treenode[j],
                                              treenode[spp+j-1]);
               IF root^.ancestor <> NIL THEN root := root^.ancestor;
               END;
            IF noroot
            THEN reroot (treenode[outgrno]);
            didreroot := outgropt AND noroot;
            evaluate (root);
            printree;
            describe;
            FOR j := 2 TO spp DO
               remove (treenode[j], dummy);
            END;
         IF trout
         THEN BEGIN
            WRITELN(OUTPUT, ' Trees also written onto file');
            WRITELN(OUTPUT);
            END;
         END
      ELSE BEGIN
         READLN(infile, numtrees);
         IF numtrees > maxuser
         THEN BEGIN
            WRITE(OUTPUT, ' TOO MANY USER-DEFINED TREES');
            WRITELN(OUTPUT, '  only the first', maxuser:4, ' used');
            numtrees := maxuser;
            END;
         WRITE(OUTPUT, ' User-defined tree');
         IF numtrees > 1
         THEN WRITE(OUTPUT, 's');
         WRITELN(OUTPUT, ':');
         FOR which := 1 TO numtrees DO BEGIN
            treeread;
            didreroot := outgropt AND noroot;
            IF noroot
            THEN reroot (treenode[outgrno]);
            evaluate (root);
            printree;
            describe;
            END;
         WRITELN(OUTPUT);
         WRITELN(OUTPUT);
         IF (numtrees > 1) AND (chars > 1)
         THEN BEGIN
            WRITE(OUTPUT, ' Tree    Steps     Diff Steps    Its S.D.');
            WRITELN(OUTPUT, '   Significantly worse?');
            WRITELN(OUTPUT);
            FOR which := 1 TO numtrees DO BEGIN
               WRITE(OUTPUT, which:4, nsteps[which]:10:3);
               IF which = minwhich
               THEN WRITELN(OUTPUT, '  <------- best')
               ELSE BEGIN
                  sum := 0.0;
                  sum2 := 0.0;
                  FOR j := 1 TO chars DO BEGIN
                     sum := sum + fsteps[which][j]-fsteps[minwhich][j];
                     sum2 := sum2
                           + SQR(fsteps[which][j]-fsteps[minwhich][j]);
                     END;
                  sd := SQRT((chars/(chars-1))*(sum2-SQR(sum/chars)));
                  WRITE(OUTPUT, (nsteps[which]-minsteps):13:5, sd:13:5);
                  IF sum > 1.95996*sd
                  THEN WRITELN(OUTPUT, '           Yes')
                  ELSE WRITELN(OUTPUT, '           No');
                  END;
               END;
            WRITELN(OUTPUT);
            WRITELN(OUTPUT);
            END;
         END;
   END; (* maketree *)

BEGIN (* Mixed parsimony by uphill search *)
   (* reads in spp, chars, and the data. Then calls maketree to
      construct the tree *)
   (* ASSIGN(infile, 'INFILE');  Turbo Pascal only *)
   RESET(infile);
   doinput;
   IF trout
   THEN BEGIN
      (* ASSIGN(treefile, 'TREEFILE');  Turbo Pascal only *)
      REWRITE(treefile);
      END;
   maketree;
(* CLOSE(infile);  -- may need this here *)
   IF trout
   THEN BEGIN
   (* CLOSE(treefile);  -- may need this here *)
      END;
END. (* Mixed parsimony by uphill search *)

