PROGRAM Penny(infile, treefile, OUTPUT);
  (* version 3.1. (c) Copyright 1986, 1987, 1988 by the University of
     Washington and Joseph Felsenstein.  Written by Joseph Felsenstein.
     Permission is granted to copy and use this program provided no
     fee is charged for it and provided that this copyright
     notice is not removed. *)

CONST maxsp = 20;    (* maximum number of species               *)
   maxsz = 39;       (* size of pointer array.  >= 2*maxsp - 1  *)
   maxchr = 200;     (* max. number characterstates per species *)
   bits = 30;        (* max. number of elements in a Pascal set *)
   wrds = 7;         (* wrds * bits >= maxchr                   *)
   nmlngth = 10;     (* number of characters in species name    *)
   maxtrees = 100;   (* maximum number of trees to be printed out *)
   howoften = 100;   (* how often to notify how many trees examined *)
   howmany = 300;    (* how many multiples of howoften before stop *)

TYPE steparray = ARRAY[1..maxchr] OF INTEGER;
   bitrange = 1..bits;
   bitset = SET OF bitrange;
   bitarray = ARRAY[1..wrds] OF bitset;
   nodeptr = ^node;   (* nodes will form a binary tree *)
   node = RECORD      (* describes a tip species or an ancestor *)
      ancestor, leftdesc, rtdesc : nodeptr;
      index : INTEGER;
      tip : BOOLEAN;        (* present species are tips of tree *)
      stateone, statezero : bitarray; (* see in PROCEDURE fillin *)
      xcoord, ycoord : INTEGER    (* used by printree *)
      END;
   pointarray = ARRAY [1..maxsz] OF nodeptr;

VAR root : nodeptr;
   infile, treefile : TEXT;
   spp, nonodes, chars, words, outgrno : INTEGER;
     (* spp = number of species
        nonodes = number of nodes in tree
        chars = number of binary characters
        words = number of words needed
                to represent characters of one organism
        outgrno indicates outgroup *)
   weights, thresh, ancvar, questions, allsokal, allwagner,
      mixture, simple, trout, noroot, didreroot, outgropt : BOOLEAN;
   extras, weight : steparray;
   ancone, anczero : ARRAY[1..maxchr] OF BOOLEAN;
   treenode : pointarray;      (* pointers to all nodes in tree *)
   name : ARRAY[1..maxsp, 1..nmlngth] OF CHAR; (* names of species *)
   threshold, fracdone, fracinc : REAL;
   threshwt : ARRAY[1..maxchr] OF REAL;
   wagner : bitarray;
   added : ARRAY [1..maxsz] OF BOOLEAN;

   FUNCTION digit (ch : CHAR) : BOOLEAN;
   BEGIN (* digit *)
      digit := (ch >= '0') AND (ch <= '9');
   END; (* digit *)

   FUNCTION letter (ch : CHAR) : BOOLEAN;
      (* tests whether ch is a letter
         --  works for both ASCII and EBCDIC codes *)
   BEGIN (* letter *)
      letter := ((ch >= 'A') AND (ch <= 'I'))
                OR ((ch >= 'J') AND (ch <= 'R'))
                OR ((ch >= 'S') AND (ch <= 'Z'))
                OR ((ch >= 'a') AND (ch <= 'i'))
                OR ((ch >= 'j') AND (ch <= 'r'))
                OR ((ch >= 's') AND (ch <= 'z'));
   END; (* letter *)

   PROCEDURE uppercase (VAR ch : CHAR);
      (* convert a character to upper case -- either ASCII or EBCDIC *)
   BEGIN (* uppercase *)
      IF letter(ch) AND (('a' > 'A') AND (ch >= 'a')
                     OR  ('a' < 'A') AND (ch < 'A'))
      THEN ch := CHR(ORD(ch) + ORD('A') - ORD('a'));
   END; (* uppercase *)

   PROCEDURE newline (i, j, k : INTEGER);
      (* go to new line if i is a multiple of j, indent k spaces *)
      VAR m : INTEGER;
   BEGIN (* newline *)
      IF (((i-1) MOD j) = 0) AND (i>1)
      THEN BEGIN
         WRITELN(OUTPUT);
         FOR m := 1 TO k DO WRITE(OUTPUT, ' ');
         END;
   END; (* newline *)

   PROCEDURE doinput;
      (* reads the input data *)

      PROCEDURE inputnumbers;
         (* input the numbers of species and of characters *)
      BEGIN (* inputnumbers *)
         WRITELN(OUTPUT);
         WRITELN(OUTPUT, ' Penny algorithm, version 3.1');
         WRITE(OUTPUT, '  branch-and-bound to find all ');
         WRITELN(OUTPUT, 'most parsimonious trees');
         WRITELN(OUTPUT);
         READ(infile, spp, chars);
         WRITELN(OUTPUT, ' ',spp:2,' species, ',chars:3,'  characters');
         IF spp > maxsp
         THEN WRITELN(OUTPUT, ' TOO MANY SPECIES: adjust CONSTants');
         IF chars > maxchr
         THEN WRITELN(OUTPUT, ' TOO MANY CHARACTERS: adjust CONSTants');
         WRITELN(OUTPUT);
         words := (chars DIV bits) + 1;
         nonodes := 2 * spp - 1;
      END; (* inputnumbers *)

      PROCEDURE inputoptions;
         (* input the information on the options *)
         VAR ch : CHAR;
            extranum, i : INTEGER;

         PROCEDURE inputmixture;
            (* input mixture of methods *)
            VAR i, j, k : INTEGER;
               ch : CHAR;
               wag : BOOLEAN;
         BEGIN (* inputmixture *)
            FOR i := 1 TO nmlngth-1 DO READ(infile, ch);
            FOR i := 1 TO words DO wagner[i] := [];
            j := 0;
            k := 1;
            FOR i := 1 TO chars DO BEGIN
               REPEAT
                  IF EOLN(infile) THEN READLN(infile);
                  READ(infile, ch);
               UNTIL ch <> ' ';
               uppercase (ch);
               wag := FALSE;
               IF (ch='W') OR (ch = '?')
               THEN wag := TRUE
               ELSE IF (ch='S') OR (ch='C')
                  THEN wag := FALSE
                  ELSE WRITELN(OUTPUT, ' error: ', ch);
               j := j + 1;
               IF j > bits
               THEN BEGIN
                  j := 1;
                  k := k + 1;
                  END;
               IF wag
               THEN wagner[k] := wagner[k] + [j];
               IF NOT ancvar
               THEN ancone[i] := wag;
               END;
            READLN(infile);
            allwagner := FALSE;
            allsokal := FALSE;
            mixture := TRUE;
         END; (* inputmixture *)

         PROCEDURE printmixture;
            (* print out list of parsimony methods *)
            VAR i, k, l : INTEGER;
         BEGIN (* printmixture *)
            allwagner := TRUE;
            WRITELN(OUTPUT, ' Parsimony methods:');
            l := 0;
            k := 1;
            FOR i := 1 TO nmlngth+4 DO WRITE(OUTPUT, ' ');
            FOR i := 1 TO chars DO BEGIN
               newline (i, 50, nmlngth+4);
               l := l + 1;
               IF l > bits
               THEN BEGIN
                  l := 1;
                  k := k + 1;
                  END;
               IF l IN wagner[k]
               THEN WRITE(OUTPUT, 'W')
               ELSE BEGIN
                  WRITE(OUTPUT, 'S');
                  allwagner := allwagner AND (weight[i] = 0);
                  END;
               IF ((i MOD 5) = 0) THEN WRITE(OUTPUT, ' ');
               END;
            WRITELN(OUTPUT);
            WRITELN(OUTPUT);
         END; (* printmixture *)

         PROCEDURE inputhresh;
            (* input the threshold *)
         BEGIN (* inputhresh *)
            READLN(infile, threshold);
            thresh := TRUE;
         END; (* inputhresh *)

         PROCEDURE printhresh;
            (* print out the threshold value *)
         BEGIN (* printhresh *)
            WRITELN(OUTPUT, '    Threshold value = ', threshold:10:3);
            WRITELN(OUTPUT);
         END; (* printhresh *)

         PROCEDURE inputweights;
            (* input the character weights, 0-9 and A-Z for weights
               0 - 35 *)
            VAR ch : CHAR;
               i : INTEGER;
         BEGIN (* inputweights *)
            FOR i := 1 TO nmlngth-1 DO READ(infile, ch);
            FOR i := 1 TO chars DO BEGIN
               REPEAT
                  IF EOLN(infile) THEN READLN(infile);
                  READ(infile, ch);
               UNTIL ch <> ' ';
               weight[i] := 1;
               IF digit(ch)
               THEN weight[i] := ORD(ch) - ORD('0')
               ELSE IF letter(ch)
                  THEN BEGIN
                     uppercase (ch);
                     IF (ch >= 'A') AND (ch <= 'I')
                     THEN weight[i] := ORD(ch) - ORD('A') + 10
                     ELSE IF (ch >= 'J') AND (ch <= 'R')
                        THEN weight[i] := ORD(ch) - ORD('J') + 19
                        ELSE weight[i] := ORD(ch) - ORD('S') + 28;
                     END
                  ELSE WRITELN(OUTPUT, ' Bad weight character: ',ch);
               END;
            READLN(infile);
            weights := TRUE;
         END; (* inputweights *)

         PROCEDURE printweights;
            (* print out the weights of characters *)
            VAR i, j, k : INTEGER;
         BEGIN (* printweights *)
            WRITELN(OUTPUT,'    Characters are weighted as follows:');
            WRITE(OUTPUT, '        ');
            FOR i := 0 TO 9 DO WRITE(OUTPUT, i:3);
            WRITELN(OUTPUT);
            WRITELN(OUTPUT,'     *---------------------------------');
            FOR j := 0 TO (chars DIV 10) DO BEGIN
               WRITE(OUTPUT, 10*j:5, '!  ');
               FOR i := 0 TO 9 DO BEGIN
                  k := 10*j + i;
                  IF (k>0) AND (k<=chars)
                  THEN WRITE(OUTPUT, weight[k]:3)
                  ELSE WRITE(OUTPUT, '   ');
                  END;
               WRITELN(OUTPUT);
               END;
            WRITELN(OUTPUT);
         END; (* printweights *)

         PROCEDURE inputancestors;
            (* reads the ancestral states for each character *)
            VAR i : INTEGER;
               ch : CHAR;
         BEGIN (* inputancestors *)
            FOR i := 1 TO nmlngth-1 DO READ(infile, ch);
            ancvar := TRUE;
            FOR i := 1 TO chars DO BEGIN
               REPEAT
                  IF EOLN(infile) THEN READLN(infile);
                  READ(infile, ch);
               UNTIL ch <> ' ';
               IF ch = 'p' THEN ch := 'P';
               IF ch = 'b' THEN ch := 'B';
               CASE ch OF
                  '1' : anczero[i] := FALSE;
                  '0' : ancone[i] := FALSE;
                  'P' : BEGIN END;
                  'B' : BEGIN END;
                  '?' : BEGIN END
                  END;
               END;
            READLN(infile);
         END; (* inputancestors *)

         PROCEDURE printancestors;
            (* print out list of ancestral states *)
            VAR i : INTEGER;
         BEGIN (* printancestors *)
            WRITELN(OUTPUT, ' Ancestral states:');
            FOR i := 1 TO nmlngth+4 DO WRITE(OUTPUT, ' ');
            FOR i := 1 TO chars DO BEGIN
               newline (i, 50, nmlngth+4);
               IF ancone[i] AND anczero[i]
               THEN WRITE(OUTPUT, '?')
               ELSE IF ancone[i]
                  THEN WRITE(OUTPUT, '1')
                  ELSE WRITE(OUTPUT, '0');
               IF ((i MOD 5) = 0) THEN WRITE(OUTPUT, ' ');
               END;
            WRITELN(OUTPUT);
            WRITELN(OUTPUT);
         END; (* printancestor *)

      BEGIN (* inputoptions *)
         ancvar := FALSE;
         FOR i := 1 TO chars DO BEGIN
            anczero[i] := TRUE;
            ancone[i] := TRUE;
            END;
         allwagner := TRUE;
         allsokal := FALSE;
         mixture := FALSE;
         outgrno := 1;
         outgropt := FALSE;
         simple := TRUE;
         thresh := FALSE;
         trout := FALSE;
         weights := FALSE;
         extranum := 0;
         WHILE NOT EOLN(infile) DO BEGIN
            READ(infile, ch);
            uppercase (ch);
            IF (ch = 'A') OR (ch = 'C') OR (ch = 'M')
               OR (ch = 'O') OR (ch = 'R') OR (ch = 'S')
               OR (ch = 'T') OR (ch = 'W') OR (ch = 'Y')
            THEN CASE ch OF
                  'A' : extranum := extranum + 1;
                  'C' : allsokal := TRUE;
                  'M' : extranum := extranum + 1;
                  'O' : BEGIN
                        outgropt := TRUE;
                        extranum := extranum + 1;
                        END;
                  'R' : simple := FALSE;
                  'S' : allsokal := TRUE;
                  'T' : extranum := extranum + 1;
                  'W' : extranum := extranum + 1;
                  'Y' : trout := TRUE
                  END
            ELSE IF ch <> ' '
               THEN WRITELN(OUTPUT, ' Bad Option Character: ', ch);
            END;
         READLN(infile);
         IF allsokal THEN allwagner := FALSE;
         FOR i := 1 TO words DO
            IF allsokal
            THEN wagner[i] := []
            ELSE wagner[i] := [1..bits];
         threshold := spp;
         FOR i := 1 TO chars DO weight[i] := 1;
         FOR i := 1 TO extranum DO BEGIN
            READ(infile, ch);
            uppercase (ch);
            IF ch = 'A'
            THEN inputancestors;
            IF ch = 'M'
            THEN inputmixture;
            IF ch = 'O'
            THEN READLN(infile, outgrno);
            IF ch = 'T'
            THEN inputhresh;
            IF ch = 'W'
            THEN inputweights;
            END;
         IF allsokal
         THEN BEGIN
            IF NOT ancvar
            THEN FOR i := 1 TO chars DO ancone[i] := FALSE;
            WRITELN(OUTPUT, ' Camin-Sokal parsimony method');
            END;
         IF allwagner
         THEN WRITELN(OUTPUT, ' Wagner parsimony method');
         WRITELN(OUTPUT);
         IF mixture
         THEN printmixture;
         IF thresh
         THEN printhresh;
         IF weights
         THEN printweights;
         IF ancvar
         THEN printancestors;
         noroot := allwagner OR (threshold <= 2.0);
         questions := FALSE;
         FOR i := 1 TO chars DO BEGIN
            noroot := noroot AND ancone[i] AND anczero[i];
            questions := questions OR (ancone[i] AND anczero[i]);
            threshwt[i] := threshold * weight[i];
            END;
      END; (* inputoptions *)

      PROCEDURE inputdata;
         (* input the names and character state data for species *)
         VAR i, j, l : INTEGER;
            k : bitrange;
            charstate : CHAR;  (* possible states are
                                  '0', '1', 'P', 'B', and '?' *)
      BEGIN (* inputdata *)
         j := nmlngth - 5 + chars DIV 2;
         IF j < nmlngth - 1 THEN j := nmlngth - 1;
         IF j > 39 THEN j := 39;
         WRITE(OUTPUT, ' Name');
         FOR i := 1 TO j DO WRITE(OUTPUT, ' ');
         WRITELN(OUTPUT, 'Characters');
         WRITE(OUTPUT, ' ----');
         FOR i := 1 TO j DO WRITE(OUTPUT, ' ');
         WRITELN(OUTPUT, '----------');
         WRITELN(OUTPUT);
         FOR i := 1 TO chars DO extras[i] := 0;
         FOR i := 1 TO nonodes DO BEGIN
            NEW (treenode[i]);
            treenode[i]^.leftdesc := NIL;
            treenode[i]^.rtdesc := NIL;
            treenode[i]^.ancestor := NIL;
            treenode[i]^.index := i;
            IF i <= spp
            THEN BEGIN
               treenode[i]^.tip := TRUE;
               FOR j := 1 TO nmlngth DO READ(infile, name[i,j]);
               WRITE(OUTPUT, ' ');
               FOR j := 1 TO nmlngth DO WRITE(OUTPUT, name[i,j]);
               WRITE(OUTPUT, '   ');
               FOR j := 1 TO words DO BEGIN
                  treenode[i]^.stateone[j] := [];
                  treenode[i]^.statezero[j] := [];
                  END;
               FOR j := 1 TO chars DO BEGIN
                  k := ((j-1) MOD bits) + 1;
                  l := ((j-1) DIV bits) + 1;
                  REPEAT
                     IF EOLN(infile) THEN READLN(infile);
                     READ(infile, charstate);
                  UNTIL charstate <> ' ';
                  IF charstate = 'b' THEN charstate := 'B';
                  IF charstate = 'p' THEN charstate := 'P';
                  IF NOT ((charstate = '0')
                         OR (charstate = '1')
                         OR (charstate = '?')
                         OR (charstate = 'P')
                         OR (charstate = 'B'))
                  THEN WRITELN(OUTPUT, ' Warning -- ',
                      'bad character state: ', charstate,
                      ' at character ', j:5, ' of species ', i:3);
                  newline (j, 60, nmlngth+4);
                  WRITE(OUTPUT, charstate);
                  IF (j MOD 5) = 0 THEN WRITE(OUTPUT, ' ');
                  IF (charstate = '1')
                  THEN treenode[i]^.stateone[l]
                             := treenode[i]^.stateone[l] + [k];
                  IF (charstate = '0')
                  THEN treenode[i]^.statezero[l]
                             := treenode[i]^.statezero[l] + [k];
                  IF (charstate = 'P') OR (charstate = 'B')
                  THEN extras[j] := extras[j] + weight[j];
                  END;
               READLN(infile);
               WRITELN(OUTPUT);
               END
            ELSE treenode[i]^.tip := FALSE;
            END;
         WRITELN(OUTPUT);
         WRITELN(OUTPUT);
      END; (* inputdata *)

   BEGIN (* doinput *)
      inputnumbers;
      inputoptions;
      inputdata;
   END; (* doinput *)

   PROCEDURE maketree;
      (* tree construction recursively by branch and bound *)
      TYPE treenumbers = ARRAY[1..maxsp] OF INTEGER;
      VAR i, j, k, nextree, examined, mults : INTEGER;
         firsttime, done : BOOLEAN;
         like, bestyet : REAL;
         guess : ARRAY[1..maxchr] OF CHAR;
         numsteps : steparray;
         bestorders, bestrees : ARRAY[1..maxtrees] OF treenumbers;
         current, order : treenumbers;
         fullset : bitset;
         steps, zeroanc, oneanc : bitarray;

      PROCEDURE add (VAR below, newtip, newfork : nodeptr);
         (* inserts the nodes newfork and its left descendant, newtip,
            to the tree.  below becomes newfork's right descendant.
            The global variable root is also updated *)
      BEGIN (* add *)
         newfork^.ancestor := below^.ancestor;
         newfork^.rtdesc := below;
         newfork^.leftdesc := newtip;
         below^.ancestor := newfork;
         newtip^.ancestor := newfork;
         IF newfork^.ancestor <> NIL
         THEN IF newfork^.ancestor^.leftdesc = below
            THEN newfork^.ancestor^.leftdesc := newfork
            ELSE newfork^.ancestor^.rtdesc := newfork;
         IF root = below THEN root := newfork;
      END; (* add *)

      PROCEDURE remove (VAR item, fork : nodeptr);
         (* removes nodes item and its ancestor, fork, from the tree.
            the new descendant of fork's ancestor is made to be
            fork's second descendant (other than item).  Also
            returns pointers to the deleted nodes, item and fork.
            The global variable root is also updated *)
         VAR other, below : nodeptr;
      BEGIN (* remove *)
         IF item^.ancestor = NIL
         THEN fork := NIL
         ELSE BEGIN
            fork := item^.ancestor;
            IF item = fork^.leftdesc
            THEN other := fork^.rtdesc
            ELSE other := fork^.leftdesc;
            below := fork^.ancestor;
            other^.ancestor := below;
            IF below <> NIL
            THEN BEGIN
               IF below^.leftdesc = fork
               THEN below^.leftdesc := other
               ELSE below^.rtdesc := other;
               fork^.ancestor := NIL;
               END;
            fork^.leftdesc := NIL;
            fork^.rtdesc := NIL;
            item^.ancestor := NIL;
            END;
         IF root = fork THEN root := other;
      END; (* remove *)

      PROCEDURE fillin (p: nodeptr);
         (* Sets up for each node in the tree two statesets.
            stateone and statezero are the sets of character
            states that must be 1 or must be 0, respectively,
            in a most parsimonious reconstruction, based on the
            information at or above this node.  Note that this
            state assignment may change based on information further
            down the tree.  If a character is in both sets it is in
            state "P".  If in neither, it is "?". *)
         VAR i : INTEGER;
            l0, l1, r0, r1, st, wa, za : bitset;
      BEGIN (* fillin *)
         FOR i := 1 TO words DO BEGIN
            l0 := p^.leftdesc^.statezero[i];
            l1 := p^.leftdesc^.stateone[i];
            r0 := p^.rtdesc^.statezero[i];
            r1 := p^.rtdesc^.stateone[i];
            wa := wagner[i];
            za := zeroanc[i];
            st := l1 * r0 + l0 * r1;
            steps[i] := st;
            p^.stateone[i] := l1 + r1 - (st * (wa + za));
            p^.statezero[i] := l0 + r0 - (st * (wa + (fullset - za)));
            END;
      END; (* fillin *)

      PROCEDURE evaluate (r: nodeptr);
         (* Determines the number of steps needed for a tree.
            This is the minimum number needed to evolve chars on
            this tree *)
         VAR i, stepnum, smaller : INTEGER;
            sum : REAL;
            numsone, numszero : steparray;
            suppsteps : bitarray;

         PROCEDURE count (VAR stps : bitarray);
            (* counts the number of steps in a fork of the tree.
               The program spends much of its time in this PROCEDURE *)
            VAR i, j, l : INTEGER;
         BEGIN (* count *)
            j := 1;
            l := 0;
            FOR i := 1 TO chars DO BEGIN
               l := l + 1;
               IF l > bits
               THEN BEGIN
                  l := 1;
                  j := j + 1;
                  END;
               IF l IN stps[j]
               THEN BEGIN
                  IF l IN zeroanc[j]
                  THEN numszero[i] := numszero[i] + weight[i]
                  ELSE numsone[i] := numsone[i] + weight[i];
                  END;
               END;
         END; (* count *)

         PROCEDURE postorder (p : nodeptr);
            (* traverses a binary tree, calling PROCEDURE fillin at a
               node's descendants before calling fillin at the node *)
         BEGIN (* postorder *)
            IF NOT p^.tip
            THEN BEGIN
               postorder (p^.leftdesc);
               postorder (p^.rtdesc);
               fillin (p);
               count (steps);
               END;
         END; (* postorder *)

         PROCEDURE supplement;
            (* determine minimum number of steps more which will
               be added when rest of species are put in tree *)
            VAR i, j, k, l : INTEGER;
               defone, defzero, a : bitset;
         BEGIN (* supplement *)
            k := 0;
            FOR i := 1 TO words DO BEGIN
               defone := [];
               defzero := [];
               a := [];
               FOR l := 1 TO bits DO BEGIN
                  k := k + 1;
                  IF k <= chars
                  THEN BEGIN
                     IF NOT ancone[k]
                     THEN defzero := defzero + [l];
                     IF NOT anczero[k]
                     THEN defone := defone + [l];
                     END;
                  END;
               FOR j := 1 TO spp DO BEGIN
                  defone := defone + (treenode[j]^.stateone[i]
                                     - treenode[j]^.statezero[i]);
                  defzero := defzero + (treenode[j]^.statezero[i]
                                       - treenode[j]^.stateone[i]);
                  IF added[j]
                  THEN a := a + defone*defzero;
                  END;
               suppsteps[i] := defone*defzero - a;
               END;
         END; (* supplement *)

      BEGIN (* evaluate *)
         sum := 0;
         FOR i := 1 TO chars DO BEGIN
            numszero[i] := 0;
            numsone[i] := 0;
            END;
         supplement;
         FOR i := 1 TO words DO
            zeroanc[i] := fullset;
         postorder (r);
         count (r^.stateone);
         count (suppsteps);
         FOR i := 1 TO words DO
            zeroanc[i] := [];
         postorder (r);
         count (r^.statezero);
         count (suppsteps);
         FOR i := 1 TO chars DO BEGIN
            smaller := spp * weight[i];
            IF anczero[i]
            THEN BEGIN
               numsteps[i] := numszero[i];
               smaller := numszero[i];
               END;
            IF ancone[i] AND (numsone[i] < smaller)
            THEN numsteps[i] := numsone[i];
            stepnum := numsteps[i] + extras[i];
            IF stepnum <= threshwt[i]
            THEN sum := sum + stepnum
            ELSE sum := sum + threshwt[i];
            guess[i] := '?';
            IF (NOT ancone[i]) OR
               (anczero[i] AND (numszero[i] < numsone[i]))
            THEN guess[i] := '0'
            ELSE IF (NOT anczero[i]) OR
                  (ancone[i] AND (numsone[i] < numszero[i]))
               THEN guess[i] := '1';
            END;
         IF (examined = 0) AND (mults = 0) THEN bestyet := -1.0;
         like := sum;
      END; (* evaluate *)

      PROCEDURE addit (m : INTEGER);
         (* adds the species one by one, recursively *)
         TYPE valarray = ARRAY[1..maxsz] OF REAL;
            placearray = ARRAY[1..maxsz] OF INTEGER;
         VAR i, j, n, n1, besttoadd : INTEGER;
            valyew, bestval : valarray;
            place, bestplace : placearray;
            oldfrac, oldfdone, sum, bestsum : REAL;

         PROCEDURE addtraverse (a, b, c : nodeptr);
            (* traverse all places to add b *)
         BEGIN (* addtraverse *)
            IF NOT done
            THEN BEGIN
               IF (m <= 2) OR (NOT (noroot AND
                     ((a = root) OR (a = root^.leftdesc))))
               THEN BEGIN
                  add (a, b, c);
                  n := n + 1;
                  evaluate (root);
                  examined := examined + 1;
                  IF examined = howoften
                  THEN BEGIN
                     examined := 0;
                     mults := mults + 1;
                     IF mults = howmany
                     THEN done := TRUE;
                     WRITE(OUTPUT, ' ', mults:7);
                     IF bestyet >= 0
                     THEN WRITE(OUTPUT,  bestyet:18:5)
                     ELSE WRITE(OUTPUT, '         -        ');
                     WRITELN(OUTPUT, nextree-1:15, fracdone*100:20:2);
                     END;
                  valyew[n] := like;
                  place[n] := a^.index;
                  remove (b, c);
                  END;
               IF NOT a^.tip
               THEN BEGIN
                  addtraverse (a^.leftdesc, b, c);
                  addtraverse (a^.rtdesc, b, c);
                  END;
               END;
         END; (* addtraverse *)

         PROCEDURE shellsort (VAR a : valarray; VAR b : placearray;
                                n : INTEGER);
            (* Shell sort keeping a, b in same order *)
            VAR gap, i, j, itemp : INTEGER;
              rtemp : REAL;
         BEGIN (* shellsort *)
            gap := n DIV 2;
            WHILE (gap > 0) DO BEGIN
               FOR i := gap TO n-1 DO BEGIN
                  j := i - gap + 1;
                  WHILE (j>0) DO BEGIN
                     IF a[j] > a[j+gap]
                     THEN BEGIN
                        rtemp := a[j];
                        a[j] := a[j+gap];
                        a[j+gap] := rtemp;
                        itemp := b[j];
                        b[j] := b[j+gap];
                        b[j+gap] := itemp;
                        END;
                     j := j - gap;
                     END;
                  END;
               gap := gap DIV 2;
               END;
         END; (* shellsort *)

      BEGIN (* addit *)
         IF simple AND (NOT firsttime)
         THEN BEGIN
            n := 0;
            added[order[m]] := TRUE;
            addtraverse (root, treenode[order[m]], treenode[spp+m-1]);
            besttoadd := order[m];
            bestplace := place;
            bestval := valyew;
            END
         ELSE BEGIN
            bestsum := -1;
            FOR i := 1 TO spp DO
               IF NOT added[i]
               THEN BEGIN
                  n := 0;
                  added[i] := TRUE;
                  addtraverse (root, treenode[i], treenode[spp+m-1]);
                  added[i] := FALSE;
                  sum := 0.0;
                  FOR j := 1 TO n DO
                     sum := sum + valyew[j];
                  IF sum > bestsum
                  THEN BEGIN
                     bestsum := sum;
                     besttoadd := i;
                     bestplace := place;
                     bestval := valyew;
                     END;
                  END;
            END;
         order[m] := besttoadd;
         place := bestplace;
         valyew := bestval;
         shellsort (valyew, place, n);
         oldfrac := fracinc;
         oldfdone := fracdone;
         n1 := 0;
         FOR i := 1 TO n DO
            IF (valyew[i] <= bestyet) OR (bestyet < 0.0)
            THEN n1 := n1 + 1;
         IF n1 > 0 THEN fracinc := fracinc / n1;
         FOR i := 1 TO n DO BEGIN
            IF (valyew[i] <= bestyet) OR (bestyet < 0.0)
            THEN BEGIN
               current[m] := place[i];
               add(treenode[place[i]], treenode[besttoadd],
                                         treenode[spp+m-1]);
               added[besttoadd] := TRUE;
               IF m < spp
               THEN addit (m+1)
               ELSE BEGIN
                  IF (valyew[i] < bestyet) OR (bestyet < 0.0)
                  THEN BEGIN
                     nextree := 1;
                     bestyet := valyew[i];
                     END;
                  IF nextree <= maxtrees
                  THEN BEGIN
                     bestorders[nextree] := order;
                     bestrees[nextree] := current;
                     END;
                  nextree := nextree + 1;
                  firsttime := FALSE;
                  END;
               remove (treenode[besttoadd], treenode[spp+m-1]);
               added[besttoadd] := FALSE;
               END;
            fracdone := fracdone + fracinc;
            END;
         fracinc := oldfrac;
         fracdone := oldfdone;
      END; (* addit *)

      PROCEDURE reroot (outgroup : nodeptr);
         (* reorients tree, putting outgroup in desired position.
            Written by Kent Fiala *)
         VAR vertex, oldvertex, ante, otherbranch : nodeptr;
            left : BOOLEAN;
      BEGIN (* reroot *)
         IF NOT (outgroup^.ancestor = root)
         THEN BEGIN
            oldvertex := outgroup;
            vertex := outgroup^.ancestor;
            ante := vertex^.ancestor;
            IF vertex^.rtdesc = outgroup
            THEN BEGIN
               vertex^.rtdesc := vertex^.leftdesc;
               vertex^.leftdesc := outgroup;
               END;
            REPEAT
               IF oldvertex = vertex^.leftdesc
               THEN BEGIN
                  vertex^.leftdesc := ante;
                  left := TRUE;
                  END
               ELSE BEGIN
                  vertex^.rtdesc := ante;
                  left := FALSE;
                  END;
               vertex^.ancestor := oldvertex;
               oldvertex := vertex;
               vertex := ante;
               ante := vertex^.ancestor
            UNTIL vertex = root;
            IF root^.leftdesc = oldvertex
            THEN otherbranch := root^.rtdesc
            ELSE otherbranch := root^.leftdesc;
            IF left
            THEN oldvertex^.leftdesc := otherbranch
            ELSE oldvertex^.rtdesc := otherbranch;
            otherbranch^.ancestor := oldvertex;
            root^.rtdesc := outgroup^.ancestor;
            root^.leftdesc := outgroup;
            outgroup^.ancestor^.ancestor := root;
            outgroup^.ancestor := root;
            END;
      END; (* reroot *)

      PROCEDURE printree;
         (* prints out diagram of the tree *)
         CONST down = 2;
            over = 3;
         VAR tipx, tipy, i, rover : INTEGER;

         PROCEDURE coordinates (p : nodeptr);
            (* establishes coordinates of nodes *)
         BEGIN (* coordinates *)
            IF p^.tip
            THEN BEGIN
               p^.xcoord := tipx;
               p^.ycoord := tipy;
               tipx := tipx + rover;
               tipy := tipy + down;
               END
            ELSE BEGIN
               coordinates (p^.leftdesc);
               p^.xcoord := p^.leftdesc^.xcoord;
               coordinates (p^.rtdesc);
               p^.ycoord := p^.rtdesc^.ycoord;
               END;
         END; (* coordinates *)

         PROCEDURE drawline (i : INTEGER);
            (* draws one row of the tree diagram by moving up tree *)
            VAR p, q : nodeptr;
               n, j : INTEGER;
         BEGIN (* drawline *)
            p := root;
            WRITE(OUTPUT, '   ');
            REPEAT
               IF p^.leftdesc^.ycoord >= i
               THEN p := p^.leftdesc
               ELSE BEGIN
                  q := p^.rtdesc;
                  n := q^.xcoord - p^.xcoord;
                  IF p^.ycoord = i
                  THEN BEGIN
                     IF rover < over
                     THEN BEGIN
                        IF noroot AND
                           (i = (tipy-down)) AND (p^.xcoord = 0)
                        THEN WRITE(OUTPUT, 'L')
                        ELSE WRITE(OUTPUT, '*');
                        n := n + 1;
                        END
                     ELSE BEGIN
                        IF p^.index - spp >= 10
                        THEN WRITE(OUTPUT, (p^.index - spp):2)
                        ELSE WRITE(OUTPUT, (p^.index - spp):1, '-');
                        END;
                     FOR j := 1 TO n-2 DO WRITE(OUTPUT, '-');
                     END
                  ELSE BEGIN
                     WRITE(OUTPUT, '!');
                     FOR j := 1 TO n-1 DO WRITE(OUTPUT, ' ');
                     END;
                  p := q;
                  END;
            UNTIL p^.tip;
            IF p^.ycoord = i
            THEN FOR j := 1 TO nmlngth DO
                       WRITE(OUTPUT, name[p^.index, j]);
            WRITELN(OUTPUT);
         END; (* drawline *)

      BEGIN (* printree *)
         tipx := 0;
         tipy := 1;
         rover := 100 DIV spp;
         IF rover > over THEN rover := over;
         coordinates (root);
         WRITELN(OUTPUT);
         WRITELN(OUTPUT);
         WRITELN(OUTPUT);
         FOR i := 1 TO tipy-down DO drawline (i);
         IF noroot
         THEN BEGIN
            WRITELN(OUTPUT);
            WRITE(OUTPUT, '   remember:');
            IF didreroot
            THEN WRITE(OUTPUT, ' (although rooted by outgroup)');
            WRITELN(OUTPUT, ' this is an unrooted tree!');
            END
         ELSE BEGIN
            WRITELN(OUTPUT, '  /');
            WRITELN(OUTPUT, ' /');
            END;
         WRITELN(OUTPUT);
      END; (* printree *)

      PROCEDURE describe;
         (* prints ancestors, steps and table of numbers of steps in
            each character *)
         VAR i, j, k : INTEGER;

         PROCEDURE hypstates;
            (* fill in and describe states at interior nodes *)
            VAR i, j, k : INTEGER;
               unknown : BOOLEAN;
               dohyp : bitarray;

            PROCEDURE filltrav (r : nodeptr);
               (* traverse to fill in interior node states *)
            BEGIN (* filltrav *)
               IF NOT r^.tip
               THEN BEGIN
                  filltrav (r^.leftdesc);
                  filltrav (r^.rtdesc);
                  fillin (r);
                  END;
            END; (* filltrav *)

            PROCEDURE hyptrav (r : nodeptr);
               (*  compute, print out states at one interior node *)
               VAR i : INTEGER;
                  bottom, maybe, nonzero : BOOLEAN;
                  l0, l1, r0, r1, s0, s1, a0, a1, temp, dh, wa : bitset;
                  zerobelow, onebelow : bitarray;

               PROCEDURE hyprint;
                  (* print out states at node *)
                  VAR i, j, k  : INTEGER;
                     l : bitrange;
                     dot, a0, a1, s0, s1 : BOOLEAN;
               BEGIN (* hyprint *)
                  IF bottom
                  THEN BEGIN
                     IF (noroot AND (NOT didreroot))
                     THEN WRITE(OUTPUT, '        ')
                     ELSE WRITE(OUTPUT, ' root   ');
                     END
                  ELSE WRITE(OUTPUT,
                       (r^.ancestor^.index - spp):4, '    ');
                  IF r^.tip
                  THEN FOR i := 1 TO nmlngth DO
                     WRITE(OUTPUT, name[r^.index, i])
                  ELSE WRITE(OUTPUT, (r^.index - spp):4, '      ');
                  IF bottom AND (noroot AND (NOT didreroot))
                  THEN WRITE(OUTPUT, '          ')
                  ELSE IF nonzero
                     THEN WRITE(OUTPUT, '   yes    ')
                     ELSE If unknown
                        THEN WRITE(OUTPUT, '    ?     ')
                        ELSE IF maybe
                           THEN WRITE(OUTPUT, '  maybe   ')
                           ELSE WRITE(OUTPUT, '   no     ');
                  FOR j := 1 TO chars DO BEGIN
                     newline(j, 40, nmlngth+18);
                     k := (j-1) DIV bits + 1;
                     l := (j-1) MOD bits + 1;
                     dot := (NOT (l IN wagner[k])) AND (guess[j] = '?');
                     s0 := l IN r^.statezero[k];
                     s1 := l IN r^.stateone[k];
                     a0 := l IN zerobelow[k];
                     a1 := l IN onebelow[k];
                     dot := dot OR (((NOT bottom) OR (NOT noroot)
                        OR didreroot) AND ((a1 = s1) AND (a0 = s0)
                        AND (NOT (s0 = s1))));
                     IF dot
                     THEN WRITE(OUTPUT, '.')
                     ELSE BEGIN
                        IF s0
                        THEN WRITE(OUTPUT, '0')
                        ELSE IF s1
                           THEN WRITE(OUTPUT, '1')
                           ELSE WRITE(OUTPUT, '?');
                        END;
                     IF (j MOD 5) = 0
                     THEN WRITE(OUTPUT, ' ');
                     END;
                  WRITELN(OUTPUT);
               END; (* hyprint *)

            BEGIN (* hyptrav *)
               bottom := (r^.ancestor = NIL);
               maybe := FALSE;
               nonzero := FALSE;
               IF bottom
               THEN BEGIN
                  zerobelow := zeroanc;
                  onebelow := oneanc;
                  END
               ELSE BEGIN
                  zerobelow := r^.ancestor^.statezero;
                  onebelow := r^.ancestor^.stateone;
                  END;
               FOR i := 1 TO words DO BEGIN
                  dh := dohyp[i];
                  s0 := r^.statezero[i];
                  s1 := r^.stateone[i];
                  a0 := zerobelow[i];
                  a1 := onebelow[i];
                  IF NOT r^.tip
                  THEN BEGIN
                     wa := wagner[i];
                     l0 := r^.leftdesc^.statezero[i];
                     l1 := r^.leftdesc^.stateone[i];
                     r0 := r^.rtdesc^.statezero[i];
                     r1 := r^.rtdesc^.stateone[i];
                     s0 := wa * (a0 * l0 + a0 * r0 + l0 * r0)
                           + dh * (fullset-wa) * s0;
                     s1 := wa * (a1 * l1 + a1 * r1 + l1 * r1)
                           + dh * (fullset-wa) * s1;
                     temp := fullset - (s0 + s1 + l1 + l0 + r1 + r0);
                     s0 := s0 + temp * a0;
                     s1 := s1 + temp * a1;
                     r^.statezero[i] := s0;
                     r^.stateone[i] := s1;
                     END;
                  maybe := maybe OR ((dh * (s0 + s1)) <> (a0 + a1));
                  nonzero := nonzero OR ((s1 * a0 + s0 * a1) <> []);
                  END;
               hyprint;
               IF NOT r^.tip
               THEN BEGIN
                  hyptrav (r^.leftdesc);
                  hyptrav (r^.rtdesc);
                  END;
            END; (* hyptrav *)

         BEGIN (* hypstates *)
            FOR i := 1 TO words DO BEGIN
               zeroanc[i] := [];
               oneanc[i] := [];
               END;
            unknown := FALSE;
            FOR i := 1 TO chars DO BEGIN
               j := (i-1) DIV bits + 1;
               k := (i-1) MOD bits + 1;
               IF (guess[i] = '0')
               THEN zeroanc[j] := zeroanc[j] + [k];
               IF (guess[i] = '1')
               THEN oneanc[j] := oneanc[j] + [k];
               unknown := unknown OR ((NOT (k IN wagner[j]))
                                     AND (guess[i] = '?'));
               END;
            FOR i := 1 TO words DO
               dohyp[i] := wagner[i] + zeroanc[i] + oneanc[i];
            filltrav (root);
            WRITE(OUTPUT, ' From    To     Any Steps?    ');
            WRITELN(OUTPUT, 'State at upper node');
            WRITE(OUTPUT, '                             ');
            WRITELN(OUTPUT, ' ( . means same as in the node',
                                    ' below it on tree)');
            WRITELN(OUTPUT);
            hyptrav (root);
         END; (* hypstates *)

         PROCEDURE treeout (p : nodeptr);
            (* write out file with representation of final tree *)
            VAR i, n : INTEGER;
               c : CHAR;
         BEGIN (* treeout *)
            IF p^.tip
            THEN BEGIN
               n := 0;
               FOR i := 1 TO nmlngth DO
                  IF name[p^.index, i] <> ' '
                  THEN n := i;
               FOR i := 1 TO n DO BEGIN
                  c := name[p^.index, i];
                  IF c = ' '
                  THEN c := '_';
                  WRITE(treefile, c);
                  END;
               END
            ELSE BEGIN
               WRITE(treefile, '(');
               treeout (p^.leftdesc);
               WRITE(treefile, ',');
               treeout (p^.rtdesc);
               WRITE(treefile, ')');
               END;
            IF p = root
            THEN WRITELN(treefile, ';');
         END; (* treeout *)

      BEGIN (* describe *)
         IF weights THEN WRITE(OUTPUT, ' weighted');
         WRITELN(OUTPUT, ' steps in each character:');
         WRITE(OUTPUT, '       ');
         FOR i := 0 TO 9 DO WRITE(OUTPUT, i:4);
         WRITELN(OUTPUT);
         WRITE(OUTPUT, '      *------------------------------------');
         WRITELN(OUTPUT, '-----');
         FOR i := 0 TO (chars DIV 10) DO BEGIN
            WRITE(OUTPUT, 10*i:6);
            WRITE(OUTPUT, '!');
            FOR j := 0 TO 9 DO BEGIN
               k := 10*i + j;
               IF (k = 0) OR (k > chars)
               THEN WRITE(OUTPUT, '    ')
               ELSE WRITE(OUTPUT, numsteps[k]+extras[k]:4);
               END;
            WRITELN(OUTPUT);
            END;
         WRITELN(OUTPUT);
         IF questions AND ((NOT noroot) OR didreroot)
         THEN BEGIN
            WRITELN(OUTPUT, ' best guesses of ancestral states:');
            WRITE(OUTPUT, '       ');
            FOR i := 0 TO 9 DO WRITE(OUTPUT, i:2);
            WRITELN(OUTPUT);
            WRITELN(OUTPUT, '      *--------------------');
            FOR i := 0 TO (chars DIV 10) DO BEGIN
               WRITE(OUTPUT, ' ');
               WRITE(OUTPUT, 10*i:5, '!');
               FOR j := 0 TO 9 DO
                  IF (10*i+j = 0) OR (10*i+j > chars)
                  THEN WRITE(OUTPUT, '  ')
                  ELSE WRITE(OUTPUT, ' ', guess[10*i+j]);
               WRITELN(OUTPUT);
               END;
            WRITELN(OUTPUT);
            END;
         hypstates;
         WRITELN(OUTPUT);
         IF trout
         THEN treeout (root);
      END; (* describe *)

   BEGIN (* maketree *)
      fullset := [1..bits];
      WRITELN(OUTPUT, ' How many');
      WRITELN(OUTPUT, ' trees looked                ',
               '                       Approximate');
      WRITELN(OUTPUT, ' at so far      Length of    ',
               '    How many           percentage');
      WRITELN(OUTPUT, ' (multiples     shortest tree',
               '    trees this long    searched');
      WRITELN(OUTPUT, ' of ', howoften:4, '):', '      found so far ',
               '    found so far       so far');
      WRITELN(OUTPUT, ' ----------     ------------ ',
               '    ------------       ------------');
      done := FALSE;
      mults := 0;
      examined := 0;
      nextree := 1;
      root := treenode[1];
      firsttime := TRUE;
      FOR i := 1 TO spp DO
         added[i] := FALSE;
      added[1] := TRUE;
      order[1] := 1;
      k := 2;
      fracdone := 0.0;
      fracinc := 1.0;
      bestyet := -1;
      addit (k);
      IF done
      THEN BEGIN
         WRITELN(OUTPUT, ' Search broken off!  Not guaranteed to');
         WRITELN(OUTPUT, '  have found the most parsimonious');
         WRITELN(OUTPUT, '  trees, but here is what we found:');
         END;
      WRITELN(OUTPUT);
      WRITELN(OUTPUT, ' requires a total of ',bestyet:18:3);
      WRITELN(OUTPUT);
      IF (nextree = 2)
      THEN WRITELN(OUTPUT, ' One most parsimonious tree found:')
      ELSE WRITELN(OUTPUT, nextree-1:6, ' trees in all found');
      IF nextree > maxtrees + 1
      THEN BEGIN
         WRITELN(OUTPUT, ' here are the first', maxtrees:4, ' of them');
         nextree := maxtrees + 1;
         END;
      WRITELN(OUTPUT);
      FOR i := 1 TO spp DO
         added[i] := TRUE;
      FOR i := 1 TO nextree-1 DO BEGIN
         FOR j := k TO spp DO
            add(treenode[bestrees[i][j]], treenode[bestorders[i][j]],
                                           treenode[spp+j-1]);
         IF noroot
         THEN reroot (treenode[outgrno]);
         didreroot := outgropt AND noroot;
         evaluate (root);
         printree;
         describe;
         FOR j := k TO spp DO
            remove (treenode[bestorders[i][j]], treenode[spp+j-1]);
         END;
      IF trout
      THEN BEGIN
         WRITELN(OUTPUT, ' Trees also written onto file');
         WRITELN(OUTPUT);
         END;
   END; (* maketree *)

BEGIN (* Penny's branch-and-bound method *)
   (* Reads in the number of species, number of characters,
      options and data.  Then finds all most parsimonious trees *)
   (* ASSIGN(infile, 'INFILE');  Turbo Pascal only *)
   RESET(infile);
   doinput;
   IF trout
   THEN BEGIN
      (* ASSIGN(treefile, 'TREEFILE');  Turbo Pascal only *)
      REWRITE(treefile);
      END;
   maketree;
(* CLOSE(infile);  -- may need this here *)
   IF trout
   THEN BEGIN
      (* CLOSE(treefile);  -- may need this here *)
      END;
END. (* Penny's branch-and-bound method *)

