PROGRAM Plotgram (INPUT, OUTPUT, treefile, parmfile, fontfile, plotfile);
   (* Copyright (c) 1986, 1987, 1988 by Christopher A. Meacham,
      University Herbarium, University of California, Berkeley,
      California 94720, U.S.A.

      Permission is granted to copy, distribute, and modify this
      program provided that (1) this copyright message is not removed
      and (2) no fee is charged for this program.

      PHYLIP version 3.1.
      PLOTGRAM Version 1.4, January, 1988.
      Incorporates modifications and bug-fixes suggested by
      D. L. Swofford and J. M. Humphries.  Also modifications
      by Joe Felsenstein.  *)
   CONST maxnodes = 300;
      maxlabels = 100;
      maxnch = 100;
      fontsize = 3200;    (* Fontsize of font1: 1165; font2: 2162;  *)
      bignum = 1000000.0; (*  font3: 3190; font4: 2173; font5: 3200 *)
      ebcdic = FALSE;  (* indicates EBCDIC code for Tek driver *)
   TYPE fontname = (font1, font2, font3, font4, font5);
      modetype = (yes, no);
      diagramtype = (phylogram, phenogram, cladogram);
      verticaltype = (levels, lengths, top, bottom, distance);
      horizontaltype = (straightline, nearest, furthest);
      drawingtype = (rectangle, slant);
      sidetype = (rightside, leftside);
      posoptiontype = (atroot, right, left, above, below, center);
      penstatustype = (penup, pendown);
      plottertype = (tektronix, hp7470, bauschlomb, laser, other);
      fonttype = ARRAY [1..fontsize] OF INTEGER;
      plotstring = ARRAY [1..maxnch] OF CHAR;
      nodeptr = ^treenode;
      labelptr = ^labelrecord;
      treenode = RECORD
         ancestor, sibling, descendant, lastdesc,
            left, right : nodeptr;
         nodenum, fromtop, frombottom : INTEGER;
         level, x, y, xstem, ystem : REAL;
         nodedone, popped : BOOLEAN;
         labels : labelptr;
         END;
      labelrecord = RECORD
         nextlab : labelptr;
         lab : plotstring;
         tipnum, nch  : INTEGER;
         xlab, ylab : REAL;
         END;
      (*  Begin section added by DLS/JMH for "TranslateLabel" option *)
      translabelptr = ^translabelrecord;
      translabelrecord = RECORD
         next : translabelptr;
         old, knew : plotstring;
         nchold, nchnew : INTEGER
         END;
      (*  End section added by DLS/JMH for "TranslateLabel" option *)
      (*  stype = STRING[255];  Turbo Pascal only *)

   VAR nnodes, nlabels, ntips, maxdepth, xnow, ynow,
         oldx, oldy, oldxhigh, oldxlow,
         oldyhigh, oldylow, popval : INTEGER;
      pi, rotation, xposition, yposition,
         xunitspercm, yunitspercm, aspectratio,
         xsize, ysize, xrange, yrange, labelheight,
         labelrotation, treeheight, treewidth, spacing,
         treeline, labelline,
         savelabelheight, savespacing,   (* DLS/JMH *)
         gap : REAL;
      parmfile, treefile, fontfile, plotfile : TEXT;
      labelspresent, fontloaded, parmsok : BOOLEAN;
      diagram : diagramtype;
      vertical : verticaltype;
      horizontal : horizontaltype;
      drawing : drawingtype;
      fonttoload : fontname;
      font : fonttype;
      sizemode, leveltips, penchange, widthmode, popmode : modetype;
      plotter : plottertype;
      xoption, yoption : posoptiontype;
      root, garbnode : nodeptr;
      garblabel : labelptr;
      nodelist : ARRAY [1..maxnodes] OF nodeptr;
      labellist : ARRAY [1..maxlabels] OF labelptr;
      escape : CHAR;     (* DLS/JMH -- used in new drivers *)
      nmoves, payge : INTEGER;  (* DLS/JMH -- used in new drivers *)
      eb : ARRAY[0..127] OF INTEGER;  (*  DLS/JMH -- used in modification
                                       of Tektronix driver for EBCDIC  *)
      cchex : ARRAY [0..15] OF CHAR;
      (*  Begin section added by DLS/JMH for "TranslateLabel" and
          "Select" options *)
      translate, selekt : modetype;
      firsttranslabel, lasttranslabel, translabel : translabelptr;
      ntrans : INTEGER;
      anerror : BOOLEAN;
      (*  End section added by DLS/JMH for "TranslateLabel" option *)

   PROCEDURE initparms;
      CONST nabs = 28;
      VAR ab1, ab2 : ARRAY [0..nabs] OF CHAR;

      PROCEDURE plotrparms(ch : CHAR);
      BEGIN (* plotrparms *)
         CASE ch OF
            'T': BEGIN
               plotter := tektronix;
               xunitspercm := 50.0;
               yunitspercm := 50.0;
               xsize := 20.46;
               ysize := 15.6;
               xposition := 10.23;
               yposition := 0.0;
               xoption := center;
               yoption := above;
               rotation := 0.0;
               END;
            'H': BEGIN
               plotter := hp7470;
               penchange := yes;
               xunitspercm := 400.0;
               yunitspercm := 400.0;
               xsize := 24.0;
               ysize := 18.0;
               xposition := 12.89;
               yposition := 9.905;
               xoption := center;
               yoption := center;
               rotation := 0.0;
               END;
            'B': BEGIN
               plotter := bauschlomb;
               penchange := yes;
               xunitspercm := 100.0;
               yunitspercm := 100.0;
               xsize := 24.5;
               ysize := 17.5;
               xposition := 12.5;
               yposition := 20.0;
               xoption := center;
               yoption := above;
               rotation := 0.0;
               END;
            'L': BEGIN
               plotter := laser;
               penchange := yes;
               xunitspercm := 28.346456693;
               yunitspercm := 28.346456693;
               xsize := 15.24;
               ysize := 22.86;
               xposition := 11.43;
               yposition := 13.97;
               xoption := center;
               yoption := center;
               rotation := -90.0;
               END;
            'O': BEGIN
               plotter := other;
               (* initial parameter settings for a new
                  plotter go here *)
               END
            END;
      END; (* plotrparms *)

      PROCEDURE readparms;
         VAR parmnum : INTEGER;
            ch, c1, c2 : CHAR;

         PROCEDURE uppercase (VAR ch : CHAR);

            FUNCTION letter (ch : CHAR) : BOOLEAN;
            BEGIN (* letter -- works for both ASCII and EBCDIC codes *)
               letter := ((ch >= 'A') AND (ch <= 'I'))
                         OR ((ch >= 'J') AND (ch <= 'R'))
                         OR ((ch >= 'S') AND (ch <= 'Z'))
                         OR ((ch >= 'a') AND (ch <= 'i'))
                         OR ((ch >= 'j') AND (ch <= 'r'))
                         OR ((ch >= 's') AND (ch <= 'z'));
            END; (* letter *)

        BEGIN (* uppercase *)
           IF letter(ch) AND ((('A' < 'a') AND (ch >= 'a'))
                           OR (('A' > 'a') AND (ch < 'A')))
           THEN ch := CHR(ORD(ch) + ORD('A') - ORD('a'));
        END; (* uppercase *)

      PROCEDURE getp(VAR ch : CHAR);
         BEGIN (* getp *)
            ch := ' ';
            REPEAT
               READ(parmfile, ch)
               UNTIL (ch <> ' ') OR EOLN(parmfile);
            uppercase(ch);
         END; (* getp *)

         FUNCTION setmode : modetype;
         BEGIN (* setmode *)
            getp(ch);
            CASE ch OF
               'Y': setmode := yes;
               'N': setmode := no
               END;
         END; (* setmode *)

      BEGIN (* readparms *)
         WHILE NOT EOF(parmfile) DO BEGIN
            c1 := ' ';
            c2 := ' ';
            WHILE NOT EOLN(parmfile) AND (c1=' ') DO
               READ(parmfile, c1);
            IF c1 <> '*' THEN BEGIN      (* added by DLS/JMH *)
               IF NOT EOLN(parmfile) THEN READ(parmfile, c2);
               REPEAT READ(parmfile, ch) UNTIL EOLN(parmfile)
                                                      OR (ch=' ');
               uppercase(c1);
               uppercase(c2);
               parmnum := 0;
               WHILE NOT((c1=ab1[parmnum]) AND (c2=ab2[parmnum])) AND
                  (parmnum<=nabs) DO parmnum := parmnum+1;
               IF parmnum<=nabs
                  THEN CASE parmnum OF
                     1: BEGIN
                        getp(ch);
                        plotrparms(ch)
                        END;
                     2: BEGIN
                        getp(ch);
                        CASE ch OF
                           'C': diagram := cladogram;
                           'E': BEGIN
                              diagram := phenogram;
                              vertical := levels;
                              horizontal := nearest;
                              drawing := rectangle;
                              widthmode := no
                              END;
                           'Y': BEGIN
                              diagram := phylogram;
                              vertical := lengths;
                              horizontal := nearest;
                              drawing := rectangle;
                              leveltips := no;
                              widthmode := no
                              END
                           END;
                        END;
                     3: BEGIN
                        getp(ch);
                        CASE ch OF
                           'V': vertical := levels;
                           'G': vertical := lengths;
                           'T': vertical := top;
                           'B': vertical := bottom;
                           'D': vertical := distance
                           END;
                        END;
                     4: BEGIN
                        getp(ch);
                        CASE ch OF
                           'S': horizontal := straightline;
                           'N': horizontal := nearest;
                           'F': horizontal := furthest
                           END;
                        END;
                     5: leveltips := setmode;
                     6: BEGIN
                        getp(ch);
                        CASE ch OF
                           'R': drawing := rectangle;
                           'S': drawing := slant
                           END;
                        END;
                     7: READ(parmfile, rotation);
                     8: BEGIN
                        getp(ch);
                        CASE ch OF
                           '1': fonttoload := font1;
                           '2': fonttoload := font2;
                           '3': fonttoload := font3;
                           '4': fonttoload := font4;
                           '5': fonttoload := font5
                           END;
                        END;
                     9: READ(parmfile, xunitspercm);
                     10: READ(parmfile, yunitspercm);
                     11: READ(parmfile, xsize);
                     12: READ(parmfile, ysize);
                     13: READ(parmfile, xposition);
                     14: READ(parmfile, yposition);
                     15: BEGIN
                        getp(ch);
                        CASE ch OF
                           'O': xoption := atroot;
                           'C': xoption := center;
                           'R': xoption := right;
                           'L': xoption := left
                           END;
                        END;
                     16: BEGIN
                        getp(ch);
                        CASE ch OF
                           'O': yoption := atroot;
                           'C': yoption := center;
                           'A': yoption := above;
                           'B': yoption := below
                           END;
                        END;
                     17: penchange := setmode;
                     18: widthmode := setmode;
                     19: READ (parmfile, treeheight);
                     20: READ (parmfile, treewidth);
                     21: READ (parmfile, labelheight);
                     22: READ (parmfile, spacing);
                     23: READ (parmfile, gap);
                     24: READ (parmfile, labelrotation);
                     25: popmode := setmode;
                     26: translate := setmode;  (* DLS/JMH *)
                     27: selekt := setmode;     (* DLS/JMH *)
                     28: sizemode := setmode
                     END
                  ELSE BEGIN
                     WRITELN(OUTPUT, ' PARAMETER LABEL THAT BEGINS ''',
                        c1, c2, ''' IS NOT A LEGITIMATE LABEL');
                     parmsok := FALSE;
                     READLN(parmfile);
                     END;
               END;
            IF EOLN(parmfile) THEN READLN(parmfile);
            END;
      END; (* readparms *)

   BEGIN (* initparms *)
      ab1[0] := ' '; ab2[0] := ' ';
      ab1[1] := 'P'; ab2[1] := 'L'; ab1[2] := 'D'; ab2[2] := 'I';
      ab1[3] := 'V'; ab2[3] := 'E'; ab1[4] := 'H'; ab2[4] := 'O';
      ab1[5] := 'T'; ab2[5] := 'I'; ab1[6] := 'D'; ab2[6] := 'R';
      ab1[7] := 'R'; ab2[7] := 'O'; ab1[8] := 'F'; ab2[8] := 'O';
      ab1[9] := 'X'; ab2[9] := 'U'; ab1[10] := 'Y'; ab2[10] := 'U';
      ab1[11] := 'X'; ab2[11] := 'S'; ab1[12] := 'Y'; ab2[12] := 'S';
      ab1[13] := 'X'; ab2[13] := 'P'; ab1[14] := 'Y'; ab2[14] := 'P';
      ab1[15] := 'X'; ab2[15] := 'O'; ab1[16] := 'Y'; ab2[16] := 'O';
      ab1[17] := 'P'; ab2[17] := 'E'; ab1[18] := 'W'; ab2[18] := 'I';
      ab1[19] := 'T'; ab2[19] := 'H'; ab1[20] := 'T'; ab2[20] := 'W';
      ab1[21] := 'L'; ab2[21] := 'H'; ab1[22] := 'S'; ab2[22] := 'P';
      ab1[23] := 'G'; ab2[23] := 'A'; ab1[24] := 'L'; ab2[24] := 'R';
      ab1[25] := 'P'; ab2[25] := 'O';
      ab1[26] := 'T'; ab2[26] := 'R';     (* DLS/JMH *)
      ab1[27] := 'S'; ab2[27] := 'E';     (* DLS/JMH *)
      ab1[28] := 'S'; ab2[28] := 'I';
      garbnode := NIL;
      garblabel := NIL;
      diagram := cladogram;
      vertical := distance;
      horizontal := straightline;
      leveltips := yes;
      translate := no;        (* DLS/JMH *)
      selekt := no;           (* DLS/JMH *)
      popmode := no;
      drawing := slant;
      widthmode := yes;
      pi := 3.1415926536;
      parmsok := TRUE;
      labelspresent := FALSE;
      fontloaded := FALSE;
      penchange := no;
      sizemode := yes;
      fonttoload := font1;
      treeheight := 12.5;
      treewidth := 25.5;
      labelrotation := 0.0;
      labelheight := 0.5;
      spacing := 1.0;
      gap := 1.0;
      plotrparms('H'); (* default plotter designation *)
      readparms;
      IF popmode = yes
         THEN popval := 1
         ELSE popval := 0;
      gap := 0.5 * gap * labelheight;
      spacing := 0.75 * spacing * labelheight;
      rotation := rotation * pi/180.0;
      labelrotation := rotation - pi/2.0 + labelrotation*pi/180.0;
      aspectratio := yunitspercm/xunitspercm;
      savespacing := spacing;  (* DLS/JMH -- needed for multiple plots *)
      savelabelheight := labelheight;    (* DLS/JMH -- ditto *)
   END; (* initparms *)

   PROCEDURE ReadTranslateTable (VAR first, last: translabelptr);
      (* written by David L. Swofford and Julian M. Humphries *)
      VAR ch : CHAR;
         i : INTEGER;

      PROCEDURE GetLabel (mode: INTEGER; VAR lab: plotstring;
                          VAR nch: INTEGER);
         VAR done : BOOLEAN;
         BEGIN (* GetLabel *)
            (*  If some of this code looks strange, it's because of the    *)
            (*  funny lookahead implementation of EOLN in Turbo Pascal.    *)
            (*  This works on both Turbo and other compilers.              *)
            REPEAT
               READ(treefile, ch)
               UNTIL (ch <> ' ') OR EOLN(treefile);
            anerror := (ch = ' ') AND EOLN(treefile);
            IF NOT anerror THEN BEGIN
               nch := 0;
               REPEAT
                  IF nch < maxnch THEN BEGIN
                     nch := nch + 1;
                     lab[nch] := ch
                     END;
                  IF EOLN(treefile)
                     THEN done := TRUE
                     ELSE BEGIN
                        READ(treefile, ch);
                        done := (ch = ' ')
                        END
                  UNTIL done;
               anerror := (mode = 1) AND EOLN(treefile);
               IF anerror
               THEN WRITELN (OUTPUT, ' ERROR IN LABEL TRANSLATION TABLE')
               END
         END; (* GetLabel *)

      BEGIN (* ReadTranslateTable *)
         READLN(treefile, ntrans);
         first := NIL;
         FOR i := 1 TO ntrans DO BEGIN
            NEW(translabel);
            WITH translabel^ DO BEGIN
               ch := ' ';
               GetLabel(1, old, nchold);
               IF NOT anerror THEN BEGIN
                  GetLabel(2, knew, nchnew);
                  IF NOT anerror THEN BEGIN
                     READLN(treefile);
                     IF first = NIL
                        THEN first := translabel
                        ELSE last^.next := translabel;
                     last := translabel;
                     next := NIL
                     END
                  END
               END
            END;
         firsttranslabel := first;
         lasttranslabel := last;
      END; (* ReadTranslateTable *)

   PROCEDURE newnode(VAR p : nodeptr);
   BEGIN (* newnode *)
      IF garbnode <> NIL
         THEN BEGIN
            p := garbnode;
            garbnode := garbnode^.descendant;
            END
         ELSE NEW(p);
      nnodes := nnodes + 1;
      nodelist[nnodes] := p;
      p^.nodenum := nnodes;
      p^.ancestor := NIL;
      p^.descendant := NIL;
      p^.sibling := NIL;
      p^.lastdesc := NIL;
      p^.labels := NIL;
      p^.nodedone := FALSE;
      p^.popped := FALSE;
      p^.level := 0.0;
      p^.ystem := 0.0;
   END; (* newnode *)

   PROCEDURE newlabel(VAR lp : labelptr);
   BEGIN (* newlabel *)
      IF garblabel <> NIL
         THEN BEGIN
            lp := garblabel;
            garblabel := garblabel^.nextlab;
            END
         ELSE NEW(lp);
      nlabels := nlabels+1;
      labellist[nlabels] := lp;
      lp^.nextlab := NIL;
      lp^.nch := 0;
   END; (* newlabel *)

   PROCEDURE readtree;
      VAR ch : CHAR;

      PROCEDURE addnode(anc : nodeptr; VAR linktonode,
         anclastdesc : nodeptr);
         CONST blank = ' ';
            point = '.';
            comma = ',';
            colon = ':';
            leftp = '(';
            rightp = ')';
            semi = ';';
            under = '_';
         VAR node : nodeptr;
            doingdescendants : BOOLEAN;

         PROCEDURE getch(VAR ch : CHAR);
         BEGIN (* getch *)
            IF EOLN(treefile) THEN READLN(treefile);
            IF EOF(treefile)
               THEN BEGIN
                  WRITELN(OUTPUT, '  NO '';'' AT END OF TREE ',
                     'DESCRIPTION, '';'' ASSUMED');
                  ch := ';';
                  END
               ELSE BEGIN
                  IF EOLN(treefile) THEN READLN(treefile);
                  READ(treefile, ch);
                  END;
         END; (* getch *)

         PROCEDURE processnodelabel(node : nodeptr);
            (*  DLS/JMH -- This routine heavily modified for the
                "TranslateLabel" option. *)
            VAR lp, lp2 : labelptr;
               lab: plotstring;        (* DLS/JMH *)
               i, nch : INTEGER;          (* DLS/JMH *)

            PROCEDURE TranslateLabel(VAR lab: plotstring;
                                     VAR nch: INTEGER);
               (* DLS/JMH -- PROCEDURE TranslateLabel *)
               VAR match : BOOLEAN;
                  i : integer;
               BEGIN (* TranslateLabel *)
                  translabel := firsttranslabel;
                  REPEAT
                     match := (nch = translabel^.nchold);
                     IF match THEN BEGIN
                        i := 0;
                        REPEAT
                           i := i + 1;
                           match := (translabel^.old[i] = lab[i]);
                           UNTIL (NOT match) OR (i = nch)
                        END;
                        IF (NOT match) THEN
                           translabel := translabel^.next
                     UNTIL match OR (translabel = NIL);
                  IF match THEN BEGIN
                     nch := translabel^.nchnew;
                     FOR i := 1 TO nch DO
                        lab[i] := translabel^.knew[i]
                     END
               END; (* TranslateLabel *)

         BEGIN (* processnodelabel *)
            nch := 0;                  (* DLS/JMH *)
            newlabel(lp);
            IF node^.labels=NIL
               THEN BEGIN
                  labelspresent := TRUE;
                  node^.labels := lp;
                  END
               ELSE BEGIN
                  lp2 := node^.labels;
                  WHILE lp2^.nextlab <> NIL DO lp2 := lp2^.nextlab;
                  lp2^.nextlab := lp;
                  END;
            REPEAT
               (*  Begin section added by DLS/JMH *)
               IF nch < maxnch THEN BEGIN
                  nch := nch + 1;
                  lab[nch] := ch;
                  END;
                  getch(ch);
               UNTIL (ch=blank) OR (ch=leftp) OR (ch=comma) OR
                  (ch=rightp) OR (ch=colon) OR (ch=semi);
               IF translate = yes THEN translatelabel(lab, nch);
               FOR i := 1 TO nch DO
                  IF lab[i] = under
                     THEN lp^.lab[i] := blank
                     ELSE lp^.lab[i] := lab[i];
               lp^.nch := nch
               (*  End section added by DLS/JMH *)
         END; (* processnodelabel *)

         PROCEDURE processlength(node : nodeptr);
            VAR digit, ordzero, ordnine : INTEGER;
               zzzvalue, divisor : REAL; (* "value" is reserved in VS-Pascal *)
               pointread : BOOLEAN;
         BEGIN (* processlength *)
            ordzero := ORD('0');
            ordnine := ORD('9');
            pointread := FALSE;
            zzzvalue := 0.0;
            divisor := 1.0;
            REPEAT getch(ch) UNTIL ch <> blank;
            digit := ORD(ch)-ordzero;
            WHILE ((digit >= 0) AND (digit <= 9)) OR (ch=point)
               DO BEGIN
               IF ch = point
                  THEN pointread := TRUE
                  ELSE BEGIN
                     zzzvalue := zzzvalue*10.0 + digit;
                     IF pointread THEN divisor := divisor*10.0;
                     END;
               getch(ch);
               digit := ORD(ch)-ordzero;
               END;
            node^.level := zzzvalue/divisor;
         END; (* processlength *)

      BEGIN (* addnode *)
         doingdescendants := FALSE;
         newnode(node);
         linktonode := node;
         anclastdesc := node;
         node^.ancestor := anc;
         WHILE ch=blank DO getch(ch);
         WHILE (ch <> semi) AND (((ch <> comma) AND
            (ch <> rightp)) OR doingdescendants)
            DO BEGIN
            IF ch = leftp
               THEN BEGIN
                  doingdescendants := TRUE;
                  getch(ch);
                  addnode(node, node^.descendant,
                     node^.lastdesc);
                  END
               ELSE IF ch = comma
                  THEN BEGIN
                     getch(ch);
                     addnode(node, node^.lastdesc^.sibling,
                        node^.lastdesc);
                     END
                  ELSE IF ch = rightp
                     THEN BEGIN
                        doingdescendants := FALSE;
                        REPEAT getch(ch) UNTIL ch <> blank;
                        END
                     ELSE IF ch = colon
                        THEN processlength(node)
                        ELSE IF ch <> semi
                           THEN processnodelabel(node);
            WHILE ch = blank DO getch(ch);
            END;
      END; (* addnode *)

   BEGIN (* readtree *)
      ch := ' ';
      addnode(NIL, root, root);
      IF NOT EOF(treefile) THEN
      READLN(treefile);
   END; (* readtree *)

   PROCEDURE calctree(root : nodeptr);
      VAR count : INTEGER;
         fullscale, maxlevel, minlevel : REAL;

      PROCEDURE popall(p : nodeptr);
         VAR newp, prev : nodeptr;
            lp : labelptr;
      BEGIN (* popall *)
         prev := NIL;
         lp := p^.labels;
         WHILE lp <> NIL DO BEGIN
            newnode(newp);
            newp^.popped := TRUE;
            lp := lp^.nextlab;
            IF p^.descendant = NIL
               THEN p^.lastdesc := newp
               ELSE newp^.sibling := p^.descendant;
            p^.descendant := newp;
            newp^.ancestor := p;
            IF vertical = levels THEN newp^.level := p^.level;
            END;
         lp := p^.labels;
         p^.labels := NIL;
         newp := p^.descendant;
         WHILE lp <> NIL DO BEGIN
            newp^.labels := lp;
            lp := lp^.nextlab;
            newp^.labels^.nextlab := NIL;
            newp := newp^.sibling;
            END;
      END; (* popall *)

      PROCEDURE popout(p : nodeptr);
      BEGIN (* popout *)
         IF p^.descendant <> NIL
            THEN BEGIN
               popout(p^.descendant);
               IF p^.labels <> NIL THEN popall(p);
               END
            ELSE IF p^.labels^.nextlab <> NIL
                  THEN popall(p);
         IF p^.sibling <> NIL THEN popout(p^.sibling);
      END; (* popout *)

      PROCEDURE counttips(p : nodeptr);
         VAR ptr : nodeptr;
      BEGIN (* counttips *)
         p^.fromtop := 0;
         IF p = root
            THEN p^.frombottom := 0
            ELSE BEGIN
               IF p^.popped
                  THEN p^.frombottom := p^.ancestor^.frombottom +
                                                              popval
                  ELSE p^.frombottom := p^.ancestor^.frombottom + 1;
               END;
         IF p^.frombottom > maxdepth THEN maxdepth := p^.frombottom;
         IF p^.descendant = NIL
            THEN BEGIN
               ntips := ntips + 1;
               IF p^.labels<>NIL THEN p^.labels^.tipnum := ntips;
               END
            ELSE BEGIN
               counttips(p^.descendant);
               ptr := p^.descendant;
               WHILE ptr <> NIL DO BEGIN
                  IF ptr^.popped
                     THEN BEGIN
                        IF p^.fromtop < ptr^.fromtop+popval
                           THEN p^.fromtop := ptr^.fromtop+popval;
                        END
                     ELSE BEGIN
                        IF ptr^.fromtop+1 > p^.fromtop
                           THEN p^.fromtop := ptr^.fromtop+1;
                        END;
                  ptr := ptr^.sibling;
                  END;
               ptr := p^.descendant;
               WHILE ptr <> NIL DO BEGIN
                  IF ptr^.popped THEN ptr^.fromtop := p^.fromtop-popval;
                  ptr := ptr^.sibling;
                  END;
               END;
         IF p^.sibling <> NIL THEN counttips(p^.sibling);
      END; (* counttips *)

      PROCEDURE placenodes(p : nodeptr; VAR count : INTEGER);
      BEGIN (* placenodes *)
         IF p^.descendant = NIL
            THEN BEGIN
               p^.left := p;
               p^.right := p;
               count := count+1;
               p^.x := (count-1)/(ntips-1);
               CASE vertical OF
                  levels : ; 
                  lengths: ; 
                  top: p^.level := 1.0-p^.fromtop/maxdepth; 
                  bottom: p^.level := p^.frombottom/maxdepth; 
                  distance: p^.level := 0.0
                  END; 
               p^.nodedone := TRUE;
               END
            ELSE BEGIN
               placenodes(p^.descendant, count);
               p^.left := p^.descendant^.left;
               p^.right := p^.lastdesc^.right;
               IF horizontal = furthest
                  THEN p^.x := (p^.left^.x+p^.right^.x)/2.0
                  ELSE p^.x := (p^.descendant^.x+p^.lastdesc^.x)/2.0;
               CASE vertical OF
                  levels : ;
                  lengths: ;
                  top: p^.level := 1.0-p^.fromtop/maxdepth;
                  bottom: p^.level := p^.frombottom/maxdepth;
                  distance: p^.level := -(p^.right^.x-p^.left^.x)
                  END;
               END;
         IF p^.sibling <> NIL THEN placenodes(p^.sibling, count);
      END; (* placenodes *)

      PROCEDURE addlengths(p : nodeptr);
      BEGIN (* addlengths *)
         IF p <> root THEN p^.level := p^.level + p^.ancestor^.level;
         IF p^.descendant <> NIL THEN addlengths(p^.descendant);
         IF p^.sibling <> NIL THEN addlengths(p^.sibling);
      END; (* addlengths *)

      PROCEDURE lineupnodes(p : nodeptr);

         PROCEDURE singleline(p : nodeptr; side : sidetype;
               pstart, pend : nodeptr);
         BEGIN (* singleline *)
            WHILE p <> NIL DO BEGIN
               p^.x := (p^.y-pstart^.y)*(pend^.x-pstart^.x)/
                                  (pend^.y-pstart^.y)+pstart^.x;
               p^.nodedone := TRUE;
               IF side = leftside
                  THEN p := p^.descendant
                  ELSE p := p^.lastdesc;
               END;
         END; (* singleline *)

      BEGIN (* lineupnodes *)
         IF p <> NIL THEN BEGIN
            IF NOT p^.nodedone THEN BEGIN
               IF ABS(p^.ancestor^.x-p^.right^.x) >=
                       ABS(p^.ancestor^.x-p^.left^.x)
                  THEN singleline(p, rightside, p^.ancestor,
                     p^.right)
                  ELSE singleline(p, leftside, p^.ancestor,
                     p^.left);
               END;
            lineupnodes(p^.descendant);
            lineupnodes(p^.sibling);
            END;
      END; (* lineupnodes *)

      PROCEDURE levellimits(p : nodeptr);
      BEGIN (* levellimits *)
         IF p^.descendant <> NIL THEN levellimits(p^.descendant);
         IF p^.sibling <> NIL THEN levellimits(p^.sibling);
         IF p^.level > maxlevel THEN maxlevel := p^.level;
         IF p^.level < minlevel THEN minlevel := p^.level;
      END; (* levellimits *)

      PROCEDURE scalenodes(p : nodeptr);
      BEGIN (* scalenodes *)
         IF p^.descendant <> NIL
               THEN scalenodes(p^.descendant);
         IF (leveltips = yes) AND (p^.descendant = NIL)
            THEN p^.y := 1.0
            ELSE p^.y := ABS(p^.level-root^.level)/fullscale;
         IF p^.labels<>NIL THEN p^.labels^.ylab := p^.y;
         IF p^.sibling <> NIL THEN scalenodes(p^.sibling);
      END; (* scalenodes *)

      PROCEDURE setstems(p : nodeptr);
         VAR ptr : nodeptr;
      BEGIN (* setstems *)
         IF p^.descendant <> NIL THEN setstems(p^.descendant);
         p^.xstem := p^.x;
         ptr := p^.descendant;
         WHILE ptr <> NIL DO BEGIN
            ptr^.ystem := p^.y;
            ptr := ptr^.sibling;
            END;
         IF p^.sibling <> NIL THEN setstems(p^.sibling);
      END; (* setstems *)

   BEGIN (* calctree *)
      popout(root);
      counttips(root);
      count := 0;
      placenodes(root, count);
      IF vertical = lengths THEN addlengths(root);
      maxlevel := -bignum;
      minlevel := bignum;
      levellimits(root);
      fullscale := maxlevel-minlevel;
      scalenodes(root);
      IF horizontal = straightline THEN BEGIN
         root^.x := 0.5;
         root^.nodedone := TRUE;
         lineupnodes(root);
         END;
      IF drawing = rectangle THEN setstems(root);
   END; (* calctree *)

   PROCEDURE hardwarecoordinates;
      VAR i : INTEGER;
         factor, xmax, xmin, ymax, ymin, thislength,
            dsin, dcos, lsin, lcos, xtrans, ytrans,
            xcorrect, ycorrect, dummy : REAL;
         lp : labelptr;
         p : nodeptr;

      FUNCTION lengthtext(pstring : plotstring; nchars : INTEGER;
         height, compress : REAL; VAR font : fonttype) : REAL;
         VAR i, j, code : INTEGER;
            sumlength, heightfont, widthfont : REAL;
      BEGIN (* lengthext *)
         sumlength := 0.0;
         FOR i := 1 TO nchars DO BEGIN
            code := ORD(pstring[i]);
            j := 1;
            WHILE (font[j+1] <> code) AND (font[j] <> 0) DO
               j := font[j];
            IF font[j+1] = code THEN BEGIN
               heightfont := font[j+2];
               widthfont := font[j+3];
               sumlength := sumlength + (height/heightfont)
                     *widthfont;
               END;
            END;
            lengthtext := sumlength*compress;
      END; (* lengthtext *)

      PROCEDURE rotate(VAR x, y : REAL; xdisp, ydisp, xcenter, ycenter,
             sinrot, cosrot : REAL);
      BEGIN (* rotate *)
         x := xcenter+xdisp*cosrot+ydisp*sinrot;
         IF xmin>x THEN xmin := x;
         IF xmax<x THEN xmax := x;
         y := ycenter-xdisp*sinrot+ydisp*cosrot;
         IF ymin>y THEN ymin := y;
         IF ymax<y THEN ymax := y;
      END; (* rotate *)

   BEGIN (* hardwarecoordinates *)
      xmin := bignum;
      xmax := -bignum;
      ymin := bignum;
      ymax := -bignum;
      dsin := SIN(rotation);
      dcos := COS(rotation);
      lsin := SIN(labelrotation);
      lcos := COS(labelrotation);
      IF widthmode = yes
         THEN BEGIN
            factor := treewidth / ((ntips-1)*(labelheight+spacing));
            IF factor < 1.0
               THEN BEGIN
                  labelheight := labelheight*factor;
                  spacing := spacing*factor;
                  END
               ELSE spacing := (treewidth/(ntips-1)) - labelheight;
            END
         ELSE treewidth := (ntips-1)*(labelheight+spacing);
      xcorrect := -labelheight*(1+SIN(labelrotation-rotation)/2.0);
      ycorrect := -labelheight*COS(labelrotation-rotation)/2.0;
      FOR i := 1 TO nlabels DO BEGIN
         lp := labellist[i];
         lp^.xlab := lp^.tipnum*labelheight +
                           (lp^.tipnum-1)*spacing + xcorrect;
         lp^.ylab := lp^.ylab*treeheight + gap + ycorrect;
         thislength := lengthtext(lp^.lab, lp^.nch, labelheight,
               1.0, font);
         rotate(lp^.xlab,lp^.ylab,lp^.xlab,lp^.ylab,0.0,0.0,dsin,dcos);
         rotate(dummy,dummy,0.0,labelheight,lp^.xlab,lp^.ylab,
               lsin,lcos);
         rotate(dummy,dummy,thislength,labelheight,lp^.xlab,lp^.ylab,
               lsin,lcos);
         rotate(dummy,dummy,thislength,0.0,lp^.xlab,lp^.ylab,lsin,lcos);
         END;
      FOR i := 1 TO nnodes DO BEGIN
         p := nodelist[i];
         p^.x := p^.x * treewidth;
         p^.y := p^.y * treeheight;
         rotate(p^.x,p^.y,p^.x,p^.y,0.0,0.0,dsin,dcos);
         IF drawing = rectangle THEN BEGIN
            p^.xstem := p^.xstem * treewidth;
            p^.ystem := p^.ystem * treeheight;
            rotate(p^.xstem,p^.ystem,p^.xstem,p^.ystem,0.0,0.0,
                  dsin,dcos);
            END;
         END;
      xrange := xmax - xmin;
      yrange := ymax - ymin;
      IF sizemode = yes
         THEN BEGIN
            IF ysize/xsize > yrange/xrange
               THEN factor := xsize/xrange
               ELSE factor := ysize/yrange;
            END
         ELSE factor := 1.0;
      CASE xoption OF
         atroot: ;
         right: xtrans := xposition - xmin*factor;
         left: xtrans := xposition - xmax*factor;
         center: xtrans := xposition + (xrange/2.0 - xmax)*factor
         END;
      CASE yoption OF
         atroot: ;
         above: ytrans := yposition - ymin*factor;
         below: ytrans := yposition - ymax*factor;
         center: ytrans := yposition + (yrange/2.0 - ymax)*factor
         END;
      FOR i := 1 TO nlabels DO BEGIN
         lp := labellist[i];
         lp^.xlab := (lp^.xlab*factor+xtrans)*xunitspercm;
         lp^.ylab := (lp^.ylab*factor+ytrans)*xunitspercm;
         END;
      FOR i := 1 TO nnodes DO BEGIN
         p := nodelist[i];
         p^.x := (p^.x*factor+xtrans)*xunitspercm;
         p^.y := (p^.y*factor+ytrans)*xunitspercm;
         IF (drawing = rectangle) AND (i <> 1) THEN BEGIN
            p^.xstem := (p^.xstem*factor+xtrans)*xunitspercm;
            p^.ystem := (p^.ystem*factor+ytrans)*xunitspercm;
            END;
         END;
      labelheight := labelheight * factor * xunitspercm;
      labelline := labelheight * 0.05;
      treeline := labelheight * 0.2;
      xrange := xrange * factor * xunitspercm;
      yrange := yrange * factor * xunitspercm;
   END; (* hardwarecoordinates *)

   PROCEDURE loadfont(VAR font : fonttype);
      VAR i, charstart, hersheynum : INTEGER;
         ch : CHAR;
   BEGIN (* loadfont *)
      ch := 'A';
      i := 0;
      WHILE NOT(EOF(fontfile) OR (ch = ' ')) DO BEGIN
         charstart := i+1;
         READ(fontfile, ch, ch, hersheynum,
            font[charstart+2], font[charstart+3]);
         font[charstart+1] := ORD(ch);
         i := charstart+3;
         REPEAT
            IF (i-(charstart+3)) MOD 10=0 THEN READLN(fontfile);
            i := i+1;
            READ(fontfile,font[i]);
            UNTIL ABS(font[i]) >= 10000;
         READLN(fontfile);
         font[charstart] := i+1;
         END;
      font[charstart] := 0;
      WRITELN(OUTPUT, '  FONT REQUIRES', i:5, ' INTEGERS OUT OF',
         fontsize:5, ' AVAILABLE');
   END; (* loadfont *)

   PROCEDURE plot(penstatus : penstatustype; xabs, yabs : REAL);
      VAR i, n, xrel, yrel, xinc, yinc, xlast, ylast,
         xhigh, yhigh, xlow, ylow, newx, newy : INTEGER;
         quadrant : CHAR;

      PROCEDURE pout(n : INTEGER);
      BEGIN (* pout *)
         IF n >= 1000
            THEN WRITE(plotfile, n:4)
            ELSE IF n >= 100
               THEN WRITE(plotfile, n:3)
               ELSE IF n >= 10
                  THEN WRITE(plotfile, n:2)
                  ELSE IF n >= 0
                     THEN WRITE(plotfile, n:1)
                     ELSE IF n >= -9
                        THEN WRITE(plotfile, n:2)
                        ELSE IF n >= -99
                           THEN WRITE(plotfile, n:3)
                           ELSE IF n >= -999
                              THEN WRITE(plotfile, n:4)
                              ELSE WRITE(plotfile, n:5);
         END; (* pout *)

      BEGIN (* plot *)
         yabs := yabs * aspectratio;
         CASE plotter OF
            tektronix: BEGIN
               IF penstatus = penup
                  THEN WRITE(plotfile, CHR(29));
               xnow := ROUND(xabs);
               ynow := ROUND(yabs);
               xhigh := xnow DIV 32;
               yhigh := ynow DIV 32;
               xlow := xnow MOD 32;
               ylow := ynow MOD 32;
               IF NOT ebcdic
               THEN BEGIN
                  IF yhigh <> oldyhigh THEN WRITE(plotfile,
                                                    CHR(yhigh+32));
                  IF (ylow <> oldylow) OR (xhigh <> oldxhigh)
                     THEN WRITE(plotfile, CHR(ylow+96));
                  IF xhigh <> oldxhigh THEN WRITE(plotfile,
                                                    CHR(xhigh+32));
                  WRITE(plotfile, CHR(xlow+64));
                  END
               ELSE BEGIN (* DLS/JMH -- for systems that use EBCDIC coding *)
                  IF yhigh <> oldyhigh THEN WRITE(plotfile,
                                                    CHR(eb[yhigh+32]));
                  IF (ylow <> oldylow) OR (xhigh <> oldxhigh)
                     THEN WRITE(plotfile, CHR(eb[ylow+96]));
                  IF xhigh <> oldxhigh THEN WRITE(plotfile,
                                                    CHR(eb[xhigh+32]));
                  WRITE(plotfile, CHR(eb[xlow+64]));
                  END;

               (*  DLS/JMH -- Put CR-LFs in appropriate places.  This is
                              needed if output is going to a file or if
                              the terminal is treated as a file with a
                              maximum record length.                     *)
             (*   Commented out because it won't work on Unix, but
                  may be helpful if you have problems with lines in
                  plotfile being too long with Tektronix option
               nmoves := nmoves + 1;
               IF nmoves = 12 THEN BEGIN
                  nmoves := 0;
                  WRITELN(plotfile);
                  END;
             end of section commented out *)
               oldxhigh := xhigh;
               oldxlow := xlow;
               oldyhigh := yhigh;
               oldylow := ylow;
               END;
            hp7470: BEGIN
               IF penstatus=pendown
                  THEN WRITE(plotfile, 'PD')
                  ELSE WRITE(plotfile, 'PU');
               pout(ROUND(xabs));
               WRITE(plotfile, ',');
               pout(ROUND(yabs));
               WRITELN(plotfile, ';');
               END;
            bauschlomb: BEGIN
               IF penstatus = pendown
                  THEN WRITE(plotfile, 'D ')
                  ELSE WRITE(plotfile, 'U ');
               pout(ROUND(xabs));
               WRITE(plotfile, ',');
               pout(ROUND(yabs));
               WRITELN(plotfile);
               END;
            laser: BEGIN
               IF penstatus=pendown
                  THEN WRITELN(plotfile, xabs:8:2, yabs:8:2,
                     ' lineto ')
                  ELSE WRITELN(plotfile, 'stroke ', xabs:8:2, yabs:8:2,
                     ' moveto ');
               END;
            other : BEGIN
               (* code for a pen move on a new plotter goes here *)
               END
            END;
      END; (* plot *)

      PROCEDURE plotlabels;
         TYPE pentype = (treepen, labelpen);
         VAR i : INTEGER;
            lp : labelptr;

         PROCEDURE changepen(pen : pentype);
         BEGIN (* changepen *)
            CASE pen OF
               treepen: CASE plotter OF
                  tektronix: ;  (* DLS/JMH -- avoids noisy warnings *)
                  hp7470: WRITELN(plotfile, 'SP1;');
                  bauschlomb: WRITELN(plotfile, 'P1');
                  laser: BEGIN
                     WRITELN(plotfile, 'stroke ',
                        treeline:8:2, ' setlinewidth ');
                     IF drawing=slant
                        THEN WRITELN(plotfile,
                           ' 1 setlinecap 1 setlinejoin ')
                        ELSE WRITELN(plotfile,
                           ' 2 setlinecap 0 setlinejoin ');
                     END;
                  other : BEGIN
                     (* code for a new plotter goes here *)
                     END
                  END;
               labelpen: CASE plotter OF
                  tektronix: ;  (* DLS/JMH -- avoids noisy warnings *)
                  hp7470: WRITELN(plotfile, 'SP2;');
                  bauschlomb: WRITELN(plotfile, 'P2');
                  laser: BEGIN
                     WRITELN(plotfile, ' stroke',
                        labelline:8:2, ' setlinewidth ');
                     WRITELN(plotfile, '1 setlinecap 1 setlinejoin ');
                     END;
                  other : BEGIN
                     (* code for a new plotter goes here *)
                     END
                  END;
               END;
         END; (* changepen *)

         PROCEDURE plottext(pstring : plotstring; nchars : INTEGER;
            height, compress : REAL; VAR x, y : REAL;
            slope : REAL; VAR font : fonttype);
            CONST xstart = 10;
               ystart = 35;
            VAR i, j, code, coord : INTEGER;
               heightfont, widthfont, xfactor, yfactor, xfont, yfont,
               xplot, yplot, sinslope, cosslope : REAL;
               penstatus : penstatustype;

            PROCEDURE plotchar(VAR place : INTEGER);
            BEGIN (* plotchar *)
               heightfont := font[place+2];
               widthfont := font[place+3];
               yfactor := height/heightfont;
               xfactor := yfactor*compress;
               place := place+3;
               REPEAT
                  place := place+1;
                  coord := font[place];
                  IF coord > 0
                     THEN penstatus := pendown
                     ELSE penstatus := penup;
                  coord := ABS(coord);
                  coord := coord MOD 10000;
                  xfont := ((coord DIV 100) - xstart)*xfactor;
                  yfont := ((coord MOD 100) - ystart)*yfactor;
                  IF slope = 0.0
                     THEN BEGIN
                        xplot := x + xfont;
                        yplot := y + yfont;
                        END
                     ELSE BEGIN
                        xplot := x + xfont*cosslope
                                    + yfont*sinslope;
                        yplot := y - xfont*sinslope
                                    + yfont*cosslope;
                        END;
                  plot(penstatus, xplot, yplot);
                  UNTIL ABS(font[place]) >= 10000;
               x := xplot;
               y := yplot;
            END; (* plotchar *)

         BEGIN (* plottext *)
            IF slope <> 0.0 THEN BEGIN
               sinslope := SIN(slope);
               cosslope := COS(slope);
               END;
            FOR i := 1 TO nchars DO BEGIN
               code := ORD(pstring[i]);
               j := 1;
               WHILE (font[j+1] <> code) AND (font[j] <> 0) DO
                  j := font[j];
               IF font[j+1] = code THEN plotchar(j);
               END;
         END; (* plottext *)

      BEGIN (* plotlabels *)
         IF penchange = yes THEN changepen(labelpen);
         FOR i := 1 TO nlabels DO BEGIN
            lp := labellist[i];
            plottext(lp^.lab,lp^.nch,labelheight,
               1.0,lp^.xlab,lp^.ylab,labelrotation,font);
            END;
         IF penchange = yes THEN changepen(treepen);
      END; (* plotlabels *)

   PROCEDURE slantdrawing(p : nodeptr);
   BEGIN (* slantdrawing *)
      IF p <> NIL THEN BEGIN
         IF p <> root THEN BEGIN
            plot(penup, p^.ancestor^.x, p^.ancestor^.y);
            plot(pendown, p^.x, p^.y);
            END;
         slantdrawing(p^.descendant);
         slantdrawing(p^.sibling);
         END;
   END; (* slantdrawing *)

   PROCEDURE rectdrawing(p : nodeptr);
   BEGIN (* rectdrawing *)
      IF p <> NIL THEN BEGIN
         IF p <> root THEN BEGIN
            IF p = p^.ancestor^.descendant
               THEN BEGIN
                  plot(penup, p^.ancestor^.lastdesc^.xstem,
                         p^.ancestor^.lastdesc^.ystem);
                  plot(pendown, p^.xstem, p^.ystem);
                  plot(pendown, p^.x, p^.y);
                  END
               ELSE BEGIN
                  plot(penup, p^.xstem, p^.ystem);
                  plot(pendown, p^.x, p^.y);
                  END;
            END;
         rectdrawing(p^.descendant);
         rectdrawing(p^.sibling);
         END;
   END; (* rectdrawing *)

   PROCEDURE initplotter;
   BEGIN (* initplotter *)
      CASE plotter OF
         tektronix: BEGIN
            oldxhigh := -1;
            oldxlow := -1;
            oldyhigh := -1;
            oldylow := -1;
            nmoves := 0;  (* DLS/JMH -- See Procedure PLOT *)
            (*  DLS/JMH -- "escape" in next statement was originally
                           CHR(27), but we need to handle EBCDIC as
                           well as ASCII.  CHR(12) is same in both codes. *)
            WRITE(plotfile, escape,  CHR(12));   (* DLS/JMH *)
            END;
         hp7470: WRITELN(plotfile, 'IN;SP1;VS10.0;');
         bauschlomb : ;
         laser: BEGIN
            IF drawing=slant
               THEN WRITELN(plotfile, ' 1 setlinecap 1 setlinejoin ')
               ELSE WRITELN(plotfile, ' 2 setlinecap 0 setlinejoin ');
            WRITELN(plotfile, treeline:8:2, ' setlinewidth ',
               'newpath ');
            END;
         other : BEGIN
            (* initialization code for a new plotter goes here *)
            END
         END;
      END; (* initplotter *)

      PROCEDURE finishplotter;
         VAR trash : CHAR;  (* DLS/JMH for ega, screen drivers *)
      BEGIN (* finishplotter *)
         CASE plotter OF
            tektronix: WRITELN(plotfile);
            hp7470: BEGIN
               plot(penup,0.0,0.0);
               WRITELN(plotfile,'SP;');
               END;
            bauschlomb : ;
            laser: BEGIN
                   WRITELN(plotfile, 'stroke showpage ');
                   WRITELN(plotfile);
                   END;
            other : BEGIN
               (* termination code for a new plotter goes here *)
               END
            END;
      END; (* finishplotter *)

   PROCEDURE garbagestack;
      VAR i : INTEGER;
   BEGIN (* garbagestack *)
      FOR i := 1 TO nnodes DO BEGIN
         nodelist[i]^.descendant := garbnode;
         garbnode := nodelist[i];
         END;
      FOR i := 1 TO nlabels DO BEGIN
         labellist[i]^.nextlab := garblabel;
         garblabel := labellist[i];
         END;
   END; (* garbagestack *)

BEGIN (* plotgram *)
   WRITELN(OUTPUT, ' PLOTGRAM VERSION 1.4');
   WRITELN(OUTPUT, ' DISTRIBUTED WITH PHYLIP 3.1');
   WRITELN(OUTPUT, ' WRITTEN BY C. MEACHAM, UNIVERSITY HERBARIUM, ',
      'UNIVERSITY OF CALIFORNIA, ');
   WRITELN(OUTPUT, '  BERKELEY, CA 94720, U.S.A');
   IF ebcdic
   THEN escape := CHR(39)      (* EBCDIC *)    (* DLS/JMH *)
   ELSE escape := CHR(27);
   IF ebcdic
   THEN BEGIN
   eb[  0]:=  0; eb[  1]:=  1; eb[  2]:=  2; eb[  3]:=  3; eb[  4]:= 55;
   eb[  5]:= 45; eb[  6]:= 46; eb[  7]:= 47; eb[  8]:= 22; eb[  9]:=  5;
   eb[ 10]:= 37; eb[ 11]:= 11; eb[ 12]:= 12; eb[ 13]:= 13; eb[ 14]:= 14;
   eb[ 15]:= 15; eb[ 16]:= 16; eb[ 17]:= 17; eb[ 18]:= 18; eb[ 19]:= 19;
   eb[ 20]:= 60; eb[ 21]:= 61; eb[ 22]:= 50; eb[ 23]:= 38; eb[ 24]:= 24;
   eb[ 25]:= 25; eb[ 26]:= 63; eb[ 27]:= 39; eb[ 28]:= 28; eb[ 29]:= 29;
   eb[ 30]:= 30; eb[ 31]:= 31; eb[ 32]:= 64; eb[ 33]:= 90; eb[ 34]:=127;
   eb[ 35]:=123; eb[ 36]:= 91; eb[ 37]:=108; eb[ 38]:= 80; eb[ 39]:=125;
   eb[ 40]:= 77; eb[ 41]:= 93; eb[ 42]:= 92; eb[ 43]:= 78; eb[ 44]:=107;
   eb[ 45]:= 96; eb[ 46]:= 75; eb[ 47]:= 97; eb[ 48]:=240; eb[ 49]:=241;
   eb[ 50]:=242; eb[ 51]:=243; eb[ 52]:=244; eb[ 53]:=245; eb[ 54]:=246;
   eb[ 55]:=247; eb[ 56]:=248; eb[ 57]:=249; eb[ 58]:=122; eb[ 59]:= 94;
   eb[ 60]:= 76; eb[ 61]:=126; eb[ 62]:=110; eb[ 63]:=111; eb[ 64]:=124;
   eb[ 65]:=193; eb[ 66]:=194; eb[ 67]:=195; eb[ 68]:=196; eb[ 69]:=197;
   eb[ 70]:=198; eb[ 71]:=199; eb[ 72]:=200; eb[ 73]:=201; eb[ 74]:=209;
   eb[ 75]:=210; eb[ 76]:=211; eb[ 77]:=212; eb[ 78]:=213; eb[ 79]:=214;
   eb[ 80]:=215; eb[ 81]:=216; eb[ 82]:=217; eb[ 83]:=226; eb[ 84]:=227;
   eb[ 85]:=228; eb[ 86]:=229; eb[ 87]:=230; eb[ 88]:=231; eb[ 89]:=232;
   eb[ 90]:=233; eb[ 91]:=173; eb[ 92]:=224; eb[ 93]:=189; eb[ 94]:= 95;
   eb[ 95]:=109; eb[ 96]:=121; eb[ 97]:=129; eb[ 98]:=130; eb[ 99]:=131;
   eb[100]:=132; eb[101]:=133; eb[102]:=134; eb[103]:=135; eb[104]:=136;
   eb[105]:=137; eb[106]:=145; eb[107]:=146; eb[108]:=147; eb[109]:=148;
   eb[110]:=149; eb[111]:=150; eb[112]:=151; eb[113]:=152; eb[114]:=153;
   eb[115]:=162; eb[116]:=163; eb[117]:=164; eb[118]:=165; eb[119]:=166;
   eb[120]:=167; eb[121]:=168; eb[122]:=169; eb[123]:=192; eb[124]:= 79;
   eb[125]:=208; eb[126]:=161; eb[127]:=  7;
      END;
   WRITELN(OUTPUT, ' READING OF PARAMETERS BEGINS');
   (* ASSIGN(parmfile, 'PARMFILE');  Turbo Pascal only *)
   RESET(parmfile);
   initparms;
   (*  CLOSE(parmfile);   may need this here *)
   WRITELN(OUTPUT, ' READING OF PARAMETERS COMPLETED');
   IF parmsok THEN BEGIN
      (* ASSIGN(treefile, 'TREEFILE');  Turbo Pascal only *)
      RESET(treefile);
      IF translate = yes
      THEN BEGIN
         WRITELN(OUTPUT, ' READING OF TRANSLATION TABLE BEGINS');
         ReadTranslateTable (firsttranslabel, lasttranslabel);
         WRITELN(OUTPUT, ' READING OF TRANSLATION TABLE ENDS');
         END;
      (* ASSIGN(plotfile, 'PLOTFILE');   Turbo Pascal only *)
      REWRITE(plotfile);
      WHILE NOT (EOF(treefile) OR anerror) DO BEGIN
         nnodes := 0;
         ntips := 0;
         nlabels := 0;
         WRITELN(OUTPUT, ' READING OF DIAGRAM BEGINS');
         readtree;
         WRITELN(OUTPUT, ' READING OF DIAGRAM COMPLETED');
         WRITELN(OUTPUT, ' CALCULATION OF DIAGRAM BEGINS');
         calctree(root);
         WRITELN(OUTPUT, ' CALCULATION OF DIAGRAM COMPLETED');
         IF NOT(fontloaded) AND labelspresent THEN BEGIN
            WRITELN(OUTPUT, ' LOADING OF FONT BEGINS');
            (* ASSIGN(fontfile, 'FONTFILE');  Turbo Pascal only *)
            RESET(fontfile);
            loadfont(font);
            (* CLOSE(fontfile);   may need this here *)
            WRITELN(OUTPUT, ' LOADING OF FONT COMPLETED');
            fontloaded := TRUE;
            END;
         hardwarecoordinates;
         WRITELN(OUTPUT, ' DRAWING OF DIAGRAM BEGINS');
         initplotter;
         IF drawing = slant
            THEN slantdrawing(root)
            ELSE rectdrawing(root);
         plotlabels;
         finishplotter;
         WRITELN(OUTPUT, ' DRAWING OF DIAGRAM COMPLETED');
         garbagestack;
         IF NOT EOF(treefile)
         THEN IF EOLN(treefile)
            THEN READLN(treefile);
         END;
      END;
   (* CLOSE(parmfile);   may need this here *)
   (* CLOSE(fontfile);   may need this here *)
   (* CLOSE(treefile);   may need this here *)
   (* CLOSE(plotfile);   may need this here *)
   WRITELN(OUTPUT, ' EXECUTION TERMINATED');
END. (* plotgram *)
