PROGRAM Plotree(parmfile, treefile, fontfile, plotfile, INPUT, OUTPUT);
   (* Copyright (c) 1986, 1987, 1988 by Christopher A. Meacham,
      University Herbarium, University of California, Berkeley,
      California 94720, U.S.A.

      Permission is granted to copy, distribute, and modify this
      program provided that (1) this copyright message is not removed
      and (2) no fee is charged for this program.

      PHYLIP version 3.1.
      PLOTREE Version 1.4, January, 1988.
      Incorporates modifications and bug-fixes suggested by
      D. L. Swofford and J. M. Humphries.  Also modifications
      by Joe Felsenstein.  *)

   CONST bignum = 1000000.0;
      maxnodes = 300;
      maxnch = 30;
      maxlablen = 3;
      fontsize = 3200;  (* Required for font1: 1165; font2: 2162;  *)
      doswitch = TRUE;  (*   font3: 3190; font4: 2173; font5: 3200 *)
      dontswitch = FALSE;
      ebcdic = FALSE;  (* indicates EBCDIC code for Tek driver *)

   TYPE fontname = (font1, font2, font3, font4, font5);
      modetype = (yes, no);
      posoptiontype = (atroot, right, left, above, below, center);
      penstatustype = (penup, pendown);
      plottertype = (tektronix, hp7470, bauschlomb, laser, other);
      fonttype = ARRAY [1..fontsize] OF INTEGER;
      plotstring = ARRAY [1..maxlablen] OF CHAR;
      nodeptr = ^treenode;
      treenode = RECORD
         ancestor, sibling, descendant, lastdesc,
            left, right : nodeptr;
         c, nbranches, nch : INTEGER;
         nodesize, len, spread, r, theta, x, y,
         height, width : REAL;
         ndlabel : plotstring
         END;
      (*  Begin section added by DLS/JMH for "TranslateLabel" option *)
      translabelptr = ^translabelrecord;
      translabelrecord = RECORD
         next : translabelptr;
         old, knew : plotstring;
         nchold, nchnew : INTEGER
         END;
      (*  End section added by DLS/JMH for "TranslateLabel" option *)
      (* stype = STRING[255];  Turbo Pascal only *)

   VAR nnodes, xnow, ynow, oldx, oldy, oldxhigh, oldxlow, oldyhigh,
         oldylow, longestlablen : INTEGER;
      pi, smallnode, radius, absradius, diameter, spanunit,
         padding, spaceval, scrunchval, heightval,
         scrunchlimit, rotation, xposition, yposition,
         xunitspercm, yunitspercm, aspectratio,
         xsize, ysize, xrange, yrange, maxdiam, labelheight,
         savelabelheight, savespacing,   (* DLS/JMH *)
         treeline, labelline, labelsize : REAL;
      xcircle, ycircle : ARRAY [1..30] OF REAL;
      parmfile, treefile, fontfile, plotfile : TEXT;
      donetrees, labelspresent, fontloaded, candraw, parmsok : BOOLEAN;
      fonttoload : fontname;
      font : fonttype;
      sizemode, lengthmode, circles, flipmode, monitor, penchange : modetype;
      plotter : plottertype;
      xoption, yoption : posoptiontype;
      root, garbage : nodeptr;
      nodelist : ARRAY [1..maxnodes] OF nodeptr;
      translate, selekt : modetype;
      firsttranslabel, lasttranslabel, translabel : translabelptr;
      ntrans : INTEGER;
      anerror : BOOLEAN;
      escape : CHAR;     (* DLS/JMH -- used in new drivers *)
      nmoves, payge : INTEGER;  (* DLS/JMH -- used in new drivers *)
      eb : ARRAY[0..127] OF INTEGER;  (*  DLS/JMH -- used in modification
                                       of Tektronix driver for EBCDIC  *)
      cchex : ARRAY [0..15] OF CHAR;

   PROCEDURE initparms;
      CONST nabs = 26;
      VAR ab1, ab2 : ARRAY [0..nabs] OF CHAR;

      PROCEDURE plotrparms(ch : CHAR);
      BEGIN (* plotrparms *)
         CASE ch OF
            'H': BEGIN
               plotter := hp7470;
               penchange := yes;
               xunitspercm := 400.0;
               yunitspercm := 400.0;
               xsize := 24.0;
               ysize := 18.0;
               xposition := 12.85;
               yposition := 9.55;
               xoption := center;
               yoption := center;
               rotation := 90.0;
               END;
            'T': BEGIN
               plotter := tektronix;
               xunitspercm := 50.0;
               yunitspercm := 50.0;
               xsize := 20.46;
               ysize := 15.6;
               xposition := 10.23;
               yposition := 0.0;
               xoption := center;
               yoption := above;
               rotation := 0.0;
               END;
            'B': BEGIN
               plotter := bauschlomb;
               penchange := yes;
               xunitspercm := 100.0;
               yunitspercm := 100.0;
               xsize := 24.5;
               ysize := 17.5;
               xposition := 12.5;
               yposition := 20.0;
               xoption := center;
               yoption := above;
               rotation := 0.0;
               END;
            'L': BEGIN
               plotter := laser;
               penchange := yes;
               xunitspercm := 28.346456693;
               yunitspercm := 28.346456693;
               xsize := 15.24;
               ysize := 22.86;
               xposition := 11.43;
               yposition := 13.97;
               xoption := center;
               yoption := center;
               rotation := -90.0;
               END;
            'O': BEGIN
               plotter := other;
               (* initial parameter settings for a new
                  plotter go here *)
               END
            END;
         END; (* plotrparms *)

      PROCEDURE readparms;
         VAR parmnum : INTEGER;
            ch, c1, c2 : CHAR;

         PROCEDURE uppercase (VAR ch : CHAR);

            FUNCTION letter (ch : CHAR) : BOOLEAN;
            BEGIN (* letter -- works for both ASCII and EBCDIC codes *)
               letter := ((ch >= 'A') AND (ch <= 'I'))
                         OR ((ch >= 'J') AND (ch <= 'R'))
                         OR ((ch >= 'S') AND (ch <= 'Z'))
                         OR ((ch >= 'a') AND (ch <= 'i'))
                         OR ((ch >= 'j') AND (ch <= 'r'))
                         OR ((ch >= 's') AND (ch <= 'z'));
               END; (* letter *)

         BEGIN (* uppercase *)
            IF letter(ch) AND ((('A' < 'a') AND (ch >= 'a'))
                            OR (('A' > 'a') AND (ch < 'A')))
            THEN ch := CHR(ORD(ch) + ORD('A') - ORD('a'));
            END; (* uppercase *)

         PROCEDURE getp(VAR ch : CHAR);
         BEGIN (* getp *)
            ch := ' ';
            REPEAT
               READ(parmfile, ch)
               UNTIL (ch <> ' ') OR EOLN(parmfile);
            uppercase(ch);
            END; (* getp *)

         FUNCTION setmode : modetype;
         BEGIN (* setmode *)
            getp(ch);
            CASE ch OF
               'Y': setmode := yes;
               'N': setmode := no
               END;
            END; (* setmode *)

      BEGIN (* readparms *)
         WHILE NOT EOF(parmfile) DO BEGIN
            c1 := ' ';
            c2 := ' ';
            WHILE NOT EOLN(parmfile) AND (c1=' ') DO
               READ(parmfile, c1);
               IF c1 <> '*' THEN BEGIN      (* added by DLS/JMH *)
               IF NOT EOLN(parmfile) THEN READ(parmfile, c2);
               REPEAT READ(parmfile, ch) UNTIL EOLN(parmfile) OR (ch=' ');
               uppercase(c1);
               uppercase(c2);
               parmnum := 0;
               WHILE NOT((c1=ab1[parmnum]) AND (c2=ab2[parmnum])) AND
                  (parmnum<=nabs) DO parmnum := parmnum+1;
               IF parmnum <= nabs
                  THEN CASE parmnum OF
                     1: BEGIN
                        getp(ch);
                        plotrparms(ch);
                        END;
                     2: lengthmode := setmode;
                     3: sizemode := setmode;
                     4: READ(parmfile, diameter);
                     5: READ(parmfile, spanunit);
                     6: READ(parmfile, smallnode);
                     7: READ(parmfile, rotation);
                     8: BEGIN
                        getp(ch);
                        CASE ch OF
                          '1': fonttoload := font1;
                          '2': fonttoload := font2;
                          '3': fonttoload := font3;
                          '4': fonttoload := font4;
                          '5': fonttoload := font5
                          END;
                        END;
                     9: READ(parmfile, xunitspercm);
                     10: READ(parmfile, yunitspercm);
                     11: READ(parmfile, xsize);
                     12: READ(parmfile, ysize);
                     13: READ(parmfile, xposition);
                     14: READ(parmfile, yposition);
                     15: BEGIN
                        getp(ch);
                        CASE ch OF
                           'O': xoption := atroot;
                           'C': xoption := center;
                           'R': xoption := right;
                           'L': xoption := left
                           END;
                        END;
                     16: BEGIN
                        getp(ch);
                        CASE ch OF
                           'O': yoption := atroot;
                           'C': yoption := center;
                           'A': yoption := above;
                           'B': yoption := below
                           END;
                        END;
                     17: READ(parmfile, padding);
                     18: READ(parmfile, maxdiam);
                     19: monitor := setmode;
                     20: penchange := setmode;
                     21: READ(parmfile, labelsize);
                     22: translate := setmode;  (* DLS/JMH *)
                     23: selekt := setmode;     (* DLS/JMH *)
                     24: sizemode := setmode;
                     25: circles := setmode;
                     26: flipmode := setmode
                     END
                  ELSE BEGIN
                     WRITELN(OUTPUT, ' PARAMETER LABEL THAT BEGINS ''',
                        c1, c2, ''' IS NOT A LEGITIMATE LABEL');
                     parmsok := FALSE;
                     READLN(parmfile);
                     END;
               END;
            IF EOLN(parmfile) THEN READLN(parmfile);
            READLN(parmfile);
            END;
         END; (* readparms *)

   BEGIN (* initparms *)
      ab1[0] := ' '; ab2[0] := ' ';
      ab1[1] := 'P'; ab2[1] := 'L'; ab1[2] := 'L'; ab2[2] := 'E';
      ab1[3] := 'S'; ab2[3] := 'I'; ab1[4] := 'N'; ab2[4] := 'O';
      ab1[5] := 'I'; ab2[5] := 'N'; ab1[6] := 'S'; ab2[6] := 'M';
      ab1[7] := 'R'; ab2[7] := 'O'; ab1[8] := 'F'; ab2[8] := 'O';
      ab1[9] := 'X'; ab2[9] := 'U'; ab1[10] := 'Y'; ab2[10] := 'U';
      ab1[11] := 'X'; ab2[11] := 'S'; ab1[12] := 'Y'; ab2[12] := 'S';
      ab1[13] := 'X'; ab2[13] := 'P'; ab1[14] := 'Y'; ab2[14] := 'P';
      ab1[15] := 'X'; ab2[15] := 'O'; ab1[16] := 'Y'; ab2[16] := 'O';
      ab1[17] := 'P'; ab2[17] := 'A'; ab1[18] := 'M'; ab2[18] := 'A';
      ab1[19] := 'M'; ab2[19] := 'O'; ab1[20] := 'P'; ab2[20] := 'E';
      ab1[21] := 'L'; ab2[21] := 'A';
      ab1[22] := 'T'; ab2[22] := 'R';     (* DLS/JMH *)
      ab1[23] := 'S'; ab2[23] := 'E';     (* DLS/JMH *)
      ab1[24] := 'S'; ab2[24] := 'I';
      ab1[25] := 'T'; ab2[25] := 'I'; ab1[26] := 'F'; ab2[26] := 'L';
      garbage := NIL;
      pi := 3.1415926536;
      heightval := 0.9;
      scrunchval := 0.8;
      scrunchlimit := 0.01;
      padding := 1.1;
      parmsok := TRUE;
      labelspresent := FALSE;
      fontloaded := FALSE;
      monitor := no;
      translate := no;        (* DLS/JMH *)
      selekt := no;           (* DLS/JMH *)
      penchange := no;
      lengthmode := no;
      sizemode := yes;
      circles := no;
      flipmode := yes;
      fonttoload := font1;
      labelsize := 1.0;
      maxdiam := 4.0;
      smallnode := 0.3;
      spanunit := 3.0;
      diameter := 1.0;
      plotrparms('H'); (* default plotter designation *)
      readparms;
      padding := padding*diameter;
      IF (SQR(spanunit)-SQR(padding/2.0)) < 0.0
         THEN BEGIN
            WRITELN(OUTPUT, ' INTERNODE LENGTH IS TOO SMALL IN ',
               'RELATION TO NODE DIAMETER');
            parmsok := FALSE;
            END
         ELSE spaceval := 2.0*ARCTAN((padding/2.0)/SQRT(SQR(spanunit)-
            SQR(padding/2.0)));
      radius := diameter/2.0;
      labelsize := labelsize * 0.9;
      rotation := rotation * pi/180.0;
      aspectratio := yunitspercm/xunitspercm;
      savelabelheight := labelheight;    (* DLS/JMH -- ditto *)
      END; (* initparms *)

   PROCEDURE ReadTranslateTable (VAR first, last: translabelptr);
      (* written by David L. Swofford and Julian M. Humphries *)
      VAR ch : CHAR;
         i : INTEGER;

      PROCEDURE GetLabel (mode: INTEGER; VAR lab: plotstring;
                          VAR nch: INTEGER);
         VAR done : BOOLEAN;
         BEGIN (* GetLabel *)
            (*  If some of this code looks strange, it's because of the    *)
            (*  funny lookahead implementation of EOLN in Turbo Pascal.    *)
            (*  This works on both Turbo and other compilers.              *)
            REPEAT
               READ(treefile, ch)
               UNTIL (ch <> ' ') OR EOLN(treefile);
            anerror := (ch = ' ') AND EOLN(treefile);
            IF NOT anerror THEN BEGIN
               nch := 0;
               REPEAT
                  IF nch < maxnch THEN BEGIN
                     nch := nch + 1;
                     lab[nch] := ch
                     END;
                  IF EOLN(treefile)
                     THEN done := TRUE
                     ELSE BEGIN
                        READ(treefile, ch);
                        done := (ch = ' ')
                        END
                  UNTIL done;
               anerror := (mode = 1) AND EOLN(treefile);
               IF anerror
               THEN WRITELN (OUTPUT, ' ERROR IN LABEL TRANSLATION TABLE')
               END
         END; (* GetLabel *)

      BEGIN (* ReadTranslateTable *)
         READLN(treefile, ntrans);
         first := NIL;
         FOR i := 1 TO ntrans DO BEGIN
            NEW(translabel);
            WITH translabel^ DO BEGIN
               ch := ' ';
               GetLabel(1, old, nchold);
               IF NOT anerror THEN BEGIN
                  GetLabel(2, knew, nchnew);
                  IF NOT anerror THEN BEGIN
                     READLN(treefile);
                     IF first = NIL
                        THEN first := translabel
                        ELSE last^.next := translabel;
                     last := translabel;
                     next := NIL
                     END
                  END
               END
            END;
         firsttranslabel := first;
         lasttranslabel := last;
      END; (* ReadTranslateTable *)

   PROCEDURE initnodes(p, bottom : nodeptr);
   BEGIN (* initnodes *)
      IF p <> NIL
      THEN BEGIN
         initnodes(p^.descendant, bottom);
         IF p <> bottom THEN initnodes(p^.sibling, bottom);
         p^.left := p;
         p^.right := p;
         p^.height := 0.0;
         p^.width := 0.0;
         p^.r := 0.0;
         p^.theta := 0.0;
         END;
      END; (* initnodes *)

   PROCEDURE readtree;
      VAR i, count : INTEGER;
         ch : CHAR;
         node : nodeptr;

      PROCEDURE newnode(VAR p : nodeptr);
      BEGIN (* newnode *)
         IF garbage <> NIL
            THEN BEGIN
               p := garbage;
               garbage := garbage^.descendant;
               END
            ELSE NEW(p);
         count := count+1;
         nodelist[count] := p;
         p^.c := count;
         p^.ancestor := NIL;
         p^.descendant := NIL;
         p^.sibling := NIL;
         p^.lastdesc := NIL;
         p^.nch := 0;
         p^.nbranches := 0;
         p^.nodesize := 1.0;
         p^.len := spanunit;
         p^.spread := pi;
         END; (* newnode *)

      PROCEDURE addnode(anc : nodeptr; VAR linktonode,
         anclastdesc : nodeptr);
         CONST blank = ' ';
            point = '.';
            comma = ',';
            colon = ':';
            leftp = '(';
            rightp = ')';
            semi = ';';
            under = '_';
         VAR node : nodeptr;
            doingdescendants : BOOLEAN;

         PROCEDURE getch(VAR ch : CHAR);
         BEGIN (* getch *)
            IF EOF(treefile)
               THEN BEGIN
                  WRITELN(OUTPUT, '  NO '';'' AT END OF TREE ',
                     'DESCRIPTION, '';'' ASSUMED');
                  ch := ';';
                  donetrees := TRUE;
                  END
               ELSE BEGIN
                  IF EOLN(treefile) THEN READLN(treefile);
                  READ(treefile, ch);
                  END;
            END; (* getch *)

         PROCEDURE processnodelabel(node : nodeptr);
            (*  DLS/JMH -- This routine heavily modified for the
                "TranslateLabel" option. *)
            VAR lab: plotstring;        (* DLS/JMH *)
               i, nch : INTEGER;          (* DLS/JMH *)

            PROCEDURE translateLabel(VAR lab: plotstring;
                                     VAR nch: INTEGER);
               (* DLS/JMH -- PROCEDURE TranslateLabel *)
               VAR match : BOOLEAN;
                  i : integer;
               BEGIN (* translateLabel *)
                  translabel := firsttranslabel;
                  REPEAT
                     match := (nch = translabel^.nchold);
                     IF match THEN BEGIN
                        i := 0;
                        REPEAT
                           i := i + 1;
                           match := (translabel^.old[i] = lab[i]);
                           UNTIL (NOT match) OR (i = nch)
                        END;
                     IF (NOT match) THEN
                        translabel := translabel^.next
                     UNTIL match OR (translabel = NIL);
                  IF match THEN BEGIN
                     nch := translabel^.nchnew;
                     FOR i := 1 TO nch DO
                        lab[i] := translabel^.knew[i]
                     END
               END; (* translateLabel *)

         BEGIN (* processnodelabel *)
            IF node^.nch=0
               THEN BEGIN
                  labelspresent := TRUE;
                  REPEAT
                   (*  Begin section added by DLS/JMH *)
                     IF node^.nch<maxlablen THEN BEGIN
                        node^.nch := node^.nch+1;
                        IF ch = under THEN ch := blank;
                        node^.ndlabel[node^.nch] := ch;
                        END;
                     getch(ch);
                     UNTIL (ch=blank) OR (ch=leftp) OR (ch=comma) OR
                        (ch=rightp) OR (ch=colon) OR (ch=semi);
                     IF translate = yes
                     THEN translatelabel(node^.ndlabel, node^.nch);
                     FOR i := 1 TO node^.nch DO
                        IF node^.ndlabel[i] = under
                        THEN node^.ndlabel[i] := blank;
                    (*  End section added by DLS/JMH *)
                  END
               ELSE REPEAT getch(ch) UNTIL (ch=blank) OR (ch=leftp)
                  OR (ch=comma) OR (ch=rightp) OR (ch=colon) OR
                  (ch=semi);
            END; (* processnodelabel *)

         PROCEDURE processlength(node : nodeptr);
            VAR digit, ordzero : INTEGER;
               zzzvalue, divisor : REAL; (* "value" is reserved in VS-Pascal *)
               pointread : BOOLEAN;
         BEGIN (* processlength *)
            ordzero := ORD('0');
            pointread := FALSE;
            zzzvalue := 0.0;
            divisor := 1.0;
            REPEAT getch(ch) UNTIL ch <> blank;
            digit := ORD(ch)-ordzero;
            WHILE ((digit >= 0) AND (digit <= 9)) OR (ch=point)
               DO BEGIN
               IF ch = point
                  THEN pointread := TRUE
                  ELSE BEGIN
                     zzzvalue := zzzvalue*10.0 + digit;
                     IF pointread THEN divisor := divisor*10.0;
                     END;
               getch(ch);
               digit := ORD(ch)-ordzero;
               END;
            IF lengthmode = yes
               THEN node^.len := (zzzvalue/divisor)*spanunit;
            END; (* processlength *)

      BEGIN (* addnode *)
         doingdescendants := FALSE;
         newnode(node);
         linktonode := node;
         anclastdesc := node;
         node^.ancestor := anc;
         WHILE ch=blank DO getch(ch);
         WHILE (ch <> semi) AND (((ch <> comma) AND
            (ch <> rightp)) OR doingdescendants) DO BEGIN
            IF ch = leftp
               THEN BEGIN
                  doingdescendants := TRUE;
                  getch(ch);
                  addnode(node, node^.descendant,
                     node^.lastdesc);
                  node^.nbranches := node^.nbranches+1;
                  END
               ELSE IF ch = comma
                  THEN BEGIN
                     getch(ch);
                     addnode(node, node^.lastdesc^.sibling,
                        node^.lastdesc);
                     node^.nbranches := node^.nbranches+1;
                     END
                  ELSE IF ch = rightp
                     THEN BEGIN
                        doingdescendants := FALSE;
                        REPEAT getch(ch) UNTIL ch <> blank;
                        END
                     ELSE IF ch = colon
                        THEN processlength(node)
                        ELSE IF ch <> semi
                           THEN 
                           processnodelabel(node);
            WHILE ch = blank DO getch(ch);
            END;
         END; (* addnode *)

   BEGIN (* readtree *)
      count := 0;
      ch := ' ';
      addnode(NIL, root, root);
      IF NOT EOF(treefile)
      THEN READLN(treefile);
      nnodes := count;
      initnodes(root, root);
      FOR i := 1 TO nnodes DO BEGIN
         node := nodelist[i];
         IF node^.nch = 0 THEN node^.nodesize := smallnode;
         IF (node^.nch = 1) AND (node^.ndlabel[1] = ' ') THEN
            node^.nch := 0;
         END;
      IF lengthmode = yes THEN
         FOR i := 1 TO nnodes DO BEGIN
            node := nodelist[i];
            IF node <> root THEN node^.len := node^.len +
               (node^.nodesize + node^.ancestor^.nodesize)*radius;
            END;
      END; (* readtree *)

   PROCEDURE calctree(p : nodeptr);
      VAR fitok : BOOLEAN;
         current : nodeptr;

         PROCEDURE fanflip(p, bottom : nodeptr; angle : REAL;
            flip : BOOLEAN);
            VAR current : nodeptr;
         BEGIN (* fanflip *)
            IF p <> NIL
            THEN BEGIN
               fanflip(p^.descendant, bottom, angle, flip);
               IF p <> bottom THEN fanflip(p^.sibling, bottom,
                                                 angle, flip);
               IF flip
                  THEN BEGIN
                     p^.theta := angle-p^.theta;
                     current := p^.descendant;
                     p^.descendant := p^.lastdesc;
                     WHILE current <> p^.descendant DO BEGIN
                        p^.lastdesc^.sibling := current;
                        p^.lastdesc := p^.lastdesc^.sibling;
                        current := current^.sibling;
                        p^.lastdesc^.sibling := NIL;
                        END;
                     current := p^.left;
                     p^.left := p^.right;
                     p^.right := current;
                     END
                  ELSE p^.theta := angle + p^.theta;
               IF bottom^.ancestor <> NIL 
               THEN BEGIN
                  IF bottom^.ancestor^.right <> NIL
                  THEN IF p^.theta > bottom^.ancestor^.right^.theta
                     THEN bottom^.ancestor^.right := p;
                  IF bottom^.ancestor^.left <> NIL
                  THEN IF p^.theta < bottom^.ancestor^.left^.theta
                     THEN bottom^.ancestor^.left := p;
                  END;
               END;
            END; (* fanflip *)

      PROCEDURE arrange(p : nodeptr; VAR fitok : BOOLEAN);
         VAR nlong : INTEGER;
            maxheight, angle, angleused, spacing : REAL;
            flip : BOOLEAN;
            previous, current, next : nodeptr;

         PROCEDURE switchbranches(VAR ance, prev, curr, next
               : nodeptr;
            switch : BOOLEAN);
         BEGIN (* switchbranches *)
            IF switch
               THEN BEGIN
                  IF prev = ance
                     THEN ance^.descendant := next
                     ELSE prev^.sibling := next;
                  curr^.sibling := next^.sibling;
                  next^.sibling := curr;
                  prev := ance;
                  curr := ance^.descendant;
                  next := curr^.sibling;
                  END
               ELSE BEGIN
                  prev := curr;
                  curr := next;
                  next := next^.sibling;
                  END;
            END; (* switchbranches *)

      BEGIN (* arrange *)
         previous := p;
         current := p^.descendant;
         next := current^.sibling;
         WHILE next <> NIL DO
            IF (current^.height < next^.height) OR
               ((current^.height = next^.height) AND
               (current^.width < next^.width))
               THEN switchbranches(p, previous, current, next, doswitch)
               ELSE switchbranches(p, previous, current,
                                                   next, dontswitch);
         nlong := 0;
         current := p^.descendant;
         maxheight := current^.height;
         WHILE current <> NIL DO BEGIN
            IF current^.height >= heightval*maxheight
               THEN nlong := nlong + 1;
            current := current^.sibling;
            END;
         angleused := 0.0;
         current := p^.descendant;
         WHILE current <> NIL DO BEGIN
            angleused := angleused + current^.width;
            current := current^.sibling;
            END;
         spacing := (p^.spread-angleused)/(p^.nbranches+1);
         fitok := spacing >= spaceval;
         current := p^.descendant;
         IF ODD(nlong)
            THEN angle := 0.0
            ELSE angle := spacing/2.0 + current^.right^.theta;
         p^.right := current^.left;
         p^.left := current^.right;
         flip := TRUE;
         fanflip(current, current, angle, flip);
         current := current^.sibling;
         WHILE current <> NIL DO BEGIN
            IF ABS(p^.left^.theta) > ABS(p^.right^.theta)
               THEN angle := p^.right^.theta
                             + spacing + current^.right^.theta
               ELSE angle := p^.left^.theta - spacing
                             - current^.right^.theta;
            flip := angle >= 0.0;
            fanflip(current, current, angle, flip);
            current := current^.sibling;
            END;
         previous := p;
         current := p^.descendant;
         next := current^.sibling;
         WHILE next <> NIL DO
            IF (current^.theta > next^.theta)
               THEN switchbranches(p, previous, current, next, doswitch)
               ELSE switchbranches(p, previous, current, next,
                                                     dontswitch);
         p^.lastdesc := current;
         END; (* arrange *)

      PROCEDURE scrunch(p, bottom : nodeptr);
      BEGIN (* scrunch *)
         IF p <> NIL THEN BEGIN
            IF (monitor = yes) AND (p = bottom)
               THEN WRITELN(OUTPUT, '    SCRUNCHING AT NODE', p^.c:4);
            scrunch(p^.descendant, bottom);
            IF p <> bottom THEN scrunch(p^.sibling, bottom);
            p^.spread := p^.spread * scrunchval;
            IF p^.spread < scrunchlimit THEN candraw := FALSE;
            END;
         END; (* scrunch *)

      PROCEDURE polartrans(p, bottom : nodeptr; len : REAL);
         VAR pr, ptheta : REAL;
      BEGIN (* polartrans *)
         IF p <> NIL THEN BEGIN
            polartrans(p^.descendant, bottom, len);
            IF p <> bottom THEN polartrans(p^.sibling, bottom, len);
            pr := p^.r;
            ptheta := p^.theta;
            p^.r := SQRT(SQR(len) + SQR(pr)+ 2*len*pr*COS(ptheta));
            p^.theta := ARCTAN(pr*SIN(ptheta)/(pr*COS(ptheta) + len));
            IF p^.r  > bottom^.height THEN bottom^.height := p^.r;
            IF p^.theta > bottom^.right^.theta THEN bottom^.right := p;
            IF p^.theta < bottom^.left^.theta THEN bottom^.left := p;
            END;
         END; (* polartrans *)

   BEGIN (* calctree *)
      IF (p <> NIL) AND candraw THEN BEGIN
         IF monitor = yes THEN WRITELN(OUTPUT, '     AT NODE',p^.c:4);
         REPEAT
            calctree(p^.descendant);
            IF p^.descendant <> NIL
               THEN arrange(p,fitok)
               ELSE fitok := TRUE;
            IF fitok
               THEN BEGIN
                  IF p <> root THEN BEGIN
                     p^.right := p;
                     p^.left := p;
                     polartrans(p, p, p^.len);
                     p^.width := p^.right^.theta - p^.left^.theta;
                     END;
                  END
               ELSE BEGIN
                  current := p^.descendant;
                  WHILE current <> NIL DO BEGIN
                     scrunch(current, current);
                     initnodes(current, current);
                     current := current^.sibling;
                     END;
                  END;
            UNTIL fitok OR NOT candraw;
         IF candraw THEN BEGIN
            IF ABS(p^.right^.theta) > ABS(p^.left^.theta)
               THEN fanflip(p, p, 0.0, TRUE);
            calctree(p^.sibling);
            END;
         END;
      END; (* calctree *)

   PROCEDURE hardwarecoordinates;
      VAR i : INTEGER;
         factor, angle, xmax, xmin, ymax, ymin, thisx, thisy,
         thisrad, xtrans, ytrans : REAL;
         p : nodeptr;
   BEGIN (* hardwarecoordinates *)
      IF rotation <> 0.0 THEN BEGIN
         FOR i := 1 TO nnodes DO
            nodelist[i]^.theta := nodelist[i]^.theta + rotation;
         END;
      xmax := -bignum;
      xmin := bignum;
      ymax := -bignum;
      ymin := bignum;
      FOR i := 1 TO nnodes DO BEGIN
         p := nodelist[i];
         p^.x := p^.r*SIN(p^.theta);
         p^.y := p^.r*COS(p^.theta);
         thisx := p^.x;
         thisy := p^.y;
         thisrad := p^.nodesize*radius;
         IF thisx + thisrad > xmax THEN xmax := thisx + thisrad;
         IF thisx - thisrad < xmin THEN xmin := thisx - thisrad;
         IF thisy + thisrad > ymax THEN ymax := thisy + thisrad;
         IF thisy - thisrad < ymin THEN ymin := thisy - thisrad;
         END;
      xrange := xmax - xmin;
      yrange := ymax - ymin;
      IF sizemode = yes
         THEN BEGIN
            IF ysize/xsize > yrange/xrange
               THEN factor := xsize/xrange
               ELSE factor := ysize/yrange;
            IF 2.0*radius*factor > maxdiam
               THEN factor := maxdiam/(2.0*radius);
            END
         ELSE factor := 1.0;
      CASE xoption OF
         atroot: ;
         right: xtrans := xposition - xmin*factor;
         left: xtrans := xposition - xmax*factor;
         center: xtrans := xposition + (xrange/2.0 - xmax)*factor
         END;
      CASE yoption OF
         atroot: ;
         above: ytrans := yposition - ymin*factor;
         below: ytrans := yposition - ymax*factor;
         center: ytrans := yposition + (yrange/2.0 - ymax)*factor
         END;
      FOR i := 1 TO nnodes DO BEGIN
         p := nodelist[i];
         p^.x := (p^.x*factor+xtrans)*xunitspercm;
         p^.y := (p^.y*factor+ytrans)*xunitspercm
         END;
      absradius := radius * factor * xunitspercm;
      xrange := xrange * factor * xunitspercm;
      yrange := yrange * factor * xunitspercm;
      FOR i := 1 TO 30 DO BEGIN
         angle := i * pi/15.0;
         xcircle[i] := SIN(angle) * absradius;
         ycircle[i] := -COS(angle) * absradius;
         END;
      END; (* hardwarecoordinates *)

   PROCEDURE loadfont(VAR font : fonttype);
      VAR i, charstart, hersheynum : INTEGER;
         ch : CHAR;
   BEGIN (* loadfont *)
      ch := 'A';
      i := 0;
      WHILE NOT(EOF(fontfile) OR (ch=' ')) DO BEGIN
         charstart := i+1;
         READ(fontfile, ch, ch, hersheynum,
            font[charstart+2], font[charstart+3]);
         font[charstart+1] := ORD(ch);
         i := charstart+3;
         REPEAT
            IF (i-(charstart+3)) MOD 10=0 THEN READLN(fontfile);
            i := i+1;
            READ(fontfile,font[i]);
            UNTIL ABS(font[i]) >= 10000;
         READLN(fontfile);
         font[charstart] := i+1;
         END;
      font[charstart] := 0;
      WRITELN(OUTPUT, '  FONT REQUIRES', i:5, ' INTEGERS OUT OF',
         fontsize:5, ' AVAILABLE');
      END; (* loadfont *)

   PROCEDURE labelsetup;
      VAR i : INTEGER;
   BEGIN (* labelsetup *)
      longestlablen := 0;
      FOR i := 1 TO nnodes DO
         IF nodelist[i]^.nch > longestlablen
            THEN longestlablen := nodelist[i]^.nch;
      IF longestlablen > 0
         THEN labelheight := 2.0*absradius/
                                   SQRT(SQR(longestlablen)+1.0)
         ELSE labelspresent := FALSE;
      END; (* labelsetup *)

   PROCEDURE plot(penstatus : penstatustype; xabs, yabs : REAL);
      VAR i, n, xrel, yrel, xinc, yinc, xlast, ylast,
         xhigh, yhigh, xlow, ylow, newx, newy : INTEGER;
         quadrant : CHAR;

      PROCEDURE pout(n : INTEGER);
      BEGIN (* pout *)
         IF n >= 1000
            THEN WRITE(plotfile, n:4)
            ELSE IF n >= 100
               THEN WRITE(plotfile, n:3)
               ELSE IF n >= 10
                  THEN WRITE(plotfile, n:2)
                  ELSE IF n >= 0
                     THEN WRITE(plotfile, n:1)
                     ELSE IF n >= -9
                        THEN WRITE(plotfile, n:2)
                        ELSE IF n >= -99
                           THEN WRITE(plotfile, n:3)
                           ELSE IF n >= -999
                              THEN WRITE(plotfile, n:4)
                              ELSE WRITE(plotfile, n:5);
         END; (* pout *)

      BEGIN (* plot *)
         yabs := yabs * aspectratio;
         CASE plotter OF
            tektronix: BEGIN
               IF penstatus = penup
                  THEN WRITE(plotfile, CHR(29));
               xnow := ROUND(xabs);
               ynow := ROUND(yabs);
               xhigh := xnow DIV 32;
               yhigh := ynow DIV 32;
               xlow := xnow MOD 32;
               ylow := ynow MOD 32;
               IF NOT ebcdic
               THEN BEGIN
                  IF yhigh <> oldyhigh THEN WRITE(plotfile,
                                                    CHR(yhigh+32));
                  IF (ylow <> oldylow) OR (xhigh <> oldxhigh)
                     THEN WRITE(plotfile, CHR(ylow+96));
                  IF xhigh <> oldxhigh THEN WRITE(plotfile,
                                                    CHR(xhigh+32));
                  WRITE(plotfile, CHR(xlow+64));
                  END
               ELSE BEGIN (* DLS/JMH -- for systems that use EBCDIC coding *)
                  IF yhigh <> oldyhigh THEN WRITE(plotfile,
                                                    CHR(eb[yhigh+32]));
                  IF (ylow <> oldylow) OR (xhigh <> oldxhigh)
                     THEN WRITE(plotfile, CHR(eb[ylow+96]));
                  IF xhigh <> oldxhigh THEN WRITE(plotfile,
                                                    CHR(eb[xhigh+32]));
                  WRITE(plotfile, CHR(eb[xlow+64]));
                  END;

               (*  DLS/JMH -- Put CR-LFs in appropriate places.  This is
                              needed if output is going to a file or if
                              the terminal is treated as a file with a
                              maximum record length.                     *)
             (*   Commented out because it won't work on Unix, but
                  may be helpful if you have problems with lines in
                  plotfile being too long with Tektronix option
               nmoves := nmoves + 1;
               IF nmoves = 12 THEN BEGIN
                  nmoves := 0;
                  WRITELN(plotfile);
                  END;
             end of section commented out *)
               WRITE(plotfile, CHR(xlow+64));
               oldxhigh := xhigh;
               oldxlow := xlow;
               oldyhigh := yhigh;
               oldylow := ylow;
               END;
            hp7470: BEGIN
               IF penstatus=pendown
                  THEN WRITE(plotfile, 'PD')
                  ELSE WRITE(plotfile, 'PU');
               pout(ROUND(xabs));
               WRITE(plotfile, ',');
               pout(ROUND(yabs));
               WRITELN(plotfile, ';');
               END;
            bauschlomb: BEGIN
               IF penstatus = pendown
                  THEN WRITE(plotfile, 'D ')
                  ELSE WRITE(plotfile, 'U ');
               pout(ROUND(xabs));
               WRITE(plotfile, ',');
               pout(ROUND(yabs));
               WRITELN(plotfile);
               END;
            laser: BEGIN
               IF penstatus=pendown
                  THEN WRITELN(plotfile, xabs:8:2, yabs:8:2,
                     ' lineto ')
                  ELSE WRITELN(plotfile, 'stroke ', xabs:8:2, yabs:8:2,
                     ' moveto ');
               END;
            other: BEGIN
               (* code for a pen move on a new plotter goes here *)
               END
            END;
         END; (* plot *)

      PROCEDURE drawinternode(x1, y1, x2, y2, ndsize1,
                                           ndsize2 : REAL);
         VAR slope, xdiff, ydiff, xstart, ystart, xend,
                yend : REAL;

         FUNCTION trimend(thisend, farend, amount : REAL) : REAL;
         BEGIN (* trimend *)
            IF thisend > farend
               THEN trimend := thisend - amount
               ELSE trimend := thisend + amount;
            END; (* trimend *)

      BEGIN (* drawinternode *)
         IF x1 = x2
            THEN BEGIN
               xdiff := 0.0;
               ydiff := absradius;
               END
            ELSE IF y1 = y2
               THEN BEGIN
                  xdiff := absradius;
                  ydiff := 0.0;
                  END
               ELSE BEGIN
                  slope := (y2 - y1)/(x2 - x1);
                  xdiff := absradius/SQRT(SQR(slope) + 1.0);
                  ydiff := absradius/SQRT((1.0/SQR(slope)) + 1.0);
                  END;
         xstart := trimend(x1, x2, xdiff*ndsize1);
         ystart := trimend(y1, y2, ydiff*ndsize1);
         xend := trimend(x2, x1, xdiff*ndsize2);
         yend := trimend(y2, y1, ydiff*ndsize2);
         plot(penup, xstart, ystart);
         plot(pendown, xend, yend);
         END; (* drawinternode *)

      PROCEDURE plotlabel(p : nodeptr);
         TYPE pentype = (treepen, labelpen);
         VAR textlen, compress, x, y, xlab, ylab : REAL;

         PROCEDURE changepen(pen : pentype);
         BEGIN (* changepen *)
            CASE pen OF
               treepen: CASE plotter OF
                  tektronix: ;  (* DLS/JMH -- avoids noisy warnings *)
                  hp7470: WRITELN(plotfile, 'SP1;');
                  bauschlomb: WRITELN(plotfile, 'P1');
                  laser: BEGIN
                     WRITELN(plotfile, 'stroke ',
                        treeline:8:2, ' setlinewidth ');
                     WRITELN(plotfile,
                        ' 1 setlinecap 1 setlinejoin ');
                     END;
                  other : BEGIN
                     (* code for a new plotter goes here *)
                     END
                  END;
               labelpen: CASE plotter OF
                  tektronix: ;  (* DLS/JMH -- avoids noisy warnings *)
                  hp7470: WRITELN(plotfile, 'SP2;');
                  bauschlomb: WRITELN(plotfile, 'P2');
                  laser: BEGIN
                     WRITELN(plotfile, ' stroke',
                        labelline:8:2, ' setlinewidth ');
                     WRITELN(plotfile, '1 setlinecap 1 setlinejoin ');
                     END;
                  other : BEGIN
                     (* code for a new plotter goes here *)
                     END
                  END;
               END;
            END; (* changepen *)

         PROCEDURE plottext(pstring : plotstring; nchars : INTEGER;
            height, compress : REAL; VAR x, y : REAL;
            slope : REAL; VAR font : fonttype);
            CONST xstart = 10;
               ystart = 35;
            VAR i, j, code : INTEGER;
               sinslope, cosslope : REAL;

            PROCEDURE plotchar(VAR place : INTEGER);
               VAR coord : INTEGER;
                  heightfont, widthfont, xfactor, yfactor,
                  xfont, yfont, xplot, yplot : REAL;
                  penstatus : penstatustype;
            BEGIN (* plotchar *)
               heightfont := font[place+2];
               widthfont := font[place+3];
               yfactor := height/heightfont;
               xfactor := yfactor*compress;
               place := place+3;
               REPEAT
                  place := place+1;
                  coord := font[place];
                  IF coord > 0
                     THEN penstatus := pendown
                     ELSE penstatus := penup;
                  coord := ABS(coord);
                  coord := coord MOD 10000;
                  xfont := ((coord DIV 100) - xstart)*xfactor;
                  yfont := ((coord MOD 100) - ystart)*yfactor;
                  IF slope = 0.0
                     THEN BEGIN
                        xplot := x + xfont;
                        yplot := y + yfont;
                        END
                     ELSE BEGIN
                        xplot := x + xfont*cosslope
                                    + yfont*sinslope;
                        yplot := y - xfont*sinslope
                                    + yfont*cosslope;
                        END;
                  plot(penstatus, xplot, yplot);
                  UNTIL ABS(font[place]) >= 10000;
               x := xplot;
               y := yplot;
               END; (* plotchar *)

         BEGIN (* plottext *)
            IF slope <> 0.0 THEN BEGIN
               sinslope := SIN(slope);
               cosslope := COS(slope);
               END;
            FOR i := 1 TO nchars DO BEGIN
               code := ORD(pstring[i]);
               j := 1;
               WHILE (font[j+1] <> code) AND (font[j] <> 0) DO
                  j := font[j];
               IF font[j+1] = code THEN plotchar(j);
               END;
            END; (* plottext *)

         FUNCTION lengthtext(pstring : plotstring; nchars : INTEGER;
            height, compress : REAL; VAR font : fonttype) : REAL;
            VAR i, j, code : INTEGER;
               sumlength, heightfont, widthfont : REAL;
         BEGIN (* lengthtext *)
            sumlength := 0.0;
            FOR i := 1 TO nchars DO BEGIN
               code := ORD(pstring[i]);
               j := 1;
               WHILE (font[j+1] <> code) AND (font[j] <> 0) DO
                  j := font[j];
               IF font[j+1] = code THEN BEGIN
                  heightfont := font[j+2];
                  widthfont := font[j+3];
                  sumlength := sumlength + (height/heightfont)
                        *widthfont;
                  END;
               END;
               lengthtext := sumlength*compress;
            END; (* lengthtext *)

      BEGIN (* plotlabel *)
         IF penchange = yes THEN changepen(labelpen);
         compress := 1.0;
         textlen := lengthtext(p^.ndlabel, p^.nch,
            labelheight, compress, font);
         If textlen > p^.nch*labelheight THEN BEGIN
            compress := p^.nch*labelheight/textlen;
            textlen := p^.nch*labelheight;
            END;
         xlab := -textlen*labelsize/2.0;
         ylab := -labelheight*labelsize/2.0;
         x := p^.x + xlab*COS(rotation) + ylab*SIN(rotation);
         y := p^.y - xlab*SIN(rotation) + ylab*COS(rotation);
         plottext(p^.ndlabel, p^.nch, labelheight*labelsize,
                      compress, x, y, rotation, font);
         IF penchange = yes THEN changepen(treepen);
         END; (* plotlabel *)

      PROCEDURE drawtree(p : nodeptr);

         PROCEDURE drawnode(x, y, ndsize : REAL);
            VAR i : INTEGER;
         BEGIN (* drawnode *)
            plot(penup, x, y - absradius*ndsize);
            IF (circles = yes) OR (p^.descendant <> NIL) THEN
               FOR i := 1 TO 30 DO plot(pendown, x+xcircle[i]*ndsize,
                  y+ycircle[i]*ndsize);
         END; (* drawnode *)

      BEGIN (* drawtree *)
         IF p <> NIL THEN BEGIN
            IF p^.ancestor <> NIL THEN drawinternode(p^.ancestor^.x,
               p^.ancestor^.y, p^.x, p^.y, p^.ancestor^.nodesize,
               p^.nodesize);
            drawnode(p^.x, p^.y, p^.nodesize);
            IF labelspresent AND (p^.nch > 0) THEN plotlabel(p);
            drawtree(p^.descendant);
            drawtree(p^.sibling);
            END;
         END; (* drawtree *)

   PROCEDURE initplotter;
   BEGIN (* initplotter *)
      CASE plotter OF
         tektronix: BEGIN
            oldxhigh := -1;
            oldxlow := -1;
            oldyhigh := -1;
            oldylow := -1;
            nmoves := 0;  (* DLS/JMH -- See Procedure PLOT *)
            (*  DLS/JMH -- "escape" in next statement was originally
                           CHR(27), but we need to handle EBCDIC as
                           well as ASCII.  CHR(12) is same in both codes. *)
            WRITE(plotfile, escape,  CHR(12));   (* DLS/JMH *)
            END;
         bauschlomb : ;
         laser: BEGIN
            WRITELN(plotfile, ' 1 setlinecap 1 setlinejoin ');
            WRITELN(plotfile, treeline:8:2, ' setlinewidth ',
               'newpath ');
            END;
         hp7470: WRITELN(plotfile, 'IN;SP1;VS10.0;');
         other : BEGIN
            (* initialization code for a new plotter goes here *)
            END
         END;
      END; (* initplotter *)

      PROCEDURE finishplotter;
         VAR trash : CHAR;  (* DLS/JMH for ega, screen drivers *)
      BEGIN (* finishplotter *)
         CASE plotter OF
            tektronix: WRITELN(plotfile);
            bauschlomb : ;
            laser: BEGIN
                   WRITELN(plotfile, 'stroke showpage ');
                   WRITELN(plotfile);
                   END;
            hp7470: BEGIN
               plot(penup,0.0,0.0);
               WRITELN(plotfile,'SP;');
               END;
            other : BEGIN
               (* termination code for a new plotter goes here *)
               END
            END;
         END; (* finishplotter *)

   PROCEDURE garbagestack;
      VAR i : INTEGER;
   BEGIN (* garbagestack *)
      FOR i := 1 TO nnodes DO BEGIN
         nodelist[i]^.descendant := garbage;
         garbage := nodelist[i];
         END;
      END; (* garbagestack *)

BEGIN (* plotree *)
   WRITELN(OUTPUT, ' PLOTREE VERSION 1.4');
   WRITELN(OUTPUT, ' DISTRIBUTED WITH PHYLIP 3.1');
   WRITELN(OUTPUT, ' WRITTEN BY C. MEACHAM, UNIVERSITY HERBARIUM, ',
      'UNIVERSITY OF CALIFORNIA, ');
   WRITELN(OUTPUT, '  BERKELEY, CA 94720, U.S.A');
   WRITELN(OUTPUT, ' READING OF PARAMETERS BEGINS');
   IF ebcdic
   THEN escape := CHR(39)      (* EBCDIC *)    (* DLS/JMH *)
   ELSE escape := CHR(27);
   IF ebcdic
   THEN BEGIN
     eb[  0]:=  0; eb[  1]:=  1; eb[  2]:=  2; eb[  3]:=  3; eb[  4]:= 55;
     eb[  5]:= 45; eb[  6]:= 46; eb[  7]:= 47; eb[  8]:= 22; eb[  9]:=  5;
     eb[ 10]:= 37; eb[ 11]:= 11; eb[ 12]:= 12; eb[ 13]:= 13; eb[ 14]:= 14;
     eb[ 15]:= 15; eb[ 16]:= 16; eb[ 17]:= 17; eb[ 18]:= 18; eb[ 19]:= 19;
     eb[ 20]:= 60; eb[ 21]:= 61; eb[ 22]:= 50; eb[ 23]:= 38; eb[ 24]:= 24;
     eb[ 25]:= 25; eb[ 26]:= 63; eb[ 27]:= 39; eb[ 28]:= 28; eb[ 29]:= 29;
     eb[ 30]:= 30; eb[ 31]:= 31; eb[ 32]:= 64; eb[ 33]:= 90; eb[ 34]:=127;
     eb[ 35]:=123; eb[ 36]:= 91; eb[ 37]:=108; eb[ 38]:= 80; eb[ 39]:=125;
     eb[ 40]:= 77; eb[ 41]:= 93; eb[ 42]:= 92; eb[ 43]:= 78; eb[ 44]:=107;
     eb[ 45]:= 96; eb[ 46]:= 75; eb[ 47]:= 97; eb[ 48]:=240; eb[ 49]:=241;
     eb[ 50]:=242; eb[ 51]:=243; eb[ 52]:=244; eb[ 53]:=245; eb[ 54]:=246;
     eb[ 55]:=247; eb[ 56]:=248; eb[ 57]:=249; eb[ 58]:=122; eb[ 59]:= 94;
     eb[ 60]:= 76; eb[ 61]:=126; eb[ 62]:=110; eb[ 63]:=111; eb[ 64]:=124;
     eb[ 65]:=193; eb[ 66]:=194; eb[ 67]:=195; eb[ 68]:=196; eb[ 69]:=197;
     eb[ 70]:=198; eb[ 71]:=199; eb[ 72]:=200; eb[ 73]:=201; eb[ 74]:=209;
     eb[ 75]:=210; eb[ 76]:=211; eb[ 77]:=212; eb[ 78]:=213; eb[ 79]:=214;
     eb[ 80]:=215; eb[ 81]:=216; eb[ 82]:=217; eb[ 83]:=226; eb[ 84]:=227;
     eb[ 85]:=228; eb[ 86]:=229; eb[ 87]:=230; eb[ 88]:=231; eb[ 89]:=232;
     eb[ 90]:=233; eb[ 91]:=173; eb[ 92]:=224; eb[ 93]:=189; eb[ 94]:= 95;
     eb[ 95]:=109; eb[ 96]:=121; eb[ 97]:=129; eb[ 98]:=130; eb[ 99]:=131;
     eb[100]:=132; eb[101]:=133; eb[102]:=134; eb[103]:=135; eb[104]:=136;
     eb[105]:=137; eb[106]:=145; eb[107]:=146; eb[108]:=147; eb[109]:=148;
     eb[110]:=149; eb[111]:=150; eb[112]:=151; eb[113]:=152; eb[114]:=153;
     eb[115]:=162; eb[116]:=163; eb[117]:=164; eb[118]:=165; eb[119]:=166;
     eb[120]:=167; eb[121]:=168; eb[122]:=169; eb[123]:=192; eb[124]:= 79;
     eb[125]:=208; eb[126]:=161; eb[127]:=  7;
        END;
   (* ASSIGN(parmfile, 'PARMFILE');  Turbo Pascal only *)
   RESET(parmfile);
   initparms;
   (* CLOSE(parmfile);   may have to change this *)
   WRITELN(OUTPUT, ' READING OF PARAMETERS COMPLETED');
   IF parmsok THEN BEGIN
      (* ASSIGN(treefile, 'TREEFILE');  Turbo Pascal only *)
      RESET(treefile);
      IF translate = yes
      THEN BEGIN
         WRITELN(OUTPUT, ' READING OF TRANSLATION TABLE BEGINS');
         ReadTranslateTable (firsttranslabel, lasttranslabel);
         WRITELN(OUTPUT, ' READING OF TRANSLATION TABLE ENDS');
         END;
      (* ASSIGN(plotfile, 'PLOTFILE');   Turbo Pascal only *)
      REWRITE(plotfile);
      donetrees := FALSE;
      anerror := FALSE;
      WHILE NOT (donetrees OR anerror) DO BEGIN
         candraw := TRUE;
         WRITELN(OUTPUT, ' READING OF TREE BEGINS');
         readtree;
         donetrees := EOF(treefile);
         WRITELN(OUTPUT, ' READING OF TREE COMPLETED');
         WRITELN(OUTPUT, ' CALCULATION OF TREE BEGINS');
         calctree(root);
         WRITELN(OUTPUT, ' CALCULATION OF TREE COMPLETED');
         IF candraw
            THEN BEGIN
               hardwarecoordinates;
               IF labelspresent THEN BEGIN
                  labelsetup;
                  IF NOT(fontloaded) AND labelspresent THEN BEGIN
                     WRITELN(OUTPUT, ' LOADING OF FONT BEGINS');
                     (* ASSIGN(fontfile, 'FONTFILE');  Turbo Pascal only *)
                     RESET(fontfile);
                     loadfont(font);
                     (* CLOSE(fontfile);   may have to change this *)
                     WRITELN(OUTPUT, ' LOADING OF FONT COMPLETED');
                     fontloaded := TRUE;
                     END;
                  END;
               WRITELN(OUTPUT, ' DRAWING OF TREE BEGINS');
               initplotter;
               drawtree(root);
               finishplotter;
               WRITELN(OUTPUT, ' DRAWING OF TREE COMPLETED');
               END
            ELSE BEGIN
               WRITELN(OUTPUT, ' TREE CANNOT BE DRAWN WITH CURRENT ',
                         'PARAMETERS');
               WRITELN(OUTPUT, '  NODE DIAMETER MUST BE SMALLER OR ',
                         'INTERNODE LENGTH LARGER');
               END;
         garbagestack;
         IF NOT EOF(treefile)
         THEN IF EOLN(treefile) 
            THEN READLN(treefile);
         END;
      END;
(*   CLOSE(parmfile);   may need this here *)
(*   CLOSE(fontfile);   may need this here *)
(*   CLOSE(treefile);   may have to change this *)
(*   CLOSE(plotfile);   may have to change this *)
   WRITELN(OUTPUT, ' EXECUTION TERMINATED');
   END. (* plotree *)

