PROGRAM Shuffle (infile, outfile, OUTPUT);
  (* version 3.1. (c) Copyright 1988 by the University of
     Washington and Joseph Felsenstein.  Written by Joseph Felsenstein.
     Permission is granted to copy and use this program provided no
     fee is charged for it and provided that this copyright
     notice is not removed. *)

CONST maxsp = 30;   (* maximum number of species *)
   namelength = 10; (* number of characters max. in species name *)

TYPE vector = ARRAY [1..maxsp] OF REAL;
     intvec = ARRAY[1..maxsp] OF INTEGER;
   naym = ARRAY[1..namelength] OF CHAR;

VAR infile, outfile : TEXT;
   numsp, seed, outgrno, outgrnew : INTEGER;
   power : REAL;
   x : ARRAY[1..maxsp] OF vector;
   reps : ARRAY[1..maxsp] OF intvec;
   names : ARRAY[1..maxsp] OF naym;
   order : ARRAY[1..maxsp] OF INTEGER;
   lower, upper, outgropt, replicates : BOOLEAN;

   FUNCTION letter (ch : CHAR) : BOOLEAN;
   BEGIN (* letter -- works for both ASCII and EBCDIC codes *)
      letter := ((ch >= 'A') AND (ch <= 'I'))
                OR ((ch >= 'J') AND (ch <= 'R'))
                OR ((ch >= 'S') AND (ch <= 'Z'))
                OR ((ch >= 'a') AND (ch <= 'i'))
                OR ((ch >= 'j') AND (ch <= 'r'))
                OR ((ch >= 's') AND (ch <= 'z'));
   END; (* letter *)

   PROCEDURE uppercase (VAR ch : CHAR);
   BEGIN (* uppercase *)
      IF letter(ch) AND (('a' > 'A') AND (ch >= 'a')
                     OR  ('a' < 'A') AND (ch < 'A'))
      THEN ch := CHR(ORD(ch) + ORD('A') - ORD('a'));
   END; (* uppercase *)

   PROCEDURE getnums;
   BEGIN (* getnums *)
      READ(infile, numsp);
      WRITE(outfile, numsp:5);
      IF numsp > maxsp
      THEN WRITELN(OUTPUT, ' TOO MANY SPECIES: adjust CONSTants');
   END; (* getnums *)

   PROCEDURE getoptions;
      VAR ch : CHAR;
         i, extranum : INTEGER;
   BEGIN (* getoptions *)
      lower := FALSE;
      upper := FALSE;
      outgropt := FALSE;
      power := 2;
      replicates := FALSE;
      extranum := 0;
      WHILE NOT EOLN(infile) DO BEGIN
         READ(infile, ch);
         uppercase (ch);
         WRITE(outfile, ch);
         IF (ch = 'G') OR (ch = 'J') OR (ch = 'L')
            OR (ch = 'N') OR (ch = 'P') OR (ch = 'R') 
            OR (ch = 'O') OR (ch = 'S') OR (ch = 'Y')
         THEN CASE ch OF
               'G' : BEGIN END;
               'J' : extranum := extranum + 1;
               'L' : lower := TRUE;
               'N' : BEGIN END;
               'O' : BEGIN
                     outgropt := TRUE;
                     extranum := extranum + 1;
                     END;
               'P' : extranum := extranum + 1;
               'R' : upper := TRUE;
               'S' : replicates := TRUE;
               'Y' : BEGIN END
               END
         ELSE IF ch <> ' '
            THEN WRITELN(OUTPUT, ' Bad Option Character: ', ch);
         END;
      FOR i := 1 TO extranum DO BEGIN
         READLN(infile);
         READ(infile, ch);
         uppercase (ch);
         IF (ch <> 'O')
         THEN BEGIN
            WRITELN(outfile);
            WRITE(outfile, ch);
            END;
         IF (ch = 'J') OR (ch = 'O') OR (ch = 'P')
         THEN CASE ch OF
            'J' : BEGIN
                  READ(infile, seed);
                  WRITE(outfile, seed:6);
                  END;
            'O' : READ(infile, outgrno);
            'P' : BEGIN
                  READ(infile, power);
                  WRITE(outfile, power:8:2);
                  END
            END;
         END;
   END; (* getoptions *)

   PROCEDURE getdata;
      VAR i, j : INTEGER;
         skipit, skipother : BOOLEAN;
   BEGIN (* getdata *)
      FOR i := 1 TO numsp DO BEGIN
         x[i][i] := 0.0;
         READLN(infile);
         FOR j := 1 TO namelength DO
            IF EOLN(infile)
            THEN names[i][j] := ' '
            ELSE READ(infile, names[i][j]);
         FOR j := 1 TO numsp DO BEGIN
            skipit := (lower AND (j >= i)) OR (upper AND (j <= i));
            skipother := (lower AND (i >= j)) OR
                           (upper AND (i <= j));
            IF NOT skipit
            THEN BEGIN
               IF EOLN(infile)
               THEN READLN(infile);
               READ(infile, x[i][j]);
               IF replicates
               THEN BEGIN
                  IF EOLN(infile)
                  THEN READLN(infile);
                  READ(infile, reps[i][j]);
                  END;
               END;
            IF (NOT skipit) AND skipother
            THEN BEGIN
               x[j][i] := x[i][j];
               reps[j][i] := reps[i][j];
               END;
            END;
         END;
   END; (* getdata *)

   PROCEDURE getorder;
      VAR i : INTEGER;
   BEGIN (* getorder *)
      FOR i := 1 TO numsp DO BEGIN
         IF EOLN(infile) THEN READLN(infile);
         READ(infile, order[i]);
         IF order[i] = outgrno
         THEN outgrnew := i;
         END;
      outgrno := outgrnew; 
   END; (* getorder *)

   PROCEDURE printdata;
      VAR i, j : INTEGER;
   BEGIN (* printdata *)
      WRITELN(outfile);
      IF outgropt
      THEN WRITELN(outfile, 'O ', outgrno:3);
      FOR i := 1 TO numsp DO BEGIN
         FOR j := 1 TO namelength DO 
            WRITE(outfile, names[order[i]][j]);
         WRITE(outfile, ' ');
         FOR j := 1 TO numsp DO BEGIN
            IF upper AND (j <= i) 
            THEN WRITE(outfile, ' ')
            ELSE IF NOT (lower AND (j >= i))
               THEN BEGIN
                  WRITE(outfile, x[order[i]][order[j]]:10:5);
                  IF replicates
                  THEN WRITE(outfile, reps[order[i]][order[j]]:4);
                  END;
            IF ((NOT replicates) AND ((j MOD 7) = 0) AND (j < numsp))
               OR (replicates AND ((j MOD 5) = 0) AND (j < numsp))
            THEN WRITELN(outfile);
            END;
         WRITELN(outfile);
         END;
   END; (* printdata *)

BEGIN (* shuffle *)
   (* ASSIGN(infile, 'INFILE'); -- Turbo Pascal only *)
   RESET(infile);
   (* ASSIGN(outfile, 'OUTFILE'); -- Turbo Pascal only *)
   REWRITE(outfile);
   getnums;
   getoptions;
   getdata;
   getorder;
   printdata;
   (* CLOSE(outfile)  --- may need this here *)
END. (* shuffle *)

